Package: sensortowerR
Title: Interface to 'Sensor Tower' Mobile App Intelligence API
Version: 1.0.1
Description: Interface to the 'Sensor Tower' API <https://app.sensortower.com/api/docs/app_analysis>
    for mobile app analytics and market intelligence. Provides a small, consistent
    set of functions to retrieve app metadata, publisher information, download and
    revenue estimates, active user metrics, category rankings, aggregate game
    market denominators, and market trends.
    Four core verbs ('st_metrics', 'st_rankings', 'st_app'/'st_apps', 'st_filter')
    cover the common workflows with standardized parameters and tidyverse-friendly
    output. Supports both iOS and Android app ecosystems with unified data structures
    for cross-platform analysis.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: dplyr, glue, httr, httr2, jsonlite, lubridate, openssl, purrr,
        rlang, stats, stringr, tibble, tidyr, utils
Suggests: gt, gtExtras, knitr, pkgbuild, rcmdcheck, rmarkdown, rhub,
        testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
Authors@R: 
    person("Phillip", "Black",
           email = "pblack@gameeconomistconsulting.com",
           role = c("aut", "cre"))
Author: Phillip Black [aut, cre]
Maintainer: Phillip Black <pblack@gameeconomistconsulting.com>
URL: https://github.com/econosopher/sensortowerR
BugReports: https://github.com/econosopher/sensortowerR/issues
Depends: R (>= 4.1.0)
NeedsCompilation: no
Packaged: 2026-05-14 10:00:06 UTC; phillip
Repository: CRAN
Date/Publication: 2026-05-14 10:20:02 UTC
Built: R 4.6.0; ; 2026-05-14 14:39:55 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.character,st_filter)
S3method(c,st_filter)
S3method(format,st_filter)
S3method(print,st_filter)
export(calculate_yoy_growth)
export(example_sensortower_data)
export(format_arpu)
export(format_currency)
export(format_downloads)
export(format_large_number)
export(format_market_share)
export(format_percent)
export(format_retention)
export(format_users)
export(lookup_category_names)
export(st_active_users)
export(st_analyze_filter)
export(st_api_diagnostics)
export(st_app)
export(st_app_details)
export(st_app_enriched)
export(st_app_info)
export(st_app_lookup)
export(st_app_tag)
export(st_apps)
export(st_batch_app_lookup)
export(st_batch_metrics)
export(st_build_filter_url)
export(st_build_web_url)
export(st_cache_info)
export(st_categories)
export(st_category_rankings)
export(st_clear_app_cache)
export(st_clear_id_cache)
export(st_combine_filters)
export(st_compare_filter_results)
export(st_create_simple_filter)
export(st_custom_fields_filter)
export(st_custom_fields_filter_by_id)
export(st_custom_fields_values)
export(st_demographics)
export(st_discover_fields)
export(st_extract_filter_id)
export(st_extract_url_params)
export(st_facets_metrics)
export(st_filter)
export(st_filter_by_date)
export(st_filter_by_genre)
export(st_filter_by_monetization)
export(st_filter_by_publisher)
export(st_filter_by_sdk)
export(st_game_summary)
export(st_generate_example_filter_ids)
export(st_get_app_names)
export(st_get_filter_collection)
export(st_get_filtered_apps)
export(st_get_unified_mapping)
export(st_gt_dashboard)
export(st_is_valid_filter_id)
export(st_market_metrics)
export(st_metrics)
export(st_parse_web_url)
export(st_publisher_apps)
export(st_publisher_portfolio)
export(st_rankings)
export(st_ratings_facets)
export(st_retention)
export(st_retention_facets)
export(st_reviews_by_rating_facets)
export(st_sales_report)
export(st_session_metrics)
export(st_smart_metrics)
export(st_test_filter)
export(st_top_charts)
export(st_top_publishers)
export(st_unified_sales_report)
export(st_yoy_metrics)
import(dplyr)
import(tidyr)
importFrom(dplyr,"%>%")
importFrom(dplyr,all_of)
importFrom(dplyr,any_of)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,coalesce)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,lag)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,transmute)
importFrom(glue,glue)
importFrom(graphics,title)
importFrom(httr,GET)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,http_error)
importFrom(httr,http_status)
importFrom(httr,status_code)
importFrom(httr,stop_for_status)
importFrom(httr2,req_body_json)
importFrom(httr2,req_error)
importFrom(httr2,req_headers)
importFrom(httr2,req_method)
importFrom(httr2,req_perform)
importFrom(httr2,req_retry)
importFrom(httr2,req_timeout)
importFrom(httr2,req_url_path_append)
importFrom(httr2,req_url_query)
importFrom(httr2,req_user_agent)
importFrom(httr2,request)
importFrom(httr2,resp_body_raw)
importFrom(httr2,resp_body_string)
importFrom(httr2,resp_check_status)
importFrom(httr2,resp_status)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(lubridate,ceiling_date)
importFrom(lubridate,day)
importFrom(lubridate,days)
importFrom(lubridate,floor_date)
importFrom(lubridate,month)
importFrom(lubridate,year)
importFrom(purrr,map)
importFrom(purrr,map2_chr)
importFrom(purrr,map_chr)
importFrom(purrr,map_df)
importFrom(purrr,map_int)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(rlang,`%||%`)
importFrom(rlang,abort)
importFrom(rlang,sym)
importFrom(rlang,warn)
importFrom(stats,na.omit)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(stringr,str_detect)
importFrom(stringr,str_to_lower)
importFrom(tibble,as_tibble)
importFrom(tibble,as_tibble_row)
importFrom(tibble,tibble)
importFrom(tidyr,nest)
importFrom(tidyr,unnest)
importFrom(utils,URLencode)
importFrom(utils,data)
importFrom(utils,head)
importFrom(utils,tail)

# sensortowerR

<p align="center"><img src="inst/images/SensorTowerR_sticker.png" width="200"></p>

`sensortowerR` is a tidyverse-first R client for Sensor Tower's mobile app intelligence API. Four core verbs — **`st_metrics()`**, **`st_rankings()`**, **`st_app()` / `st_apps()`**, **`st_filter()`** — cover the vast majority of workflows with consistent parameters and long-format tibble output.

> **Upgrading from 0.9.x?** See [`vignette("migrating-to-1.0")`](vignettes/migrating-to-1.0.Rmd). Revenue is now in dollars by default (was cents). Most 0.9.x functions are `.Defunct()` stubs that will tell you exactly what to replace them with.

## Installation

```r
# From GitHub
remotes::install_github("econosopher/sensortowerR")

# From CRAN (once 1.0.0 is released)
install.packages("sensortowerR")
```

## Authentication

Store your Sensor Tower API token as an environment variable:

```r
usethis::edit_r_environ()
# SENSORTOWER_AUTH_TOKEN="YOUR_SECRET_TOKEN_HERE"
```

Restart your R session after updating `.Renviron`.

## The four core verbs

### `st_metrics()` — revenue and downloads

```r
library(sensortowerR)
library(dplyr)

# Single app, unified across iOS + Android
royal_match <- st_apps("Royal Match") |> slice(1) |> pull(app_id)

sales <- st_metrics(
  app_id      = royal_match,
  metrics     = c("revenue", "downloads"),
  countries   = c("US", "JP", "GB"),
  date_from   = "2025-01-01",
  date_to     = "2025-12-31",
  granularity = "monthly"
)

# Returns long format by default: app_id, os, country, date, metric, value
# Revenue is in dollars. Pass `revenue_unit = "cents"` for raw integer cents.
```

### `st_rankings()` — top charts, publishers, categories

```r
top_ios_games <- st_rankings(
  entity   = "app",
  os       = "ios",
  category = 6014,
  country  = "US",
  limit    = 50
)

top_publishers <- st_rankings(
  entity  = "publisher",
  os      = "unified",
  country = "US"
)
```

### `st_app()` and `st_apps()` — lookup vs. search

```r
# Direct ID lookup
st_app(app_id = "55c5022f02ac64f9c0001f9f")

# Discovery by name
st_apps(query = "candy crush", os = "ios", limit = 10)

# Discovery by filter
rpg_filter <- st_filter(genre = "rpg", monetization = "in_app_purchases")
rpg_apps   <- st_apps(filter = rpg_filter, limit = 50)
```

### `st_filter()` — the builder

`st_filter()` returns an `st_filter` S3 object that `st_apps()`, `st_rankings()`, and `st_get_filtered_apps()` all accept. Compose filters with `c()`.

```r
us_rpgs_2025 <- st_filter(
  date_from    = "2025-01-01",
  date_to      = "2025-12-31",
  genre        = "rpg",
  monetization = "free"
)

print(us_rpgs_2025)  # Human-readable summary
as.character(us_rpgs_2025)  # Server-side filter ID

# Combine two filters
combined <- c(us_rpgs_2025, st_filter(sdk = "unity"))
```

### `st_market_metrics()` — tidy aggregate game market totals

Use `st_market_metrics()` when the analysis needs market/category denominator
series in a tidy schema. It wraps Sensor Tower's aggregate
`/v1/{os}/games_breakdown` endpoint through `st_game_summary()` and returns
stable snake_case columns.

```r
casino_market <- st_market_metrics(
  category    = c("7006", "game_casino"),
  countries   = "WW",
  os          = "unified",
  date_from   = "2025-01-01",
  date_to     = "2025-12-31",
  granularity = "monthly",
  shape       = "wide"
)
```

### `st_game_summary()` — low-level aggregate game market totals

Use `st_game_summary()` when the analysis needs market/category/genre denominator series rather than title-level cohorts. It calls Sensor Tower's aggregate `/v1/{os}/games_breakdown` endpoint for `os = "ios"` or `os = "android"`; `os = "unified"` fetches the iOS and Android aggregate rows and sums them by date/country.

```r
market <- st_game_summary(
  categories        = 7001,
  countries         = c("US", "JP", "GB"),
  os                = "unified",
  date_granularity  = "monthly",
  start_date        = "2025-01-01",
  end_date          = "2025-12-31"
)

# Enriched revenue columns are in dollars. Raw endpoint revenue is cents.
```

Do not build denominator time series by batching top-N rankings, top charts, custom-filter pagination, or 1,500-app rosters. Those are title cohorts, not market totals, and they will drift with roster selection and chart cutoffs. Use `st_rankings()` and `st_apps(filter = ...)` for discovery, leaderboards, and peer cohorts; use `st_market_metrics()` for category-backed denominators.

## Function index

### Core (4 unified verbs)

| Function | Purpose |
|---|---|
| `st_metrics()` | Revenue / downloads across one or many apps, countries, dates |
| `st_rankings()` | Top charts / publishers / categories |
| `st_app()` / `st_apps()` | ID lookup / name search / filter-based discovery |
| `st_filter()` | Build a reusable filter object |
| `st_market_metrics()` | Aggregate game category market revenue/download denominators |

### Analytics

| Function | Purpose |
|---|---|
| `st_active_users()` | DAU / WAU / MAU time series |
| `st_retention()` | D1–D90 retention curves |
| `st_retention_facets()` | Demographic retention breakdowns |
| `st_ratings_facets()` | Ratings by demographic facets |
| `st_reviews_by_rating_facets()` | Review counts per star rating |
| `st_session_metrics()` | Session counts, duration, time spent |
| `st_demographics()` | Age / gender breakdowns |
| `st_app_enriched()` | Multi-metric enrichment for known apps |
| `st_yoy_metrics()` | Year-over-year comparison helper |
| `st_game_summary()` | Low-level game category / market summary from `games_breakdown` |

### Publishers

| Function | Purpose |
|---|---|
| `st_publisher_apps()` | All apps owned by a publisher |
| `st_publisher_portfolio()` | End-to-end publisher portfolio analysis |

### Low-level / power-user

| Function | Purpose |
|---|---|
| `st_facets_metrics()` | Direct `/facets/metrics` access |
| `st_get_filtered_apps()` | Raw filter-ID app listing |
| `st_test_filter()`, `st_analyze_filter()` | Validate and inspect filters |
| `st_discover_fields()`, `st_custom_fields_values()` | Introspect available filter fields |
| `st_api_diagnostics()` | Debug API request issues |
| `st_get_unified_mapping()`, `st_batch_app_lookup()` | Platform ID ↔ unified-ID resolution |
| `st_categories()` | Full category list |

### Utilities

| Function | Purpose |
|---|---|
| `st_gt_dashboard()` | Formatted `gt` table output |
| `st_cache_info()`, `st_clear_id_cache()`, `st_clear_app_cache()` | Cache management |
| `st_build_web_url()`, `st_parse_web_url()` | App-ID ↔ web URL |
| `format_currency()`, `format_downloads()`, `format_percent()`, `format_retention()`, `format_market_share()`, `format_arpu()`, `format_users()`, `format_large_number()` | Human-readable number formatters |
| `calculate_yoy_growth()`, `lookup_category_names()` | Analysis helpers |
| `example_sensortower_data()` | Sample data for examples (no API call) |

## Data notes

| Data Type | Function | Time Series | Coverage |
|---|---|---|---|
| Revenue / downloads | `st_metrics()` | Yes | Country-level |
| DAU / WAU / MAU | `st_active_users()` | Yes | Country-level |
| Retention | `st_retention()`, `st_retention_facets()` | Snapshot | US / WW primarily |
| Ratings / reviews | `st_ratings_facets()`, `st_reviews_by_rating_facets()` | Snapshot | WW |
| Rankings | `st_rankings()` | Point-in-time | Country-level |

## Parameter conventions (v1.0.0)

Every function uses the same parameter names:

| Parameter | Meaning |
|---|---|
| `app_id` | Scalar or vector of app IDs (unified, iOS, or Android — auto-resolved) |
| `os` | `"ios"`, `"android"`, or `"unified"` |
| `country` / `countries` | ISO-2 code (scalar) or vector of codes |
| `date_from`, `date_to` | Dates or ISO strings |
| `granularity` | `"daily"`, `"weekly"`, `"monthly"`, `"quarterly"` |
| `metrics` | Character vector of metric names |
| `auth_token` | Optional override of the env-var token |
| `limit` | Row cap |

## Learn more

- Vignette: `vignette("migrating-to-1.0", package = "sensortowerR")`
- Vignette: `vignette("tidy-active-users", package = "sensortowerR")`
- Vignette: `vignette("custom-fields", package = "sensortowerR")`
- Changelog: [NEWS.md](https://github.com/econosopher/sensortowerR/blob/main/NEWS.md)

## License

MIT (`LICENSE` file).

#!/usr/bin/env Rscript

# Cross-platform CRAN check script for sensortowerR
# This script helps run CRAN checks across multiple platforms

# Load required packages
required_packages <- c("devtools", "rcmdcheck", "rhub", "pkgbuild")
for (pkg in required_packages) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    message(sprintf("Installing %s...", pkg))
    install.packages(pkg)
  }
}

# Function to display separator
separator <- function(title) {
  cat("\n", paste(rep("=", 50), collapse = ""), "\n", sep = "")
  cat(title, "\n")
  cat(paste(rep("=", 50), collapse = ""), "\n\n", sep = "")
}

# Main function to coordinate all checks
run_all_platform_checks <- function() {
  
  cat("SENSORTOWER R PACKAGE - CROSS-PLATFORM CRAN CHECKS\n")
  cat("==================================================\n\n")
  
  # 1. Clean up old build artifacts
  separator("CLEANUP")
  old_tarballs <- list.files(pattern = "*.tar.gz", full.names = TRUE)
  if (length(old_tarballs) > 0) {
    cat("Removing old tarballs:\n")
    for (tb in old_tarballs) {
      cat("  -", tb, "\n")
      unlink(tb)
    }
  }
  
  # 2. Build the package
  separator("BUILD PACKAGE")
  cat("Building package...\n")
  pkg_path <- pkgbuild::build(quiet = FALSE)
  cat("\nPackage built:", pkg_path, "\n")
  
  # 3. Local CRAN check
  separator("LOCAL CRAN CHECK")
  cat("Running R CMD check --as-cran\n\n")
  
  check_result <- rcmdcheck::rcmdcheck(
    path = pkg_path,
    args = "--as-cran",
    error_on = "never"
  )
  
  cat("\nResults:\n")
  cat("- Errors:  ", length(check_result$errors), "\n")
  cat("- Warnings:", length(check_result$warnings), "\n")
  cat("- Notes:   ", length(check_result$notes), "\n")
  
  # 4. rhub checks
  separator("RHUB CHECKS (Multiple Platforms)")
  
  if (requireNamespace("rhub", quietly = TRUE)) {
    cat("rhub v2 is available for cross-platform testing.\n\n")
    cat("To run rhub checks:\n")
    cat("1. First time setup: rhub::rhub_setup()\n")
    cat("2. Submit checks:    rhub::rhub_check()\n\n")
    cat("This will test on:\n")
    cat("  - Ubuntu Linux (multiple R versions)\n")
    cat("  - Windows (multiple R versions)\n")
    cat("  - macOS (multiple R versions)\n")
    cat("  - Debian Linux\n")
    cat("  - Fedora Linux\n\n")
    
    # Ask user if they want to run rhub
    if (interactive()) {
      response <- readline("Do you want to submit to rhub now? (y/n): ")
      if (tolower(response) == "y") {
        rhub::rhub_check()
      }
    }
  }
  
  # 5. Win-builder checks
  separator("WIN-BUILDER CHECKS")
  cat("Win-builder tests on Windows with different R versions.\n\n")
  cat("Available commands:\n")
  cat("  devtools::check_win_devel()      # R-devel\n")
  cat("  devtools::check_win_release()    # R-release\n")
  cat("  devtools::check_win_oldrelease() # R-oldrel\n\n")
  
  if (interactive()) {
    response <- readline("Submit to win-builder? (y/n): ")
    if (tolower(response) == "y") {
      which_version <- readline("Which version? (devel/release/oldrelease/all): ")
      
      if (which_version %in% c("devel", "all")) {
        cat("\nSubmitting to win-builder R-devel...\n")
        devtools::check_win_devel()
      }
      if (which_version %in% c("release", "all")) {
        cat("\nSubmitting to win-builder R-release...\n")
        devtools::check_win_release()
      }
      if (which_version %in% c("oldrelease", "all")) {
        cat("\nSubmitting to win-builder R-oldrelease...\n")
        devtools::check_win_oldrelease()
      }
    }
  }
  
  # 6. macOS builder
  separator("MACOS BUILDER")
  cat("macOS builder tests on real Mac hardware.\n")
  cat("Submit package at: https://mac.r-project.org/macbuilder/submit.html\n")
  cat("Upload the tarball:", pkg_path, "\n\n")
  
  # 7. Additional platform-specific notes
  separator("PLATFORM-SPECIFIC TESTING NOTES")
  
  cat("SOLARIS:\n")
  cat("  - No longer required for CRAN (Solaris support ended)\n\n")
  
  cat("LINUX DISTRIBUTIONS:\n") 
  cat("  - rhub covers major distributions\n")
  cat("  - Consider Docker for specific distro testing\n\n")
  
  cat("ARCHITECTURES:\n")
  cat("  - Most checks cover x86_64\n")
  cat("  - ARM64 (Apple Silicon) covered by macOS builder\n\n")
  
  # 8. Summary
  separator("SUMMARY")
  cat("Package:", pkg_path, "\n\n")
  cat("Checks completed:\n")
  cat("  ✓ Local CRAN check\n")
  cat("  - rhub (if submitted)\n")
  cat("  - win-builder (if submitted)\n")
  cat("  - macOS builder (manual submission required)\n\n")
  
  cat("Next steps:\n")
  cat("1. Review any errors, warnings, or notes\n")
  cat("2. Fix issues and re-run checks\n")
  cat("3. Wait for email results from remote builders\n")
  cat("4. Update cran-comments.md with test results\n\n")
  
  return(invisible(pkg_path))
}

# Run the checks if script is executed directly
if (!interactive() || length(commandArgs(trailingOnly = TRUE)) > 0) {
  run_all_platform_checks()
} else {
  cat("Cross-platform check script loaded.\n")
  cat("Run: run_all_platform_checks()\n")
}

#!/usr/bin/env Rscript

# Quick platform check script - runs basic checks without interaction
# Usage: Rscript inst/scripts/quick_platform_check.R

message("Starting quick cross-platform checks for sensortowerR...")

# Ensure we're in the package root
if (!file.exists("DESCRIPTION")) {
  stop("This script must be run from the package root directory")
}

# Load required packages
if (!requireNamespace("rcmdcheck", quietly = TRUE)) {
  install.packages("rcmdcheck")
}

# Clean up old artifacts
old_files <- list.files(pattern = "*.tar.gz|*.Rcheck", full.names = TRUE)
if (length(old_files) > 0) {
  message("Cleaning up old files...")
  unlink(old_files, recursive = TRUE)
}

# Run local CRAN check
message("\n=== Running local CRAN check ===")
result <- rcmdcheck::rcmdcheck(args = "--as-cran", error_on = "never")

# Display results
message("\n=== Check Results ===")
message("Errors: ", length(result$errors))
message("Warnings: ", length(result$warnings))
message("Notes: ", length(result$notes))

if (length(result$errors) > 0) {
  message("\nERRORS:")
  for (e in result$errors) message("  - ", e)
}

if (length(result$warnings) > 0) {
  message("\nWARNINGS:")
  for (w in result$warnings) message("  - ", w)
}

if (length(result$notes) > 0) {
  message("\nNOTES:")
  for (n in result$notes) message("  - ", n)
}

# Platform test commands
message("\n=== Platform Test Commands ===")
message("\nFor Windows testing:")
message("  devtools::check_win_devel()")
message("  devtools::check_win_release()")
message("  devtools::check_win_oldrelease()")

message("\nFor multi-platform testing (rhub):")
message("  rhub::rhub_setup()  # First time only")
message("  rhub::rhub_check()  # Submit checks")

message("\nFor macOS testing:")
message("  Upload tarball to: https://mac.r-project.org/macbuilder/submit.html")

message("\nFor comprehensive testing:")
message("  source('inst/scripts/check_all_platforms.R')")
message("  run_all_platform_checks()")

# Exit status based on errors
quit(status = length(result$errors))

# Comprehensive Analysis of st_top_charts API Response
# This example demonstrates the rich data structure returned by the Sensor Tower API
# when using st_top_charts with custom filters

library(sensortowerR)
suppressPackageStartupMessages(library(dplyr))

# Example URL from Sensor Tower web interface
# This URL includes a custom filter and requests DAU (Daily Active Users) data
test_url <- "https://app.sensortower.com/market-analysis/top-apps?os=unified&edit=1&granularity=weekly&start_date=2025-06-29&end_date=2025-07-28&duration=P30D&measure=DAU&comparison_attribute=absolute&category=0&device=iphone&device=ipad&device=android&metric=revenue&page=1&page_size=25&custom_fields_filter_mode=include_unified_apps&period=day&custom_fields_filter_id=60746340241bc16eb8a65d76&country=US&country=CA&country=AU&country=NZ&country=GB&country=CH&country=SE&country=NO&country=NL&country=LU&country=IE&country=DE&country=FR&country=FI&country=DK&country=BE&country=AT&country=ES&country=IT&country=PT"

# Parse URL and make API call
cat("=== SENSOR TOWER API RESPONSE ANALYSIS ===\n\n")
params <- st_parse_web_url(test_url, verbose = FALSE)
data <- suppressMessages(do.call(st_top_charts, params))

# 1. RESPONSE OVERVIEW
cat("1. RESPONSE OVERVIEW\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
cat("Total apps returned:", nrow(data), "\n")
cat("Total data columns:", ncol(data), "\n")
cat("Primary metric:", params$measure, "(Daily Active Users)\n")
cat("Date range:", params$date, "to", params$end_date, "\n")
cat("Countries:", length(strsplit(params$regions, ",")[[1]]), "countries\n")
cat("Custom filter applied:", if(!is.null(params$custom_fields_filter_id)) "Yes" else "No", "\n\n")

# 2. TOP APPS BY PRIMARY METRIC (DAU)
cat("2. TOP 10 APPS BY DAILY ACTIVE USERS\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
top_apps <- data %>% 
  select(
    Rank = rank,
    App = unified_app_name,
    Publisher = publisher_name,
    DAU = entities.users_absolute
  ) %>%
  arrange(desc(DAU)) %>%
  mutate(DAU = format(round(DAU), big.mark = ",")) %>%
  head(10)
print(top_apps, row.names = FALSE)

# 3. DATA CATEGORIES BREAKDOWN
cat("\n3. DATA CATEGORIES (", ncol(data), " columns total)\n", sep = "")
cat(paste(rep("-", 50), collapse = ""), "\n")

# Count columns by category
categories <- list(
  "Core App Info" = length(grep("^(unified_app_|publisher_|rank)", names(data))),
  "Entity Metrics" = length(grep("^entities\\.", names(data))),
  "Custom Tags" = length(grep("custom_tags", names(data))),
  "Revenue Data" = length(grep("revenue", names(data), ignore.case = TRUE)),
  "Download Data" = length(grep("download|units", names(data), ignore.case = TRUE)),
  "User Metrics" = length(grep("dau|wau|mau|users", names(data), ignore.case = TRUE)),
  "Retention Rates" = length(grep("retention", names(data), ignore.case = TRUE)),
  "Monetization" = length(grep("rpd|arpu|arpdau", names(data), ignore.case = TRUE)),
  "Demographics" = length(grep("age|gender|male|female", names(data), ignore.case = TRUE)),
  "Ratings" = length(grep("rating", names(data), ignore.case = TRUE)),
  "Dates" = length(grep("date|release", names(data), ignore.case = TRUE))
)

for (cat_name in names(categories)) {
  cat(sprintf("%-20s: %3d columns\n", cat_name, categories[[cat_name]]))
}

# 4. RICH METRICS FOR TOP APP
cat("\n4. SAMPLE METRICS FOR TOP APP\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
top_app <- data[which.max(data$entities.users_absolute), ]
cat("App:", top_app$unified_app_name, "\n")
cat("Publisher:", top_app$publisher_name, "\n")
cat("App ID:", top_app$unified_app_id, "\n\n")

# Display key metrics
metrics <- list(
  "Daily Active Users (DAU)" = list(value = top_app$entities.users_absolute, format = "number"),
  "30-Day Revenue (WW)" = list(value = top_app$revenue_30d_ww, format = "currency"),
  "30-Day Downloads (WW)" = list(value = top_app$downloads_30d_ww, format = "number"),
  "All-Time Revenue (WW)" = list(value = top_app$revenue_alltime_ww, format = "currency"),
  "All-Time Downloads (WW)" = list(value = top_app$downloads_alltime_ww, format = "number"),
  "7-Day Retention (US)" = list(value = top_app$retention_7d_us, format = "percent"),
  "30-Day Retention (US)" = list(value = top_app$retention_30d_us, format = "percent"),
  "Revenue Per Download (US)" = list(value = top_app$rpd_alltime_us, format = "currency"),
  "Monthly Active Users (WW)" = list(value = top_app$mau_month_ww, format = "number"),
  "Average Age (US)" = list(value = top_app$age_us, format = "decimal")
)

cat("Key Performance Metrics:\n")
for (metric_name in names(metrics)) {
  metric <- metrics[[metric_name]]
  if (!is.null(metric$value) && !is.na(metric$value)) {
    formatted_value <- switch(metric$format,
      "currency" = paste0("$", format(round(metric$value), big.mark = ",")),
      "number" = format(round(metric$value), big.mark = ","),
      "percent" = sprintf("%.1f%%", metric$value * 100),
      "decimal" = sprintf("%.1f", metric$value)
    )
    cat(sprintf("  %-30s: %s\n", metric_name, formatted_value))
  }
}

# 5. CUSTOM FILTER ANALYSIS
cat("\n5. CUSTOM FILTER ANALYSIS\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
if (!is.null(params$custom_fields_filter_id)) {
  cat("Filter ID:", params$custom_fields_filter_id, "\n")
  cat("Filter Mode:", params$custom_tags_mode, "\n")
  cat("Apps returned:", nrow(data), "(filtered from thousands)\n\n")
  
  # Publisher distribution
  cat("Publisher Distribution (Top 5):\n")
  pub_count <- data %>% 
    count(publisher_name) %>% 
    arrange(desc(n)) %>%
    head(5)
  
  for (i in 1:nrow(pub_count)) {
    cat(sprintf("  %-30s: %d apps\n", pub_count$publisher_name[i], pub_count$n[i]))
  }
} else {
  cat("No custom filter applied\n")
}

# 6. GEOGRAPHIC COVERAGE
cat("\n6. GEOGRAPHIC COVERAGE\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
countries <- strsplit(params$regions, ",")[[1]]
cat("Countries included (", length(countries), "):\n", sep = "")
cat(paste(strwrap(paste(countries, collapse = ", "), width = 70), collapse = "\n"), "\n")

# 7. EXAMPLE COLUMN STRUCTURE
cat("\n7. COLUMN NAMING STRUCTURE\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
cat("The API returns nested JSON that gets flattened using dot notation:\n\n")
example_cols <- c(
  "unified_app_name" = "App name (unified across platforms)",
  "entities.users_absolute" = "Primary metric value (DAU in this case)",
  "entities.delta" = "Change from previous period",
  "entities.custom_tags.Revenue" = "Custom account-specific metrics",
  "revenue_30d_ww" = "30-day revenue worldwide",
  "retention_7d_us" = "7-day retention rate in US"
)

for (col in names(example_cols)) {
  cat(sprintf("  %-35s: %s\n", col, example_cols[col]))
}

# 8. DAU STATISTICS
cat("\n8. DAILY ACTIVE USERS STATISTICS\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
dau_stats <- data %>%
  summarise(
    Min = min(entities.users_absolute, na.rm = TRUE),
    Q1 = quantile(entities.users_absolute, 0.25, na.rm = TRUE),
    Median = median(entities.users_absolute, na.rm = TRUE),
    Mean = mean(entities.users_absolute, na.rm = TRUE),
    Q3 = quantile(entities.users_absolute, 0.75, na.rm = TRUE),
    Max = max(entities.users_absolute, na.rm = TRUE)
  )

cat("DAU Distribution:\n")
cat(sprintf("  Minimum:    %s\n", format(round(dau_stats$Min), big.mark = ",")))
cat(sprintf("  Q1:         %s\n", format(round(dau_stats$Q1), big.mark = ",")))
cat(sprintf("  Median:     %s\n", format(round(dau_stats$Median), big.mark = ",")))
cat(sprintf("  Mean:       %s\n", format(round(dau_stats$Mean), big.mark = ",")))
cat(sprintf("  Q3:         %s\n", format(round(dau_stats$Q3), big.mark = ",")))
cat(sprintf("  Maximum:    %s\n", format(round(dau_stats$Max), big.mark = ",")))

# 9. DATA EXPORT OPTIONS
cat("\n9. DATA EXPORT\n")
cat(paste(rep("-", 50), collapse = ""), "\n")
cat("Saving data for further analysis:\n")

# Save full dataset
saveRDS(data, "top_charts_api_response.rds")
cat("  - Full dataset saved to: top_charts_api_response.rds\n")

# Save column names
writeLines(names(data), "api_column_names.txt")
cat("  - Column names saved to: api_column_names.txt\n")

# Save top apps summary
write.csv(top_apps, "top_apps_by_dau.csv", row.names = FALSE)
cat("  - Top apps summary saved to: top_apps_by_dau.csv\n")

# 10. SUMMARY
cat("\n10. SUMMARY\n")
cat(paste(rep("=", 50), collapse = ""), "\n")
cat("This API response demonstrates the comprehensive data available:\n\n")
cat("✓", nrow(data), "apps matching the custom filter\n")
cat("✓", ncol(data), "data points per app\n")
cat("✓ Primary metric: Daily Active Users (DAU)\n")
cat("✓ Coverage: Revenue, downloads, retention, demographics\n")
cat("✓ Time periods: 30-day, 180-day, all-time metrics\n")
cat("✓ Geographic scope:", length(countries), "countries\n")
cat("✓ Custom metrics specific to your account/filter\n")

cat("\nThe flattened structure makes it easy to analyze in R using\n")
cat("standard data manipulation tools like dplyr and ggplot2.\n")

# Demonstration of ID Optimization Features in sensortowerR
# This shows how the package minimizes API calls through intelligent caching

library(sensortowerR)

# Enable verbose mode to see the optimization in action
options(sensortowerR.verbose = TRUE)

cat("=== sensortowerR ID Optimization Demo ===\n\n")

# 1. First, let's clear any existing cache to start fresh
st_clear_id_cache()

# 2. Show cache info (should be empty)
cat("\nInitial cache status:\n")
st_cache_info()

# 3. First request - will make API calls
cat("\n\nFIRST REQUEST (Cold Cache):\n")
cat("==========================\n")

apps <- c(
  "553834731",                    # Candy Crush iOS
  "com.king.candycrushsaga",      # Candy Crush Android
  "5ba4585f539ce75b97db6bcb",     # Star Trek unified
  "1195621598",                   # Homescapes iOS
  "com.playrix.homescapes"        # Homescapes Android
)

# This will make API calls to resolve IDs
metrics1 <- st_smart_metrics(
  app_ids = apps,
  metrics = c("revenue", "downloads"),
  start_date = "2024-01-01",
  end_date = "2024-01-07",
  countries = "US"
)

# 4. Check cache after first request
cat("\n\nCache after first request:\n")
st_cache_info()

# 5. Second request with same apps - should use cache
cat("\n\nSECOND REQUEST (Warm Cache):\n")
cat("============================\n")

# This should NOT make any ID lookup API calls
metrics2 <- st_smart_metrics(
  app_ids = apps,
  metrics = c("revenue", "downloads", "dau"),
  start_date = "2024-02-01",
  end_date = "2024-02-07",
  countries = "US"
)

cat("\nNotice: No ID lookup API calls needed - all IDs resolved from cache!\n")

# 6. Request with partially cached apps
cat("\n\nTHIRD REQUEST (Mixed Cache):\n")
cat("===========================\n")

mixed_apps <- c(
  "553834731",                    # Cached
  "1094591345",                   # New - Pokemon GO iOS
  "com.nianticlabs.pokemongo"     # New - Pokemon GO Android
)

metrics3 <- st_smart_metrics(
  app_ids = mixed_apps,
  metrics = c("revenue"),
  start_date = "2024-03-01",
  end_date = "2024-03-07",
  countries = "WW"
)

# 7. Show final cache statistics
cat("\n\nFinal cache statistics:\n")
st_cache_info()

# 8. Demonstrate batch resolution efficiency
cat("\n\nBATCH RESOLUTION DEMO:\n")
cat("=====================\n")

# Many apps at once - uses batch API endpoint when possible
many_apps <- c(
  "529479190",                    # Clash of Clans iOS
  "com.supercell.clashofclans",   # Clash of Clans Android
  "1053012308",                   # MONOPOLY GO! iOS
  "com.scopely.monopolygo",       # MONOPOLY GO! Android
  "1354664240",                   # Royal Match iOS
  "com.dreamgames.royalmatch"     # Royal Match Android
)

cat("Resolving", length(many_apps), "app IDs...\n")
batch_metrics <- st_smart_metrics(
  app_ids = many_apps,
  metrics = c("revenue"),
  start_date = "2024-01-01",
  end_date = "2024-01-01",
  countries = "US"
)

# 9. Show how the cache persists between sessions
cat("\n\nCACHE PERSISTENCE:\n")
cat("==================\n")

# Save current cache stats
cache_stats_before <- st_cache_info()

# To persist cache between sessions, explicitly call save_id_cache()
cat("\nTo save cache for future sessions, call save_id_cache()\n")
cat("Cache location (CRAN-compliant): ", tools::R_user_dir("sensortowerR", "cache"), "\n")
cat("To load cached data in a new session, call load_id_cache()\n")

# 10. ID Type Detection Demo
cat("\n\nID TYPE DETECTION:\n")
cat("==================\n")

test_ids <- c(
  "553834731",                    # iOS numeric
  "com.king.candycrushsaga",      # Android package
  "5ba4585f539ce75b97db6bcb",     # Unified hex
  "pub123456",                    # Publisher ID
  "unknown_format_123"            # Unknown
)

for (id in test_ids) {
  cat(sprintf("%-30s -> %s\n", id, detect_id_type(id)))
}

# 11. Configuration options
cat("\n\nCONFIGURATION OPTIONS:\n")
cat("=====================\n")

cat("Current settings:\n")
cat("  Cache enabled:", getOption("sensortowerR.use_cache"), "\n")
cat("  Cache max age:", getOption("sensortowerR.cache_max_age_days"), "days\n")
cat("  Auto resolve:", getOption("sensortowerR.auto_resolve"), "\n")
cat("  Verbose mode:", getOption("sensortowerR.verbose"), "\n")

# Summary
cat("\n\n=== SUMMARY ===\n")
cat("The ID optimization features provide:\n")
cat("1. Automatic caching of ID lookups (30-day expiry)\n")
cat("2. Batch resolution to minimize API calls\n")
cat("3. Smart ID type detection\n")
cat("4. Persistent cache between R sessions\n")
cat("5. Configurable behavior via options()\n")
cat("\nResult: Dramatically reduced API calls and faster performance!\n")

# sensortowerR v0.2.3+ Revenue Standardization Example
# This script demonstrates the automatic revenue standardization feature

library(tidyverse)
library(sensortowerR)
library(scales)

cat("sensortowerR Revenue Standardization Example\n")
cat("============================================\n\n")

# Check package version
pkg_version <- packageVersion("sensortowerR")
cat(sprintf("Current sensortowerR version: %s\n", pkg_version))

if (pkg_version < "0.2.3") {
  cat("\nWARNING: Please update to sensortowerR v0.2.3+ for automatic revenue standardization\n")
  cat("Install the latest version with: devtools::install_github('econosopher/SensorTowerR')\n\n")
}

# Example 1: st_top_charts() with standardized revenue
cat("\n1. Fetching top games with standardized revenue units\n")
cat("=====================================================\n")

top_games <- st_top_charts(
  measure = "revenue",
  category = 6000,  # Games
  os = "ios",
  regions = "US",
  time_range = "month",
  limit = 5
)

# Display the standardized revenue column
cat("\nStandardized revenue values (in base currency units):\n")
if ("revenue" %in% names(top_games)) {
  display_data <- top_games %>%
    select(unified_app_name, revenue, revenue_absolute) %>%
    mutate(
      revenue_formatted = dollar(revenue, accuracy = 0.1, scale = 1e-6, suffix = "M"),
      cents_formatted = comma(revenue_absolute, suffix = "¢"),
      conversion_check = round(revenue_absolute / revenue, 0)
    )
  
  print(display_data, n = 5)
  
  cat("\nRevenue unit attributes:\n")
  cat(sprintf("- revenue unit: %s\n", attr(top_games$revenue, "unit") %||% "base_currency"))
  cat(sprintf("- revenue_absolute unit: %s\n", attr(top_games$revenue_absolute, "unit") %||% "cents"))
} else {
  cat("Note: 'revenue' column not found. Please update to sensortowerR v0.2.3+\n")
}

# Example 2: Demonstrating unit consistency across functions
cat("\n\n2. Unit Consistency Across Functions\n")
cat("====================================\n")

# Get a single app for detailed analysis
if (nrow(top_games) > 0) {
  sample_app <- top_games$app_id[1]
  sample_name <- top_games$unified_app_name[1]
  
  cat(sprintf("\nAnalyzing: %s (App ID: %s)\n", sample_name, sample_app))
  
  # Fetch sales report data
  cat("\nFetching sales report data...\n")
  sales_data <- st_sales_report(
    app_ids = sample_app,
    os = "ios",
    countries = "US",
    start_date = Sys.Date() - 30,
    end_date = Sys.Date() - 1,
    date_granularity = "monthly"
  )
  
  if (!is.null(sales_data) && nrow(sales_data) > 0) {
    cat("\nRevenue comparison:\n")
    cat(sprintf("- Top charts revenue (30d): %s\n", 
                dollar(top_games$revenue[1], accuracy = 0.1, scale = 1e-6, suffix = "M")))
    cat(sprintf("- Sales report total: %s\n", 
                dollar(sales_data$total_revenue[1], accuracy = 0.1, scale = 1e-6, suffix = "M")))
    cat("\nNote: Values may differ due to timing and data sources\n")
  }
}

# Example 3: Working with the standardized data
cat("\n\n3. Analysis with Standardized Revenue\n")
cat("=====================================\n")

# Category comparison
cat("\nComparing game subcategories (standardized revenue):\n")

categories <- list(
  "Puzzle" = 7003,
  "Action" = 7001,
  "RPG" = 7014
)

category_data <- map_dfr(names(categories), function(cat_name) {
  cat_data <- st_top_charts(
    measure = "revenue",
    category = categories[[cat_name]],
    os = "ios",
    regions = "US",
    time_range = "month",
    limit = 10
  )
  
  if (!is.null(cat_data) && nrow(cat_data) > 0 && "revenue" %in% names(cat_data)) {
    tibble(
      category = cat_name,
      total_revenue = sum(cat_data$revenue, na.rm = TRUE),
      avg_revenue = mean(cat_data$revenue, na.rm = TRUE),
      top_game_revenue = max(cat_data$revenue, na.rm = TRUE),
      n_games = nrow(cat_data)
    )
  } else {
    NULL
  }
})

if (nrow(category_data) > 0) {
  category_summary <- category_data %>%
    mutate(
      total_formatted = dollar(total_revenue, accuracy = 0.1, scale = 1e-6, suffix = "M"),
      avg_formatted = dollar(avg_revenue, accuracy = 0.1, scale = 1e-6, suffix = "M"),
      top_formatted = dollar(top_game_revenue, accuracy = 0.1, scale = 1e-6, suffix = "M")
    ) %>%
    select(category, n_games, total_formatted, avg_formatted, top_formatted)
  
  print(category_summary)
}

# Example 4: Best practices
cat("\n\n4. Best Practices\n")
cat("==================\n")

cat("
1. Always use the standardized 'revenue' column for analysis
2. The original 'revenue_absolute' (cents) is preserved for reference
3. Check for the 'revenue' column to ensure you're using v0.2.3+
4. Revenue units are consistent across all sensortowerR functions:
   - st_top_charts(): 'revenue' column
   - st_sales_report(): 'total_revenue', 'iphone_revenue', etc.
   - st_top_publishers(): 'revenue_usd' column

5. Example pattern for backward compatibility:
")

cat('
# Use standardized revenue if available, fallback to manual conversion
revenue_millions <- if ("revenue" %in% names(data)) {
  data$revenue / 1e6
} else {
  data$revenue_absolute / 100 / 1e6  # Convert cents to millions
}
')

cat("\n\nFor more information, see: inst/docs/revenue_units_guide.md\n")

#!/usr/bin/env Rscript

# Monitor rhub GitHub Actions workflow

monitor_rhub_run <- function(run_id = NULL) {
  if (is.null(run_id)) {
    # Get latest run
    runs <- system2("gh", 
                    args = c("run", "list", "--workflow=rhub.yaml", 
                             "--repo", "econosopher/SensorTowerR", 
                             "--limit", "1", "--json", "databaseId,status"),
                    stdout = TRUE)
    run_info <- jsonlite::fromJSON(runs)
    if (nrow(run_info) == 0) {
      stop("No rhub runs found")
    }
    run_id <- run_info$databaseId[1]
  }
  
  cat("Monitoring rhub run:", run_id, "\n")
  cat("View on GitHub: https://github.com/econosopher/SensorTowerR/actions/runs/", run_id, "\n\n", sep = "")
  
  # Monitor until complete
  while (TRUE) {
    status <- system2("gh", 
                      args = c("run", "view", run_id, 
                               "--repo", "econosopher/SensorTowerR", 
                               "--json", "status,conclusion,jobs"),
                      stdout = TRUE)
    
    run_data <- jsonlite::fromJSON(status)
    
    # Display job status
    cat("\033[2J\033[H") # Clear screen
    cat("=== R-hub Cross-Platform Check Status ===\n")
    cat("Run ID:", run_id, "\n")
    cat("Status:", run_data$status, "\n")
    if (!is.null(run_data$conclusion)) {
      cat("Conclusion:", run_data$conclusion, "\n")
    }
    cat("\nJobs:\n")
    
    jobs <- run_data$jobs
    for (i in seq_len(nrow(jobs))) {
      job <- jobs[i, ]
      icon <- if (job$status == "completed") {
        if (job$conclusion == "success") "✓" else "✗"
      } else if (job$status == "in_progress") {
        "⋯"
      } else {
        "○"
      }
      cat(sprintf("  %s %s (%s)\n", icon, job$name, job$status))
    }
    
    # Check if all done
    if (run_data$status == "completed") {
      cat("\nWorkflow completed with conclusion:", run_data$conclusion, "\n")
      
      # Show summary
      if (run_data$conclusion == "success") {
        cat("\n✅ All platform checks passed!\n")
      } else {
        cat("\n❌ Some checks failed. Review the logs at:\n")
        cat("   https://github.com/econosopher/SensorTowerR/actions/runs/", run_id, "\n", sep = "")
      }
      break
    }
    
    Sys.sleep(10) # Check every 10 seconds
  }
}

# If run directly
if (!interactive()) {
  args <- commandArgs(trailingOnly = TRUE)
  run_id <- if (length(args) > 0) args[1] else NULL
  monitor_rhub_run(run_id)
}

#' Diagnose API Issues
#'
#' This function helps diagnose common API issues by testing various ID formats
#' and endpoints to determine the best approach for fetching data.
#'
#' @param app_id Character string. The app ID to diagnose (can be unified, iOS, or Android)
#' @param verbose Logical. Show detailed diagnostic output. Default is TRUE.
#' @param auth_token Character string. Your Sensor Tower API authentication token.
#'
#' @return A list with diagnostic results including:
#'   - `id_type`: Detected ID type
#'   - `platform_ids`: Resolved platform-specific IDs
#'   - `endpoint_results`: Results from testing various endpoints
#'   - `recommendations`: Suggested approach for this app
#'
#' @examples
#' \dontrun{
#' # Diagnose Star Trek Fleet Command
#' diagnosis <- st_api_diagnostics("5ba4585f539ce75b97db6bcb")
#' 
#' # Check iOS app
#' diagnosis <- st_api_diagnostics("1427744264")
#' }
#' 
#' @export
st_api_diagnostics <- function(app_id, 
                              verbose = TRUE,
                              auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN")) {
  
  auth_token <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token is required. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )
  
  results <- list(
    id_type = "unknown",
    platform_ids = list(ios = NULL, android = NULL),
    endpoint_results = list(),
    recommendations = character()
  )
  
  # Step 1: Detect ID type
  if (verbose) message("=== API DIAGNOSTICS FOR: ", app_id, " ===\n")
  
  if (grepl("^[a-f0-9]{24}$", app_id)) {
    results$id_type <- "sensor_tower_hex"
    if (verbose) message("Detected: Sensor Tower hex ID")
  } else if (grepl("^\\d+$", app_id)) {
    results$id_type <- "ios_numeric"
    if (verbose) message("Detected: iOS numeric ID")
    results$platform_ids$ios <- app_id
  } else if (grepl("^(com|net|org|io)\\.", app_id)) {
    results$id_type <- "android_package"
    if (verbose) message("Detected: Android package name")
    results$platform_ids$android <- app_id
  }
  
  # Step 2: Try to resolve platform IDs if needed
  if (results$id_type == "sensor_tower_hex") {
    if (verbose) message("\nResolving platform IDs...")
    
    lookup_result <- tryCatch({
      st_app_lookup_impl(app_id, verbose = FALSE, auth_token = auth_token)
    }, error = function(e) NULL)
    
    if (!is.null(lookup_result)) {
      results$platform_ids$ios <- lookup_result$ios_app_id
      results$platform_ids$android <- lookup_result$android_app_id
      if (verbose) {
        message("  iOS ID: ", results$platform_ids$ios %||% "not found")
        message("  Android ID: ", results$platform_ids$android %||% "not found")
      }
    } else {
      if (verbose) message("  Could not resolve platform IDs")
    }
  }
  
  # Step 3: Test various endpoints
  if (verbose) message("\nTesting endpoints...")
  
  # Test unified endpoint
  test_endpoint <- function(endpoint_name, test_func) {
    if (verbose) message("  Testing ", endpoint_name, "... ", appendLF = FALSE)
    
    result <- tryCatch({
      data <- test_func()
      if (is.null(data) || nrow(data) == 0) {
        if (verbose) message("EMPTY")
        list(status = "empty", rows = 0)
      } else {
        if (verbose) message("SUCCESS (", nrow(data), " rows)")
        list(status = "success", rows = nrow(data))
      }
    }, error = function(e) {
      if (verbose) message("ERROR: ", e$message)
      list(status = "error", message = e$message)
    })
    
    return(result)
  }
  
  # Test sales report with detected OS
  os_type <- if (results$id_type == "ios_numeric") "ios" 
             else if (results$id_type == "android_package") "android" 
             else "ios"  # default
  
  # Determine the appropriate parameter based on os_type
  results$endpoint_results$sales_report <- test_endpoint(
    paste0(toupper(os_type), " Sales Report"),
    function() {
      if (os_type == "ios") {
        st_sales_report_impl(
          ios_app_id = app_id,
          os = os_type,
          date_granularity = "monthly",
          start_date = format(Sys.Date() - 90, "%Y-%m-%d"),
          end_date = format(Sys.Date() - 1, "%Y-%m-%d"),
          countries = "US",
          auth_token = auth_token
        )
      } else {
        st_sales_report_impl(
          android_app_id = app_id,
          os = os_type,
          date_granularity = "monthly",
          start_date = format(Sys.Date() - 90, "%Y-%m-%d"),
          end_date = format(Sys.Date() - 1, "%Y-%m-%d"),
          countries = "US",
          auth_token = auth_token
        )
      }
    }
  )

  # Test platform-specific if we have IDs
  if (!is.null(results$platform_ids$ios)) {
    results$endpoint_results$ios_sales <- test_endpoint(
      "iOS Sales Report",
      function() {
        st_sales_report_impl(
          ios_app_id = results$platform_ids$ios,
          os = "ios",
          date_granularity = "monthly",
          start_date = format(Sys.Date() - 90, "%Y-%m-%d"),
          end_date = format(Sys.Date() - 1, "%Y-%m-%d"),
          countries = "US",
          auth_token = auth_token
        )
      }
    )
  }

  if (!is.null(results$platform_ids$android)) {
    results$endpoint_results$android_sales <- test_endpoint(
      "Android Sales Report",
      function() {
        st_sales_report_impl(
          android_app_id = results$platform_ids$android,
          os = "android",
          date_granularity = "monthly",
          start_date = format(Sys.Date() - 90, "%Y-%m-%d"),
          end_date = format(Sys.Date() - 1, "%Y-%m-%d"),
          countries = "US",
          auth_token = auth_token
        )
      }
    )
  }

  # Test app details
  if (results$id_type %in% c("ios_numeric", "android_package")) {
    os <- ifelse(results$id_type == "ios_numeric", "ios", "android")
    results$endpoint_results$app_details <- test_endpoint(
      paste0(toupper(os), " App Details"),
      function() {
        st_app_details_impl(
          app_ids = app_id,
          os = os,
          auth_token = auth_token
        )
      }
    )
  }
  
  # Step 4: Generate recommendations
  if (verbose) message("\nGenerating recommendations...")
  
  if (results$id_type == "sensor_tower_hex") {
    if (!is.null(results$platform_ids$ios) || !is.null(results$platform_ids$android)) {
      results$recommendations <- c(
        "Use platform-specific IDs for best results",
        sprintf("iOS: %s", results$platform_ids$ios %||% "not available"),
        sprintf("Android: %s", results$platform_ids$android %||% "not available"),
        "Unified endpoints may return empty data for this app"
      )
    } else {
      results$recommendations <- c(
        "Could not resolve platform IDs",
        "Try searching by app name using st_app_info_impl()",
        "Unified endpoints unlikely to work with this ID"
      )
    }
  } else {
    results$recommendations <- c(
      sprintf("This is a %s ID", gsub("_", " ", results$id_type)),
      "Use directly with platform-specific endpoints",
      "For combined iOS+Android data, find the corresponding ID for the other platform"
    )
  }
  
  if (verbose) {
    message("\n=== RECOMMENDATIONS ===")
    for (rec in results$recommendations) {
      message("- ", rec)
    }
  }
  
  return(results)
}

# Helper operator
`%||%` <- function(x, y) if (is.null(x)) y else x

#' Smart Metrics Fetching with Automatic ID Resolution
#'
#' Enhanced metrics fetching that automatically handles ID resolution,
#' caching, and fallbacks to minimize API calls.
#'
#' @param app_ids Character vector. Can be any mix of iOS IDs, Android IDs,
#'   or Sensor Tower unified IDs.
#' @param metrics Character vector. Metrics to fetch (e.g., "revenue", "downloads", "dau").
#' @param start_date Date or character string. Start date for metrics.
#' @param end_date Date or character string. End date for metrics.
#' @param countries Character vector. Country codes (default "WW").
#' @param granularity Character. Date granularity ("daily", "weekly", "monthly").
#' @param auto_resolve Logical. Automatically resolve IDs using cache/API (default TRUE).
#' @param use_cache Logical. Use ID cache to minimize lookups (default TRUE).
#' @param parallel Logical. Use parallel processing (default TRUE).
#' @param auth_token Character string. Your Sensor Tower API authentication token.
#' @param verbose Logical. Print progress messages.
#'
#' @return A tibble with metrics in long format
#'
#' @examples
#' \dontrun{
#' # Mixed ID types - automatically resolved
#' metrics <- st_smart_metrics(
#'   app_ids = c(
#'     "553834731", # Candy Crush iOS
#'     "com.king.candycrushsaga", # Candy Crush Android
#'     "5ba4585f539ce75b97db6bcb" # Star Trek unified ID
#'   ),
#'   metrics = c("revenue", "downloads", "dau"),
#'   start_date = "2024-01-01",
#'   end_date = "2024-12-31"
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_smart_metrics_impl <- function(
  app_ids,
  metrics = c("revenue", "downloads"),
  start_date = Sys.Date() - 30,
  end_date = Sys.Date() - 1,
  countries = "WW",
  granularity = "daily",
  auto_resolve = TRUE,
  use_cache = TRUE,
  parallel = TRUE,
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
  verbose = TRUE
) {
  # Load cache if needed
  if (use_cache && !exists("id_cache", envir = .sensortowerR_env)) {
    load_id_cache()
  }

  # Step 1: Analyze and categorize IDs
  if (verbose) message("Analyzing ", length(app_ids), " app IDs...")

  id_analysis <- tibble::tibble(
    input_id = app_ids,
    id_type = sapply(app_ids, detect_id_type),

  )

  if (verbose) {
    type_counts <- table(id_analysis$id_type)
    message(
      "  ID types detected: ",
      paste(names(type_counts), type_counts, sep = ":", collapse = ", ")
    )
  }

  # Step 2: Resolve IDs if needed
  if (auto_resolve) {
    # Get all unique IDs that need resolution
    needs_resolution <- unique(app_ids)

    # Batch resolve with caching
    resolved <- batch_resolve_ids(needs_resolution, auth_token, use_cache, verbose)

    # Create enhanced app list with all available IDs
    app_list <- lapply(app_ids, function(id) {
      res <- resolved[[id]]
      if (is.null(res)) {
        # Fallback to basic type detection
        id_type <- detect_id_type(id)
        list(
          input_id = id,
          ios_id = if (id_type == "ios") id else NULL,
          android_id = if (id_type == "android") id else NULL,
          unified_id = if (id_type == "unified_hex") id else NULL,
          app_name = NULL
        )
      } else {
        list(
          input_id = id,
          ios_id = res$ios_id,
          android_id = res$android_id,
          unified_id = res$unified_id,
          app_name = res$app_name
        )
      }
    })
  } else {
    # Use IDs as-is based on type detection
    app_list <- lapply(seq_along(app_ids), function(i) {
      id <- app_ids[i]
      id_type <- id_analysis$id_type[i]

      list(
        input_id = id,
        ios_id = if (id_type == "ios") id else NULL,
        android_id = if (id_type == "android") id else NULL,
        unified_id = if (id_type == "unified_hex") id else NULL,
        app_name = NULL
      )
    })
  }

  # Convert resolved app metadata into a data frame for grouping
  to_safe_char <- function(x) {
    if (is.null(x)) {
      return(NA_character_)
    }
    val <- as.character(x)[1]
    if (is.na(val) || !nzchar(val)) {
      return(NA_character_)
    }
    val
  }
  resolved_apps <- dplyr::bind_rows(lapply(app_list, function(app) {
    tibble::tibble(
      input_id = to_safe_char(app$input_id),
      ios_id = to_safe_char(app$ios_id),
      android_id = to_safe_char(app$android_id),
      unified_id = to_safe_char(app$unified_id),
      app_name = to_safe_char(app$app_name),

    )
  }))

  valid_ios <- !is.na(resolved_apps$ios_id) & nzchar(resolved_apps$ios_id)
  valid_android <- !is.na(resolved_apps$android_id) & nzchar(resolved_apps$android_id)
  valid_unified <- !is.na(resolved_apps$unified_id) & nzchar(resolved_apps$unified_id)

  fetch_groups <- list()
  fetch_groups$both <- resolved_apps[valid_ios & valid_android, , drop = FALSE]
  fetch_groups$ios <- resolved_apps[valid_ios & !valid_android, , drop = FALSE]
  fetch_groups$android <- resolved_apps[!valid_ios & valid_android, , drop = FALSE]
  fetch_groups$unified <- resolved_apps[!valid_ios & !valid_android & valid_unified, , drop = FALSE]

  # Remove empty groups
  fetch_groups <- fetch_groups[vapply(fetch_groups, nrow, integer(1)) > 0]
  unresolved <- resolved_apps[!valid_ios & !valid_android & !valid_unified, , drop = FALSE]

  if (verbose) {
    message("\nOptimized fetch groups:")
    if (length(fetch_groups) == 0) {
      message("  None - no resolvable IDs")
    } else {
      for (group in names(fetch_groups)) {
        message("  ", group, ": ", nrow(fetch_groups[[group]]), " apps")
      }
    }
    if (nrow(unresolved) > 0) {
      message("  Skipped ", nrow(unresolved), " app(s) with unresolved identifiers")
    }
  }

  # Step 4: Fetch metrics using st_batch_metrics
  all_results <- list()

  for (group_name in names(fetch_groups)) {
    group_apps <- fetch_groups[[group_name]]
    if (nrow(group_apps) == 0) next
    if (verbose) message("\nFetching ", group_name, " group...")

    app_list_df <- tibble::tibble(
      app_id = group_apps$input_id,
      ios_id = group_apps$ios_id,
      android_id = group_apps$android_id,
      unified_id = group_apps$unified_id,
      app_name = group_apps$app_name,

    )

    result <- switch(group_name,
      both = st_batch_metrics_impl(
        os = "unified",
        app_list = app_list_df,
        metrics = metrics,
        date_range = list(start_date = start_date, end_date = end_date),
        countries = countries,
        granularity = granularity,
        parallel = parallel,
        auth_token = auth_token,
        verbose = verbose
      ),
      ios = st_batch_metrics_impl(
        os = "ios",
        app_list = app_list_df,
        metrics = metrics,
        date_range = list(start_date = start_date, end_date = end_date),
        countries = countries,
        granularity = granularity,
        parallel = parallel,
        auth_token = auth_token,
        verbose = verbose
      ),
      android = st_batch_metrics_impl(
        os = "android",
        app_list = app_list_df,
        metrics = metrics,
        date_range = list(start_date = start_date, end_date = end_date),
        countries = countries,
        granularity = granularity,
        parallel = parallel,
        auth_token = auth_token,
        verbose = verbose
      ),
      unified = st_batch_metrics_impl(
        os = "unified",
        app_list = app_list_df,
        metrics = metrics,
        date_range = list(start_date = start_date, end_date = end_date),
        countries = countries,
        granularity = granularity,
        parallel = parallel,
        auth_token = auth_token,
        verbose = verbose
      )
    )

    if (!is.null(result) && nrow(result) > 0) {
      result$group <- group_name
      all_results[[group_name]] <- result
    }
  }

  final_results <- if (length(all_results) > 0) {
    dplyr::bind_rows(all_results)
  } else {
    tibble::tibble()
  }

  if (nrow(final_results) > 0) {
    resolved_apps$input_position <- seq_len(nrow(resolved_apps))
    ordering <- resolved_apps[, c("input_id", "input_position"), drop = FALSE]
    names(ordering)[1] <- "original_id"
    final_results <- final_results %>%
      dplyr::left_join(ordering, by = "original_id")
    if ("date" %in% names(final_results)) {
      final_results <- final_results %>%
        dplyr::arrange(.data$input_position, .data$original_id, .data$date)
    } else {
      final_results <- final_results %>%
        dplyr::arrange(.data$input_position, .data$original_id)
    }
    final_results <- final_results %>% dplyr::select(-"input_position")
  }

  # Save cache if it's grown
  if (use_cache) {
    save_id_cache()
  }

  if (verbose) {
    message("\nFetch complete:")
    message("  Total records: ", nrow(final_results))
    unique_apps <- if ("app_id" %in% names(final_results)) length(unique(final_results$app_id)) else 0
    message("  Unique apps: ", unique_apps)
    if ("metric" %in% names(final_results) && nrow(final_results) > 0) {
      message("  Metrics: ", paste(unique(final_results$metric), collapse = ", "))
    }
  }

  return(final_results)
}

#' Clear App ID Cache
#'
#' Clears the in-memory and on-disk cache of app ID mappings
#'
#' @param disk Logical. Also remove the on-disk cache file (default TRUE).
#' @return No return value, called for side effects (clearing the cache).
#'
#' @export
st_clear_id_cache <- function(disk = TRUE) {
  # Clear in-memory cache
  .sensortowerR_env$id_cache <- list()

  # Clear disk cache
  if (disk) {
    cache_file <- file.path(get_cache_dir(), "id_cache.rds")
    if (file.exists(cache_file)) {
      unlink(cache_file)
      message("Cleared app ID cache")
    }
  }

  invisible(TRUE)
}

#' Show App ID Cache Statistics
#'
#' Display information about the current app ID cache
#'
#' @return No return value, called for side effects (displaying cache statistics).
#' @export
st_cache_info <- function() {
  cache <- get_id_cache()

  if (length(cache) == 0) {
    message("App ID cache is empty")
    return(invisible(NULL))
  }

  # Analyze cache
  cache_ages <- sapply(cache, function(x) {
    if (!is.null(x$cached_at)) {
      as.numeric(difftime(Sys.time(), x$cached_at, units = "days"))
    } else {
      NA
    }
  })

  has_ios <- sum(sapply(cache, function(x) !is.null(x$ios_id)))
  has_android <- sum(sapply(cache, function(x) !is.null(x$android_id)))
  has_both <- sum(sapply(cache, function(x) !is.null(x$ios_id) && !is.null(x$android_id)))

  cat("App ID Cache Statistics:\n")
  cat("  Total entries: ", length(cache), "\n")
  cat("  Apps with iOS ID: ", has_ios, "\n")
  cat("  Apps with Android ID: ", has_android, "\n")
  cat("  Apps with both platforms: ", has_both, "\n")
  cat("  Average cache age: ", round(mean(cache_ages, na.rm = TRUE), 1), " days\n")
  cat("  Oldest entry: ", round(max(cache_ages, na.rm = TRUE), 1), " days\n")
  cat("  Cache location: ", file.path(get_cache_dir(), "id_cache.rds"), "\n")

  invisible(cache)
}

#' Publisher Portfolio Analysis
#'
#' Fetches comprehensive portfolio data for a publisher including revenue,
#' downloads, MAU, and rankings. Returns a tidy data frame ready for
#' visualization or GT table creation.
#'
#' @param publisher Character. Publisher name to search for (e.g., "Lilith Games",
#'   "Supercell", "King"). The function will search for the publisher and use
#'   the first match.
#' @param publisher_id Character. Optional. If provided, skips the publisher
#'   search and uses this unified_publisher_id directly.
#' @param start_date Date or character. Start date for metrics (default: "2023-01-01").
#' @param end_date Date or character. End date for metrics (default: last day of
#'   previous month).
#' @param countries Character. Countries for metrics (default: "WW" for worldwide).
#' @param metrics Character vector. Which metrics to fetch. Options: "revenue",
#'   "downloads", "mau". Default: all three.
#' @param include_rankings Logical. Whether to fetch subgenre rankings from top
#'   charts. Default: TRUE.
#' @param include_portfolio_total Logical. Whether to add a portfolio total row.
#'   Default: TRUE.
#' @param granularity Character. How to aggregate the data: "yearly" (default),
#'   "quarterly", or "monthly".
#' @param min_revenue Numeric. Minimum revenue threshold to include an app.
#'   Default: 100000 (apps with at least $100K in any year).
#' @param auth_token Character. Sensor Tower API token. Defaults to
#'   SENSORTOWER_AUTH_TOKEN environment variable
#' @param verbose Logical. Print progress messages. Default: TRUE.
#' @param use_cache Logical. Use cached data if available. Default: FALSE.
#'   When TRUE, requires cache_dir to be specified.
#' @param cache_dir Character. Directory for cached data. Default: NULL (no caching).
#'   Must be explicitly set to enable caching. Use tempdir() for temporary caching.
#'
#' @return A tibble with portfolio data including:
#'   - app_name: Game name
#'   - subgenre: Game sub-genre
#'   - subgenre_rank: Rank within sub-genre
#'   - revenue_\{year\}: Revenue by year
#'   - downloads_\{year\}: Downloads by year
#'   - mau_\{year\}: Average MAU by year (if requested)
#'   - revenue_yoy, downloads_yoy, mau_yoy: Year-over-year growth percentages
#'
#' @examples
#' \dontrun{
#' # Simple usage - just provide publisher name
#' lilith_portfolio <- st_publisher_portfolio("Lilith Games")
#'
#' # Piped workflow
#' library(dplyr)
#'
#' "Supercell" %>%
#'   st_publisher_portfolio(
#'     start_date = "2023-01-01",
#'     metrics = c("revenue", "downloads")
#'   ) %>%
#'   filter(revenue_2024 > 1000000) %>%
#'   arrange(desc(revenue_2024))
#'
#' # Custom date range and countries
#' portfolio <- st_publisher_portfolio(
#'   publisher = "King",
#'   start_date = "2022-01-01",
#'   end_date = "2024-12-31",
#'   countries = c("US", "GB", "DE"),
#'   metrics = c("revenue", "downloads", "mau"),
#'   include_rankings = TRUE
#' )
#' }
#'
#' @import dplyr
#' @import tidyr
#' @importFrom lubridate year month floor_date days
#' @importFrom glue glue
#' @importFrom utils tail
#' @export
st_publisher_portfolio <- function(publisher = NULL,
                                    publisher_id = NULL,
                                    start_date = "2023-01-01",
                                    end_date = NULL,
                                    countries = "WW",
                                    metrics = c("revenue", "downloads", "mau"),
                                    include_rankings = TRUE,
                                    include_portfolio_total = TRUE,
                                    granularity = "yearly",
                                    min_revenue = 100000,
                                    auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                                    verbose = TRUE,
                                    use_cache = FALSE,
                                    cache_dir = NULL) {


  # --- Input Validation ---
  if (is.null(publisher) && is.null(publisher_id)) {
    rlang::abort("Either 'publisher' (name) or 'publisher_id' must be provided")
  }

  # Validate cache settings - cache_dir must be provided if caching is enabled

  if (use_cache && is.null(cache_dir)) {
    rlang::abort(c(
      "cache_dir must be specified when use_cache = TRUE",
      "i" = "Use cache_dir = tempdir() for temporary caching",
      "i" = "Or set use_cache = FALSE to disable caching"
    ))
  }

  metrics <- tolower(metrics)
  valid_metrics <- c("revenue", "downloads", "mau")
  if (!all(metrics %in% valid_metrics)) {
    rlang::abort(glue::glue(
      "Invalid metrics: {paste(setdiff(metrics, valid_metrics), collapse = ', ')}. ",
      "Valid options: {paste(valid_metrics, collapse = ', ')}"
    ))
  }

  granularity <- match.arg(granularity, c("yearly", "quarterly", "monthly"))

  # Default end_date to last complete month
  if (is.null(end_date)) {
    end_date <- lubridate::floor_date(Sys.Date(), "month") - lubridate::days(1)
  }
  start_date <- as.Date(start_date)
  end_date <- as.Date(end_date)

  # Create cache directory if needed (only when user explicitly provides cache_dir)
  if (use_cache && !is.null(cache_dir) && !dir.exists(cache_dir)) {
    dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
  }

  # --- Step 1: Find Publisher ---
  if (is.null(publisher_id)) {
    if (verbose) message("Step 1: Finding publisher '", publisher, "'...")

    publisher_search <- st_app_info_impl(
      term = publisher,
      entity_type = "publisher",
      auth_token = auth_token
    )

    if (is.null(publisher_search) || nrow(publisher_search) == 0) {
      rlang::abort(glue::glue("Publisher '{publisher}' not found"))
    }

    publisher_id <- publisher_search$unified_publisher_id[1]
    publisher_name <- publisher_search$unified_publisher_name[1]

    if (verbose) message("  Found: ", publisher_name, " (", publisher_id, ")")
  } else {
    publisher_name <- publisher %||% "Unknown Publisher"
  }

  # --- Step 2: Get Publisher Apps ---
  if (verbose) message("\nStep 2: Fetching publisher apps...")

  cache_file_apps <- file.path(cache_dir, paste0(publisher_id, "_apps.rds"))

  if (use_cache && file.exists(cache_file_apps)) {
    if (verbose) message("  Loading cached apps...")
    apps <- readRDS(cache_file_apps)
  } else {
    apps <- st_publisher_apps(
      unified_id = publisher_id,
      aggregate_related = TRUE,
      auth_token = auth_token,
      verbose = verbose
    )

    if (use_cache && nrow(apps) > 0) {
      saveRDS(apps, cache_file_apps)
    }
  }

  if (nrow(apps) == 0) {
    rlang::warn("No apps found for publisher")
    return(tibble::tibble())
  }

  if (verbose) message("  Found ", nrow(apps), " apps")

  app_ids <- apps$unified_app_id

  # --- Step 3: Fetch Sales Data (Revenue & Downloads) ---
  sales_data <- NULL
  if (any(c("revenue", "downloads") %in% metrics)) {
    if (verbose) message("\nStep 3: Fetching sales data...")

    cache_file_sales <- file.path(
      cache_dir,
      paste0(publisher_id, "_sales_", format(start_date, "%Y%m%d"), "_",
             format(end_date, "%Y%m%d"), ".rds")
    )

    if (use_cache && file.exists(cache_file_sales)) {
      if (verbose) message("  Loading cached sales data...")
      sales_data <- readRDS(cache_file_sales)
    } else {
      # Fetch in batches
      batch_size <- 10
      n_batches <- ceiling(length(app_ids) / batch_size)
      sales_list <- list()

      for (i in seq_len(n_batches)) {
        start_idx <- (i - 1) * batch_size + 1
        end_idx <- min(i * batch_size, length(app_ids))
        batch_ids <- app_ids[start_idx:end_idx]

        if (verbose) message(glue::glue("  Batch {i}/{n_batches}: apps {start_idx}-{end_idx}"))

        result <- tryCatch({
          st_unified_sales_report_impl(
            unified_app_id = batch_ids,
            countries = countries,
            start_date = start_date,
            end_date = end_date,
            date_granularity = "monthly",
            auth_token = auth_token,
            verbose = FALSE
          )
        }, error = function(e) {
          if (verbose) message("    Warning: ", e$message)
          NULL
        })

        if (!is.null(result) && nrow(result) > 0) {
          sales_list[[i]] <- result
        }

        Sys.sleep(0.3)  # Rate limiting
      }

      if (length(sales_list) > 0) {
        sales_data <- dplyr::bind_rows(sales_list)
        if (use_cache) saveRDS(sales_data, cache_file_sales)
      }
    }

    if (!is.null(sales_data)) {
      if (verbose) message("  Sales data: ", nrow(sales_data), " records")
    }
  }

  # --- Step 4: Fetch MAU Data ---
  mau_data <- NULL
  if ("mau" %in% metrics) {
    if (verbose) message("\nStep 4: Fetching MAU data...")

    cache_file_mau <- file.path(
      cache_dir,
      paste0(publisher_id, "_mau_", format(start_date, "%Y%m%d"), "_",
             format(end_date, "%Y%m%d"), ".rds")
    )

    if (use_cache && file.exists(cache_file_mau)) {
      if (verbose) message("  Loading cached MAU data...")
      mau_data <- readRDS(cache_file_mau)
    } else {
      mau_data <- tryCatch({
        # Use parallel=FALSE to avoid race conditions when apps are in
        # different platform groups (both, ios, android)
        st_batch_metrics_impl(
          os = "unified",
          app_list = app_ids,
          metrics = "mau",
          date_range = list(start_date = start_date, end_date = end_date),
          countries = countries,
          granularity = "monthly",
          auth_token = auth_token,
          parallel = FALSE
        )
      }, error = function(e) {
        if (verbose) message("  Warning: Could not fetch MAU: ", e$message)
        NULL
      })

      if (!is.null(mau_data) && nrow(mau_data) > 0 && use_cache) {
        saveRDS(mau_data, cache_file_mau)
      }
    }

    if (!is.null(mau_data)) {
      if (verbose) message("  MAU data: ", nrow(mau_data), " records")
    }
  }

  # --- Step 5: Fetch Rankings ---
  rank_data <- NULL
  if (include_rankings) {
    if (verbose) message("\nStep 5: Fetching rankings data...")

    cache_file_ranks <- file.path(cache_dir, "top_charts_rankings.rds")

    if (use_cache && file.exists(cache_file_ranks)) {
      file_age <- difftime(Sys.time(), file.info(cache_file_ranks)$mtime, units = "days")
      if (file_age < 1) {
        if (verbose) message("  Loading cached rankings...")
        rank_data <- readRDS(cache_file_ranks)
      }
    }

    if (is.null(rank_data)) {
      rank_data <- tryCatch({
        top_games <- st_top_charts_impl(
          measure = "revenue",
          os = "unified",
          category = "6014",  # Games
          regions = "WW",
          time_range = "month",
          date = format(lubridate::floor_date(Sys.Date(), "month") - lubridate::days(1), "%Y-%m-%d"),
          limit = 1500,
          auth_token = auth_token
        )

        top_games %>%
          dplyr::mutate(global_rank = dplyr::row_number()) %>%
          dplyr::group_by(`aggregate_tags.Game Sub-genre`) %>%
          dplyr::mutate(subgenre_rank = dplyr::row_number()) %>%
          dplyr::ungroup() %>%
          dplyr::select(
            unified_app_id, unified_app_name, global_rank, subgenre_rank,
            subgenre = `aggregate_tags.Game Sub-genre`
          )
      }, error = function(e) {
        if (verbose) message("  Warning: Could not fetch rankings: ", e$message)
        NULL
      })

      if (!is.null(rank_data) && use_cache) {
        saveRDS(rank_data, cache_file_ranks)
      }
    }

    if (!is.null(rank_data)) {
      if (verbose) message("  Rankings data: ", nrow(rank_data), " apps")
    }
  }

  # --- Step 6: Build Portfolio Summary ---
  if (verbose) message("\nStep 6: Building portfolio summary...")

  # Get app name mapping
  app_names <- apps %>%
    dplyr::select(dplyr::any_of(c("unified_app_id", "unified_app_name"))) %>%
    dplyr::distinct()

  # If rank_data has more complete names, use those
  if (!is.null(rank_data)) {
    app_names <- dplyr::bind_rows(
      app_names,
      rank_data %>% dplyr::select(unified_app_id, unified_app_name)
    ) %>%
      dplyr::distinct(unified_app_id, .keep_all = TRUE)
  }

  # Determine current year and month for YTD filtering
  current_year <- lubridate::year(Sys.Date())
  current_month <- lubridate::month(Sys.Date()) - 1  # Last complete month
  if (current_month == 0) {
    current_month <- 12
    current_year <- current_year - 1
  }

  # Process sales data
  if (!is.null(sales_data) && nrow(sales_data) > 0) {
    sales_summary <- sales_data %>%
      dplyr::left_join(app_names, by = "unified_app_id") %>%
      dplyr::mutate(
        Year = lubridate::year(date),
        Month = lubridate::month(date)
      ) %>%
      # YTD filter: include full years for past years, YTD for current year
      dplyr::filter(Month <= current_month | Year < current_year) %>%
      dplyr::group_by(unified_app_name, Year) %>%
      dplyr::summarise(
        revenue = sum(revenue, na.rm = TRUE),
        downloads = sum(downloads, na.rm = TRUE),
        .groups = "drop"
      )

    # Pivot to wide format
    portfolio_data <- sales_summary %>%
      tidyr::pivot_wider(
        names_from = Year,
        values_from = c(revenue, downloads),
        names_sep = "_",
        values_fill = 0
      )
  } else {
    portfolio_data <- app_names %>%
      dplyr::select(unified_app_name) %>%
      dplyr::distinct()
  }

  # Add MAU data
  if (!is.null(mau_data) && nrow(mau_data) > 0) {
    mau_summary <- mau_data %>%
      dplyr::left_join(app_names, by = c("original_id" = "unified_app_id")) %>%
      dplyr::mutate(
        Year = lubridate::year(date),
        Month = lubridate::month(date)
      ) %>%
      dplyr::filter(Month <= current_month | Year < current_year) %>%
      dplyr::group_by(unified_app_name, Year) %>%
      dplyr::summarise(
        mau = mean(value, na.rm = TRUE),
        .groups = "drop"
      )

    mau_wide <- mau_summary %>%
      tidyr::pivot_wider(
        names_from = Year,
        values_from = mau,
        names_prefix = "mau_",
        values_fill = 0
      )

    portfolio_data <- portfolio_data %>%
      dplyr::left_join(mau_wide, by = "unified_app_name")
  }

  # Add rankings
  if (!is.null(rank_data)) {
    portfolio_data <- portfolio_data %>%
      dplyr::left_join(
        rank_data %>%
          dplyr::select(unified_app_name, subgenre_rank, subgenre) %>%
          dplyr::distinct(unified_app_name, .keep_all = TRUE),
        by = "unified_app_name"
      )
  }

  # Filter by minimum revenue
  years_in_data <- grep("^revenue_\\d{4}$", names(portfolio_data), value = TRUE)
  if (length(years_in_data) > 0) {
    portfolio_data <- portfolio_data %>%
      dplyr::filter(
        dplyr::if_any(dplyr::all_of(years_in_data), ~ . >= min_revenue)
      ) %>%
      dplyr::arrange(dplyr::desc(!!rlang::sym(tail(years_in_data, 1))))
  }

  # Calculate YoY growth
  if (length(years_in_data) >= 2) {
    latest_year <- max(as.numeric(gsub("revenue_", "", years_in_data)))
    prev_year <- latest_year - 1

    rev_latest <- paste0("revenue_", latest_year)
    rev_prev <- paste0("revenue_", prev_year)
    dl_latest <- paste0("downloads_", latest_year)
    dl_prev <- paste0("downloads_", prev_year)
    mau_latest <- paste0("mau_", latest_year)
    mau_prev <- paste0("mau_", prev_year)

    portfolio_data <- portfolio_data %>%
      dplyr::mutate(
        revenue_yoy = dplyr::if_else(
          !!rlang::sym(rev_prev) > 0,
          round((!!rlang::sym(rev_latest) - !!rlang::sym(rev_prev)) /
                  !!rlang::sym(rev_prev) * 100, 0),
          NA_real_
        )
      )

    if (dl_prev %in% names(portfolio_data) && dl_latest %in% names(portfolio_data)) {
      portfolio_data <- portfolio_data %>%
        dplyr::mutate(
          downloads_yoy = dplyr::if_else(
            !!rlang::sym(dl_prev) > 0,
            round((!!rlang::sym(dl_latest) - !!rlang::sym(dl_prev)) /
                    !!rlang::sym(dl_prev) * 100, 0),
            NA_real_
          )
        )
    }

    if (mau_prev %in% names(portfolio_data) && mau_latest %in% names(portfolio_data)) {
      portfolio_data <- portfolio_data %>%
        dplyr::mutate(
          mau_yoy = dplyr::if_else(
            !!rlang::sym(mau_prev) > 0,
            round((!!rlang::sym(mau_latest) - !!rlang::sym(mau_prev)) /
                    !!rlang::sym(mau_prev) * 100, 0),
            NA_real_
          )
        )
    }
  }

  # Add rank column
  portfolio_data <- portfolio_data %>%
    dplyr::mutate(rank = dplyr::row_number())

  # Add portfolio total row
  if (include_portfolio_total && nrow(portfolio_data) > 0) {
    numeric_cols <- names(portfolio_data)[sapply(portfolio_data, is.numeric)]
    numeric_cols <- setdiff(numeric_cols, c("rank", "subgenre_rank"))

    total_row <- portfolio_data %>%
      dplyr::summarise(
        dplyr::across(dplyr::all_of(numeric_cols), ~ sum(., na.rm = TRUE))
      ) %>%
      dplyr::mutate(
        unified_app_name = "PORTFOLIO TOTAL",
        rank = NA_integer_,
        subgenre = NA_character_,
        subgenre_rank = NA_integer_
      )

    # Recalculate YoY for total
    if ("revenue_yoy" %in% names(total_row) && length(years_in_data) >= 2) {
      latest_year <- max(as.numeric(gsub("revenue_", "", years_in_data)))
      prev_year <- latest_year - 1
      rev_latest <- paste0("revenue_", latest_year)
      rev_prev <- paste0("revenue_", prev_year)

      total_row <- total_row %>%
        dplyr::mutate(
          revenue_yoy = round((!!rlang::sym(rev_latest) - !!rlang::sym(rev_prev)) /
                                !!rlang::sym(rev_prev) * 100, 0)
        )

      if ("downloads_yoy" %in% names(total_row)) {
        dl_latest <- paste0("downloads_", latest_year)
        dl_prev <- paste0("downloads_", prev_year)
        total_row <- total_row %>%
          dplyr::mutate(
            downloads_yoy = round((!!rlang::sym(dl_latest) - !!rlang::sym(dl_prev)) /
                                    !!rlang::sym(dl_prev) * 100, 0)
          )
      }

      if ("mau_yoy" %in% names(total_row)) {
        mau_latest <- paste0("mau_", latest_year)
        mau_prev <- paste0("mau_", prev_year)
        if (mau_latest %in% names(total_row) && mau_prev %in% names(total_row)) {
          total_row <- total_row %>%
            dplyr::mutate(
              mau_yoy = round((!!rlang::sym(mau_latest) - !!rlang::sym(mau_prev)) /
                                !!rlang::sym(mau_prev) * 100, 0)
            )
        }
      }
    }

    portfolio_data <- dplyr::bind_rows(total_row, portfolio_data)
  }

  # Rename columns for cleaner output
  portfolio_data <- portfolio_data %>%
    dplyr::rename(app_name = unified_app_name)

  if (verbose) message("\nPortfolio summary complete: ", nrow(portfolio_data), " apps")

  # Add metadata as attributes
  attr(portfolio_data, "publisher") <- publisher_name

  attr(portfolio_data, "publisher_id") <- publisher_id
  attr(portfolio_data, "date_range") <- list(start = start_date, end = end_date)
  attr(portfolio_data, "countries") <- countries

  return(portfolio_data)
}

#' Fetch Rating Metrics from Sensor Tower's Facets API
#'
#' Retrieves rating metrics from Sensor Tower's new facets-based ratings
#' endpoint. This wrapper targets `"/v1/facets/metrics?facets=ratings"` and
#' returns a tidy tibble with the live response columns preserved.
#'
#' @param app_ids Character vector of iOS app IDs or Android package names.
#'   Sensor Tower documents a maximum of 1,000 IDs per request.
#' @param bundle Character string. Rating bundle to request:
#'   `"ratings_incremental"` or `"ratings_cumulative"`.
#' @param breakdown Breakdown fields or a comma-separated breakdown string.
#'   Supported combinations are `"app_id"`, `"app_id,date"`, `"region"`,
#'   `"region,date"`, and `"app_version"`.
#' @param start_date Start date in `YYYY-MM-DD` format or as `Date`.
#' @param end_date End date in `YYYY-MM-DD` format or as `Date`.
#' @param date_granularity Optional date granularity. Required when
#'   `breakdown` includes `date`.
#' @param regions Optional character vector of region codes.
#' @param android_localized_estimates Logical. Whether to apply Android country
#'   weighting. Defaults to `TRUE` to match the current documented default.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'   Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.
#' @param verbose Logical. If `TRUE`, prints the request URL with the auth token
#'   redacted.
#'
#' @return A [tibble][tibble::tibble] containing rating metrics such as
#'   `rating_average_incremental`, `rating_count_incremental`, or their
#'   cumulative counterparts.
#'
#' @details
#' Validated against the live Sensor Tower API on March 17, 2026.
#'
#' @examples
#' \dontrun{
#' ratings <- st_ratings_facets(
#'   app_ids = "553834731",
#'   bundle = "ratings_incremental",
#'   breakdown = c("app_id", "date"),
#'   start_date = "2024-01-01",
#'   end_date = "2024-01-07",
#'   date_granularity = "day"
#' )
#' }
#'
#' @seealso [st_facets_metrics()] for raw facets access
#'
#' @export
st_ratings_facets <- function(app_ids,
                              bundle = c("ratings_incremental", "ratings_cumulative"),
                              breakdown = c("app_id", "date"),
                              start_date,
                              end_date,
                              date_granularity = NULL,
                              regions = NULL,
                              android_localized_estimates = TRUE,
                              auth_token = NULL,
                              verbose = FALSE) {
  if (missing(app_ids) || is.null(app_ids) || length(app_ids) == 0) {
    rlang::abort("`app_ids` must contain at least one app ID.")
  }

  if (length(app_ids) > 1000) {
    rlang::abort("`app_ids` supports a maximum of 1,000 values per request.")
  }

  bundle <- match.arg(bundle)
  dates <- validate_facets_date_inputs(start_date, end_date)
  breakdown <- normalize_ratings_facets_breakdown(breakdown)
  date_granularity <- normalize_optional_facets_date_granularity(
    date_granularity = date_granularity,
    breakdown = breakdown
  )

  response <- st_facets_metrics(
    params = list(
      facets = "ratings",
      bundle = bundle,
      breakdown = breakdown,
      date_granularity = date_granularity,
      app_ids = app_ids,
      regions = regions,
      start_date = dates$start_date,
      end_date = dates$end_date,
      android_localized_estimates = android_localized_estimates
    ),
    auth_token = auth_token,
    verbose = verbose
  )

  result <- coerce_facets_data_tibble(response)

  if (nrow(result) == 0) {
    return(result)
  }

  result <- normalize_facets_result_columns(
    result = result,
    character_cols = c("app_id", "region", "app_version"),
    integer_cols = c(
      "rating_count_incremental",
      "rating_1_star_count_incremental",
      "rating_2_star_count_incremental",
      "rating_3_star_count_incremental",
      "rating_4_star_count_incremental",
      "rating_5_star_count_incremental",
      "rating_count_cumulative",
      "rating_1_star_count_cumulative",
      "rating_2_star_count_cumulative",
      "rating_3_star_count_cumulative",
      "rating_4_star_count_cumulative",
      "rating_5_star_count_cumulative"
    ),
    numeric_cols = c(
      "rating_average_incremental",
      "rating_average_cumulative"
    )
  )

  sort_facets_result(result, c("app_id", "region", "app_version", "date"))
}

#' Fetch Review Metrics by Rating from Sensor Tower's Facets API
#'
#' Retrieves review metrics broken down by star rating from Sensor Tower's new
#' facets-based review endpoint. This wrapper targets
#' `"/v1/facets/metrics?facets=reviews_by_rating"`.
#'
#' @param app_id Single iOS app ID or Android package name.
#' @param breakdown Breakdown fields or a comma-separated breakdown string.
#'   Supported combinations are `"review_rating"`, `"date,review_rating"`,
#'   `"region,review_rating"`, `"language,review_rating"`, and
#'   `"app_version,review_rating"`.
#' @param start_date Start date in `YYYY-MM-DD` format or as `Date`.
#' @param end_date End date in `YYYY-MM-DD` format or as `Date`.
#' @param date_granularity Optional date granularity. Required when
#'   `breakdown` includes `date`.
#' @param regions Optional character vector of iOS region codes.
#' @param languages Optional character vector of Android language codes.
#' @param review_keywords Optional character vector of review-keyword filters.
#' @param review_sentiments Optional character vector of sentiment filters.
#' @param review_tags Optional character vector of review-tag filters.
#' @param search_terms Optional character vector of content search terms.
#' @param rating_filters Optional character vector or integer vector of star
#'   filters (`1` through `5`).
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'   Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.
#' @param verbose Logical. If `TRUE`, prints the request URL with the auth token
#'   redacted.
#'
#' @return A [tibble][tibble::tibble] containing fields such as
#'   `review_rating`, `review_rating_count`, `review_rating_percentage`, and
#'   `review_rating_average`.
#'
#' @details
#' Validated against the live Sensor Tower API on March 17, 2026.
#'
#' @examples
#' \dontrun{
#' reviews <- st_reviews_by_rating_facets(
#'   app_id = "553834731",
#'   breakdown = c("date", "review_rating"),
#'   start_date = "2024-01-01",
#'   end_date = "2024-01-07",
#'   date_granularity = "day",
#'   regions = "US"
#' )
#' }
#'
#' @seealso [st_facets_metrics()] for raw facets access
#'
#' @export
st_reviews_by_rating_facets <- function(app_id,
                                        breakdown = c("date", "review_rating"),
                                        start_date,
                                        end_date,
                                        date_granularity = NULL,
                                        regions = NULL,
                                        languages = NULL,
                                        review_keywords = NULL,
                                        review_sentiments = NULL,
                                        review_tags = NULL,
                                        search_terms = NULL,
                                        rating_filters = NULL,
                                        auth_token = NULL,
                                        verbose = FALSE) {
  if (missing(app_id) || is.null(app_id) || length(app_id) != 1 || !nzchar(trimws(as.character(app_id)))) {
    rlang::abort("`app_id` must be a single non-empty app ID.")
  }

  dates <- validate_facets_date_inputs(start_date, end_date)
  breakdown <- normalize_reviews_by_rating_facets_breakdown(breakdown)
  date_granularity <- normalize_optional_facets_date_granularity(
    date_granularity = date_granularity,
    breakdown = breakdown
  )

  response <- st_facets_metrics(
    params = list(
      facets = "reviews_by_rating",
      bundle = "reviews_by_rating",
      breakdown = breakdown,
      date_granularity = date_granularity,
      app_ids = app_id,
      regions = regions,
      languages = languages,
      start_date = dates$start_date,
      end_date = dates$end_date,
      review_keywords = review_keywords,
      review_sentiments = review_sentiments,
      review_tags = review_tags,
      search_terms = search_terms,
      rating_filters = rating_filters
    ),
    auth_token = auth_token,
    verbose = verbose
  )

  result <- coerce_facets_data_tibble(response)

  if (nrow(result) == 0) {
    return(result)
  }

  result <- normalize_facets_result_columns(
    result = result,
    character_cols = c("region", "language", "app_version"),
    integer_cols = c("review_rating", "review_rating_count"),
    numeric_cols = c("review_rating_percentage", "review_rating_average")
  )

  sort_facets_result(
    result,
    c("date", "region", "language", "app_version", "review_rating")
  )
}

validate_facets_date_inputs <- function(start_date, end_date) {
  start_date <- as.Date(start_date)
  end_date <- as.Date(end_date)

  if (is.na(start_date) || is.na(end_date)) {
    rlang::abort("`start_date` and `end_date` must be valid dates.")
  }

  if (end_date < start_date) {
    rlang::abort("`end_date` must be on or after `start_date`.")
  }

  list(start_date = start_date, end_date = end_date)
}

parse_facets_breakdown_input <- function(breakdown) {
  if (is.null(breakdown)) {
    return(character())
  }

  breakdown <- as.character(breakdown)

  if (length(breakdown) == 1 && grepl(",", breakdown, fixed = TRUE)) {
    breakdown <- strsplit(breakdown, ",", fixed = TRUE)[[1]]
  }

  breakdown <- trimws(tolower(breakdown))
  unique(breakdown[nzchar(breakdown)])
}

normalize_optional_facets_date_granularity <- function(date_granularity, breakdown) {
  has_date_breakdown <- "date" %in% breakdown

  if (has_date_breakdown && (is.null(date_granularity) || !nzchar(trimws(as.character(date_granularity[1]))))) {
    rlang::abort("`date_granularity` is required when `breakdown` includes `date`.")
  }

  if (is.null(date_granularity) || !nzchar(trimws(as.character(date_granularity[1])))) {
    return(NULL)
  }

  match.arg(as.character(date_granularity[1]), c("day", "week", "month"))
}

normalize_ratings_facets_breakdown <- function(breakdown) {
  breakdown <- parse_facets_breakdown_input(breakdown)

  if (identical(sort(breakdown), "app_id")) {
    return("app_id")
  }

  if (identical(sort(breakdown), c("app_id", "date"))) {
    return(c("app_id", "date"))
  }

  if (identical(sort(breakdown), "region")) {
    return("region")
  }

  if (identical(sort(breakdown), c("date", "region"))) {
    return(c("region", "date"))
  }

  if (identical(sort(breakdown), "app_version")) {
    return("app_version")
  }

  breakdown_key <- paste(breakdown, collapse = ",")
  rlang::abort(
    sprintf(
      "Unsupported `breakdown` for ratings facets: %s.",
      if (nzchar(breakdown_key)) breakdown_key else "<blank>"
    )
  )
}

normalize_reviews_by_rating_facets_breakdown <- function(breakdown) {
  breakdown <- parse_facets_breakdown_input(breakdown)

  if (identical(sort(breakdown), "review_rating")) {
    return("review_rating")
  }

  if (identical(sort(breakdown), c("date", "review_rating"))) {
    return(c("date", "review_rating"))
  }

  if (identical(sort(breakdown), c("region", "review_rating"))) {
    return(c("region", "review_rating"))
  }

  if (identical(sort(breakdown), c("language", "review_rating"))) {
    return(c("language", "review_rating"))
  }

  if (identical(sort(breakdown), c("app_version", "review_rating"))) {
    return(c("app_version", "review_rating"))
  }

  breakdown_key <- paste(breakdown, collapse = ",")
  rlang::abort(
    sprintf(
      "Unsupported `breakdown` for reviews-by-rating facets: %s.",
      if (nzchar(breakdown_key)) breakdown_key else "<blank>"
    )
  )
}

normalize_facets_result_columns <- function(result,
                                            character_cols = character(),
                                            integer_cols = character(),
                                            numeric_cols = character()) {
  if ("date" %in% names(result)) {
    result$date <- as.Date(result$date)
  }

  for (col in intersect(character_cols, names(result))) {
    result[[col]] <- as.character(result[[col]])
  }

  for (col in intersect(integer_cols, names(result))) {
    result[[col]] <- as.integer(result[[col]])
  }

  for (col in intersect(numeric_cols, names(result))) {
    result[[col]] <- as.numeric(result[[col]])
  }

  tibble::as_tibble(result)
}

sort_facets_result <- function(result, sort_cols) {
  sort_cols <- intersect(sort_cols, names(result))

  if (length(sort_cols) == 0) {
    return(tibble::as_tibble(result))
  }

  order_index <- do.call(order, c(result[sort_cols], list(na.last = TRUE)))
  tibble::as_tibble(result[order_index, , drop = FALSE])
}

#' Fetch Unified Sales Report Estimates (All Regional SKUs Aggregated)
#'
#' Retrieves download and revenue estimates using the unified API endpoint,
#' which properly aggregates ALL regional SKUs (app variants from different
#' publishers/regions) within a unified_app_id.
#'
#' This function solves the problem where apps like "Watcher of Realms" have
#' multiple regional versions (e.g., Moonton, Shanghai Moonton, Vizta Games,
#' Skystone Games publishers) that need to be combined for accurate totals.
#'
#' @param unified_app_id Character string or vector. Sensor Tower unified app ID(s)
#'   (24-character hex format, e.g., "67ec0bf3e540b65904256cc4").
#' @param countries Character vector. Country codes (e.g., c("US", "GB", "JP"),
#'   or "WW" for worldwide). Required.
#' @param start_date Date or character string. Start date in "YYYY-MM-DD" format. Required.
#' @param end_date Date or character string. End date in "YYYY-MM-DD" format. Required.
#' @param date_granularity Character string. One of "daily", "weekly", "monthly", "quarterly". Required.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param verbose Logical. If TRUE, prints progress messages.
#'
#' @return A tibble with columns:
#'   - `date`: Date of the data point
#'   - `country`: Country code
#'   - `unified_app_id`: The unified app ID
#'   - `revenue`: Revenue in dollars (converted from cents)
#'   - `downloads`: Number of downloads
#'
#' @details
#' **Why use this instead of st_sales_report()?**
#'
#' When a game has multiple regional SKUs (same game published under different
#' publishers or app IDs in different regions), the standard `st_sales_report()`
#' function with ID resolution only fetches data for the FIRST iOS and Android
#' app ID. This can result in significantly undercounted revenue/downloads.
#'
#' Example: "Watcher of Realms" has 4 iOS apps and 3 Android apps across
#' different publishers (Moonton, Vizta Games, Skystone Games, etc.). Using
#' `st_sales_report()` with the unified_app_id might only fetch data for
#' 2 of these 7 apps.
#'
#' This function uses the `/v1/unified/sales_report_estimates` endpoint which
#' automatically aggregates ALL app IDs within the unified entity.
#'
#' **API Response Fields:**
#' The unified API returns `unified_revenue` and `unified_units` which are
#' automatically converted to `revenue` (dollars) and `downloads`.
#'
#' @examples
#' \dontrun{
#' # Get unified sales data for Watcher of Realms
#' sales <- st_unified_sales_report(
#'   unified_app_id = "67ec0bf3e540b65904256cc4",
#'   countries = "WW",
#'   start_date = "2024-01-01",
#'   end_date = "2024-12-31",
#'   date_granularity = "monthly"
#' )
#'
#' # Multiple apps at once
#' sales <- st_unified_sales_report(
#'   unified_app_id = c("67ec0bf3e540b65904256cc4", "5ba4585f539ce75b97db6bcb"),
#'   countries = c("US", "GB", "JP"),
#'   start_date = "2024-01-01",
#'   end_date = "2024-12-31",
#'   date_granularity = "monthly"
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_unified_sales_report_impl <- function(unified_app_id,
                                    countries,
                                    start_date,
                                    end_date,
                                    date_granularity,
                                    auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                                    verbose = TRUE) {

  # Validate required parameters
  if (missing(unified_app_id) || is.null(unified_app_id) || length(unified_app_id) == 0) {
    rlang::abort("'unified_app_id' parameter is required. Provide one or more 24-character hex IDs.")
  }

  # Validate unified_app_id format
  invalid_ids <- unified_app_id[!grepl("^[a-f0-9]{24}$", unified_app_id)]
  if (length(invalid_ids) > 0) {
    rlang::abort(paste0(
      "Invalid unified_app_id format. Expected 24-character hex IDs.\n",
      "Invalid IDs: ", paste(invalid_ids, collapse = ", "), "\n",
      "Use st_app_info_impl() to search for apps and get their unified_app_id."
    ))
  }

  if (missing(countries) || is.null(countries) || length(countries) == 0) {
    rlang::abort("'countries' parameter is required. Specify country codes (e.g., 'US', 'GB', 'JP', or 'WW' for worldwide).")
  }

  if (missing(start_date) || is.null(start_date)) {
    rlang::abort("'start_date' parameter is required. Specify in YYYY-MM-DD format.")
  }

  if (missing(end_date) || is.null(end_date)) {
    rlang::abort("'end_date' parameter is required. Specify in YYYY-MM-DD format.")
  }

  if (missing(date_granularity) || is.null(date_granularity)) {
    rlang::abort("'date_granularity' parameter is required. Specify one of: 'daily', 'weekly', 'monthly', 'quarterly'.")
  }

  date_granularity <- match.arg(date_granularity, c("daily", "weekly", "monthly", "quarterly"))

  # Convert dates to Date objects
  start_date <- as.Date(start_date)
  end_date <- as.Date(end_date)

  # Authentication
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  if (verbose) {
    message("=== Unified Sales Report (All Regional SKUs) ===")
    message("  Apps: ", length(unified_app_id))
    message("  Countries: ", paste(countries, collapse = ", "))
    message("  Date range: ", start_date, " to ", end_date)
    message("  Granularity: ", date_granularity)
    message("================================================")
  }

  # Build query parameters
  query_params <- list(
    auth_token = auth_token_val,
    app_ids = paste(unified_app_id, collapse = ","),
    date_granularity = date_granularity,
    start_date = format(start_date, "%Y-%m-%d"),
    end_date = format(end_date, "%Y-%m-%d"),
    countries = paste(countries, collapse = ",")
  )

  # Build and perform request using the unified endpoint
  endpoint <- st_endpoint_relative_path("unified_sales_report_estimates")

  result <- fetch_data_core(
    endpoint = endpoint,
    params = query_params,
    auth_token = auth_token_val,
    verbose = verbose,
    enrich_response = FALSE,
    processor = process_unified_sales_response
  )

  if (is.null(result) || nrow(result) == 0) {
    if (verbose) message("No data returned from unified API.")
    return(tibble::tibble(
      date = as.Date(character()),
      country = character(),
      unified_app_id = character(),
      revenue = numeric(),
      downloads = numeric()
    ))
  }

  if (verbose) {
    message(sprintf("Retrieved %d records from unified API.", nrow(result)))
  }

  return(result)
}

#' Process unified sales report response
#' @noRd
process_unified_sales_response <- function(resp) {
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(NULL)
  }

  body_text <- rawToChar(body_raw)
  result <- jsonlite::fromJSON(body_text, flatten = TRUE)

  if (length(result) == 0 || nrow(result) == 0) {
    return(NULL)
  }

  # Convert to tibble
  result_tbl <- tibble::as_tibble(result)

  # Rename and process unified API columns
  # The unified API returns: app_id, date, country, unified_revenue (cents), unified_units
  result_tbl <- result_tbl %>%
    dplyr::rename(
      unified_app_id = "app_id",
      downloads = "unified_units"
    ) %>%
    dplyr::mutate(
      date = as.Date(date),
      # Convert revenue from cents to dollars
      revenue = unified_revenue / 100
    ) %>%
    dplyr::select(date, country, unified_app_id, revenue, downloads)

  return(result_tbl)
}

#' Example Data for sensortowerR Package
#'
#' Creates example data structures that mimic the API responses for testing
#' and demonstration purposes without requiring actual API calls.
#'
#' @param type Character string specifying the type of example data to create.
#'   Options: "app_info", "rankings", "details", "metrics", "publisher_apps",
#'   "top_charts", "game_summary"
#' @param n Number of example records to generate (default: 5)
#'
#' @return A tibble with example data matching the structure of actual API responses
#' 
#' @examples
#' # Get example app search results
#' example_apps <- example_sensortower_data("app_info")
#' head(example_apps)
#'
#' # Get example ranking data
#' example_rankings <- example_sensortower_data("rankings", n = 10)
#' print(example_rankings)
#'
#' @importFrom stats runif
#' @keywords internal
#' @export
example_sensortower_data <- function(type = "app_info", n = 5) {
  
  type <- match.arg(type, c("app_info", "rankings", "details", "metrics", 
                           "publisher_apps", "top_charts", "game_summary"))
  
  # Sample app names and publishers for realistic data
  app_names <- c("Candy Crush Saga", "Clash Royale", "Pokemon GO", 
                 "Roblox", "Among Us", "Genshin Impact", "PUBG Mobile",
                 "Call of Duty Mobile", "Minecraft", "Fortnite")
  
  publishers <- c("King", "Supercell", "Niantic", "Roblox Corporation",
                 "InnerSloth", "miHoYo", "Tencent", "Activision",
                 "Mojang", "Epic Games")
  
  app_ids_ios <- c("553834731", "1053012308", "834731712", "431946152",
                   "1351168404", "1517712324", "1330667917", "1476862022",
                   "479516143", "1261357398")
  
  n <- min(n, length(app_names))
  
  switch(type,
    app_info = tibble::tibble(
      unified_app_id = paste0("ua_", seq_len(n)),
      unified_app_name = app_names[seq_len(n)],
      category_details = rep(list(tibble::tibble(
        platform = "ios",
        category_id = 6014,
        category_name = "Games"
      )), n)
    ),
    
    rankings = tibble::tibble(
      rank = seq_len(n),
      app_id = app_ids_ios[seq_len(n)],
      category = 6014,
      country = "US",
      date = Sys.Date(),
      chart_type = "topfreeapplications",
      os = "ios"
    ),
    
    details = tibble::tibble(
      app_id = app_ids_ios[seq_len(n)],
      app_name = app_names[seq_len(n)],
      publisher_name = publishers[seq_len(n)],
      rating = round(runif(n, 4.0, 5.0), 2),
      rating_count = sample(10000:1000000, n),
      price = rep(0, n),
      description = paste("Example description for", app_names[seq_len(n)]),
      os = "ios"
    ),
    
    metrics = tibble::tibble(
      unified_app_id = paste0("ua_", seq_len(n)),
      date = Sys.Date() - seq_len(n) + 1,
      country = "US",
      revenue = round(runif(n, 10000, 1000000), 0),
      downloads = round(runif(n, 1000, 100000), 0),
      active_users = round(runif(n, 5000, 500000), 0)
    ),
    
    publisher_apps = tibble::tibble(
      unified_app_id = paste0("ua_", seq_len(n)),
      unified_app_name = app_names[seq_len(n)],
      publisher_id = rep("pub_123456", n),
      publisher_name = publishers[1]
    ),
    
    top_charts = tibble::tibble(
      unified_app_id = paste0("ua_", seq_len(n)),
      unified_app_name = app_names[seq_len(n)],
      revenue_180d_ww = round(runif(n, 1e6, 1e8), 0),
      downloads_180d_ww = round(runif(n, 1e5, 1e7), 0),
      mau_month_ww = round(runif(n, 1e5, 1e7), 0),
      retention_1d_us = round(runif(n, 20, 60), 1),
      retention_7d_us = round(runif(n, 10, 40), 1)
    ),
    
    game_summary = tibble::tibble(
      Date = rep(Sys.Date() - 0:2, each = n/3)[seq_len(n)],
      Country = rep(c("US", "GB", "JP"), length.out = n),
      `iOS Revenue` = round(runif(n, 1e6, 1e8), 0),
      `iOS Downloads` = round(runif(n, 1e5, 1e7), 0),
      `Android Revenue` = round(runif(n, 5e5, 5e7), 0),
      `Android Downloads` = round(runif(n, 2e5, 2e7), 0)
    )
  )
}

#' Fetch Session Metrics Time Series Data
#'
#' Retrieves session metrics time series data (session count, session duration,
#' time spent) for apps from the Sensor Tower Usage Intelligence API.
#'
#' @param unified_app_id Character string or vector. Sensor Tower unified app ID(s)
#'   (24-character hex). Maximum 100 apps per request.
#' @param ios_app_id Character string or vector. iOS app ID(s) for non-unified queries.
#' @param android_app_id Character string or vector. Android package name(s) for
#'   non-unified queries.
#' @param start_date Date or character string. Start date in "YYYY-MM-DD" format.
#'   Data is available from 2021-01-01 onward.
#' @param end_date Date or character string. End date in "YYYY-MM-DD" format.
#' @param metrics Character vector. Metrics to retrieve. Options include:
#'   - "time_spent" (average seconds per user per day)
#'   - "total_time_spent" (total seconds across all users)
#'   - "session_duration" (average session length in seconds)
#'   - "session_count" (average sessions per user per day)
#'   - "total_session_count" (total sessions across all users)
#'   Default is c("session_count", "session_duration", "time_spent").
#' @param regions Character vector. Region/country codes (e.g., "US", "GB").
#'   Default is "US". Use NULL for all regions.
#' @param time_period Character string. Session metrics time period.
#'   Options: "day", "week". Default is "week".
#'   Returns averaged session metrics for each period within a month.
#' @param date_granularity Character string. Aggregate data by granularity.
#'   Options: "daily", "weekly", "monthly". Default is "monthly".
#'   Note: "daily" granularity may not be supported by the API for all apps;
#'   use "weekly" or "monthly" if you receive empty results with "daily".
#' @param os Character string. Filter by platform for unified apps.
#'   Options: "ios", "android", or NULL for both. Default is NULL.
#' @param breakdown Character string. Fields for data aggregation.
#'   Options: "unified_app_id", "app_id", "region". Default is "unified_app_id".
#' @param auth_token Character string. Your Sensor Tower API token.
#'   Defaults to environment variable SENSORTOWER_AUTH_TOKEN.
#' @param verbose Logical. If TRUE, prints progress messages.
#'
#' @return A [tibble][tibble::tibble] with session metrics including:
#'   - **unified_app_id** or **app_id**: The app identifier
#'   - **date**: Date of the data point
#'   - **time_spent**: Average seconds spent per user per day
#'   - **total_time_spent**: Total seconds across all users
#'   - **session_duration**: Average session length in seconds
#'   - **session_count**: Average sessions per user per day
#'   - **total_session_count**: Total session count across all users
#'
#' @section Data Availability:
#' - Data is available from 2021-01-01 onward
#' - Session metrics require Usage Intelligence subscription
#' - Maximum 100 apps per request
#'
#' @examples
#' \dontrun{
#' # Get session metrics for a unified app
#' sessions <- st_session_metrics(
#'   unified_app_id = "5fbc3849d0b8414136857afc",
#'   start_date = "2024-01-01",
#'   end_date = "2024-12-01"
#' )
#'
#' # Get specific metrics with weekly granularity
#' sessions <- st_session_metrics(
#'   unified_app_id = "5fbc3849d0b8414136857afc",
#'   start_date = "2024-01-01",
#'   end_date = "2024-03-01",
#'   metrics = c("session_count", "session_duration"),
#'   date_granularity = "weekly"
#' )
#'
#' # Get session data for Android app directly
#' sessions <- st_session_metrics(
#'   android_app_id = "com.example.app",
#'   start_date = "2024-01-01",
#'   end_date = "2024-06-01"
#' )
#' }
#'
#' @seealso [st_retention()] for retention metrics,
#'   [st_demographics()] for user demographics,
#'   [st_batch_metrics()] for MAU/DAU/WAU metrics
#'
#' @export
#' @importFrom rlang %||% abort
#' @importFrom httr2 request req_url_path_append req_url_query req_headers
#' @importFrom httr2 req_perform resp_body_raw resp_status req_retry
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr bind_rows mutate select
st_session_metrics <- function(unified_app_id = NULL,
                                ios_app_id = NULL,
                                android_app_id = NULL,
                                start_date,
                                end_date,
                                metrics = c("session_count", "session_duration", "time_spent"),
                                regions = "US",
                                time_period = "week",
                                date_granularity = "monthly",
                                os = NULL,
                                breakdown = "unified_app_id",
                                auth_token = NULL,
                                verbose = TRUE) {

  # Validate that at least one app ID is provided
  if (is.null(unified_app_id) && is.null(ios_app_id) && is.null(android_app_id)) {
    rlang::abort(c(
      "At least one app ID must be provided.",
      "i" = "Use unified_app_id, ios_app_id, or android_app_id parameter.",
      "i" = "Use st_app_info_impl('app name') to find app IDs."
    ))
  }

  # Validate dates
  start_date <- as.Date(start_date)
  end_date <- as.Date(end_date)

  if (start_date < as.Date("2021-01-01")) {
    rlang::warn("Session metrics data is only available from 2021-01-01 onward. Adjusting start_date.")
    start_date <- as.Date("2021-01-01")
  }

  # Validate metrics
valid_metrics <- c("time_spent", "total_time_spent", "session_duration",
                     "session_count", "total_session_count")
  metrics <- match.arg(metrics, valid_metrics, several.ok = TRUE)

  # Validate time_period
  time_period <- match.arg(time_period, c("day", "week"))

  # Validate date_granularity
  date_granularity <- match.arg(date_granularity, c("daily", "weekly", "monthly"))

  # Authentication
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  # Determine which endpoint to use based on app ID type
  if (!is.null(unified_app_id)) {
    # Use the unified_apps endpoint
    result <- fetch_session_metrics_unified(
      app_ids = unified_app_id,
      start_date = start_date,
      end_date = end_date,
      metrics = metrics,
      regions = regions,
      time_period = time_period,
      date_granularity = date_granularity,
      os = os,
      breakdown = breakdown,
      auth_token = auth_token_val,
      verbose = verbose
    )
  } else {
    # Use the non-unified endpoint for platform-specific IDs
    app_ids <- c(ios_app_id, android_app_id)
    app_ids <- app_ids[!is.null(app_ids) & !is.na(app_ids)]

    result <- fetch_session_metrics_platform(
      app_ids = app_ids,
      start_date = start_date,
      end_date = end_date,
      metrics = metrics,
      regions = regions,
      time_period = time_period,
      date_granularity = date_granularity,
      breakdown = "app_id",
      auth_token = auth_token_val,
      verbose = verbose
    )
  }

  return(result)
}

#' Fetch session metrics for unified apps
#' @noRd
fetch_session_metrics_unified <- function(app_ids,
                                           start_date,
                                           end_date,
                                           metrics,
                                           regions,
                                           time_period,
                                           date_granularity,
                                           os,
                                           breakdown,
                                           auth_token,
                                           verbose = FALSE) {

  # Build query parameters
  query_params <- list(
    auth_token = auth_token,
    start_date = format(start_date, "%Y-%m-%d"),
    end_date = format(end_date, "%Y-%m-%d"),
    time_period = time_period,
    date_granularity = date_granularity,
    breakdown = breakdown
  )

  # Add app_ids as comma-separated
  query_params$app_ids <- paste(app_ids, collapse = ",")

  # Add metrics as comma-separated (API uses 'timeseries' param)
  query_params$timeseries <- paste(metrics, collapse = ",")

  # Add regions if specified
  if (!is.null(regions) && length(regions) > 0) {
    query_params$regions <- paste(regions, collapse = ",")
  }

  # Add OS filter if specified
  if (!is.null(os)) {
    query_params$os <- os
  }

  # Build request
  base_url <- st_api_base_url()

  if (verbose) {
    message("Fetching session metrics for ", length(app_ids), " unified app(s)...")
  }

  req <- build_request(
    base_url = base_url,
    path_segments = st_endpoint_segments("apps_timeseries_unified"),
    query_params = query_params
  ) %>%
    httr2::req_headers(
      "Accept" = "application/json"
    ) %>%
    httr2::req_timeout(60) %>%
    httr2::req_retry(max_tries = 3, backoff = function(i) 2^i)

  # Perform request
  resp <- tryCatch({
    httr2::req_perform(req)
  }, error = function(e) {
    if (verbose) message("API request failed: ", e$message)
    return(NULL)
  })

  if (is.null(resp)) {
    return(tibble::tibble())
  }

  # Check status
  status <- httr2::resp_status(resp)
  if (status != 200) {
    body <- tryCatch(httr2::resp_body_string(resp), error = function(.) "")
    if (verbose) message(sprintf("API returned status %d: %s", status, body))
    return(tibble::tibble())
  }

  # Parse response
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(tibble::tibble())
  }

  body_text <- rawToChar(body_raw)
  result <- tryCatch({
    jsonlite::fromJSON(body_text, flatten = TRUE)
  }, error = function(e) {
    if (verbose) message("Failed to parse response: ", e$message)
    return(NULL)
  })

  if (is.null(result)) {
    return(tibble::tibble())
  }

  # Process the response
  # Expected format: { "unified_apps": [ { "unified_app_id": "...", "timeseries": [...] } ] }
  process_session_response(result, "unified", verbose)
}

#' Fetch session metrics for platform-specific apps
#' @noRd
fetch_session_metrics_platform <- function(app_ids,
                                            start_date,
                                            end_date,
                                            metrics,
                                            regions,
                                            time_period,
                                            date_granularity,
                                            breakdown,
                                            auth_token,
                                            verbose = FALSE) {

  # Build query parameters
  query_params <- list(
    auth_token = auth_token,
    start_date = format(start_date, "%Y-%m-%d"),
    end_date = format(end_date, "%Y-%m-%d"),
    time_period = time_period,
    date_granularity = date_granularity,
    breakdown = breakdown
  )

  # Add app_ids as comma-separated
  query_params$app_ids <- paste(app_ids, collapse = ",")

  # Add metrics as comma-separated
  query_params$timeseries <- paste(metrics, collapse = ",")

  # Add regions if specified
  if (!is.null(regions) && length(regions) > 0) {
    query_params$regions <- paste(regions, collapse = ",")
  }

  # Build request
  base_url <- st_api_base_url()

  if (verbose) {
    message("Fetching session metrics for ", length(app_ids), " platform-specific app(s)...")
  }

  req <- build_request(
    base_url = base_url,
    path_segments = st_endpoint_segments("apps_timeseries"),
    query_params = query_params
  ) %>%
    httr2::req_headers(
      "Accept" = "application/json"
    ) %>%
    httr2::req_timeout(60) %>%
    httr2::req_retry(max_tries = 3, backoff = function(i) 2^i)

  # Perform request
  resp <- tryCatch({
    httr2::req_perform(req)
  }, error = function(e) {
    if (verbose) message("API request failed: ", e$message)
    return(NULL)
  })

  if (is.null(resp)) {
    return(tibble::tibble())
  }

  # Check status
  status <- httr2::resp_status(resp)
  if (status != 200) {
    body <- tryCatch(httr2::resp_body_string(resp), error = function(.) "")
    if (verbose) message(sprintf("API returned status %d: %s", status, body))
    return(tibble::tibble())
  }

  # Parse response
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(tibble::tibble())
  }

  body_text <- rawToChar(body_raw)
  result <- tryCatch({
    jsonlite::fromJSON(body_text, flatten = TRUE)
  }, error = function(e) {
    if (verbose) message("Failed to parse response: ", e$message)
    return(NULL)
  })

  if (is.null(result)) {
    return(tibble::tibble())
  }

  # Process the response
  process_session_response(result, "platform", verbose)
}

#' Process session metrics API response into a tibble
#' @noRd
process_session_response <- function(result, response_type, verbose = FALSE) {
  # Handle both unified_apps and apps response structures
  apps_data <- NULL

  if ("unified_apps" %in% names(result)) {
    apps_data <- result$unified_apps
  } else if ("apps" %in% names(result)) {
    apps_data <- result$apps
  } else if (is.list(result) && length(result) > 0) {
    # Try to use the result directly if it's a list of apps
    apps_data <- result
  }

  if (is.null(apps_data) || length(apps_data) == 0) {
    if (verbose) message("No session data found in response.")
    return(tibble::tibble())
  }

  # Process each app's timeseries data
  all_rows <- list()

  # Handle different data structures
  if (is.data.frame(apps_data)) {
    for (i in seq_len(nrow(apps_data))) {
      app_row <- apps_data[i, ]
      app_id <- app_row$unified_app_id %||% app_row$app_id %||% NA_character_

      timeseries <- app_row$timeseries
      if (is.list(timeseries) && length(timeseries) > 0) {
        ts_df <- timeseries[[1]]
        if (is.data.frame(ts_df) && nrow(ts_df) > 0) {
          ts_df$app_id <- app_id
          if ("unified_app_id" %in% names(app_row)) {
            ts_df$unified_app_id <- app_id
          }
          all_rows[[length(all_rows) + 1]] <- ts_df
        }
      }
    }
  } else if (is.list(apps_data)) {
    for (app_item in apps_data) {
      if (is.null(app_item)) next

      app_id <- app_item$unified_app_id %||% app_item$app_id %||% NA_character_
      timeseries <- app_item$timeseries

      if (!is.null(timeseries) && length(timeseries) > 0) {
        if (is.data.frame(timeseries)) {
          ts_df <- timeseries
        } else if (is.list(timeseries)) {
          ts_df <- tryCatch(
            dplyr::bind_rows(timeseries),
            error = function(e) tibble::tibble()
          )
        } else {
          next
        }

        if (nrow(ts_df) > 0) {
          ts_df$app_id <- app_id
          if (!is.null(app_item$unified_app_id)) {
            ts_df$unified_app_id <- app_id
          }
          all_rows[[length(all_rows) + 1]] <- ts_df
        }
      }
    }
  }

  if (length(all_rows) == 0) {
    if (verbose) message("No timeseries data found.")
    return(tibble::tibble())
  }

  # Combine all rows
  result_tbl <- dplyr::bind_rows(all_rows)

  # Convert date column
  if ("date" %in% names(result_tbl)) {
    result_tbl$date <- as.Date(result_tbl$date)
  }

  # Reorder columns to put identifiers first
  id_cols <- intersect(c("unified_app_id", "app_id", "date"), names(result_tbl))
  other_cols <- setdiff(names(result_tbl), id_cols)
  result_tbl <- result_tbl[, c(id_cols, other_cols)]

  if (verbose) {
    message(sprintf("Retrieved %d session data points.", nrow(result_tbl)))
  }

  return(tibble::as_tibble(result_tbl))
}

#' Low-Level Access to Sensor Tower Facets Metrics
#'
#' Performs a GET request against Sensor Tower's `"/v1/facets/metrics"` route.
#' This helper stays intentionally low-level so package code can work with new
#' facets-based endpoints while higher-level wrappers are added incrementally.
#'
#' @param query Character vector of raw query fragments appended verbatim after
#'   encoded `params`. This is useful when Sensor Tower documents nested or
#'   repeated parameters that are awkward to express as a regular named R list.
#'   Example: `c("filters[app_ids][]=553834731")`.
#' @param params Named list of regular query parameters. Vector values are
#'   serialized as comma-separated strings to match Sensor Tower's
#'   `style=form, explode=false` usage in the OpenAPI spec.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'   Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.
#' @param host Character string. Which host to target: `"api"` (default) or
#'   `"app"`. The `app` host uses `https://app.sensortower.com/api`.
#' @param verbose Logical. If `TRUE`, prints the request URL with the auth token
#'   redacted.
#'
#' @return A parsed JSON response. Rectangular top-level responses are returned
#'   as a tibble; nested responses are returned as a named list.
#'
#' @details
#' As of March 17, 2026, the facets route is live and the retention contract is
#' validated against production. The machine-readable Sensor Tower docs remain
#' gated behind a signed-in web session, so this helper still provides a stable
#' escape hatch for raw or partially documented facets requests.
#'
#' @examples
#' \dontrun{
#' # Retention request using regular query parameters
#' response <- st_facets_metrics(
#'   params = list(
#'     facets = "retention",
#'     bundle = "retention_daily",
#'     breakdown = c("date", "app_id"),
#'     start_date = "2025-01-01",
#'     end_date = "2025-01-31",
#'     app_ids = "553834731"
#'   )
#' )
#' }
#'
#' @export
st_facets_metrics <- function(query = character(),
                              params = list(),
                              auth_token = NULL,
                              host = c("api", "app"),
                              verbose = FALSE) {
  if (!is.character(query)) {
    rlang::abort("`query` must be a character vector of raw query fragments.")
  }

  if (!is.list(params)) {
    rlang::abort("`params` must be a named list.")
  }

  if (length(params) > 0 && (is.null(names(params)) || any(!nzchar(names(params))))) {
    rlang::abort("`params` must be a named list.")
  }

  host <- match.arg(host)
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  base_url <- switch(
    host,
    api = st_api_base_url(),
    app = "https://app.sensortower.com/api"
  )

  query <- trimws(query)
  query <- query[nzchar(query)]
  params <- params[!vapply(params, is.null, logical(1))]

  encoded_params <- if (length(params) == 0) {
    character()
  } else {
    stats::setNames(
      lapply(params, format_facets_query_value),
      names(params)
    )
  }

  encoded_params <- encoded_params[!vapply(encoded_params, is.null, logical(1))]
  encoded_query <- if (length(encoded_params) == 0) {
    character()
  } else {
    paste0(
      names(encoded_params),
      "=",
      vapply(
        encoded_params,
        utils::URLencode,
        reserved = TRUE,
        FUN.VALUE = character(1)
      )
    )
  }

  raw_query <- c(
    paste0("auth_token=", utils::URLencode(auth_token_val, reserved = TRUE)),
    encoded_query,
    query
  )

  url <- paste0(
    base_url,
    "/",
    st_endpoint_path("facets_metrics"),
    "?",
    paste(raw_query, collapse = "&")
  )

  if (verbose) {
    message("Requesting: ", sub(auth_token_val, "HIDDEN", url, fixed = TRUE))
  }

  req <- httr2::request(url) %>%
    httr2::req_headers("Accept" = "application/json") %>%
    httr2::req_timeout(30)

  resp <- tryCatch(
    httr2::req_perform(req),
    error = function(e) {
      if (inherits(e, "httr2_http") && !is.null(e$resp)) {
        body <- tryCatch(httr2::resp_body_string(e$resp), error = function(...) "")
        status <- tryCatch(httr2::resp_status(e$resp), error = function(...) NA_integer_)

        rlang::abort(
          c(
            sprintf("Facets metrics request failed with status %s.", status),
            if (nzchar(body)) c("i" = body) else NULL
          ),
          parent = e
        )
      }

      rlang::abort(
        "An unexpected error occurred during the facets metrics request.",
        parent = e
      )
    }
  )

  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(tibble::tibble())
  }

  parsed <- jsonlite::fromJSON(
    rawToChar(body_raw),
    simplifyVector = TRUE,
    flatten = TRUE
  )

  if (inherits(parsed, "data.frame")) {
    return(tibble::as_tibble(parsed))
  }

  parsed
}

format_facets_query_value <- function(value) {
  if (length(value) == 0 || all(is.na(value))) {
    return(NULL)
  }

  if (inherits(value, "Date")) {
    value <- format(value, "%Y-%m-%d")
  }

  if (is.logical(value)) {
    value <- tolower(as.character(value))
  } else {
    value <- as.character(value)
  }

  value <- trimws(value)
  value <- value[nzchar(value)]

  if (length(value) == 0) {
    return(NULL)
  }

  paste(value, collapse = ",")
}

#' Column validation and mapping helpers
#' 
#' @description Helper functions to handle column name variations and missing columns

#' Get available columns matching a pattern
#'
#' @keywords internal
#' @noRd
find_column <- function(data, pattern, prefer = NULL) {
  cols <- names(data)
  
  # Check preferred columns first
  if (!is.null(prefer)) {
    for (pref in prefer) {
      if (pref %in% cols) return(pref)
    }
  }
  
  # Fall back to pattern matching
  matches <- grep(pattern, cols, value = TRUE, ignore.case = TRUE)
  if (length(matches) > 0) return(matches[1])
  
  return(NULL)
}

#' Select columns safely with fallbacks
#'
#' @keywords internal
#' @noRd
select_columns_safe <- function(data, columns) {
  result <- data
  selected_cols <- c()
  
  for (new_name in names(columns)) {
    spec <- columns[[new_name]]
    
    # Handle different specification formats
    if (is.character(spec)) {
      # Simple column name or pattern
      if (spec %in% names(data)) {
        selected_cols[new_name] <- spec
      } else {
        # Try pattern matching
        found <- find_column(data, spec)
        if (!is.null(found)) {
          selected_cols[new_name] <- found
        }
      }
    } else if (is.list(spec)) {
      # Advanced specification with fallbacks
      found <- find_column(data, 
                          pattern = spec$pattern %||% "", 
                          prefer = spec$prefer)
      if (!is.null(found)) {
        selected_cols[new_name] <- found
      } else if (!is.null(spec$default)) {
        # Add column with default value
        result[[new_name]] <- spec$default
        selected_cols[new_name] <- new_name
      }
    }
  }
  
  # Select and rename columns
  if (length(selected_cols) > 0) {
    result <- result %>%
      dplyr::select(all_of(selected_cols)) %>%
      dplyr::rename(!!!stats::setNames(selected_cols, names(selected_cols)))
  }
  
  return(result)
}

#' Map region-specific columns intelligently
#'
#' @keywords internal
#' @noRd
map_region_columns <- function(data, requested_region = "US") {
  # Common patterns for region-specific columns
  metric_patterns <- c("revenue", "download", "dau", "wau", "mau", 
                      "retention", "rpd", "arpu", "age")
  
  # Check what regions are actually available
  available_regions <- c()
  for (pattern in metric_patterns) {
    cols <- grep(paste0(pattern, ".*_(us|ww)$"), names(data), 
                value = TRUE, ignore.case = TRUE)
    if (length(cols) > 0) {
      regions <- unique(gsub(".*_([a-z]+)$", "\\1", cols))
      available_regions <- unique(c(available_regions, regions))
    }
  }
  
  message(sprintf("Available regions in data: %s", 
                  paste(toupper(available_regions), collapse = ", ")))
  
  # If requested region not available, use what we have
  actual_region <- tolower(requested_region)
  if (!actual_region %in% available_regions) {
    rlang::abort(sprintf("Requested region '%s' not found in data. Available regions: %s",
                 toupper(requested_region),
                 ifelse(length(available_regions) > 0, paste(toupper(available_regions), collapse = ", "), "none")),
         call. = FALSE)
  }
  
  return(data)
}

#' Validate required columns exist
#'
#' @keywords internal
#' @noRd
validate_columns <- function(data, required, context = "data") {
  missing <- setdiff(required, names(data))
  
  if (length(missing) > 0) {
    available <- names(data)
    suggestions <- c()
    
    for (col in missing) {
      # Find similar column names
      pattern <- gsub("_", ".*", col)
      matches <- grep(pattern, available, value = TRUE, ignore.case = TRUE)
      if (length(matches) > 0) {
        suggestions <- c(suggestions, 
                        sprintf("  - '%s' not found, did you mean: %s?", 
                                col, paste(matches[1:min(3, length(matches))], 
                                         collapse = ", ")))
      } else {
        suggestions <- c(suggestions, sprintf("  - '%s' not found", col))
      }
    }
    
    rlang::abort(sprintf("Missing required columns in %s:\n%s\n\nAvailable columns: %s",
                 context,
                 paste(suggestions, collapse = "\n"),
                 paste(available[1:min(10, length(available))], collapse = ", ")),
         call. = FALSE)
  }
  
  return(TRUE)
}

#' Get column specification for common Sensor Tower metrics
#'
#' @keywords internal
#' @noRd
get_column_spec <- function(metric_type = NULL, time_period = NULL, region = NULL) {
  specs <- list()
  
  # App identification columns
  specs$app_name <- list(
    prefer = c("unified_app_name", "app_name", "name", "humanized_name"),
    pattern = "(app_)?name"
  )
  
  specs$app_id <- list(
    prefer = c("unified_app_id", "app_id", "id"),
    pattern = "(app_)?id"
  )
  
  # Build metric columns dynamically
  if (!is.null(metric_type)) {
    base_metrics <- list(
      revenue = c("revenue", "rev"),
      downloads = c("downloads", "units", "download"),
      retention = c("retention", "ret"),
      dau = c("dau", "daily.*active.*users"),
      mau = c("mau", "monthly.*active.*users"),
      wau = c("wau", "weekly.*active.*users")
    )
    
    for (type in metric_type) {
      if (type %in% names(base_metrics)) {
        patterns <- base_metrics[[type]]
        
        # Build column variations
        for (pattern in patterns) {
          if (!is.null(time_period)) {
            for (tp in time_period) {
              if (!is.null(region)) {
                for (r in region) {
                  col_name <- sprintf("%s_%s_%s", type, tp, r)
                  specs[[col_name]] <- list(
                    prefer = c(
                      sprintf("%s_%s_%s", pattern, tp, r),
                      sprintf("%s_%s_%s", pattern, tp, toupper(r))
                    ),
                    pattern = sprintf("%s.*%s.*%s", pattern, tp, r)
                  )
                }
              }
            }
          }
        }
      }
    }
  }
  
  return(specs)
}

#' Handle column mapping errors gracefully
#'
#' @keywords internal
#' @noRd
try_column_operation <- function(expr, data, context = "operation") {
  tryCatch({
    eval(expr)
  }, error = function(e) {
    # Extract column name from error if possible
    col_match <- regmatches(e$message, regexpr("`[^`]+`", e$message))
    if (length(col_match) > 0) {
      col_name <- gsub("`", "", col_match[1])
      warning(sprintf("Column '%s' not found in %s. Available columns: %s",
                      col_name, context,
                      paste(names(data)[1:min(5, length(names(data)))], 
                            collapse = ", ")),
              call. = FALSE)
    } else {
      warning(sprintf("Error in %s: %s", context, e$message), call. = FALSE)
    }
    return(NULL)
  })
}

#' Custom Fields Utility Functions
#'
#' Utility functions for common custom fields filtering scenarios in Sensor Tower.
#' These functions provide pre-built filters for frequently used queries.
#'
#' @name st_custom_fields_utils
#' @rdname st_custom_fields_utils
NULL

#' Create Genre-Based Filter
#'
#' Creates a filter for apps in specific game genres or sub-genres.
#'
#' @param genres Character vector. Game genres to filter (e.g., "Puzzle", "Action")
#' @param sub_genres Character vector. Game sub-genres to filter (e.g., "Word", "Match-3")
#' @param exclude_genres Logical. Whether to exclude these genres (FALSE = include)
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return Character string containing the filter ID
#'
#' @examples
#' \dontrun{
#' # Get Puzzle and Word games
#' filter_id <- st_filter_by_genre(
#'   genres = "Puzzle",
#'   sub_genres = "Word"
#' )
#' 
#' # Exclude Action and Shooter games
#' filter_id <- st_filter_by_genre(
#'   genres = c("Action", "Shooter"),
#'   exclude_genres = TRUE
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_filter_by_genre_impl <- function(
    genres = NULL,
    sub_genres = NULL,
    exclude_genres = FALSE,
    auth_token = NULL
) {
  
  if (is.null(genres) && is.null(sub_genres)) {
    rlang::abort("At least one of 'genres' or 'sub_genres' must be provided")
  }
  
  custom_fields <- list()
  
  # Add genre filter
  if (!is.null(genres)) {
    custom_fields <- append(custom_fields, list(list(
      name = "Game Genre",
      values = as.list(genres),
      global = TRUE,
      exclude = exclude_genres
    )))
  }
  
  # Add sub-genre filter
  if (!is.null(sub_genres)) {
    custom_fields <- append(custom_fields, list(list(
      name = "Game Sub-genre",
      values = as.list(sub_genres),
      global = TRUE,
      exclude = exclude_genres
    )))
  }
  
  st_custom_fields_filter_impl(
    custom_fields = custom_fields,
    auth_token = auth_token
  )
}

#' Create Monetization-Based Filter
#'
#' Creates a filter for apps based on their monetization model.
#'
#' @param free_only Logical. Only free apps
#' @param has_iap Logical. Has in-app purchases
#' @param has_ads Logical. Contains ads
#' @param has_subscription Logical. Has subscription model
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return Character string containing the filter ID
#'
#' @examples
#' \dontrun{
#' # Get free apps with IAP but no ads
#' filter_id <- st_filter_by_monetization(
#'   free_only = TRUE,
#'   has_iap = TRUE,
#'   has_ads = FALSE
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_filter_by_monetization_impl <- function(
    free_only = NULL,
    has_iap = NULL,
    has_ads = NULL,
    has_subscription = NULL,
    auth_token = NULL
) {
  
  custom_fields <- list()
  
  if (!is.null(free_only)) {
    custom_fields <- append(custom_fields, list(list(
      name = "Free",
      global = TRUE,
      true = free_only
    )))
  }
  
  if (!is.null(has_iap)) {
    custom_fields <- append(custom_fields, list(list(
      name = "In-App Purchases",
      global = TRUE,
      true = has_iap
    )))
  }
  
  if (!is.null(has_ads)) {
    custom_fields <- append(custom_fields, list(list(
      name = "Contains Ads",
      global = TRUE,
      true = has_ads
    )))
  }
  
  if (!is.null(has_subscription)) {
    custom_fields <- append(custom_fields, list(list(
      name = "In-App Subscription",
      global = TRUE,
      true = has_subscription
    )))
  }
  
  if (length(custom_fields) == 0) {
    rlang::abort("At least one monetization parameter must be specified")
  }
  
  st_custom_fields_filter_impl(
    custom_fields = custom_fields,
    auth_token = auth_token
  )
}

#' Create Publisher-Based Filter
#'
#' Creates a filter for apps from specific publishers.
#'
#' @param publisher_names Character vector. Publisher names to filter
#' @param exclude Logical. Whether to exclude these publishers (FALSE = include)
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return Character string containing the filter ID
#'
#' @examples
#' \dontrun{
#' # Get apps from specific publishers
#' filter_id <- st_filter_by_publisher(
#'   publisher_names = c("Electronic Arts", "Activision")
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_filter_by_publisher_impl <- function(
    publisher_names,
    exclude = FALSE,
    auth_token = NULL
) {
  
  if (missing(publisher_names) || is.null(publisher_names)) {
    rlang::abort("publisher_names is required")
  }
  
  st_custom_fields_filter_impl(
    custom_fields = list(
      list(
        name = "Publisher",
        values = as.list(publisher_names),
        global = TRUE,
        exclude = exclude
      )
    ),
    auth_token = auth_token
  )
}

#' Create SDK-Based Filter
#'
#' Creates a filter for apps using specific SDKs or technologies.
#'
#' @param sdk_names Character vector. SDK names to filter (e.g., "Unity", "Firebase")
#' @param exclude Logical. Whether to exclude apps with these SDKs
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return Character string containing the filter ID
#'
#' @examples
#' \dontrun{
#' # Get Unity-based games
#' filter_id <- st_filter_by_sdk(sdk_names = "Unity")
#' 
#' # Get apps using both Firebase and AdMob
#' filter_id <- st_filter_by_sdk(
#'   sdk_names = c("Firebase", "AdMob")
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_filter_by_sdk_impl <- function(
    sdk_names,
    exclude = FALSE,
    auth_token = NULL
) {
  
  if (missing(sdk_names) || is.null(sdk_names)) {
    rlang::abort("sdk_names is required")
  }
  
  # Create filters for each SDK
  custom_fields <- lapply(sdk_names, function(sdk) {
    list(
      name = paste0("SDK: ", sdk),
      values = list(),
      global = TRUE,
      exclude = exclude,
      true = !exclude
    )
  })
  
  st_custom_fields_filter_impl(
    custom_fields = custom_fields,
    auth_token = auth_token
  )
}

#' Create Date-Based Filter
#'
#' Creates a filter for apps based on release date criteria.
#'
#' @param released_after Date or character. Apps released after this date
#' @param released_before Date or character. Apps released before this date
#' @param region Character. Region for release date ("US", "WW", "JP", "CN")
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return Character string containing the filter ID
#'
#' @examples
#' \dontrun{
#' # Get apps released in 2024
#' filter_id <- st_filter_by_date(
#'   released_after = "2024-01-01",
#'   released_before = "2024-12-31",
#'   region = "US"
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_filter_by_date_impl <- function(
    released_after = NULL,
    released_before = NULL,
    region = "US",
    auth_token = NULL
) {
  
  if (is.null(released_after) && is.null(released_before)) {
    rlang::abort("At least one of 'released_after' or 'released_before' must be provided")
  }
  
  # Format dates
  if (!is.null(released_after)) {
    released_after <- format(as.Date(released_after), "%Y-%m-%d")
  }
  if (!is.null(released_before)) {
    released_before <- format(as.Date(released_before), "%Y-%m-%d")
  }
  
  # Determine field name based on region
  field_name <- switch(region,
    "US" = "Release Date (US)",
    "WW" = "Release Date (WW)",
    "JP" = "Release Date (JP)",
    "CN" = "Release Date (CN)",
    "Release Date (US)"  # Default
  )
  
  # Create date range values
  values <- list()
  if (!is.null(released_after) && !is.null(released_before)) {
    values <- list(paste(released_after, released_before, sep = " to "))
  } else if (!is.null(released_after)) {
    values <- list(paste("after", released_after))
  } else {
    values <- list(paste("before", released_before))
  }
  
  st_custom_fields_filter_impl(
    custom_fields = list(
      list(
        name = field_name,
        values = values,
        global = TRUE,
        exclude = FALSE
      )
    ),
    auth_token = auth_token
  )
}

#' Get Pre-Built Filter Collections
#'
#' Returns commonly used filter IDs for quick access to pre-defined app segments.
#'
#' @param collection Character. Name of the collection:
#'   - "top_genres": Major game genres
#'   - "monetization_models": Different monetization approaches
#'   - "platform_exclusive": Platform-specific apps
#'   - "market_segments": Market segment filters
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return A named list of filter IDs
#'
#' @examples
#' \dontrun{
#' # Get filter IDs for top game genres
#' genre_filters <- st_get_filter_collection("top_genres")
#' 
#' # Use a filter from the collection
#' puzzle_apps <- st_get_filtered_apps(
#'   filter_id = genre_filters$puzzle,
#'   measure = "DAU",
#'   regions = "US"
#' )
#' }
#'
#' @export
st_get_filter_collection <- function(
    collection = c("top_genres", "monetization_models", "platform_exclusive", "market_segments"),
    auth_token = NULL
) {
  
  collection <- match.arg(collection)
  
  filters <- switch(collection,
    "top_genres" = list(
      puzzle = st_filter_by_genre_impl(genres = "Puzzle", auth_token = auth_token),
      action = st_filter_by_genre_impl(genres = "Action", auth_token = auth_token),
      strategy = st_filter_by_genre_impl(genres = "Strategy", auth_token = auth_token),
      casual = st_filter_by_genre_impl(genres = "Casual", auth_token = auth_token),
      simulation = st_filter_by_genre_impl(genres = "Simulation", auth_token = auth_token),
      rpg = st_filter_by_genre_impl(genres = "Role Playing", auth_token = auth_token),
      word = st_filter_by_genre_impl(sub_genres = "Word", auth_token = auth_token),
      match3 = st_filter_by_genre_impl(sub_genres = "Match-3", auth_token = auth_token)
    ),
    
    "monetization_models" = list(
      free_with_ads = st_filter_by_monetization_impl(
        free_only = TRUE, has_ads = TRUE, auth_token = auth_token
      ),
      free_with_iap = st_filter_by_monetization_impl(
        free_only = TRUE, has_iap = TRUE, has_ads = FALSE, auth_token = auth_token
      ),
      premium = st_filter_by_monetization_impl(
        free_only = FALSE, has_iap = FALSE, auth_token = auth_token
      ),
      subscription = st_filter_by_monetization_impl(
        has_subscription = TRUE, auth_token = auth_token
      ),
      hybrid = st_filter_by_monetization_impl(
        has_iap = TRUE, has_ads = TRUE, auth_token = auth_token
      )
    ),
    
    "platform_exclusive" = list(
      ios_only = st_create_simple_filter_impl(
        field_name = "Platform",
        field_values = "iOS",
        auth_token = auth_token
      ),
      android_only = st_create_simple_filter_impl(
        field_name = "Platform", 
        field_values = "Android",
        auth_token = auth_token
      ),
      cross_platform = st_create_simple_filter_impl(
        field_name = "Is Unified",
        field_values = list(),
        auth_token = auth_token
      )
    ),
    
    "market_segments" = list(
      editors_choice = st_create_simple_filter_impl(
        field_name = "Editors' Choice",
        field_values = list(),
        auth_token = auth_token
      ),
      new_releases = st_filter_by_date_impl(
        released_after = Sys.Date() - 30,
        region = "US",
        auth_token = auth_token
      ),
      established = st_filter_by_date_impl(
        released_before = Sys.Date() - 365,
        region = "US",
        auth_token = auth_token
      )
    )
  )
  
  return(filters)
}

#' Combine Multiple Filters
#'
#' Creates a compound filter by combining multiple filter criteria.
#'
#' @param filter_ids Character vector. Existing filter IDs to combine
#' @param operator Character. How to combine filters ("AND" or "OR")
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return Character string containing the combined filter ID
#'
#' @note OR operations may not be supported by all Sensor Tower endpoints
#'
#' @examples
#' \dontrun{
#' # Get Word Puzzle games (AND operation)
#' word_filter <- st_filter_by_genre(sub_genres = "Word")
#' puzzle_filter <- st_filter_by_genre(genres = "Puzzle")
#' combined <- st_combine_filters(
#'   filter_ids = c(word_filter, puzzle_filter),
#'   operator = "AND"
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_combine_filters_impl <- function(
    filter_ids,
    operator = c("AND", "OR"),
    auth_token = NULL
) {
  
  operator <- match.arg(operator)
  
  if (length(filter_ids) < 2) {
    rlang::abort("At least two filter_ids must be provided to combine")
  }
  
  # Get details of each filter
  all_fields <- list()
  
  for (filter_id in filter_ids) {
    details <- st_custom_fields_filter_by_id_impl(
      id = filter_id,
      auth_token = auth_token
    )
    
    if (!is.null(details$custom_fields)) {
      # Convert data frame to list format
      for (i in seq_len(nrow(details$custom_fields))) {
        field <- list(
          name = details$custom_fields$name[i],
          values = details$custom_fields$values[[i]],
          global = details$custom_fields$global[i],
          exclude = details$custom_fields$exclude[i]
        )
        
        if (operator == "AND") {
          # For AND, simply add all fields
          all_fields <- append(all_fields, list(field))
        } else {
          # OR operation might need special handling
          # This is a simplified approach
          all_fields <- append(all_fields, list(field))
        }
      }
    }
  }
  
  if (length(all_fields) == 0) {
    rlang::abort("No valid fields found in provided filter IDs")
  }
  
  # Create combined filter
  st_custom_fields_filter_impl(
    custom_fields = all_fields,
    auth_token = auth_token
  )
}

# Global variable declarations to avoid R CMD check warnings
# This file is loaded last (zzz is conventional name for this purpose)

# Declare global variables used in dplyr pipelines and other non-standard evaluation contexts
utils::globalVariables(c(
  # st_ytd_metrics variables
  "country", "daily_users", "total_user_days", "n_days", "dau",
  "weekly_users", "total_user_weeks", "n_weeks", "wau",
  "monthly_users", "total_user_months", "n_months", "mau",
  "entity_id", "entity_name", "date_start", "date_end", "metric", "value",
  
  # st_metrics variables  
  "total_revenue", "total_downloads", "revenue", "downloads", "platform",
  "category_id", "revenue_usd", "revenue_cents",
  
  # st_top_publishers variables
  "revenue_absolute", "apps"
))

#' Fetch Sensor Tower Rankings
#'
#' Thin v1.0.0 facade that standardizes the app, publisher, and category ranking
#' entrypoints while dispatching to the existing implementations.
#'
#' @param entity Ranking entity to fetch: `"app"`, `"publisher"`, or
#'   `"category"`.
#' @param os Operating system. One of `"ios"`, `"android"`, or `"unified"`.
#' @param category Optional category identifier forwarded to the legacy ranking
#'   implementation.
#' @param country Two-letter country code.
#' @param chart_type Chart type for category rankings. Ignored for app and
#'   publisher rankings.
#' @param date Ranking date. Accepts a `Date` object or ISO date string.
#' @param limit Positive integer row limit.
#' @param filter Optional `st_filter` object or 24-character filter ID string.
#' @param auth_token Optional Sensor Tower API token.
#'
#' @return A tibble with standardized columns `rank`, `id`, `name`, `os`,
#'   `category`, `country`, and `date`, plus any entity-specific columns returned
#'   by the dispatched implementation.
#'
#' @examples
#' \dontrun{
#' st_rankings(entity = "app", os = "ios", category = 6014, country = "US")
#'
#' st_rankings(
#'   entity = "category",
#'   os = "android",
#'   category = "game",
#'   chart_type = "topgrossing",
#'   country = "GB"
#' )
#' }
#'
#' @export
st_rankings <- function(entity = c("app", "publisher", "category"),
                        os = "ios",
                        category = NULL,
                        country = "US",
                        chart_type = "topfreeapplications",
                        date = Sys.Date() - 1,
                        limit = 100,
                        filter = NULL,
                        auth_token = NULL) {
  entity <- match.arg(entity)
  os <- normalize_os(os)
  country <- normalize_country(country)
  auth_token <- get_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required.",
      "Set SENSORTOWER_AUTH_TOKEN or pass `auth_token`."
    )
  )

  ranking_date <- tryCatch(as.Date(date), error = function(e) NA)
  if (is.na(ranking_date)) {
    rlang::abort("`date` must be a Date or ISO date string (YYYY-MM-DD).")
  }

  limit <- suppressWarnings(as.integer(limit[1]))
  if (is.na(limit) || limit < 1L) {
    rlang::abort("`limit` must be a positive integer.")
  }

  filter_id <- .st_filter_resolve_id(filter)
  custom_tags_mode <- if (!is.null(filter_id) && os == "unified") {
    "include_unified_apps"
  } else {
    NULL
  }

  raw <- switch(
    entity,
    app = st_top_charts_impl(
      measure = "revenue",
      os = os,
      time_range = "day",
      date = ranking_date,
      end_date = ranking_date,
      category = if (!is.null(category)) category else if (!is.null(filter_id)) 0 else NULL,
      regions = country,
      limit = limit,
      custom_fields_filter_id = filter_id,
      custom_tags_mode = custom_tags_mode,
      auth_token = auth_token
    ),
    publisher = {
      if (!is.null(filter_id)) {
        rlang::abort("`filter` is not supported for `entity = 'publisher'`.")
      }
      .st_rankings_fetch_publishers(
        os = os,
        category = category %||% 0,
        country = country,
        date = ranking_date,
        limit = limit,
        auth_token = auth_token
      )
    },
    category = st_category_rankings_impl(
      os = os,
      category = category,
      chart_type = chart_type,
      country = country,
      date = ranking_date,
      limit = limit,
      custom_fields_filter_id = filter_id,
      custom_tags_mode = custom_tags_mode,
      auth_token = auth_token
    )
  )

  .st_rankings_standardize(
    data = raw,
    entity = entity,
    os = os,
    category = category,
    country = country,
    date = ranking_date
  )
}

.st_rankings_fetch_publishers <- function(os,
                                          category,
                                          country,
                                          date,
                                          limit,
                                          auth_token) {
  page_size <- 10L
  offsets <- seq.int(0L, max(limit - 1L, 0L), by = page_size)

  pages <- lapply(offsets, function(offset) {
    st_top_publishers_impl(
      measure = "revenue",
      os = os,
      category = category,
      time_range = "day",
      comparison_attribute = "absolute",
      date = date,
      country = country,
      limit = min(page_size, limit - offset),
      offset = offset,
      include_apps = FALSE,
      auth_token = auth_token
    )
  })

  dplyr::bind_rows(pages)
}

.st_rankings_standardize <- function(data, entity, os, category, country, date) {
  if (is.null(data) || !nrow(data)) {
    return(
      tibble::tibble(
        rank = integer(),
        id = character(),
        name = character(),
        os = character(),
        category = character(),
        country = character(),
        date = as.Date(character())
      )
    )
  }

  data <- tibble::as_tibble(data)

  id <- switch(
    entity,
    app = dplyr::coalesce(
      .st_chr_col(data, "unified_app_id"),
      .st_chr_col(data, "app_id")
    ),
    publisher = dplyr::coalesce(
      .st_chr_col(data, "publisher_id"),
      .st_chr_col(data, "unified_publisher_id"),
      .st_chr_col(data, "app_id")
    ),
    category = dplyr::coalesce(
      .st_chr_col(data, "app_id"),
      .st_chr_col(data, "unified_app_id")
    )
  )

  name <- switch(
    entity,
    app = dplyr::coalesce(
      .st_chr_col(data, "app_name"),
      .st_chr_col(data, "unified_app_name"),
      .st_chr_col(data, "name")
    ),
    publisher = dplyr::coalesce(
      .st_chr_col(data, "publisher_name"),
      .st_chr_col(data, "unified_publisher_name"),
      .st_chr_col(data, "name")
    ),
    category = dplyr::coalesce(
      .st_chr_col(data, "app_name"),
      .st_chr_col(data, "unified_app_name"),
      .st_chr_col(data, "name")
    )
  )

  standardized <- tibble::tibble(
    rank = if ("rank" %in% names(data)) as.integer(data$rank) else seq_len(nrow(data)),
    id = id,
    name = name,
    os = rep(os, nrow(data)),
    category = rep(as.character(category %||% NA), nrow(data)),
    country = rep(country, nrow(data)),
    date = rep(as.Date(date), nrow(data))
  )

  drop_cols <- c(
    "rank", "app_id", "unified_app_id", "name", "app_name", "unified_app_name",
    "publisher_id", "publisher_name", "unified_publisher_id",
    "unified_publisher_name", "os", "category", "country", "date"
  )

  extras <- data[, setdiff(names(data), drop_cols), drop = FALSE]
  dplyr::bind_cols(standardized, extras)
}

.st_chr_col <- function(data, column) {
  if (!column %in% names(data)) {
    return(rep(NA_character_, nrow(data)))
  }

  as.character(data[[column]])
}

# Test OS parameter behavior on v1.0.0 unified functions

test_that("st_metrics validates OS parameter values", {
  expect_error(
    st_metrics(
      app_id     = "5ba4585f539ce75b97db6bcb",
      os         = "invalid",
      countries  = "US",
      auth_token = "dummy"
    ),
    "must be one of"
  )
})

test_that("st_metrics accepts the three allowed OS values", {
  # Each `os` value is accepted at the arg layer. Further execution may
  # succeed (return a tibble) or fail (network/auth) — we only care that
  # the "must be one of" validation path isn't triggered.
  for (v in c("ios", "android", "unified")) {
    res <- tryCatch(
      st_metrics(
        app_id     = "5ba4585f539ce75b97db6bcb",
        os         = v,
        countries  = "US",
        date_from  = "2024-01-01",
        date_to    = "2024-01-01",
        auth_token = "dummy"
      ),
      error = function(e) e
    )
    msg <- if (inherits(res, "condition")) conditionMessage(res) else ""
    expect_false(grepl("must be one of", msg))
  }
})

test_that("st_rankings validates entity and os", {
  expect_error(
    st_rankings(
      entity     = "bogus",
      os         = "ios",
      category   = 6000,
      country    = "US",
      auth_token = "dummy"
    )
  )
  expect_error(
    st_rankings(
      entity     = "app",
      os         = "invalid",
      category   = 6000,
      country    = "US",
      auth_token = "dummy"
    ),
    "must be one of"
  )
})

test_that("ID resolution works based on OS parameter", {
  skip_on_cran()
  skip_if_no_auth()
  # Live-API test. Skipped in CRAN checks.
})

test_that("API calls retrieve data", {
  skip_on_cran()
  skip_if_no_auth()

  auth_token <- Sys.getenv("SENSORTOWER_AUTH_TOKEN")
  # st_apps() - search by name (replaces st_app_info)
  apps <- st_apps(
    query = "Pokemon",
    os = "unified",
    limit = 1,
    auth_token = auth_token
  )
  expect_s3_class(apps, "tbl_df")
  expect_true(nrow(apps) > 0)

  # st_publisher_apps
  publisher_apps <- st_publisher_apps(
    auth_token = auth_token,
    publisher_id = "560c48b48ac350643900b82d" # Supercell
  )
  expect_s3_class(publisher_apps, "tbl_df")
  expect_true(nrow(publisher_apps) > 0)

  # st_metrics (new v1.0.0 signature)
  metrics <- st_metrics(
    app_id      = "602c795c912b51622f233ffe", # Pokemon GO
    os          = "unified",
    countries   = "US",
    date_from   = "2021-09-22",
    date_to     = "2021-09-22",
    granularity = "daily",
    auth_token  = auth_token
  )
  expect_s3_class(metrics, "tbl_df")

  # st_rankings (replaces st_top_charts)
  top_ios <- st_rankings(
    entity     = "app",
    os         = "ios",
    category   = 6000,
    country    = "US",
    date       = "2023-10-01",
    limit      = 5,
    auth_token = auth_token
  )
  expect_s3_class(top_ios, "tbl_df")
})

test_that("clean_numeric_values removes special characters correctly", {
  # Create test data with special characters in numeric columns
  test_data <- tibble::tibble(
    downloads_180d_ww = c("1,234,567", "$2,500", "45%", "100.5", NA),
    revenue_30d_us = c("$1,000.50", "2,500%", "45", "", "N/A"),
    retention_7d_us = c("15.5%", "25%", "0", "45.2%", NA),
    aggregate_tags.test_metric = c("$100", "200%", "1,500", "", NA),
    app_name = c("Test App", "Another App", "Game", "Tool", "App"),
    category = c("Games", "Social", "Productivity", "Entertainment", "Utilities")
  )

  cleaned_data <- sensortowerR:::clean_numeric_values(test_data)

  expect_true(is.numeric(cleaned_data$downloads_180d_ww))
  expect_true(is.numeric(cleaned_data$revenue_30d_us))
  expect_true(is.numeric(cleaned_data$retention_7d_us))
  expect_true(is.numeric(cleaned_data$`aggregate_tags.test_metric`))

  expect_true(is.character(cleaned_data$app_name))
  expect_true(is.character(cleaned_data$category))

  expect_equal(cleaned_data$downloads_180d_ww[1], 1234567)
  expect_equal(cleaned_data$downloads_180d_ww[2], 2500)
  expect_equal(cleaned_data$downloads_180d_ww[3], 45)
  expect_equal(cleaned_data$revenue_30d_us[1], 1000.50)
  expect_equal(cleaned_data$retention_7d_us[1], 0.155)
  expect_equal(cleaned_data$retention_7d_us[2], 0.25)

  expect_true(is.na(cleaned_data$downloads_180d_ww[5]))
  expect_true(is.na(cleaned_data$retention_7d_us[5]))
})

test_that("st_rankings rejects invalid entity", {
  expect_error(
    st_rankings(
      entity     = "bogus",
      os         = "ios",
      category   = 6000,
      country    = "US",
      auth_token = "dummy"
    )
  )
})

test_that("st_rankings rejects invalid os", {
  expect_error(
    st_rankings(
      entity     = "app",
      os         = "nope",
      category   = 6000,
      country    = "US",
      auth_token = "dummy"
    ),
    "must be one of"
  )
})

test_that("st_rankings accepts the three allowed entities", {
  # Each errors at the network layer, not the arg layer
  for (e in c("app", "publisher", "category")) {
    res <- tryCatch(
      st_rankings(
        entity     = e,
        os         = "ios",
        category   = 6000,
        country    = "US",
        date       = "2024-01-01",
        auth_token = "dummy"
      ),
      error = function(err) err
    )
    # The entity passed validation if we did NOT get an entity-arg error
    msg <- conditionMessage(res) %||% ""
    expect_false(grepl("'arg' should be one of", msg))
  }
})

#' Fetch Retention Metrics from Sensor Tower's Facets API
#'
#' Retrieves retention curves from Sensor Tower's new facets-based retention
#' endpoint. This wrapper targets `"/v1/facets/metrics?facets=retention"` and
#' returns a tidy tibble with the live response columns preserved.
#'
#' @param app_ids Character vector of platform-specific app IDs. May contain iOS
#'   numeric IDs or Android bundle IDs. Supply exactly one of `app_ids` or
#'   `unified_app_ids`.
#' @param unified_app_ids Character vector of Sensor Tower unified app IDs.
#'   Supply exactly one of `app_ids` or `unified_app_ids`.
#' @param bundle Character string. Retention bundle to request:
#'   `"retention_daily"`, `"retention_weekly"`, or `"retention_monthly"`.
#' @param breakdown Character vector of breakdown fields. Supported combinations
#'   are the ones documented by Sensor Tower:
#'   `"date"`, `"app_id"`, `"unified_app_id"`, `"date,app_id"`,
#'   `"date,unified_app_id"`, `"unified_app_id,app_id"`, and
#'   `"date,unified_app_id,app_id"`.
#' @param start_date Start date in `YYYY-MM-DD` format or as `Date`.
#' @param end_date End date in `YYYY-MM-DD` format or as `Date`.
#' @param regions Optional character vector of region codes. When omitted,
#'   Sensor Tower returns worldwide estimates.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'   Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.
#' @param verbose Logical. If `TRUE`, prints the request URL with the auth token
#'   redacted.
#'
#' @return A [tibble][tibble::tibble] with one row per requested breakdown
#'   combination. The response preserves Sensor Tower's live metric column names
#'   such as `est_retention_d1`, `est_retention_d14`, `est_retention_w52`, or
#'   `est_retention_m12`.
#'
#' @details
#' Validated against the live Sensor Tower API on March 17, 2026. The production
#' response currently returns daily retention columns including
#' `est_retention_d14` and `est_retention_d365`.
#'
#' @examples
#' \dontrun{
#' retention <- st_retention_facets(
#'   app_ids = "553834731",
#'   bundle = "retention_daily",
#'   breakdown = c("date", "app_id"),
#'   start_date = "2025-01-01",
#'   end_date = "2025-01-31"
#' )
#' }
#'
#' @seealso [st_retention()] for the legacy retention endpoint,
#'   [st_facets_metrics()] for raw facets access
#'
#' @export
st_retention_facets <- function(app_ids = NULL,
                                unified_app_ids = NULL,
                                bundle = c("retention_daily", "retention_weekly", "retention_monthly"),
                                breakdown = c("date", "app_id"),
                                start_date,
                                end_date,
                                regions = NULL,
                                auth_token = NULL,
                                verbose = FALSE) {
  bundle <- match.arg(bundle)
  start_date <- as.Date(start_date)
  end_date <- as.Date(end_date)

  if (is.na(start_date) || is.na(end_date)) {
    rlang::abort("`start_date` and `end_date` must be valid dates.")
  }

  if (end_date < start_date) {
    rlang::abort("`end_date` must be on or after `start_date`.")
  }

  has_app_ids <- !is.null(app_ids) && length(app_ids) > 0
  has_unified_ids <- !is.null(unified_app_ids) && length(unified_app_ids) > 0

  if ((has_app_ids + has_unified_ids) != 1) {
    rlang::abort("Exactly one of `app_ids` or `unified_app_ids` must be supplied.")
  }

  breakdown <- normalize_retention_facets_breakdown(breakdown)

  params <- list(
    facets = "retention",
    bundle = bundle,
    breakdown = breakdown,
    start_date = start_date,
    end_date = end_date,
    regions = regions,
    app_ids = app_ids,
    unified_app_ids = unified_app_ids
  )

  response <- st_facets_metrics(
    params = params,
    auth_token = auth_token,
    verbose = verbose
  )

  result <- coerce_facets_data_tibble(response)

  if (nrow(result) == 0) {
    return(result)
  }

  if ("date" %in% names(result)) {
    result$date <- as.Date(result$date)
  }

  id_cols <- intersect(c("app_id", "unified_app_id"), names(result))
  for (col in id_cols) {
    result[[col]] <- as.character(result[[col]])
  }

  retention_cols <- grep("^est_retention_", names(result), value = TRUE)
  for (col in retention_cols) {
    result[[col]] <- as.numeric(result[[col]])
  }

  sort_cols <- intersect(c("date", "unified_app_id", "app_id"), names(result))
  if (length(sort_cols) > 0) {
    order_index <- do.call(order, c(result[sort_cols], list(na.last = TRUE)))
    result <- result[order_index, , drop = FALSE]
  }

  tibble::as_tibble(result)
}

normalize_retention_facets_breakdown <- function(breakdown) {
  breakdown <- trimws(tolower(as.character(breakdown)))
  breakdown <- unique(breakdown[nzchar(breakdown)])

  valid_breakdowns <- c("date", "app_id", "unified_app_id")
  invalid_breakdowns <- setdiff(breakdown, valid_breakdowns)

  if (length(breakdown) == 0) {
    rlang::abort("`breakdown` must contain at least one supported field.")
  }

  if (length(invalid_breakdowns) > 0) {
    rlang::abort(
      sprintf(
        "Unsupported `breakdown` field(s): %s.",
        paste(invalid_breakdowns, collapse = ", ")
      )
    )
  }

  ordered_breakdown <- c("date", "unified_app_id", "app_id")
  breakdown <- ordered_breakdown[ordered_breakdown %in% breakdown]

  valid_combinations <- c(
    "date",
    "app_id",
    "unified_app_id",
    "date,app_id",
    "date,unified_app_id",
    "unified_app_id,app_id",
    "date,unified_app_id,app_id"
  )

  breakdown_key <- paste(breakdown, collapse = ",")
  if (!breakdown_key %in% valid_combinations) {
    rlang::abort(
      sprintf(
        "Unsupported `breakdown` combination: %s.",
        breakdown_key
      )
    )
  }

  breakdown
}

coerce_facets_data_tibble <- function(response) {
  if (inherits(response, "data.frame")) {
    return(tibble::as_tibble(response))
  }

  if (!is.list(response) || is.null(response$data)) {
    return(tibble::tibble())
  }

  data <- response$data

  if (inherits(data, "data.frame")) {
    return(tibble::as_tibble(data))
  }

  data_json <- jsonlite::toJSON(data, auto_unbox = TRUE, null = "null")
  tibble::as_tibble(jsonlite::fromJSON(data_json, flatten = TRUE))
}

#' Get All Apps from a Publisher
#'
#' Retrieves a list of apps associated with a specified unified publisher ID
#' from the Sensor Tower API. Targets the `/v1/unified/publishers/apps`
#' endpoint.
#'
#' @param unified_id Character. Unified ID to resolve apps for. May be either:
#'   - Unified Publisher ID (24-char hex)
#'   - Unified App ID (24-char hex) belonging to a publisher
#'   The API returns the unified publisher and all associated apps in both cases.
#' @param publisher_id Deprecated alias for `unified_id`.
#' @param aggregate_related Logical. If TRUE, ensures each app's unified_app_id
#'   is the canonical ID that aggregates ALL regional SKUs. This solves the problem
#'   where games like "Watcher of Realms" are published under multiple regional
#'   publishers (Moonton, Vizta Games, Skystone Games, etc.) and may return
#'   different unified_app_ids. When TRUE, the function looks up each app by name
#'   to find the true unified_app_id that combines all regional versions.
#'   Defaults to FALSE for backwards compatibility.
#' @param auth_token Character. Your Sensor Tower API authentication token.
#'   Defaults to the value stored in the `SENSORTOWER_AUTH_TOKEN` environment
#'   variable.
#' @param verbose Logical. If TRUE, prints progress messages during aggregation.
#'   Defaults to TRUE.
#'
#' @return A [tibble][tibble::tibble] containing details of the apps associated
#'   with the publisher. The exact columns depend on the API response but often
#'   include app IDs, names, platform, etc. Returns an empty tibble if the
#'   publisher ID is invalid, has no apps, or an error occurs.
#'
#' @section Solving Regional Publisher Issues:
#' Many publishers have regional subsidiaries or partners that publish the same
#' game under different app IDs in different regions. For example, Moonton's
#' "Watcher of Realms" is published by Moonton in some regions, Vizta Games in
#' others, and Skystone Games in others.
#'
#' When `aggregate_related = TRUE`, this function ensures you get the unified_app_id
#' that represents the FULL game across all regional publishers, which is required
#' for accurate revenue/download aggregation via `st_unified_sales_report()`.
#'
#' @section API Endpoint Used:
#'   - `GET /v1/unified/publishers/apps`
#'
#' @examples
#' \dontrun{
#' # Ensure SENSORTOWER_AUTH_TOKEN is set in your environment
#' # Sys.setenv(SENSORTOWER_AUTH_TOKEN = "your_secure_auth_token_here")
#'
#' # Basic usage - get publisher's apps
#' apps_list <- st_publisher_apps(unified_id = "647eb849d9d91f31a54f1792")
#'
#' # With regional SKU aggregation - ensures canonical unified_app_ids
#' apps_list <- st_publisher_apps(
#'   unified_id = "647eb849d9d91f31a54f1792",
#'   aggregate_related = TRUE
#' )
#'
#' # Then use with st_unified_sales_report() for accurate data
#' sales <- st_unified_sales_report(
#'   unified_app_id = apps_list$unified_app_id,
#'   countries = "WW",
#'   start_date = "2024-01-01",
#'   end_date = "2024-12-31",
#'   date_granularity = "monthly"
#' )
#' }
#'
#' @import dplyr
#' @importFrom httr GET add_headers stop_for_status content http_status
#'   http_error
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble as_tibble
#' @importFrom rlang abort warn `%||%`
#' @export
st_publisher_apps <- function(unified_id = NULL,
                              publisher_id = NULL,
                              aggregate_related = FALSE,
                              auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                              verbose = TRUE) {
  # --- Input Validation & Setup ---
  # Backward compatibility: support old signature
  if (is.null(unified_id) && !is.null(publisher_id)) unified_id <- publisher_id
  if (is.null(unified_id)) {
    rlang::abort("unified_id is required (24-char hex unified publisher or app id)")
  }
  if (!is.character(unified_id)) {
    rlang::abort("unified_id must be a character string")
  }
  if (length(unified_id) != 1) {
    rlang::abort("unified_id must be a single value, not a vector")
  }
  if (!nzchar(unified_id)) {
    rlang::abort("unified_id cannot be an empty string")
  }
  # Basic format check: 24-char hex
  if (!grepl("^[a-f0-9]{24}$", unified_id)) {
    rlang::abort(paste0(
      "Invalid unified_id format: '", unified_id, "'. Expected 24-char hex (unified publisher or app id)."
    ))
  }

  auth_token <- resolve_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required.",
      "Set SENSORTOWER_AUTH_TOKEN environment variable",
      "or pass via auth_token argument."
    )
  )

  url <- paste0(st_api_base_url(), "/", st_endpoint_path("unified_publishers_apps"))

  query_params <- list(
    unified_id = unified_id,
    auth_token = auth_token
  )

  # --- API Call ---
  response <- tryCatch(
    httr::GET(
      url = url,
      query = query_params,
      httr::add_headers("Accept" = "application/json")
    ),
    error = function(e) {
      rlang::abort(paste("HTTP request failed:", e$message))
    }
  )

  # --- Response Handling ---
  status_code <- tryCatch(httr::status_code(response), error = function(e) NA_integer_)
  if (is.na(status_code) || status_code >= 400) {
    body_text <- tryCatch(httr::content(response, "text", encoding = "UTF-8"), error = function(e) "")
    msg <- paste0("API request failed (HTTP ", status_code, ")")
    if (nzchar(body_text)) {
      # Try extracting structured error
      parsed_err <- tryCatch(jsonlite::fromJSON(body_text), error = function(e) NULL)
      if (!is.null(parsed_err) && !is.null(parsed_err$error)) {
        msg <- paste0(msg, ": ", parsed_err$error)
      }
    }
    rlang::abort(msg)
  }

  content_text <- httr::content(response, "text", encoding = "UTF-8")
  if (content_text == "") {
    rlang::warn("API returned an empty response body.")
    return(tibble::tibble())
  }
  parsed_data <- jsonlite::fromJSON(content_text, flatten = TRUE)

  # --- Data Extraction & Processing ---
  if (is.null(parsed_data) || !"apps" %in% names(parsed_data)) {
    rlang::warn(
      "API response did not contain an 'apps' list for the given unified_id."
    )
    return(tibble::tibble())
  }

  apps_data <- parsed_data$apps

  if (length(apps_data) == 0) {
    message("No apps found associated with the publisher ID.")
    return(tibble::tibble())
  }

  result <- tibble::as_tibble(apps_data)


  # --- Aggregate Related SKUs ---
  if (aggregate_related && nrow(result) > 0) {
    if (verbose) {
      message("Resolving canonical unified_app_ids for ", nrow(result), " apps...")
    }

    # Get app names - handle different possible column names
    name_col <- intersect(c("unified_app_name", "name", "app_name"), names(result))[1]
    id_col <- intersect(c("unified_app_id", "app_id"), names(result))[1]

    if (is.na(name_col) || is.na(id_col)) {
      rlang::warn("Could not find name/id columns for aggregation. Returning raw results.")
      return(result)
    }

    # Look up each app by name to get canonical unified_app_id
    canonical_ids <- list()
    for (i in seq_len(nrow(result))) {
      app_name <- result[[name_col]][i]
      original_id <- result[[id_col]][i]

      # Search for the app by name
      search_result <- tryCatch({
        st_app_info_impl(
          term = app_name,
          app_store = "unified",
          entity_type = "app",
          limit = 5,
          auth_token = auth_token,
          return_all_fields = FALSE
        )
      }, error = function(e) {
        if (verbose) message("  Warning: Could not search for '", app_name, "': ", e$message)
        NULL
      })

      # Find best match - prefer exact name match
      canonical_id <- original_id
      if (!is.null(search_result) && nrow(search_result) > 0) {
        # Try exact match first
        exact_match <- search_result %>%
          dplyr::filter(tolower(.data$unified_app_name) == tolower(app_name))

        if (nrow(exact_match) > 0) {
          canonical_id <- exact_match$unified_app_id[1]
        } else {
          # Use first result as best match
          canonical_id <- search_result$unified_app_id[1]
        }

        if (canonical_id != original_id && verbose) {
          message("  '", app_name, "': ", original_id, " -> ", canonical_id)
        }
      }

      canonical_ids[[i]] <- list(
        original_id = original_id,
        canonical_id = canonical_id,
        app_name = app_name
      )

      Sys.sleep(0.2)  # Rate limiting
    }

    # Update result with canonical IDs and preserve mapping
    canonical_df <- dplyr::bind_rows(canonical_ids)

    # Store original ID before replacing
    result$original_unified_app_id <- result[[id_col]]
    result[[id_col]] <- canonical_df$canonical_id

    # Add canonical_id column for apps that were remapped
    result$canonical_id <- ifelse(
      result$original_unified_app_id != result[[id_col]],
      result[[id_col]],
      NA_character_
    )

    # Deduplicate by canonical unified_app_id (keep first occurrence with name)
    n_before <- nrow(result)
    result <- result %>%
      dplyr::distinct(dplyr::across(dplyr::all_of(id_col)), .keep_all = TRUE)
    n_after <- nrow(result)

    if (verbose && n_before != n_after) {
      message("Deduplicated: ", n_before, " -> ", n_after, " unique apps")
    }

    # Create a canonical ID -> name lookup for use by other functions
    # This includes ALL canonical IDs (including those from remapped apps)
    name_lookup <- canonical_df %>%
      dplyr::select(unified_app_id = canonical_id, app_name) %>%
      dplyr::distinct(unified_app_id, .keep_all = TRUE)

    attr(result, "canonical_name_lookup") <- name_lookup

    if (verbose) {
      message("Aggregation complete. ", nrow(result), " apps with canonical IDs.")
    }
  }

  return(result)
}


#' Get App Names from Publisher Apps Result
#'
#' Helper function to create a name lookup table from the result of
#' `st_publisher_apps()`. This handles canonical ID mapping automatically,
#' so you can join sales data (which uses canonical IDs) back to app names.
#'
#' @param apps_df A tibble returned by `st_publisher_apps()`.
#' @param include_canonical Logical. If TRUE, includes mappings for canonical IDs
#'   that were resolved during aggregation. Defaults to TRUE.
#'
#' @return A tibble with columns `unified_app_id` and `app_name` suitable for
#'   joining with sales data or other API results.
#'
#' @examples
#' \dontrun{
#' # Get apps with canonical ID resolution
#' apps <- st_publisher_apps("647eb849d9d91f31a54f1792", aggregate_related = TRUE)
#'
#' # Get name lookup table
#' name_lookup <- st_get_app_names(apps)
#'
#' # Use with sales data
#' sales <- st_unified_sales_report(apps$unified_app_id, ...)
#' sales_with_names <- sales %>%
#'   left_join(name_lookup, by = "unified_app_id")
#' }
#'
#' @export
st_get_app_names <- function(apps_df, include_canonical = TRUE) {
  if (!is.data.frame(apps_df)) {
    rlang::abort("apps_df must be a data frame from st_publisher_apps()")
  }

  # Determine column names
  id_col <- intersect(c("unified_app_id", "app_id"), names(apps_df))[1]
  name_col <- intersect(c("unified_app_name", "name", "app_name"), names(apps_df))[1]


  if (is.na(id_col) || is.na(name_col)) {
    rlang::abort("Could not find unified_app_id and app_name columns in apps_df")
  }

  # Start with the main mapping
  name_lookup <- apps_df %>%
    dplyr::select(
      unified_app_id = dplyr::all_of(id_col),
      app_name = dplyr::all_of(name_col)
    ) %>%
    dplyr::distinct()

  # Add canonical ID mappings if available
  if (include_canonical) {
    canonical_lookup <- attr(apps_df, "canonical_name_lookup")
    if (!is.null(canonical_lookup) && nrow(canonical_lookup) > 0) {
      name_lookup <- dplyr::bind_rows(name_lookup, canonical_lookup) %>%
        dplyr::distinct(unified_app_id, .keep_all = TRUE)
    }
  }

  return(name_lookup)
}

#' Fetch Detailed App Metadata
#'
#' Retrieves comprehensive metadata for one or more apps including descriptions,
#' screenshots, ratings, publisher information, and more. This function provides
#' rich app store listing data for apps when you already know their IDs.
#'
#' @param app_ids Character vector. App IDs to fetch details for. 
#'   - For iOS: numeric app IDs (e.g., "553834731")
#'   - For Android: bundle IDs (e.g., "com.king.candycrushsaga")
#'   - For unified: unified app IDs
#'   Maximum 100 apps per request.
#' @param os Character string. Required. Operating system: "ios", "android", or "unified".
#' @param include_developer_contacts Logical. Include developer contact information
#'   (email, address). Defaults to TRUE.
#' @param auth_token Character string. Sensor Tower API authentication token.
#'   Defaults to environment variable SENSORTOWER_AUTH_TOKEN.
#'
#' @return A [tibble][tibble::tibble] containing detailed app metadata with columns:
#'   - `app_id`: The app's store ID
#'   - `app_name`: The app's display name
#'   - `publisher_name`: Publisher/developer name
#'   - `publisher_id`: Publisher ID
#'   - `categories`: App store categories
#'   - `description`: Full app description
#'   - `subtitle`: App subtitle (iOS) or short description (Android)
#'   - `rating`: Current average rating
#'   - `rating_count`: Total number of ratings
#'   - `rating_current_version`: Rating for current version
#'   - `rating_count_current_version`: Rating count for current version
#'   - `content_rating`: Age rating/content rating
#'   - `price`: App price
#'   - `currency`: Price currency
#'   - `release_date`: Initial release date
#'   - `last_update`: Last update date
#'   - `version`: Current version
#'   - `size_bytes`: App size in bytes
#'   - `screenshots`: List of screenshot URLs
#'   - `icon_url`: App icon URL
#'   - `publisher_email`: Developer email (if include_developer_contacts = TRUE)
#'   - `publisher_address`: Developer address (if include_developer_contacts = TRUE)
#'   - `publisher_country`: Developer country
#'   - `ios_app_id`: Primary iOS app ID (for `os = "unified"`, when available)
#'   - `android_app_id`: Primary Android app ID (for `os = "unified"`, when available)
#'   - Additional platform-specific fields
#'
#' @section API Endpoint Used:
#'   - `GET /v1/\{os\}/apps`
#'
#' @examples
#' \dontrun{
#' # Get details for a single iOS app
#' candy_crush <- st_app_details(
#'   app_ids = "553834731",
#'   os = "ios"
#' )
#'
#' # Get details for multiple Android apps
#' android_games <- st_app_details(
#'   app_ids = c("com.king.candycrushsaga", "com.supercell.clashofclans"),
#'   os = "android"
#' )
#'
#' # Get details without developer contacts
#' apps <- st_app_details(
#'   app_ids = c("553834731", "1053012308"),
#'   include_developer_contacts = FALSE
#' )
#' }
#'
#' @importFrom rlang %||% abort
#' @importFrom httr2 resp_body_raw
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble as_tibble as_tibble_row
#' @importFrom dplyr rename bind_rows
#' @keywords internal
#' @noRd
st_app_details_impl <- function(app_ids,
                          os,
                          include_developer_contacts = TRUE,
                          auth_token = NULL) {
  
  # Validate OS parameter
  if (missing(os) || is.null(os) || !os %in% c("ios", "android", "unified")) {
    rlang::abort("'os' parameter is required and must be one of: 'ios', 'android', or 'unified'")
  }
  
  # Input validation
  if (missing(app_ids) || length(app_ids) == 0) {
    rlang::abort("At least one app_id is required.")
  }
  
  if (length(app_ids) > 100) {
    rlang::abort("Maximum 100 app IDs allowed per request.")
  }
  
  # Authentication
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Set SENSORTOWER_AUTH_TOKEN environment variable or pass via auth_token argument."
  )
  
  # Build query parameters
  query_params <- list(
    auth_token = auth_token_val,
    app_ids = paste(app_ids, collapse = ",")
  )

  # Unified app details endpoint requires app_id_type.
  if (os == "unified") {
    query_params$app_id_type <- "unified"
  }
  
  if (include_developer_contacts) {
    query_params$attributes = "contact_info"
  }
  
  # Build and perform request
  path <- st_endpoint_segments("apps", os = os)
  req <- build_request(st_api_base_url(), path, query_params)
  resp <- perform_request(req)
  
  # Process response
  result <- process_app_details_response(resp, os)
  
  return(result)
}

#' Process App Details API Response
#'
#' Internal function to process and enrich app details API responses.
#'
#' @param resp Response object from httr2
#' @param os Operating system
#'
#' @return A processed tibble with app details
#' @keywords internal
process_app_details_response <- function(resp, os) {
  
  # Get raw response
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(tibble::tibble())
  }
  
  body_text <- rawToChar(body_raw)
  result <- jsonlite::fromJSON(body_text, flatten = TRUE)
  
  if (length(result) == 0) {
    return(tibble::tibble())
  }
  
  # The API returns a list with an "apps" key containing a data frame
  if (is.list(result) && "apps" %in% names(result)) {
    result_tbl <- tibble::as_tibble(result$apps)
  } else if (is.data.frame(result)) {
    result_tbl <- tibble::as_tibble(result)
  } else {
    # Fallback - shouldn't reach here normally
    return(tibble::tibble())
  }
  
  if (nrow(result_tbl) == 0) {
    return(result_tbl)
  }
  
  # Platform-specific field mappings
  if (os == "ios") {
    # iOS field mappings
    if ("name" %in% names(result_tbl) && !"app_name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app_name = name)
    }
    
    # Handle categories - iOS uses primary_category and category_ids
    if ("primary_category" %in% names(result_tbl) || "category_ids" %in% names(result_tbl)) {
      # Categories might be a complex structure, keep as is for now
    }
    
    # Handle version-specific ratings
    if ("user_rating" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, rating = user_rating)
    }
    if ("user_rating_count" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, rating_count = user_rating_count)
    }
    
  } else if (os == "android") {
    # Android field mappings
    if ("title" %in% names(result_tbl) && !"app_name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app_name = title)
    }
    # Also check for "name" field
    if ("name" %in% names(result_tbl) && !"app_name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app_name = name)
    }
    
    # Android uses score instead of rating
    if ("score" %in% names(result_tbl) && !"rating" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, rating = score)
    }
    
    # Handle developer info
    if ("developer" %in% names(result_tbl) && !"publisher_name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, publisher_name = developer)
    }
    if ("developer_id" %in% names(result_tbl) && !"publisher_id" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, publisher_id = developer_id)
    }
    if ("developer_email" %in% names(result_tbl) && !"publisher_email" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, publisher_email = developer_email)
    }
    if ("developer_address" %in% names(result_tbl) && !"publisher_address" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, publisher_address = developer_address)
    }
  } else if (os == "unified") {
    if ("name" %in% names(result_tbl) && !"app_name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app_name = name)
    }

    if ("unified_app_id" %in% names(result_tbl) && !"app_id" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app_id = unified_app_id)
    }

    extract_first_app_id <- function(x) {
      if (is.null(x)) return(NA_character_)
      if (!is.data.frame(x) || nrow(x) == 0 || !"app_id" %in% names(x)) return(NA_character_)
      as.character(x$app_id[1])
    }

    if ("itunes_apps" %in% names(result_tbl) && !"ios_app_id" %in% names(result_tbl)) {
      result_tbl$ios_app_id <- vapply(result_tbl$itunes_apps, extract_first_app_id, character(1))
    }

    if ("android_apps" %in% names(result_tbl) && !"android_app_id" %in% names(result_tbl)) {
      result_tbl$android_app_id <- vapply(result_tbl$android_apps, extract_first_app_id, character(1))
    }
  }
  
  # Ensure app_id is character type for consistent joining
  if ("app_id" %in% names(result_tbl)) {
    result_tbl$app_id <- as.character(result_tbl$app_id)
  }
  
  # Common field processing
  # Convert dates to Date objects
  date_fields <- c("release_date", "last_update", "current_version_release_date")
  for (field in date_fields) {
    if (field %in% names(result_tbl)) {
      result_tbl[[field]] <- as.Date(result_tbl[[field]])
    }
  }
  
  # Handle numeric fields
  numeric_fields <- c("price", "size_bytes", "rating", "rating_count")
  for (field in numeric_fields) {
    if (field %in% names(result_tbl)) {
      result_tbl[[field]] <- as.numeric(result_tbl[[field]])
    }
  }
  
  # Define preferred column order
  preferred_cols <- c(
    "app_id", "app_name", "publisher_name", "publisher_id",
    "categories", "description", "subtitle", "short_description",
    "rating", "rating_count", "rating_current_version", "rating_count_current_version",
    "content_rating", "price", "currency", "release_date", "last_update",
    "version", "size_bytes", "screenshots", "icon_url",
    "publisher_email", "publisher_address", "publisher_country"
  )
  
  # Reorder columns, keeping any additional fields at the end
  existing_cols <- intersect(preferred_cols, names(result_tbl))
  other_cols <- setdiff(names(result_tbl), preferred_cols)
  
  result_tbl <- result_tbl[, c(existing_cols, other_cols)]
  
  # Add OS information
  result_tbl$os <- os
  
  return(result_tbl)
}

#' Fetch App Store Category Rankings
#'
#' Retrieves the top ranking apps for a specific category and chart type from
#' the App Store or Google Play Store. This provides the official store rankings
#' as they appear in the actual app stores.
#'
#' @param os Character string. Required. Operating system: "ios", "android", or "unified".
#' @param category Character or numeric. Category ID to fetch rankings for.
#'   Use `st_categories()` to find valid category IDs. Required unless 
#'   `custom_fields_filter_id` is provided.
#' @param chart_type Character string. The chart type to retrieve. Options vary by OS:
#'   - iOS: "topfreeapplications", "toppaidapplications", "topgrossingapplications", etc.
#'   - Android: "topselling_free", "topselling_paid", "topgrossing", etc.
#'   Defaults to "topfreeapplications" for iOS, "topselling_free" for Android.
#' @param country Character string. Two-letter country code (e.g., "US", "GB").
#'   Defaults to "US".
#' @param date Date or character string in "YYYY-MM-DD" format. Date for rankings.
#'   Defaults to NULL (uses today's date).
#' @param limit Numeric. Number of results to return (1-400). Defaults to 100.
#' @param offset Numeric. Offset for pagination. Defaults to 0.
#' @param custom_fields_filter_id Optional. Character string. ID of a Sensor
#'   Tower custom field filter to apply. Use filter IDs from the web interface
#'   at app.sensortower.com. When provided, this filter will be applied to the
#'   results. The 'category' parameter becomes optional when using a custom filter.
#' @param custom_tags_mode Optional. Character string. Required if `os` is
#'   'unified' and `custom_fields_filter_id` is provided. Specifies how the
#'   custom filter applies to unified apps. Options: "include", "exclude",
#'   "include_unified_apps". The "include_unified_apps" option includes all
#'   platform versions when any version matches the filter.
#' @param auth_token Character string. Sensor Tower API authentication token.
#'   Defaults to environment variable SENSORTOWER_AUTH_TOKEN.
#'
#' @return A [tibble][tibble::tibble] containing ranking data with columns:
#'   - `rank`: The app's position in the chart
#'   - `app_id`: The app's store ID
#'   - `category`: The category ID
#'   - `country`: The country code
#'   - `date`: The ranking date
#'   - `chart_type`: The chart type
#'   - `os`: The operating system
#'   
#' @note The API returns only app IDs, not names. To get app names and other
#'   metadata, use the app IDs with `st_app_details()`.
#'
#' @section API Endpoint Used:
#'   - `GET /v1/\{os\}/ranking`
#'
#' @examples
#' \dontrun{
#' # Get top free games in the US
#' top_games <- st_category_rankings(
#'   os = "ios",
#'   category = 6014,  # Games category
#'   chart_type = "topfreeapplications",
#'   country = "US",
#'   limit = 50
#' )
#'
#' # Get top grossing apps in UK for a specific date
#' top_grossing <- st_category_rankings(
#'   os = "android",
#'   category = "game",
#'   chart_type = "topgrossing",
#'   country = "GB",
#'   date = "2024-01-15",
#'   limit = 100
#' )
#'
#' # Use custom filter instead of category
#' filtered_rankings <- st_category_rankings(
#'   os = "ios",
#'   custom_fields_filter_id = "60746340241bc16eb8a65d76",
#'   chart_type = "topgrossingapplications",
#'   country = "US",
#'   limit = 50
#' )
#'
#' # With unified OS and custom filter
#' unified_rankings <- st_category_rankings(
#'   os = "unified",
#'   custom_fields_filter_id = "60746340241bc16eb8a65d76",
#'   custom_tags_mode = "include_unified_apps",
#'   chart_type = "topfreeapplications"
#' )
#' }
#'
#' @importFrom rlang %||% abort
#' @importFrom httr2 resp_body_raw
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr rename
#' @keywords internal
#' @noRd
st_category_rankings_impl <- function(os,
                                category = NULL,
                                chart_type = NULL,
                                country = "US",
                                date = NULL,
                                limit = 100,
                                offset = 0,
                                custom_fields_filter_id = NULL,
                                custom_tags_mode = NULL,
                                auth_token = NULL) {
  
  # Validate OS parameter
  if (missing(os) || is.null(os) || !os %in% c("ios", "android", "unified")) {
    rlang::abort("'os' parameter is required and must be one of: 'ios', 'android', or 'unified'")
  }
  
  # Load custom filter utilities if available
  if (exists("validate_custom_filter_params", mode = "function")) {
    validate_custom_filter_params(
      custom_fields_filter_id = custom_fields_filter_id,
      custom_tags_mode = custom_tags_mode,
      os = os,
      require_category = TRUE,
      category = category
    )
  } else {
    # Fallback validation if utilities not loaded
    if (is.null(category) && is.null(custom_fields_filter_id)) {
      rlang::abort("Either 'category' or 'custom_fields_filter_id' parameter is required.")
    }
  }
  
  # Set default chart type based on OS
  if (is.null(chart_type)) {
    chart_type <- switch(os,
      ios = "topfreeapplications",
      android = "topselling_free",
      unified = "topfreeapplications"
    )
  }
  
  # Validate limit
  if (limit < 1 || limit > 400) {
    rlang::abort("Limit must be between 1 and 400.")
  }
  
  # Authentication
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Set SENSORTOWER_AUTH_TOKEN environment variable or pass via auth_token argument."
  )
  
  # Build query parameters
  query_params <- list(
    auth_token = auth_token_val,
    chart_type = chart_type,
    country = country,
    limit = limit,
    offset = offset
  )
  
  # Only add category if provided
  if (!is.null(category)) {
    query_params$category <- as.character(category)
  }
  
  # Add date - API requires it even though documentation suggests optional
  if (!is.null(date)) {
    if (inherits(date, "Date")) {
      date <- format(date, "%Y-%m-%d")
    }
    query_params$date = date
  } else {
    # Default to today's date
    query_params$date = format(Sys.Date(), "%Y-%m-%d")
  }
  
  # Add custom filter parameters if available
  if (exists("add_custom_filter_params", mode = "function")) {
    query_params <- add_custom_filter_params(
      query_params,
      custom_fields_filter_id = custom_fields_filter_id,
      custom_tags_mode = custom_tags_mode,
      os = os
    )
  } else if (!is.null(custom_fields_filter_id)) {
    # Fallback if utilities not loaded
    query_params$custom_fields_filter_id <- custom_fields_filter_id
    if (!is.null(custom_tags_mode)) {
      query_params$custom_tags_mode <- custom_tags_mode
    }
  }
  
  # Build and perform request
  path <- st_endpoint_segments("ranking", os = os)
  req <- build_request(st_api_base_url(), path, query_params)
  resp <- perform_request(req)
  
  # Process response
  result <- process_ranking_response(resp, os, category, country, chart_type, date, limit)
  
  return(result)
}

#' Process Category Ranking API Response
#'
#' Internal function to process and enrich category ranking API responses.
#'
#' @param resp Response object from httr2
#' @param os Operating system
#' @param category Category ID used in request
#' @param country Country code used in request
#' @param chart_type Chart type used in request
#' @param date Date used in request (may be NULL)
#' @param limit Number of results requested
#'
#' @return A processed tibble with ranking data
#' @keywords internal
process_ranking_response <- function(resp, os, category, country, chart_type, date, limit) {
  
  # Get raw response
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(tibble::tibble())
  }
  
  body_text <- rawToChar(body_raw)
  result <- jsonlite::fromJSON(body_text, flatten = TRUE)
  
  if (length(result) == 0) {
    return(tibble::tibble())
  }
  
  # Convert to tibble
  result_tbl <- tibble::as_tibble(result)
  
  if (nrow(result_tbl) == 0) {
    return(result_tbl)
  }
  
  # Add ranking position (API returns ordered list)
  result_tbl$rank <- seq_len(nrow(result_tbl))
  
  # Reorder columns for better readability
  col_order <- c("rank")
  
  # The API returns a simple array of app IDs in the 'ranking' field
  # We need to restructure this into a more useful format
  if ("ranking" %in% names(result_tbl)) {
    # Extract app IDs from the ranking column
    app_ids <- result_tbl$ranking
    
    # Create a new tibble with proper structure
    result_tbl <- tibble::tibble(
      rank = result_tbl$rank,
      app_id = as.character(app_ids)
    )
  }
  
  # Add context columns
  result_tbl$category <- category
  result_tbl$country <- country
  result_tbl$chart_type <- chart_type
  result_tbl$os <- os
  
  # Add date if available
  if (!is.null(date)) {
    result_tbl$date <- as.Date(date)
  } else {
    result_tbl$date <- Sys.Date()  # Use today's date as approximation
  }
  
  # Define preferred column order
  preferred_cols <- c("rank", "app_id", "app_name", "publisher_name", "category", 
                     "country", "date", "chart_type", "os")
  
  # Reorder columns, keeping any additional fields at the end
  existing_cols <- intersect(preferred_cols, names(result_tbl))
  other_cols <- setdiff(names(result_tbl), preferred_cols)
  
  result_tbl <- result_tbl[, c(existing_cols, other_cols)]
  
  # Apply limit since API seems to ignore it
  if (nrow(result_tbl) > limit) {
    result_tbl <- result_tbl[1:limit, ]
  }
  
  # Add platform information
  result_tbl$platform <- os
  
  return(result_tbl)
}

#' Fetch Unified App or Publisher Information from Sensor Tower
#'
#' This function retrieves information about apps or publishers from the Sensor Tower API
#' based on a search term. It targets the `/v1/\{app_store\}/search_entities`
#' endpoint and fetches IDs and names for unified app or publisher entities.
#'
#' @param term Character string. The search term for the app or publisher.
#' @param app_store Character string. The app store to search.
#'   Defaults to "unified".
#' @param entity_type Character string. The type of entity to search for.
#'   Either "app" (default) or "publisher".
#' @param limit Numeric. The maximum number of results to return.
#'   Defaults to 20.
#' @param auth_token Character string. Your Sensor Tower API authentication token.
#' @param return_all_fields Boolean. If TRUE, returns all available fields
#'   from the API response. Defaults to FALSE.
#'
#' @return A [tibble][tibble::tibble] with entity information:
#'
#'   **For apps** (`entity_type = "app"`):
#'   - `unified_app_id`: The unified app ID (24-char hex)
#'   - `unified_app_name`: The app name
#'   - `category_details`: (when available) Nested tibble with category info
#'
#'   **For publishers** (`entity_type = "publisher"`):
#'   - `unified_publisher_id`: The unified publisher ID (24-char hex)
#'   - `unified_publisher_name`: The publisher name
#'
#'   Use the returned publisher ID with `st_publisher_apps()` to get the
#'   publisher's apps.
#'
#' @examples
#' \dontrun{
#' # Search for an app by name
#' app_info <- st_app_info(term = "Clash of Clans")
#' print(app_info)
#'
#' # Search for a publisher by name
#' publisher_info <- st_app_info(term = "Lilith", entity_type = "publisher")
#' print(publisher_info)
#'
#' # Get publisher's apps
#' lilith_apps <- st_publisher_apps(
#'   unified_id = publisher_info$unified_publisher_id[1],
#'   aggregate_related = TRUE
#' )
#'
#' # ---- Piping Workflow Examples ----
#' library(dplyr)
#'
#' # Pipe-friendly workflow: Find publisher -> Get apps -> Fetch sales
#' lilith_sales <- st_app_info("Lilith", entity_type = "publisher") %>%
#'   slice(1) %>%
#'   pull(unified_publisher_id) %>%
#'   st_publisher_apps(aggregate_related = TRUE) %>%
#'   pull(unified_app_id) %>%
#'   st_unified_sales_report(
#'     countries = "WW",
#'     start_date = "2024-01-01",
#'     end_date = "2024-12-31",
#'     date_granularity = "monthly"
#'   )
#' }
#'
#' @import dplyr
#' @importFrom dplyr all_of
#' @importFrom httr GET add_headers http_error status_code content
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble
#' @importFrom rlang abort
#' @keywords internal
#' @noRd
st_app_info_impl <- function(term,
                        app_store = "unified",
                        entity_type = "app",
                        limit = 20,
                        auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                        return_all_fields = FALSE) {
  auth_token <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token is required. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  # Define the API endpoint
  endpoint <- paste0(st_api_base_url(), "/", st_endpoint_path("search_entities", app_store = app_store))

  # Make the GET request to the API
  response <- httr::GET(
    url = endpoint,
    query = list(
      entity_type = entity_type,
      term = term,
      limit = min(limit, 250) # Limit results to a maximum of 250
    ),
    httr::add_headers(
      Authorization = sprintf("Bearer %s", auth_token)
    )
  )

  # Handle potential HTTP errors
  if (httr::http_error(response)) {
    content <- httr::content(response, "parsed")
    rlang::abort(sprintf(
      "API request failed [%s]: %s",
      httr::status_code(response),
      if (!is.null(content$error)) content$error else "Unknown error"
    ))
  }

  # Parse the response content as text
  content_text <- httr::content(response, as = "text", encoding = "UTF-8")

  # Convert JSON to R list
  content_list <- jsonlite::fromJSON(content_text, flatten = TRUE)

  # Validate the structure of the API response
  if (!is.list(content_list)) {
    rlang::abort("Unexpected API response format: The response is not a list.")
  }

  # Bind rows if the content_list contains multiple entities
  if ("data" %in% names(content_list)) {
    apps <- content_list$data
  } else {
    apps <- bind_rows(content_list)
  }

  # Ensure the response contains expected fields
  if (!"entity_type" %in% colnames(apps)) {
    rlang::abort("Unexpected API response format: Missing 'entity_type' field.")
  }

  # Determine the target entity type based on parameter
  target_entity_type <- if (entity_type == "publisher") "unified_publisher" else "unified_app"

  # Filter for the appropriate entity type
  filtered_entities <- apps %>%
    filter(.data$entity_type == target_entity_type)

  # Check if any entities are available
  if (nrow(filtered_entities) == 0) {
    warning(paste0("No ", target_entity_type, " entities found for the given search term."))
    return(tibble::tibble())
  }

  # Handle app-specific processing
  if (entity_type != "publisher") {
    # Always create nested category_details when categories are present
    if ("categories" %in% colnames(filtered_entities) && is.list(filtered_entities$categories)) {
      category_lookup <- st_categories()
      filtered_entities$category_details <- lapply(filtered_entities$categories, function(ids) {
        if (length(ids) > 0 && !all(is.na(ids))) {
          category_lookup[category_lookup$category_id %in% as.character(ids), ]
        } else {
          tibble::tibble(platform = character(0), category_id = character(0), category_name = character(0))
        }
      })
      # Remove the raw categories column in favor of the nested approach
      filtered_entities <- filtered_entities %>% dplyr::select(-"categories")
    }
  }

  # Return all fields if requested, otherwise extract and rename
  if (return_all_fields) {
    return(filtered_entities)
  } else {
    if (entity_type == "publisher") {
      # Publisher output - API returns publisher_id and publisher_name
      # Check which columns exist
      id_col <- if ("publisher_id" %in% colnames(filtered_entities)) "publisher_id" else "app_id"
      name_col <- if ("publisher_name" %in% colnames(filtered_entities)) "publisher_name" else "name"

      base_columns <- c(id_col, name_col)
      extracted_info <- filtered_entities %>%
        dplyr::select(dplyr::any_of(base_columns))

      # Rename to consistent output columns
      if (id_col %in% names(extracted_info)) {
        names(extracted_info)[names(extracted_info) == id_col] <- "unified_publisher_id"
      }
      if (name_col %in% names(extracted_info)) {
        names(extracted_info)[names(extracted_info) == name_col] <- "unified_publisher_name"
      }
    } else {
      # App output - include category_details in the simplified output if available
      base_columns <- c("app_id", "name")
      if ("category_details" %in% colnames(filtered_entities)) {
        base_columns <- c(base_columns, "category_details")
      }

      extracted_info <- filtered_entities %>%
        dplyr::select(dplyr::all_of(base_columns)) %>%
        dplyr::rename(
          unified_app_id = app_id,
          unified_app_name = name
        )
    }
    return(tibble::as_tibble(extracted_info))
  }
}

#' Parse Sensor Tower Web URL to API Parameters
#' 
#' Converts a Sensor Tower web interface URL into API-compatible parameters
#' that can be used with sensortowerR functions. This is helpful when you want
#' to replicate a web query in R.
#' 
#' @param url Character string. A Sensor Tower web interface URL
#' @param verbose Logical. Whether to print parameter mapping details. 
#'   Defaults to TRUE.
#' @return List of API-compatible parameters suitable for use with st_top_charts()
#'   and other sensortowerR functions
#' @export
#' @examples
#' \dontrun{
#' # Parse a web URL
#' url <- "https://app.sensortower.com/market-analysis/top-apps?os=unified&measure=DAU"
#' params <- st_parse_web_url(url)
#' 
#' # Use the parameters in an API call
#' data <- do.call(st_top_charts, params)
#' 
#' # Or modify parameters before using
#' params$limit <- 50
#' data <- do.call(st_top_charts, params)
#' }
st_parse_web_url <- function(url, verbose = TRUE) {
  if (is.null(url) || !is.character(url)) {
    rlang::abort("URL must be a character string")
  }
  
  # Parse the URL
  parsed <- httr::parse_url(url)
  
  if (is.null(parsed$query)) {
    rlang::abort("No query parameters found in URL")
  }
  
  query <- parsed$query
  
  if (verbose) {
    cat("=== Parsing Sensor Tower Web URL ===\n")
    cat("Path:", parsed$path, "\n")
    cat("Parameters found:", length(unique(names(query))), "\n\n")
  }
  
  # Initialize API parameters
  api_params <- list()
  
  # Direct parameter mappings
  direct_mappings <- list(
    os = "os",
    measure = "measure",
    comparison_attribute = "comparison_attribute",
    custom_fields_filter_id = "custom_fields_filter_id"
  )
  
  for (web_param in names(direct_mappings)) {
    if (!is.null(query[[web_param]])) {
      api_params[[direct_mappings[[web_param]]]] <- query[[web_param]]
      if (verbose) {
        cat(sprintf("  %s: %s\n", web_param, query[[web_param]]))
      }
    }
  }
  
  # Date handling
  if (!is.null(query$start_date)) {
    api_params$date <- query$start_date
    if (verbose) cat("  start_date ->", api_params$date, "\n")
  }
  
  if (!is.null(query$end_date)) {
    api_params$end_date <- query$end_date
    if (verbose) cat("  end_date:", query$end_date, "\n")
  }
  
  # Granularity/time range mapping
  if (!is.null(query$granularity)) {
    time_range_map <- c(
      "daily" = "day",
      "weekly" = "week", 
      "monthly" = "month",
      "quarterly" = "quarter"
    )
    
    api_params$time_range <- time_range_map[query$granularity] %||% query$granularity
    if (verbose) {
      cat(sprintf("  granularity: %s -> time_range: %s\n", 
                  query$granularity, api_params$time_range))
    }
  }
  
  # Category handling (0 = all categories)
  if (!is.null(query$category)) {
    if (query$category == "0") {
      if (verbose) cat("  category: 0 (All Categories) - omitting from API call\n")
    } else {
      api_params$category <- query$category
      if (verbose) cat("  category:", query$category, "\n")
    }
  }
  
  # Country/regions handling
  countries <- query[names(query) == "country"]
  if (length(countries) > 0) {
    unique_countries <- unique(unlist(countries))
    
    if (verbose) {
      cat(sprintf("  countries: %d unique countries specified\n", 
                  length(unique_countries)))
    }
    
    # If too many countries, suggest using WW
    if (length(unique_countries) > 50) {
      api_params$regions <- "WW"
      if (verbose) {
        cat("    -> Using 'WW' (worldwide) due to large country list\n")
      }
    } else {
      api_params$regions <- paste(unique_countries, collapse = ",")
      if (verbose && length(unique_countries) <= 10) {
        cat("    ->", api_params$regions, "\n")
      }
    }
  }
  
  # Pagination
  if (!is.null(query$page_size)) {
    api_params$limit <- as.integer(query$page_size)
    if (verbose) cat("  page_size ->", api_params$limit, "\n")
  }
  
  if (!is.null(query$page) && !is.null(query$page_size)) {
    page <- as.integer(query$page)
    limit <- as.integer(query$page_size)
    api_params$offset <- (page - 1) * limit
    if (verbose) {
      cat(sprintf("  page %d (size %d) -> offset: %d\n", 
                  page, limit, api_params$offset))
    }
  }
  
  # Custom fields filter mode
  if (!is.null(query$custom_fields_filter_mode)) {
    api_params$custom_tags_mode <- query$custom_fields_filter_mode
    if (verbose) {
      cat(sprintf("  custom_fields_filter_mode -> custom_tags_mode: %s\n",
                  api_params$custom_tags_mode))
    }
  }
  
  # Device handling (aggregate multiple device parameters)
  devices <- query[names(query) == "device"]
  if (length(devices) > 0) {
    unique_devices <- unique(unlist(devices))
    
    # If all devices are selected, it's essentially unified
    if (all(c("iphone", "ipad", "android") %in% unique_devices)) {
      if (verbose) {
        cat("  devices: all platforms selected (iphone, ipad, android)\n")
        cat("    -> Implicit in os=unified\n")
      }
    } else if (api_params$os %in% c("ios", "unified") && 
               any(c("iphone", "ipad") %in% unique_devices)) {
      # For iOS, we can specify device_type
      if (all(c("iphone", "ipad") %in% unique_devices)) {
        api_params$device_type <- "total"
      } else if ("iphone" %in% unique_devices) {
        api_params$device_type <- "iphone"
      } else if ("ipad" %in% unique_devices) {
        api_params$device_type <- "ipad"
      }
      
      if (verbose && !is.null(api_params$device_type)) {
        cat("  device_type:", api_params$device_type, "\n")
      }
    }
  }
  
  # Web-specific parameters to ignore
  ignored_params <- c("edit", "duration", "period", "metric")
  web_only <- intersect(names(query), ignored_params)
  
  if (verbose && length(web_only) > 0) {
    cat("\nWeb-only parameters (ignored):\n")
    for (param in web_only) {
      cat(sprintf("  %s: %s\n", param, query[[param]]))
    }
  }
  
  # Validation warnings
  if (verbose) {
    cat("\n=== Validation ===\n")
    
    # Check for required parameters
    if (is.null(api_params$regions)) {
      cat("!  No regions specified - you may need to add regions parameter\n")
    }
    
    # Check for custom filter without tags mode
    if (!is.null(api_params$custom_fields_filter_id) && 
        is.null(api_params$custom_tags_mode) &&
        api_params$os == "unified") {
      cat("!  custom_fields_filter_id with unified OS requires custom_tags_mode\n")
    }
    
    cat("\nReady to use with sensortowerR functions!\n")
  }
  
  # Return the API parameters
  invisible(api_params)
}

#' Extract All Parameters from Sensor Tower URL
#' 
#' Extracts and displays all parameters from a Sensor Tower web URL
#' in a readable format. Useful for understanding complex URLs.
#' 
#' @param url Character string. A Sensor Tower web interface URL
#' @return Data frame with parameter names and values
#' @export
#' @examples
#' \dontrun{
#' url <- "https://app.sensortower.com/market-analysis/top-apps?os=unified&measure=DAU"
#' params_df <- st_extract_url_params(url)
#' View(params_df)
#' }
st_extract_url_params <- function(url) {
  parsed <- httr::parse_url(url)
  
  if (is.null(parsed$query)) {
    rlang::abort("No query parameters found in URL")
  }
  
  # Convert to data frame
  params_list <- list()
  
  for (name in unique(names(parsed$query))) {
    values <- parsed$query[names(parsed$query) == name]
    if (length(values) == 1) {
      params_list[[length(params_list) + 1]] <- tibble::tibble(
        parameter = name,
        value = values[[1]],
        count = 1,

      )
    } else {
      # Multiple values (like country)
      params_list[[length(params_list) + 1]] <- tibble::tibble(
        parameter = name,
        value = paste(unlist(values), collapse = ", "),
        count = length(values),

      )
    }
  }
  
  params_df <- dplyr::bind_rows(params_list)
  
  # Add interpretation
  params_df$interpretation <- ""
  
  # Add interpretations for common parameters
  interpretations <- list(
    os = "Operating system filter",
    measure = "Metric type (DAU, WAU, MAU, revenue, units)",
    comparison_attribute = "How to compare apps",
    granularity = "Time aggregation level",
    start_date = "Beginning of date range",
    end_date = "End of date range",
    category = "App category filter (0 = all)",
    country = "Country/region filters",
    device = "Device type filters",
    page_size = "Results per page",
    page = "Current page number",
    custom_fields_filter_id = "Custom filter from web UI",
    custom_fields_filter_mode = "How custom filter applies to unified apps"
  )
  
  for (i in 1:nrow(params_df)) {
    param <- params_df$parameter[i]
    if (param %in% names(interpretations)) {
      params_df$interpretation[i] <- interpretations[[param]]
    }
  }
  
  params_df
}

#' Create Sensor Tower Web URL from Parameters
#' 
#' Builds a Sensor Tower web interface URL from API parameters.
#' This is the reverse of st_parse_web_url().
#' 
#' @param os Operating system
#' @param measure Measure type
#' @param category Category ID
#' @param regions Region codes (converted to country parameters)
#' @param start_date Start date
#' @param end_date End date
#' @param custom_fields_filter_id Custom filter ID
#' @param custom_tags_mode Custom tags mode
#' @param ... Additional parameters
#' @return Character string URL
#' @export
st_build_web_url <- function(os = "unified",
                            measure = "revenue",
                            category = NULL,
                            regions = "US",
                            start_date = NULL,
                            end_date = NULL,
                            custom_fields_filter_id = NULL,
                            custom_tags_mode = NULL,
                            ...) {
  
  base_url <- "https://app.sensortower.com/market-analysis/top-apps"
  
  # Build query parameters
  query <- list(os = os, measure = measure)
  
  # Add optional parameters
  if (!is.null(category)) query$category <- category
  if (!is.null(start_date)) query$start_date <- start_date
  if (!is.null(end_date)) query$end_date <- end_date
  if (!is.null(custom_fields_filter_id)) {
    query$custom_fields_filter_id <- custom_fields_filter_id
  }
  if (!is.null(custom_tags_mode)) {
    query$custom_fields_filter_mode <- custom_tags_mode
  }
  
  # Handle regions -> countries
  if (!is.null(regions)) {
    if (regions == "WW") {
      # Don't add country parameters for worldwide
      message("Note: 'WW' (worldwide) may require selecting all countries in web UI")
    } else {
      countries <- unlist(strsplit(regions, ","))
      # Add each country as a separate parameter
      for (country in countries) {
        query[[paste0("country[", length(query) + 1, "]")]] <- trimws(country)
      }
    }
  }
  
  # Build URL
  url <- httr::modify_url(base_url, query = query)
  
  # Fix country parameter format (remove the indexing)
  url <- gsub("country%5B\\d+%5D", "country", url)
  
  message("Generated URL:")
  message(url)
  
  invisible(url)
}

#' Validate app ID format
#'
#' Checks if an app ID matches the expected format for a given OS.
#'
#' @param app_id Character string. The app ID to validate.
#' @param os Character string. The operating system: "ios", "android", or "unified".
#' @return Logical. TRUE if the app ID format is valid for the OS, FALSE otherwise.
#' @noRd
is_valid_app_id <- function(app_id, os) {
  if (is.null(app_id) || app_id == "") {
    return(FALSE)
  }
  
  switch(os,
    ios = grepl("^[0-9]+$", app_id),  # iOS IDs are numeric strings
    android = grepl("^[a-zA-Z][a-zA-Z0-9_]*(\\.[a-zA-Z][a-zA-Z0-9_]*)*$", app_id),  # Android package names
    unified = grepl("^[a-f0-9]{24}$", app_id),  # Unified IDs are 24-char hex
    FALSE
  )
}

#' Detect app ID type
#'
#' Determines what type of app ID was provided (iOS, Android, or unified).
#'
#' @param app_id Character string. The app ID to check.
#' @return Character string. "ios", "android", "unified", or "unknown".
#' @noRd
detect_app_id_type <- function(app_id) {
  if (is.null(app_id) || app_id == "") {
    return("unknown")
  }
  
  if (grepl("^[0-9]+$", app_id)) {
    return("ios")
  } else if (grepl("^[a-zA-Z][a-zA-Z0-9_]*(\\.[a-zA-Z][a-zA-Z0-9_]*)*$", app_id)) {
    return("android")
  } else if (grepl("^[a-f0-9]{24}$", app_id)) {
    return("unified")
  } else {
    return("unknown")
  }
}

#' Validate and transform app IDs for API calls
#'
#' This function validates app IDs against the requested OS and can optionally
#' look up the correct platform-specific ID when a mismatched ID is provided.
#'
#' @param app_ids Character vector. The app IDs to validate.
#' @param os Character string. The requested operating system: "ios" or "android".
#' @param lookup_mismatched Logical. If TRUE, attempts to look up the correct 
#'   platform ID when a mismatched ID is provided. Default is TRUE.
#' @param auth_token Character string. API token for lookups.
#' @param verbose Logical. Whether to show progress messages.
#' @return Character vector of validated/transformed app IDs.
#' @noRd
validate_and_transform_app_ids <- function(app_ids, os, 
                                         lookup_mismatched = TRUE,
                                         auth_token = NULL,
                                         verbose = TRUE) {
  
  if (is.null(app_ids) || length(app_ids) == 0) {
    return(app_ids)
  }
  
  # For unified OS, we don't transform IDs
  if (os == "unified") {
    # Check if all IDs are unified format
    invalid_ids <- app_ids[!sapply(app_ids, is_valid_app_id, os = "unified")]
    if (length(invalid_ids) > 0) {
      rlang::abort(paste0(
        "Invalid unified app ID format. Unified IDs must be 24-character hexadecimal strings.\n",
        "Invalid IDs: ", paste(invalid_ids, collapse = ", "), "\n",
        "Example unified ID: '5ba4585f539ce75b97db6bcb'"
      ))
    }
    return(app_ids)
  }
  
  # For iOS/Android, check each ID
  validated_ids <- character(length(app_ids))
  
  for (i in seq_along(app_ids)) {
    app_id <- app_ids[i]
    id_type <- detect_app_id_type(app_id)
    
    # If the ID matches the requested OS, use it as-is
    if (id_type == os) {
      validated_ids[i] <- app_id
      next
    }
    
    # If it's a unified ID, we need to look it up
    if (id_type == "unified") {
      if (!lookup_mismatched) {
        rlang::abort(paste0(
          "Unified app ID provided but ", os, " ID required: ", app_id, "\n",
          "Use st_app_lookup_impl() to get platform-specific IDs."
        ))
      }
      
      if (verbose) {
        message("Looking up ", os, " ID for unified app: ", app_id)
      }
      
      # Use st_app_lookup to get platform IDs
      lookup_result <- tryCatch({
        st_app_lookup_impl(app_id, auth_token = auth_token, verbose = FALSE)
      }, error = function(e) {
        NULL
      })
      
      if (!is.null(lookup_result)) {
        platform_id <- if (os == "ios") lookup_result$ios_app_id else lookup_result$android_app_id
        
        if (!is.null(platform_id)) {
          validated_ids[i] <- platform_id
          if (verbose) {
            message("Found ", os, " ID: ", platform_id, " for app: ", 
                   lookup_result$app_name %||% app_id)
          }
        } else {
          rlang::abort(paste0(
            "App does not have an ", os, " version: ", 
            lookup_result$app_name %||% app_id
          ))
        }
      } else {
        rlang::abort(paste0("Could not look up unified app ID: ", app_id))
      }
      
    } else if (id_type == "ios" && os == "android") {
      # iOS ID provided but Android requested
      error_msg <- paste0(
        "iOS app ID provided but Android ID required: ", app_id, "\n",
        "iOS app IDs are numeric (e.g., '1234567890').\n",
        "Android app IDs are package names (e.g., 'com.example.app').\n"
      )
      
      if (lookup_mismatched) {
        error_msg <- paste0(error_msg, 
          "Note: Cross-platform lookup from iOS to Android is not supported.\n",
          "Please provide the Android package name directly."
        )
      }
      
      rlang::abort(error_msg)
      
    } else if (id_type == "android" && os == "ios") {
      # Android ID provided but iOS requested
      error_msg <- paste0(
        "Android app ID provided but iOS ID required: ", app_id, "\n",
        "Android app IDs are package names (e.g., 'com.example.app').\n",
        "iOS app IDs are numeric (e.g., '1234567890').\n"
      )
      
      if (lookup_mismatched) {
        error_msg <- paste0(error_msg, 
          "Note: Cross-platform lookup from Android to iOS is not supported.\n",
          "Please provide the iOS app ID directly."
        )
      }
      
      rlang::abort(error_msg)
      
    } else {
      # Unknown format
      rlang::abort(paste0(
        "Invalid app ID format for ", os, ": ", app_id, "\n",
        "iOS IDs are numeric (e.g., '1234567890').\n",
        "Android IDs are package names (e.g., 'com.example.app').\n",
        "Unified IDs are 24-character hex strings (e.g., '5ba4585f539ce75b97db6bcb')."
      ))
    }
  }
  
  return(validated_ids)
}

#' Create helpful error message for app ID format issues
#'
#' @param app_id The invalid app ID
#' @param expected_os The expected OS format
#' @param detected_type The detected ID type
#' @return Character string with error message
#' @noRd
create_app_id_error_message <- function(app_id, expected_os, detected_type = NULL) {
  if (is.null(detected_type)) {
    detected_type <- detect_app_id_type(app_id)
  }
  
  base_msg <- paste0("Invalid app ID for ", expected_os, " platform: '", app_id, "'\n\n")
  
  format_info <- paste0(
    "Expected formats:\n",
    "  - iOS: Numeric ID (e.g., '1234567890')\n",
    "  - Android: Package name (e.g., 'com.example.app')\n",
    "  - Unified: 24-char hex ID (e.g., '5ba4585f539ce75b97db6bcb')\n"
  )
  
  detected_msg <- if (detected_type != "unknown") {
    paste0("\nDetected format: ", detected_type, "\n")
  } else {
    "\nDetected format: unknown\n"
  }
  
  suggestion <- switch(detected_type,
    ios = if (expected_os == "android") {
      "To query this iOS app on Android, you need the Android package name.\n"
    } else "",
    android = if (expected_os == "ios") {
      "To query this Android app on iOS, you need the iOS numeric ID.\n"
    } else "",
    unified = paste0(
      "This appears to be a Sensor Tower unified ID.\n",
      "Use st_app_lookup_impl() to get platform-specific IDs, or\n",
      "use os='unified' with functions that support it.\n"
    ),
    ""
  )
  
  paste0(base_msg, format_info, detected_msg, suggestion)
}

library(testthat)
library(sensortowerR)

# Run cross-platform checks if requested
if (nzchar(Sys.getenv("RUN_CROSS_PLATFORM_CHECKS"))) {
  message("Running cross-platform checks...")
}

test_check("sensortowerR") 

#' Fetch Demographics Data for Apps
#'
#' Retrieves user demographics (age and gender breakdown) for specific apps from
#' the Sensor Tower Usage Intelligence API. This function queries the demographics
#' endpoint directly using platform-specific app IDs.
#'
#' @param unified_app_id Character string. Sensor Tower unified app ID (24-character hex).
#'   Will be resolved to platform-specific IDs automatically.
#' @param ios_app_id Character string. iOS app ID (numeric, e.g., "1234567890").
#' @param android_app_id Character string. Android package name (e.g., "com.example.app").
#' @param os Character string. Operating system: "ios" or "android". Required if using
#'   platform-specific IDs. When using unified_app_id, defaults to trying both platforms.
#' @param country Character string. Country code (e.g., "US", "GB"). Default is "US".
#'   Only single country supported per request.
#' @param date_granularity Character string. Either "all_time" (default) or "quarterly".
#'   All-time data goes back to Q4 2015. Quarterly data begins Q1 2021.
#' @param start_date Date or character string. Start date for quarterly data in "YYYY-MM-DD" format.
#'   Ignored for all_time granularity.
#' @param end_date Date or character string. End date for quarterly data in "YYYY-MM-DD" format.
#'   Ignored for all_time granularity.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'   Defaults to environment variable SENSORTOWER_AUTH_TOKEN.
#' @param verbose Logical. If TRUE, prints progress messages.
#'
#' @return A [tibble][tibble::tibble] with demographics metrics including:
#'   - **app_id**: The platform-specific app ID
#'   - **os**: Platform (ios or android)
#'   - **country**: Country code
#'   - **female_percent**: Percentage of female users (0-100)
#'   - **male_percent**: Percentage of male users (0-100)
#'   - **average_age**: Average user age
#'   - **age_13_17**, **age_18_24**, **age_25_34**, **age_35_44**, **age_45_54**,
#'     **age_55_64**, **age_65_plus**: Age group percentages
#'   - **confidence**: Data confidence level
#'
#' @section Data Availability:
#' - Quarterly data begins Q1 2021
#' - All-time data goes back to Q4 2015
#' - Demographics are primarily available for US market
#' - Data availability depends on app's user base size
#'
#' @section Recommended Workflow:
#' ```
#' # Step 1: Search for app by name
#' app <- st_app_info("Royal Match")
#'
#' # Step 2: Get demographics using unified ID
#' demographics <- st_demographics(unified_app_id = app$unified_app_id[1])
#' ```
#'
#' @examples
#' \dontrun{
#' # Get demographics for an app using unified ID
#' demo <- st_demographics(unified_app_id = "5f16a8019f7b275235017614")
#'
#' # Get demographics for iOS app directly
#' demo <- st_demographics(
#'   ios_app_id = "553834731",
#'   os = "ios",
#'   country = "US"
#' )
#' }
#'
#' @seealso [st_app_info()] for searching apps by name,
#'   [st_app_lookup()] for resolving app IDs,
#'   [st_retention()] for retention metrics
#'
#' @export
#' @importFrom rlang %||% abort
#' @importFrom httr2 request req_url_path_append req_url_query req_headers
#' @importFrom httr2 req_perform resp_body_raw resp_status
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr bind_rows mutate select
st_demographics <- function(unified_app_id = NULL,
                            ios_app_id = NULL,
                            android_app_id = NULL,
                            os = NULL,
                            country = "US",
                            date_granularity = "all_time",
                            start_date = NULL,
                            end_date = NULL,
                            auth_token = NULL,
                            verbose = TRUE) {

  # Validate that at least one app ID is provided
  if (is.null(unified_app_id) && is.null(ios_app_id) && is.null(android_app_id)) {
    rlang::abort(c(
      "At least one app ID must be provided.",
      "i" = "Use unified_app_id, ios_app_id, or android_app_id parameter.",
      "i" = "Use st_app_info_impl('app name') to find app IDs."
    ))
  }

  # Validate date_granularity
  date_granularity <- match.arg(date_granularity, c("all_time", "quarterly"))

  # Validate quarterly parameters
  if (date_granularity == "quarterly") {
    if (is.null(start_date) || is.null(end_date)) {
      rlang::abort(c(
        "start_date and end_date are required for quarterly granularity.",
        "i" = "Use date_granularity = 'all_time' for aggregate data without date ranges."
      ))
    }
    start_date <- as.Date(start_date)
    end_date <- as.Date(end_date)
  }

  # Authentication
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  # Resolve IDs - the demographics endpoint requires platform-specific IDs
  platforms_to_query <- list()

  if (!is.null(unified_app_id)) {
    # Resolve unified ID to platform-specific IDs
    if (verbose) message("Resolving unified app ID to platform IDs...")

    app_lookup <- tryCatch({
      st_app_lookup_impl(unified_app_id, auth_token = auth_token_val, verbose = verbose)
    }, error = function(e) {
      if (verbose) message("ID lookup failed: ", e$message)
      NULL
    })

    if (!is.null(app_lookup)) {
      if (!is.null(app_lookup$ios_app_id)) {
        platforms_to_query$ios <- app_lookup$ios_app_id
      }
      if (!is.null(app_lookup$android_app_id)) {
        platforms_to_query$android <- app_lookup$android_app_id
      }
    } else {
      rlang::abort(c(
        "Could not resolve unified app ID to platform IDs.",
        "i" = paste("ID:", unified_app_id),
        "i" = "Try using platform-specific IDs (ios_app_id or android_app_id) directly."
      ))
    }
  }

  # Add explicit platform IDs if provided
  if (!is.null(ios_app_id)) {
    platforms_to_query$ios <- ios_app_id
  }
  if (!is.null(android_app_id)) {
    platforms_to_query$android <- android_app_id
  }

  # Filter to specific OS if requested
  if (!is.null(os)) {
    os <- match.arg(os, c("ios", "android"))
    platforms_to_query <- platforms_to_query[names(platforms_to_query) == os]
  }

  if (length(platforms_to_query) == 0) {
    rlang::abort("No valid platform IDs found to query.")
  }

  # Fetch demographics data for each platform
  all_results <- list()

  for (platform in names(platforms_to_query)) {
    app_id <- platforms_to_query[[platform]]

    if (verbose) {
      message(sprintf("Fetching %s demographics data for app: %s", platform, app_id))
    }

    result <- fetch_demographics_for_platform(
      app_id = app_id,
      os = platform,
      country = country,
      date_granularity = date_granularity,
      start_date = start_date,
      end_date = end_date,
      auth_token = auth_token_val,
      verbose = verbose
    )

    if (!is.null(result) && nrow(result) > 0) {
      all_results[[platform]] <- result
    }
  }

  # Combine results
  if (length(all_results) == 0) {
    if (verbose) message("No demographics data found for the specified app(s).")
    return(tibble::tibble())
  }

  final_result <- dplyr::bind_rows(all_results)

  # Add unified_app_id if we have it
  if (!is.null(unified_app_id)) {
    final_result$unified_app_id <- unified_app_id
  }

  if (verbose) {
    message(sprintf("Retrieved demographics data for %d platform(s)", length(all_results)))
  }

  return(final_result)
}

#' Fetch demographics data for a single platform
#' @noRd
fetch_demographics_for_platform <- function(app_id,
                                             os,
                                             country,
                                             date_granularity,
                                             start_date,
                                             end_date,
                                             auth_token,
                                             verbose = FALSE) {

  # Build query parameters - auth_token as query param per API spec
  query_params <- list(
    auth_token = auth_token,
    app_ids = app_id,
    date_granularity = date_granularity,
    country = country
  )

  # Add date parameters - API requires dates even for all_time
  if (date_granularity == "quarterly" && !is.null(start_date) && !is.null(end_date)) {
    query_params$start_date <- format(start_date, "%Y-%m-%d")
    query_params$end_date <- format(end_date, "%Y-%m-%d")
  } else {
    # For all_time or when dates not specified, use a wide date range
    # All-time data goes back to Q4 2015
    query_params$start_date <- "2015-10-01"
    query_params$end_date <- format(Sys.Date() - 1, "%Y-%m-%d")
  }

  # Build request
  req <- build_request(
    base_url = st_api_base_url(),
    path_segments = st_endpoint_segments("usage_demographics", os = os),
    query_params = query_params
  ) %>%
    httr2::req_headers(
      "Accept" = "application/json"
    ) %>%
    httr2::req_timeout(30)

  # Perform request
  resp <- tryCatch({
    httr2::req_perform(req)
  }, error = function(e) {
    if (verbose) message("API request failed: ", e$message)
    return(NULL)
  })

  if (is.null(resp)) {
    return(NULL)
  }

  # Check status
  status <- httr2::resp_status(resp)
  if (status != 200) {
    body <- tryCatch(httr2::resp_body_string(resp), error = function(.) "")
    if (verbose) message(sprintf("API returned status %d: %s", status, body))
    return(NULL)
  }

  # Parse response
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(NULL)
  }

  body_text <- rawToChar(body_raw)
  result <- tryCatch({
    jsonlite::fromJSON(body_text, flatten = TRUE)
  }, error = function(e) {
    if (verbose) message("Failed to parse response: ", e$message)
    return(NULL)
  })

  if (is.null(result) || length(result) == 0) {
    return(NULL)
  }

  # The API returns: { app_data: [...], baseline_data: {...}, disabled_app_ids: [...] }
  app_data_list <- result$app_data

  if (is.null(app_data_list) || length(app_data_list) == 0) {
    return(NULL)
  }

  # Process each app's demographics data
  # app_data_list is typically a data frame from fromJSON
  if (is.data.frame(app_data_list)) {
    processed <- lapply(seq_len(nrow(app_data_list)), function(i) {
      app_row <- app_data_list[i, ]

      # Build the row
      row <- tibble::tibble(
        app_id = as.character(app_row$app_id),
        os = os,
        country = country
      )

      # Gender percentages (API returns as 0-1, convert to 0-100)
      row$female_percent <- (app_row$female %||% NA_real_) * 100
      row$male_percent <- (app_row$male %||% NA_real_) * 100

      # Average age
      row$average_age <- app_row$average_age_total %||% NA_real_

      # Extract age breakdown from normalized_demographics
      # The API returns format like: female_18, female_25, male_18, etc.
      # These represent female 18-24, female 25-34, etc.
      if ("normalized_demographics" %in% names(app_row)) {
        demo <- app_row$normalized_demographics
        if (is.list(demo)) {
          demo <- demo[[1]]  # Unlist if nested
        }

        if (!is.null(demo) && is.list(demo)) {
          # Sum female and male for each age group
          row$age_18_24 <- ((demo$female_18 %||% 0) + (demo$male_18 %||% 0)) * 100
          row$age_25_34 <- ((demo$female_25 %||% 0) + (demo$male_25 %||% 0)) * 100
          row$age_35_44 <- ((demo$female_35 %||% 0) + (demo$male_35 %||% 0)) * 100
          row$age_45_54 <- ((demo$female_45 %||% 0) + (demo$male_45 %||% 0)) * 100
          row$age_55_plus <- ((demo$female_55 %||% 0) + (demo$male_55 %||% 0)) * 100
        } else {
          row$age_18_24 <- NA_real_
          row$age_25_34 <- NA_real_
          row$age_35_44 <- NA_real_
          row$age_45_54 <- NA_real_
          row$age_55_plus <- NA_real_
        }
      } else {
        row$age_18_24 <- NA_real_
        row$age_25_34 <- NA_real_
        row$age_35_44 <- NA_real_
        row$age_45_54 <- NA_real_
        row$age_55_plus <- NA_real_
      }

      # Confidence level
      row$confidence <- as.integer(app_row$confidence %||% NA_integer_)

      return(row)
    })
  } else {
    # Handle list structure
    processed <- lapply(app_data_list, function(app_item) {
      if (is.null(app_item) || !is.list(app_item)) return(NULL)

      row <- tibble::tibble(
        app_id = as.character(app_item$app_id %||% app_id),
        os = os,
        country = country
      )

      row$female_percent <- (app_item$female %||% NA_real_) * 100
      row$male_percent <- (app_item$male %||% NA_real_) * 100
      row$average_age <- app_item$average_age_total %||% NA_real_

      demo <- app_item$normalized_demographics
      if (!is.null(demo) && is.list(demo)) {
        row$age_18_24 <- ((demo$female_18 %||% 0) + (demo$male_18 %||% 0)) * 100
        row$age_25_34 <- ((demo$female_25 %||% 0) + (demo$male_25 %||% 0)) * 100
        row$age_35_44 <- ((demo$female_35 %||% 0) + (demo$male_35 %||% 0)) * 100
        row$age_45_54 <- ((demo$female_45 %||% 0) + (demo$male_45 %||% 0)) * 100
        row$age_55_plus <- ((demo$female_55 %||% 0) + (demo$male_55 %||% 0)) * 100
      } else {
        row$age_18_24 <- NA_real_
        row$age_25_34 <- NA_real_
        row$age_35_44 <- NA_real_
        row$age_45_54 <- NA_real_
        row$age_55_plus <- NA_real_
      }

      row$confidence <- as.integer(app_item$confidence %||% NA_integer_)
      return(row)
    })
  }

  # Combine and return
  result_tbl <- dplyr::bind_rows(processed)

  return(result_tbl)
}

#' Standardize Revenue Units Across sensortowerR
#'
#' Ensures revenue values have consistent units across all sensortowerR functions.
#' The Sensor Tower API returns revenue in cents, but for consistency with other
#' functions like st_sales_report, this converts to base currency units.
#'
#' @param data Data frame containing revenue columns
#' @param source Character string indicating the data source
#' @param target_unit Character string: "base" (dollars/euros/etc) or "cents"
#' @return Data frame with standardized revenue values
#' @keywords internal
#' 
#' @details
#' The Sensor Tower API inconsistently returns revenue values:
#' - Top Charts API: revenue in cents
#' - Sales Report API: revenue in base currency units
#' - Publishers API: revenue in cents (but converts to base units)
#' 
#' This function ensures consistency by converting all revenue to base units
#' by default, while preserving the original values for reference.
standardize_revenue_units <- function(data, source = "unknown", target_unit = "base") {
  if (nrow(data) == 0) return(data)
  
  # Define columns that are known to be in cents
  cents_columns <- c(
    "revenue_absolute",      # From top_charts
    "current_revenue_value", # From top_charts  
    "comparison_revenue_value", # From top_charts
    "revenue_cents",         # Generic cents column
    "iphone_revenue_cents",  # From sales_report (raw)
    "ipad_revenue_cents"     # From sales_report (raw)
  )
  
  # Process each cents column
  for (col in cents_columns) {
    if (col %in% names(data)) {
      if (target_unit == "base") {
        # Convert cents to base currency units (divide by 100)
        base_col <- gsub("_cents$|_absolute$|_value$", "", col)
        if (base_col == col) base_col <- paste0(col, "_base")
        
        data[[base_col]] <- data[[col]] / 100
        
        # Special case: create a standardized "revenue" column for common access
        if (col %in% c("revenue_absolute", "current_revenue_value")) {
          data$revenue <- data[[col]] / 100
          attr(data$revenue, "unit") <- "base_currency"
        }
      }
      
      # Always add unit metadata to original column
      attr(data[[col]], "unit") <- "cents"
    }
  }
  
  # Handle columns already in base units
  base_columns <- c(
    "total_revenue",    # From sales_report
    "iphone_revenue",   # From sales_report  
    "ipad_revenue",     # From sales_report
    "revenue_usd"       # From publishers
  )
  
  for (col in base_columns) {
    if (col %in% names(data)) {
      attr(data[[col]], "unit") <- "base_currency"
    }
  }
  
  # Add informative message for top_charts
  if (source == "top_charts" && "revenue" %in% names(data)) {
    message("Revenue units standardized: 'revenue' column contains base currency units (dollars/euros/etc).")
    message("Original cents values preserved in 'revenue_absolute' column.")
  }
  
  return(data)
}

#' Resolve App IDs Based on OS Parameter
#'
#' Internal function that resolves app IDs based on the requested OS.
#' The OS parameter always controls what IDs are used for data fetching.
#'
#' @param unified_app_id Character. Unified app ID (24-char hex)
#' @param ios_app_id Character. iOS app ID (numeric)
#' @param android_app_id Character. Android package name
#' @param os Character. Required. One of "ios", "android", or "unified"
#' @param auth_token Character. Sensor Tower API token
#' @param verbose Logical. Show progress messages
#'
#' @return A list with:
#'   - resolved_ids: List of IDs to use for the API call
#'   - app_id_type: Type of ID ("ios", "android", or "unified")
#'   - lookup_performed: Whether ID lookups were needed
#'
#' @keywords internal
#' @noRd
resolve_ids_for_os <- function(
  unified_app_id = NULL,
  ios_app_id = NULL,
  android_app_id = NULL,
  os,
  auth_token,
  verbose = FALSE
) {
  # Validate OS parameter
  if (missing(os) || is.null(os) || !os %in% c("ios", "android", "unified")) {
    rlang::abort("'os' parameter is required and must be one of: 'ios', 'android', or 'unified'")
  }

  # Check that at least one ID is provided
  if (is.null(unified_app_id) && is.null(ios_app_id) && is.null(android_app_id)) {
    rlang::abort("At least one app ID must be provided (unified_app_id, ios_app_id, or android_app_id)")
  }

  # Centralized mapping using cache
  input_ids <- c(unified_app_id, ios_app_id, android_app_id)
  input_ids <- input_ids[!vapply(input_ids, is.null, logical(1))]
  input_ids <- as.character(input_ids)[1]

  # Use the cached resolution function
  mapping <- tryCatch(
    {
      resolve_app_id(input_ids, auth_token = auth_token, verbose = verbose)
    },
    error = function(e) NULL
  )

  # Extract IDs from the mapping result
  # resolve_app_id returns: unified_app_id, ios_app_id, android_app_id
  # Prefer mapped ID, but fall back to input ID if mapped ID is missing/NA
  unified_mapped <- if (!is.null(mapping) && !is.null(mapping$unified_app_id) && !is.na(mapping$unified_app_id)) mapping$unified_app_id else unified_app_id
  ios_mapped <- if (!is.null(mapping) && !is.null(mapping$ios_app_id) && !is.na(mapping$ios_app_id)) mapping$ios_app_id else ios_app_id
  android_mapped <- if (!is.null(mapping) && !is.null(mapping$android_app_id) && !is.na(mapping$android_app_id)) mapping$android_app_id else android_app_id
  lookup_performed <- !is.null(mapping)

  if (os == "ios" && !is.null(ios_mapped) && !is.na(ios_mapped) && nzchar(as.character(ios_mapped))) {
    return(list(resolved_ids = list(ios_app_id = ios_mapped), app_id_type = "ios", lookup_performed = lookup_performed))
  }
  if (os == "android" && !is.null(android_mapped) && !is.na(android_mapped) && nzchar(as.character(android_mapped))) {
    return(list(resolved_ids = list(android_app_id = android_mapped), app_id_type = "android", lookup_performed = lookup_performed))
  }
  if (os == "unified") {
    if (!is.null(unified_mapped) && !is.na(unified_mapped) && nzchar(as.character(unified_mapped))) {
      return(list(resolved_ids = list(unified_app_id = unified_mapped), app_id_type = "unified", lookup_performed = lookup_performed))
    }
    # Fallback: if we have platform IDs, return them
    if ((!is.null(ios_mapped) && !is.na(ios_mapped)) || (!is.null(android_mapped) && !is.na(android_mapped))) {
      return(list(
        resolved_ids = list(
          unified_app_id = NULL,
          ios_app_id = ios_mapped,
          android_app_id = android_mapped
        ),
        app_id_type = "unified",
        lookup_performed = lookup_performed
      ))
    }
  }

  rlang::abort("Failed to resolve appropriate IDs for OS: ", os)
}

#' Add App ID Metadata to Data Frame
#'
#' Adds app_id and app_id_type columns to a data frame
#'
#' @param data Data frame to modify
#' @param app_id The app ID used
#' @param app_id_type The type of app ID ("ios", "android", or "unified")
#'
#' @return Modified data frame with app_id and app_id_type columns
#' @keywords internal
#' @noRd
add_app_id_metadata <- function(data, app_id, app_id_type) {
  if (is.null(data) || nrow(data) == 0) {
    return(data)
  }

  data$app_id <- app_id
  data$app_id_type <- app_id_type

  # Move these columns to the front
  id_cols <- c("app_id", "app_id_type")
  other_cols <- setdiff(names(data), id_cols)
  data <- data[, c(id_cols, other_cols)]

  return(data)
}

#' Data validation functions for Sensor Tower API responses
#'
#' @description Functions to validate and clean data from API responses

#' Validate top charts data
#'
#' @keywords internal
#' @noRd
validate_top_charts_data <- function(data, measure, regions) {
  if (nrow(data) == 0) {
    return(data)
  }

  # Check for common column issues
  message("Validating data columns...")

  # 1. Handle app name column variations
  app_name_cols <- c("unified_app_name", "app_name", "app.name", "name")
  found_name_col <- intersect(app_name_cols, names(data))

  if (length(found_name_col) == 0) {
    message(
      "No app name column found. Available columns: ",
      paste(head(names(data), 10), collapse = ", ")
    )
  } else if (found_name_col[1] != "unified_app_name") {
    # Standardize to unified_app_name
    data <- data %>%
      dplyr::rename(unified_app_name = !!found_name_col[1])
    message(sprintf("Renamed '%s' to 'unified_app_name'", found_name_col[1]))
  }

  # 2. Enforce region-specific column availability (no fallbacks)
  requested_region <- tolower(regions[1])

  # Look for metric columns with 2-letter region suffixes
  metric_patterns <- c("revenue", "download", "retention", "dau", "mau", "wau", "rpd")

  for (pattern in metric_patterns) {
    cols <- grep(paste0(pattern, ".*_([a-z]{2})$"), names(data), value = TRUE)
    if (length(cols) > 0) {
      available_regions <- unique(gsub(".*_([a-z]{2})$", "\\1", cols))
      # If a specific region is requested, and region-specific metrics are missing,
      # fail loudly rather than implying WW values, but only when working with
      # active user measures where these columns are expected to be present.
      if (requested_region != "ww" && pattern %in% c("mau", "retention") &&
        toupper(measure) %in% c("DAU", "WAU", "MAU")) {
        if (!(requested_region %in% available_regions)) {
          rlang::abort(sprintf(
            "Region-specific '%s' metrics for '%s' are not available in the response. Available regions for '%s': %s",
            toupper(pattern), toupper(requested_region), toupper(pattern),
            ifelse(length(available_regions) > 0, paste(toupper(available_regions), collapse = ", "), "none")
          ), call. = FALSE)
        }
      }
    }
  }

  # 3. Validate numeric columns are actually numeric
  numeric_patterns <- c(
    "revenue", "download", "retention", "dau", "mau", "wau",
    "rpd", "arpu", "share", "rating", "age"
  )

  for (col in names(data)) {
    if (any(sapply(numeric_patterns, function(p) grepl(p, col, ignore.case = TRUE)))) {
      if (!is.numeric(data[[col]])) {
        # Try to convert to numeric
        cleaned <- clean_numeric_column(data[[col]])
        if (!all(is.na(cleaned))) {
          data[[col]] <- cleaned
          message(sprintf("Converted column '%s' to numeric", col))
        }
      }
    }
  }

  # 4. Check for essential columns based on measure
  essential_cols <- list(
    revenue = c("revenue_180d_ww", "revenue_30d_ww"),
    units = c("downloads_180d_ww", "downloads_30d_ww"),
    DAU = c("dau_30d_ww", "dau_30d_us"),
    MAU = c("mau_month_ww", "mau_month_us"),
    WAU = c("wau_4w_ww", "wau_4w_us")
  )

  if (measure %in% names(essential_cols)) {
    missing_all <- !any(essential_cols[[measure]] %in% names(data))
    if (missing_all) {
      # Only warn if we have NO revenue/download data at all
      # The API often doesn't return the 180d/30d metrics for standard sales calls
      has_any_data <- any(grepl(paste0(measure, "|absolute|value"), names(data), ignore.case = TRUE))
      if (!has_any_data) {
        warning(sprintf(
          "Expected columns for %s not found: %s. Available metric columns: %s",
          measure,
          paste(essential_cols[[measure]], collapse = ", "),
          paste(grep("(revenue|download|dau|mau|wau)", names(data),
            value = TRUE
          )[seq_len(min(5, length(grep("(revenue|download|dau|mau|wau)",
            names(data),
            value = TRUE
          ))))], collapse = ", ")
        ))
      }
    }
  }

  return(data)
}

#' Clean numeric column by removing special characters
#'
#' @keywords internal
#' @noRd
clean_numeric_column <- function(x) {
  if (is.numeric(x)) {
    return(x)
  }

  # Convert to character first
  x <- as.character(x)

  # Remove common formatting characters
  x <- gsub("[$,%]", "", x)
  x <- gsub(",", "", x)

  # Handle percentages (keep as 0-100 not 0-1)
  x <- gsub("%", "", x)

  # Convert to numeric
  result <- suppressWarnings(as.numeric(x))

  # Don't convert percentages to decimals
  # (they should stay as 0-100)

  return(result)
}

#' Validate column exists with fallback options
#'
#' @keywords internal
#' @noRd
require_column <- function(data, primary, fallbacks = NULL, operation = "operation") {
  all_options <- c(primary, fallbacks)
  found <- intersect(all_options, names(data))

  if (length(found) == 0) {
    available <- names(data)
    similar <- grep(gsub("_", ".*", primary), available, value = TRUE)

    error_msg <- sprintf("Required column '%s' not found for %s.", primary, operation)
    if (length(similar) > 0) {
      error_msg <- paste0(
        error_msg,
        sprintf(
          "\nDid you mean: %s?",
          paste(similar[seq_len(min(3, length(similar)))], collapse = ", ")
        )
      )
    }
    error_msg <- paste0(
      error_msg,
      sprintf(
        "\nAvailable columns: %s",
        paste(available[seq_len(min(10, length(available)))], collapse = ", ")
      )
    )

    rlang::abort(error_msg)
  }

  return(found[1])
}

#' Safe column selection with automatic fallbacks
#'
#' @keywords internal
#' @noRd
select_robust <- function(data, ...) {
  cols <- list(...)
  selected <- c()

  for (col_spec in cols) {
    if (length(col_spec) == 1 && col_spec %in% names(data)) {
      selected <- c(selected, col_spec)
    } else {
      # Try each option
      found <- intersect(col_spec, names(data))
      if (length(found) > 0) {
        selected <- c(selected, found[1])
      }
    }
  }

  if (length(selected) == 0) {
    return(data)
  }

  return(dplyr::select(data, all_of(selected)))
}

# --- Utility Functions for sensortowerR ---

# This file contains helper functions used by the main data-fetching functions
# in the package. They handle tasks like input validation, query parameter
# preparation, and API request execution.

#' @importFrom rlang abort
#' @importFrom stats setNames
#' @importFrom httr2 request req_user_agent req_url_path_append req_url_query
#'   req_error req_perform resp_status resp_body_raw resp_check_status
#'   resp_body_string
#' @importFrom jsonlite fromJSON
#' @importFrom tibble as_tibble
#' @importFrom utils URLencode head
#' @importFrom tidyr unnest
#' @importFrom dplyr rename all_of
#'
# --- Shared API Helpers ---

st_api_base_url <- function() {
  "https://api.sensortower.com"
}

.st_endpoint_registry <- list(
  sales_report_estimates = c("{os}", "sales_report_estimates"),
  sales_report_estimates_comparison_attributes = c("{os}", "sales_report_estimates_comparison_attributes"),
  facets_metrics = c("facets", "metrics"),
  usage_active_users = c("{os}", "usage", "active_users"),
  usage_demographics = c("{os}", "usage", "demographics"),
  usage_retention = c("{os}", "usage", "retention"),
  top_and_trending_active_users = c("{os}", "top_and_trending", "active_users"),
  top_and_trending_publishers = c("{os}", "top_and_trending", "publishers"),
  ranking = c("{os}", "ranking"),
  apps = c("{os}", "apps"),
  app_tag_apps = c("app_tag", "apps"),
  custom_fields_filter = c("custom_fields_filter"),
  custom_fields_filter_id = c("custom_fields_filter", "{id}"),
  custom_fields_filter_fields_values = c("custom_fields_filter", "fields_values"),
  apps_timeseries = c("apps", "timeseries"),
  apps_timeseries_unified = c("apps", "timeseries", "unified_apps"),
  unified_publishers_apps = c("unified", "publishers", "apps"),
  unified_sales_report_estimates = c("unified", "sales_report_estimates"),
  search_entities = c("{app_store}", "search_entities"),
  games_breakdown = c("{os}", "games_breakdown"),
  store_summary = c("{os}", "store_summary"),
  compact_sales_report_estimates = c("{os}", "compact_sales_report_estimates")
)

resolve_endpoint_segment <- function(segment, placeholders) {
  if (!grepl("^\\{[A-Za-z0-9_]+\\}$", segment)) {
    return(segment)
  }

  key <- gsub("^\\{|\\}$", "", segment)
  value <- placeholders[[key]]
  value_chr <- if (is.null(value)) "" else as.character(value[1])

  if (!nzchar(value_chr)) {
    rlang::abort(sprintf("Missing required endpoint placeholder: '%s'.", key))
  }

  value_chr
}

st_endpoint_relative_segments <- function(endpoint_key, ...) {
  template <- .st_endpoint_registry[[endpoint_key]]
  if (is.null(template)) {
    rlang::abort(sprintf("Unknown endpoint key '%s'.", endpoint_key))
  }

  placeholders <- list(...)

  vapply(
    template,
    resolve_endpoint_segment,
    placeholders = placeholders,
    FUN.VALUE = character(1)
  )
}

st_endpoint_segments <- function(endpoint_key, ...) {
  c("v1", st_endpoint_relative_segments(endpoint_key, ...))
}

st_endpoint_relative_path <- function(endpoint_key, ...) {
  paste(st_endpoint_relative_segments(endpoint_key, ...), collapse = "/")
}

st_endpoint_path <- function(endpoint_key, ...) {
  paste(st_endpoint_segments(endpoint_key, ...), collapse = "/")
}

resolve_auth_token <- function(auth_token = NULL,
                               env_var = "SENSORTOWER_AUTH_TOKEN",
                               error_message = NULL) {
  token <- if (is.null(auth_token)) Sys.getenv(env_var) else auth_token
  token <- trimws(as.character(token[1]))

  if (!nzchar(token)) {
    msg <- if (!is.null(error_message) && nzchar(error_message)) {
      error_message
    } else {
      sprintf("Authentication token not found. Set %s environment variable.", env_var)
    }
    rlang::abort(msg)
  }

  token
}

# --- Input Validation ---

validate_inputs <- function(os,
                            comparison_attribute,
                            time_range,
                            measure,
                            date,
                            category,
                            regions,
                            end_date = NULL,
                            limit = 25,
                            offset = NULL,
                            device_type = NULL,
                            custom_fields_filter_id = NULL,
                            custom_tags_mode = NULL,
                            data_model = NULL) {
  # Validation checks for common parameters
  if (!os %in% c("ios", "android", "unified")) {
    rlang::abort("`os` must be one of 'ios', 'android', or 'unified'")
  }
  if (!comparison_attribute %in% c("absolute", "delta", "transformed_delta")) {
    rlang::abort("`comparison_attribute` must be one of 'absolute', 'delta', 'transformed_delta'")
  }
  if (!is.character(time_range) || !nzchar(time_range)) {
    rlang::abort("`time_range` must be a non-empty string")
  }
  if (!is.character(measure) || !nzchar(measure)) {
    rlang::abort("`measure` must be a non-empty string")
  }
  if (is.null(date)) {
    rlang::abort("`date` must be provided")
  }
  if (is.null(category) && is.null(custom_fields_filter_id)) {
    rlang::abort("Either `category` or `custom_fields_filter_id` must be provided")
  }
  if (is.null(regions)) {
    rlang::abort("`regions` must be provided")
  }
  if (!is.numeric(limit) || limit <= 0 || limit != round(limit)) {
    rlang::abort("`limit` must be a positive integer")
  }

  # Specific validations
  if (os %in% c("ios", "unified") && is.null(device_type)) {
    message("`device_type` is not specified for `os = '", os, "'`. Defaulting to 'total'.")
    device_type <- "total"
  }

  if (!is.null(custom_fields_filter_id) && os == "unified" && is.null(custom_tags_mode)) {
    rlang::abort("`custom_tags_mode` must be provided when `os` is 'unified' and `custom_fields_filter_id` is used.")
  }
}

# --- Query Parameter Preparation ---

prepare_query_params_sales <- function(auth_token,
                                       comparison_attribute,
                                       time_range,
                                       measure,
                                       date,
                                       category,
                                       end_date,
                                       regions,
                                       limit,
                                       offset,
                                       device_type,
                                       custom_fields_filter_id,
                                       custom_tags_mode,
                                       os) {
  params <- list(
    auth_token = auth_token,
    comparison_attribute = comparison_attribute,
    time_range = time_range,
    measure = measure,
    date = as.character(date),
    category = category, # Keep category even with custom filter
    end_date = if (!is.null(end_date)) as.character(end_date) else NULL,
    regions = paste(regions, collapse = ","),
    limit = limit,
    offset = offset,
    device_type = if (os %in% c("ios", "unified")) device_type else NULL,
    custom_fields_filter_id = custom_fields_filter_id,
    custom_tags_mode = if (os == "unified") custom_tags_mode else NULL
  )
  # Remove NULLs
  params <- params[!sapply(params, is.null)]

  # Debug: print params when using custom filter
  if (!is.null(custom_fields_filter_id)) {
    message("Debug: Query params with custom filter (sales):")
    message("  custom_fields_filter_id: ", params$custom_fields_filter_id)
    message("  category: ", if (is.null(params$category)) "NULL" else params$category)
    message("  custom_tags_mode: ", params$custom_tags_mode)
  }

  params
}

prepare_query_params_active_users <- function(auth_token,
                                              comparison_attribute,
                                              time_range,
                                              measure,
                                              date,
                                              category,
                                              regions,
                                              limit,
                                              offset,
                                              device_type,
                                              custom_fields_filter_id,
                                              custom_tags_mode,
                                              data_model,
                                              os) {
  params <- list(
    auth_token = auth_token,
    comparison_attribute = comparison_attribute,
    time_range = time_range,
    measure = measure,
    date = as.character(date),
    category = category, # Keep category even with custom filter
    regions = paste(regions, collapse = ","),
    limit = limit,
    offset = offset,
    device_type = if (os %in% c("ios", "unified")) device_type else NULL,
    custom_fields_filter_id = custom_fields_filter_id,
    custom_tags_mode = if (os == "unified") custom_tags_mode else NULL,
    data_model = data_model
  )
  # Remove NULLs
  params[!sapply(params, is.null)]
}


# --- API Request Building and Performance ---

build_request <- function(base_url, path_segments, query_params) {
  req <- httr2::request(base_url) %>%
    httr2::req_user_agent("sensortowerR (https://github.com/ge-data-solutions/SensorTowerR)")

  for (segment in path_segments) {
    req <- req %>% httr2::req_url_path_append(segment)
  }

  req %>% httr2::req_url_query(!!!query_params)
}


perform_request <- function(req) {
  tryCatch(
    {
      resp <- httr2::req_perform(req)
      httr2::resp_check_status(resp) # Check for HTTP errors
      resp
    },
    httr2_error = function(e) {
      status <- httr2::resp_status(e$resp)
      body <- httr2::resp_body_string(e$resp)
      rlang::abort(
        message = sprintf("API request failed with status %d.", status),
        body = body,
        parent = e
      )
    },
    error = function(e) {
      rlang::abort("An unexpected error occurred during the API request.", parent = e)
    }
  )
}

#' Core Data Fetching Function
#'
#' A reusable function to handle the common pattern of fetching data from the API.
#' Handles URL building, request execution, error handling, and response processing.
#'
#' @param endpoint Character. API endpoint path (e.g. "ios/sales_report_estimates").
#' @param params List. Query parameters.
#' @param auth_token Character. API token.
#' @param verbose Logical. Whether to print debug messages.
#' @param enrich_response Logical. Whether to enrich the response with metadata.
#' @param processor Function. Function to process the response. Defaults to process_response.
#'
#' @return A tibble with the results.
#' @keywords internal
fetch_data_core <- function(endpoint, params, auth_token, verbose = FALSE, enrich_response = TRUE, processor = process_response) {
  base_url <- paste0(st_api_base_url(), "/v1")

  # Clean up params
  params <- params[!sapply(params, is.null)]
  params$auth_token <- auth_token

  if (verbose) {
    message("Requesting: ", endpoint)
    # Don't print auth token in debug
    debug_params <- params
    debug_params$auth_token <- "HIDDEN"
    print(debug_params)
  }

  req <- httr2::request(base_url) %>%
    httr2::req_url_path_append(endpoint) %>%
    httr2::req_url_query(!!!params) %>%
    httr2::req_user_agent("sensortowerR") %>%
    httr2::req_retry(max_tries = 3, backoff = function(i) 2^i)

  resp <- tryCatch(
    {
      httr2::req_perform(req)
    },
    error = function(e) {
      if (verbose) message("API Request Failed: ", e$message)
      return(NULL)
    }
  )

  if (is.null(resp)) {
    return(NULL)
  }

  # Call the processor function
  # Check if processor accepts enrich_response argument
  if ("enrich_response" %in% names(formals(processor))) {
    processor(resp, enrich_response = enrich_response)
  } else {
    processor(resp)
  }
}

# --- Response Processing ---

# Helper function to perform unnesting of entities
perform_unnest <- function(result_tbl) {
  # Proactively coerce app_id to character in the nested data frames
  # to prevent type errors during the unnest operation. The API can return
  # a mix of integer and character IDs, which vctrs cannot combine.
  result_tbl$entities <- lapply(result_tbl$entities, function(df) {
    if (!is.null(df) && "app_id" %in% names(df)) {
      df$app_id <- as.character(df$app_id)
    }
    df
  })

  tidyr::unnest(result_tbl, dplyr::all_of("entities"), names_sep = ".")
}

process_response <- function(resp, enrich_response = TRUE) {
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(tibble::tibble())
  }

  body_text <- rawToChar(body_raw)
  result <- jsonlite::fromJSON(body_text, flatten = TRUE)

  if (length(result) == 0 || nrow(result) == 0) {
    return(tibble::tibble())
  }

  result_tbl <- tibble::as_tibble(result)

  # Enrich with app names if requested and possible
  if (enrich_response && "entities" %in% names(result_tbl) && is.list(result_tbl$entities)) {
    # IMPORTANT: Preserve the original app_id as unified_app_id BEFORE unnesting
    # This is the true unified hex ID from the API
    if ("app_id" %in% names(result_tbl)) {
      result_tbl$unified_app_id <- result_tbl$app_id

      # OPTIMIZATION: If we already have unified hex IDs, check if unnesting is necessary
      # When using os="unified" with custom filters, the API returns unified data already
      non_na_ids <- result_tbl$app_id[!is.na(result_tbl$app_id)]
      if (length(non_na_ids) > 0 && all(grepl("^[a-f0-9]{24}$", non_na_ids))) {
        # We have unified hex IDs - check if entities just contains redundant platform data
        # If the entities only contain platform-specific versions of the same app,
        # we can skip unnesting to avoid creating duplicates that need consolidation

        # Check first entity to see if it would create duplicates
        first_entity <- result_tbl$entities[[1]]
        if (!is.null(first_entity) && is.data.frame(first_entity) && nrow(first_entity) > 1) {
          # Multiple rows in entity = iOS + Android versions = will create duplicates
          # For unified data, we can use aggregate_tags instead of unnesting entities
          message("Streamlined processing: Using unified data without unnesting platform entities")

          custom_tag_cols <- unique(unlist(lapply(result_tbl$entities, function(df) {
            if (is.null(df) || !is.data.frame(df)) {
              return(character())
            }
            names(df)[grepl("^custom_tags\\.", names(df))]
          })))
          for (col in custom_tag_cols) {
            agg_col <- sub("^custom_tags\\.", "aggregate_tags.", col)
            if (!agg_col %in% names(result_tbl)) {
              result_tbl[[agg_col]] <- rep(NA_character_, nrow(result_tbl))
            }
          }
          for (i in seq_len(nrow(result_tbl))) {
            entity_df <- result_tbl$entities[[i]]
            if (is.null(entity_df) || !is.data.frame(entity_df)) next
            for (col in custom_tag_cols) {
              agg_col <- sub("^custom_tags\\.", "aggregate_tags.", col)
              value <- entity_df[[col]][1]
              if (!is.null(value) && !is.na(value) && (is.na(result_tbl[[agg_col]][i]) || result_tbl[[agg_col]][i] == "")) {
                result_tbl[[agg_col]][i] <- value
              }
            }
          }

          # Now safe to remove entities after extracting key fields
          if ("aggregate_tags" %in% names(result_tbl)) {
            # The aggregate_tags column already contains the unified metrics
            # We've extracted gender data, so now we can remove entities
            result_tbl$entities <- NULL # Remove entities to prevent confusion
          }
        } else {
          # Single row or empty entity - safe to unnest
          result_tbl <- perform_unnest(result_tbl)
        }
      } else {
        # Not all unified IDs - need to unnest for platform resolution
        result_tbl <- perform_unnest(result_tbl)
      }
    } else {
      # No app_id column - proceed with normal unnesting
      result_tbl <- perform_unnest(result_tbl)
    }

    # Clean up duplicate columns - prefer the entities.* versions for detailed data
    base_cols <- setdiff(names(result_tbl), grep("^entities\\.", names(result_tbl), value = TRUE))
    entities_cols <- grep("^entities\\.", names(result_tbl), value = TRUE)

    # Remove base columns that have entities.* equivalents
    # But preserve unified_app_id and app_id if they're hex format unified IDs
    preserve_cols <- "unified_app_id"
    if ("app_id" %in% base_cols && all(grepl("^[a-f0-9]{24}$", result_tbl$app_id[!is.na(result_tbl$app_id)]))) {
      # If app_id contains hex unified IDs, don't remove it
      preserve_cols <- c(preserve_cols, "app_id")
    }
    duplicated_bases <- intersect(
      gsub("^entities\\.", "", entities_cols),
      setdiff(base_cols, preserve_cols)
    )
    result_tbl <- result_tbl[, !names(result_tbl) %in% duplicated_bases]

    # Handle app name column - check multiple possible sources
    if ("entities.custom_tags.unified_product_name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app.name = "entities.custom_tags.unified_product_name")
    } else if ("entities.name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app.name = "entities.name")
    } else if ("name" %in% names(result_tbl) && !"app.name" %in% names(result_tbl)) {
      result_tbl <- dplyr::rename(result_tbl, app.name = "name")
    }

    # Create unified_app_name and unified_app_id for consistency
    if ("app.name" %in% names(result_tbl)) {
      result_tbl$unified_app_name <- result_tbl$app.name
    }

    # Note: unified_app_id was already set above before unnesting
    # The entities.app_id contains platform-specific IDs, not unified IDs

    # Store platform-specific app_id separately for reference
    if ("entities.app_id" %in% names(result_tbl) && !"platform_app_id" %in% names(result_tbl)) {
      result_tbl$platform_app_id <- result_tbl$entities.app_id
    }

    # If we have app_ids but missing app names, look them up
    if ((!"unified_app_name" %in% names(result_tbl) ||
      any(is.na(result_tbl$unified_app_name))) &&
      "unified_app_id" %in% names(result_tbl)) {
      result_tbl <- lookup_app_names_by_id(result_tbl)
    }

    # Extract custom metrics with clean names
    result_tbl <- extract_custom_metrics(result_tbl)

    # Clean special characters from numeric values
    result_tbl <- clean_numeric_values(result_tbl)

    # Convert date columns to proper Date class
    result_tbl <- clean_date_values(result_tbl)
  } else {
    # Even without enrichment, clean any numeric values that might have special characters
    result_tbl <- clean_numeric_values(result_tbl)

    # Convert date columns to proper Date class
    result_tbl <- clean_date_values(result_tbl)
  }

  return(result_tbl)
}

# Helper function to extract useful custom metrics from entities.custom_tags and aggregate_tags columns
extract_custom_metrics <- function(data) {
  if (nrow(data) == 0) {
    return(data)
  }

  # Use the global constant METRIC_MAPPING
  metrics_map <- METRIC_MAPPING

  # Also add variations for custom_tags and entities.custom_tags prefixes
  metrics_map <- c(
    metrics_map,
    stats::setNames(metrics_map, gsub("^aggregate_tags\\.", "custom_tags.", names(metrics_map))),
    stats::setNames(metrics_map, gsub("^aggregate_tags\\.", "entities.custom_tags.", names(metrics_map)))
  )
  metrics_map <- metrics_map[!duplicated(names(metrics_map))]

  # Extract only the metrics that exist in the data
  available_metrics <- intersect(names(metrics_map), names(data))

  if (length(available_metrics) > 0) {
    # Create a new data frame with the extracted metrics
    extracted_data <- data[, available_metrics, drop = FALSE]
    names(extracted_data) <- metrics_map[available_metrics]

    # Remove the original entities.custom_tags columns to avoid duplication
    data <- data[, !names(data) %in% available_metrics, drop = FALSE]

    # Bind the cleaned metrics back to the data using dplyr::bind_cols for safety
    data <- dplyr::bind_cols(data, extracted_data)
  }

  return(data)
}

# Helper function to clean special characters from numeric values
# Helper function to clean special characters from numeric values
clean_numeric_values <- function(data) {
  if (nrow(data) == 0) {
    return(data)
  }

  # Define patterns of metrics that should be treated as numeric
  # These patterns match the metric names we extract in extract_custom_metrics
  numeric_metric_patterns <- NUMERIC_METRIC_PATTERNS

  # Find columns that likely contain numeric data
  numeric_cols <- names(data)[sapply(names(data), function(col_name) {
    # Skip non-character columns
    if (!is.character(data[[col_name]])) {
      return(FALSE)
    }

    # Check if column name matches our numeric patterns
    matches_pattern <- any(sapply(numeric_metric_patterns, function(pattern) {
      grepl(pattern, col_name, ignore.case = TRUE)
    }))

    # Also check for aggregate_tags columns that might contain numeric data
    is_aggregate_tag <- grepl("^aggregate_tags\\.", col_name)

    # Check for entities.custom_tags columns that might be numeric
    is_custom_tag <- grepl("^entities\\.custom_tags\\.", col_name)

    # Skip obvious text columns
    is_text_column <- grepl("\\bname$|\\burl$|\\bdate$|app_id$|country$|gender|genre|style|category", col_name, ignore.case = TRUE)

    # If column matches our criteria and isn't obviously text, check the content
    if ((matches_pattern || is_aggregate_tag || is_custom_tag) && !is_text_column) {
      sample_values <- head(data[[col_name]][!is.na(data[[col_name]]) & data[[col_name]] != ""], 10)
      if (length(sample_values) > 0) {
        # Check if values contain digits with special characters OR pure numbers
        has_numeric_with_special <- any(grepl("[0-9].*[%$,]|[%$,].*[0-9]", sample_values))
        has_pure_numeric <- any(grepl("^[0-9]+\\.?[0-9]*$", sample_values))

        # Also check for formatted numbers like "1,234" or scientific notation
        has_formatted_numeric <- any(grepl("^[0-9,]+\\.?[0-9]*$", sample_values))

        return(has_numeric_with_special || has_pure_numeric || has_formatted_numeric)
      }
    }
    return(FALSE)
  })]

  # Clean each numeric column
  for (col in numeric_cols) {
    if (is.character(data[[col]])) {
      original_values <- data[[col]]

      # Handle percentage values (convert "45%" to 45, not 0.45)
      # Most analytics metrics use percentages as whole numbers
      has_percentages <- any(grepl("%", original_values, fixed = TRUE), na.rm = TRUE)

      # Remove currency symbols, commas, spaces, and other formatting
      cleaned_values <- gsub("[$,\\s]", "", original_values)

      # Remove percentage signs (but track that they were there)
      cleaned_values <- gsub("%", "", cleaned_values)

      # Remove any remaining non-numeric characters except decimal points and minus signs
      cleaned_values <- gsub("[^0-9.-]", "", cleaned_values)

      # Convert empty strings to NA
      cleaned_values[cleaned_values == ""] <- NA

      # Convert to numeric, suppressing warnings for values that can't be converted
      numeric_values <- suppressWarnings(as.numeric(cleaned_values))

      # Only replace if we successfully converted most values
      # This prevents accidentally converting text columns that happen to match patterns
      non_na_original <- sum(!is.na(original_values))
      non_na_converted <- sum(!is.na(numeric_values))

      if (non_na_original > 0) {
        conversion_rate <- non_na_converted / non_na_original
        if (conversion_rate > 0.5) { # If more than 50% converted successfully
          data[[col]] <- numeric_values

          # Special handling for retention metrics - convert to decimals
          if (grepl("retention", col, ignore.case = TRUE) && has_percentages) {
            # Convert retention percentages to decimals (15.5% becomes 0.155)
            data[[col]] <- numeric_values / 100
            message(sprintf("Converted column '%s' to numeric (percentages to decimals)", col))
          } else {
            # Log the cleaning for transparency
            if (has_percentages) {
              message(sprintf("Converted column '%s' to numeric (removed %% symbols)", col))
            } else {
              message(sprintf("Converted column '%s' to numeric", col))
            }
          }
        }
      }
    }
  }

  return(data)
}

# Global cache for app name lookups to avoid redundant API calls across function calls
.app_name_cache <- new.env()

# Helper function to lookup app names by app ID (for sales endpoint which doesn't provide names)
lookup_app_names_by_id <- function(data) {
  # Determine which ID column to use for lookups
  id_column <- NULL
  if ("entities.app_id" %in% names(data)) {
    id_column <- "entities.app_id"
  } else if ("unified_app_id" %in% names(data)) {
    id_column <- "unified_app_id"
  } else if ("app_id" %in% names(data)) {
    id_column <- "app_id"
  } else {
    return(data) # No ID column available
  }

  if (nrow(data) == 0) {
    return(data)
  }

  # Determine which rows still need a unified app name
  needs_lookup <- if ("unified_app_name" %in% names(data)) {
    is.na(data$unified_app_name) | data$unified_app_name == ""
  } else {
    rep(TRUE, nrow(data))
  }

  candidate_ids <- unique(as.character(data[[id_column]][needs_lookup]))
  candidate_ids <- candidate_ids[!is.na(candidate_ids) & candidate_ids != ""]

  if (length(candidate_ids) == 0) {
    if (!"unified_app_name" %in% names(data)) {
      data$unified_app_name <- data[[id_column]]
    }
    return(data)
  }

  verbose_lookup <- getOption("sensortowerR.verbose", getOption("SensorTowerR.verbose", FALSE))

  # For unified hex IDs, prefer direct /v1/unified/apps lookup.
  # This avoids search-based resolution returning IDs as fallback names.
  auth_token <- Sys.getenv("SENSORTOWER_AUTH_TOKEN")
  all_hex_ids <- length(candidate_ids) > 0 && all(grepl("^[a-f0-9]{24}$", candidate_ids))
  if (all_hex_ids && nzchar(auth_token)) {
    chunk_size <- 100
    id_chunks <- split(candidate_ids, ceiling(seq_along(candidate_ids) / chunk_size))
    direct_details <- lapply(id_chunks, function(ids) {
      tryCatch(
        st_app_details_impl(
          app_ids = ids,
          os = "unified",
          include_developer_contacts = FALSE,
          auth_token = auth_token
        ),
        error = function(e) NULL
      )
    })

    direct_details <- direct_details[!vapply(direct_details, is.null, logical(1))]
    if (length(direct_details) > 0) {
      direct_tbl <- dplyr::bind_rows(direct_details)
      if (nrow(direct_tbl) > 0 && all(c("app_id", "app_name") %in% names(direct_tbl))) {
        direct_lookup <- direct_tbl %>%
          dplyr::transmute(
            lookup_id = as.character(.data$app_id),
            resolved_app_name = as.character(.data$app_name),
            resolved_unified_id = as.character(.data$app_id)
          ) %>%
          dplyr::distinct(.data$lookup_id, .keep_all = TRUE)

        data <- dplyr::left_join(data, direct_lookup, by = stats::setNames("lookup_id", id_column))

        if (!"unified_app_name" %in% names(data)) {
          data$unified_app_name <- NA_character_
        }
        fill_names <- is.na(data$unified_app_name) | data$unified_app_name == ""
        data$unified_app_name[fill_names] <- dplyr::coalesce(
          data$resolved_app_name[fill_names],
          as.character(data[[id_column]][fill_names])
        )

        if (!"unified_app_id" %in% names(data)) {
          data$unified_app_id <- NA_character_
        }
        fill_unified <- is.na(data$unified_app_id) | data$unified_app_id == ""
        data$unified_app_id[fill_unified] <- dplyr::coalesce(
          data$resolved_unified_id[fill_unified],
          data$unified_app_id[fill_unified]
        )

        data <- data %>%
          dplyr::select(-dplyr::any_of(c("resolved_app_name", "resolved_unified_id")))

        unresolved <- is.na(data$unified_app_name) | data$unified_app_name == "" |
          grepl("^[a-f0-9]{24}$", data$unified_app_name)
        if (!any(unresolved)) {
          return(data)
        }
      }
    }
  }

  # Attempt to satisfy requests from the ID cache first
  cache_entries <- lapply(candidate_ids, lookup_cached_id)
  names(cache_entries) <- candidate_ids
  missing_ids <- candidate_ids[vapply(cache_entries, is.null, logical(1))]

  # Resolve any missing IDs using the shared resolution pipeline
  if (length(missing_ids) > 0) {
    if (nzchar(auth_token)) {
      if (verbose_lookup) {
        message("Resolving ", length(missing_ids), " app IDs for name enrichment...")
      }
      batch_resolve_ids(missing_ids, auth_token = auth_token, use_cache = TRUE, verbose = verbose_lookup)
      cache_entries[missing_ids] <- lapply(missing_ids, lookup_cached_id)
    } else if (verbose_lookup) {
      message("Skipping name lookup for ", length(missing_ids), " IDs (missing auth token)")
    }
  }

  # Build a lookup table from cache results
  lookup_table <- tibble::tibble(
    lookup_id = candidate_ids,
    resolved_app_name = NA_character_,
    resolved_unified_id = NA_character_,

  )

  for (i in seq_along(candidate_ids)) {
    entry <- cache_entries[[candidate_ids[i]]]
    if (!is.null(entry)) {
      name_val <- entry$app_name
      if (!is.null(name_val)) {
        name_val <- as.character(name_val)[1]
        if (!is.na(name_val) && nzchar(name_val)) {
          lookup_table$resolved_app_name[i] <- name_val
          assign(candidate_ids[i], name_val, envir = .app_name_cache)
        }
      }
      unified_val <- entry$unified_app_id
      if (!is.null(unified_val)) {
        unified_val <- as.character(unified_val)[1]
        if (!is.na(unified_val) && nzchar(unified_val)) {
          lookup_table$resolved_unified_id[i] <- unified_val
        }
      }
    }
  }

  join_cols <- stats::setNames("lookup_id", id_column)
  data <- dplyr::left_join(data, lookup_table, by = join_cols)

  if (!"unified_app_name" %in% names(data)) {
    data$unified_app_name <- NA_character_
  }
  fill_names <- is.na(data$unified_app_name) | data$unified_app_name == ""
  data$unified_app_name[fill_names] <- dplyr::coalesce(
    data$resolved_app_name[fill_names],
    as.character(data[[id_column]][fill_names])
  )

  if (!"unified_app_id" %in% names(data)) {
    data$unified_app_id <- NA_character_
  }
  fill_unified <- is.na(data$unified_app_id) | data$unified_app_id == ""
  data$unified_app_id[fill_unified] <- dplyr::coalesce(
    data$resolved_unified_id[fill_unified],
    data$unified_app_id[fill_unified]
  )

  data <- data %>%
    dplyr::select(-dplyr::any_of(c("resolved_app_name", "resolved_unified_id")))

  data
}

# Helper function to deduplicate by a specific grouping column
deduplicate_by_group_id <- function(data, group_col) {
  if (nrow(data) == 0 || !group_col %in% names(data)) {
    return(data)
  }

  # Separate numeric and non-numeric columns
  numeric_cols <- names(data)[sapply(data, is.numeric)]
  non_numeric_cols <- setdiff(names(data), numeric_cols)

  # Metrics to sum (downloads, revenue, counts)
  sum_metrics <- numeric_cols[grepl("downloads|revenue|units|count|absolute", numeric_cols, ignore.case = TRUE)]

  # Metrics to average (DAU, MAU, WAU, retention, ratings, percentages)
  avg_metrics <- numeric_cols[grepl("dau|mau|wau|retention|rating|percentage|arpdau|rpd|avg|average", numeric_cols, ignore.case = TRUE)]

  # Everything else (first value)
  other_metrics <- setdiff(numeric_cols, c(sum_metrics, avg_metrics))

  # Group by the specified column and aggregate
  result <- data %>%
    dplyr::group_by(!!rlang::sym(group_col)) %>%
    dplyr::summarise(
      # Keep the first unified_app_name
      unified_app_name = dplyr::first(.data$unified_app_name),

      # Keep first unified_app_id
      unified_app_id = dplyr::first(.data$unified_app_id),

      # Sum metrics that should be additive
      dplyr::across(dplyr::all_of(sum_metrics), ~ sum(.x, na.rm = TRUE)),

      # Average metrics that should be averaged
      dplyr::across(dplyr::all_of(avg_metrics), ~ mean(.x, na.rm = TRUE)),

      # Keep first value for other numeric metrics
      dplyr::across(dplyr::all_of(other_metrics), ~ dplyr::first(.x[!is.na(.x)])),

      # Keep first value for non-numeric columns
      dplyr::across(
        dplyr::all_of(setdiff(non_numeric_cols, c("unified_app_name", "unified_app_id", group_col))),
        ~ dplyr::first(.x[!is.na(.x)])
      ),

      # Keep first non-NA date
      dplyr::across(where(lubridate::is.Date), ~ dplyr::first(.x[!is.na(.x)])),
      dplyr::across(where(lubridate::is.POSIXt), ~ dplyr::first(.x[!is.na(.x)])),
      .groups = "drop"
    ) %>%
    # Remove the grouping column if it starts with a dot (temporary column)
    {
      if (startsWith(group_col, ".")) dplyr::select(., -!!rlang::sym(group_col)) else .
    }

  # Convert 0 values back to NA where appropriate for averaged metrics
  for (col in avg_metrics) {
    if (col %in% names(result)) {
      result[[col]][result[[col]] == 0] <- NA
    }
  }

  return(result)
}

# Helper function to deduplicate apps by consolidating metrics for the same app name
deduplicate_apps_by_name <- function(data, fuzzy_match = TRUE) {
  if (nrow(data) == 0 || !"unified_app_name" %in% names(data)) {
    return(data)
  }

  # Create normalized names first to check for duplicates
  if (fuzzy_match) {
    # Group similar app names more aggressively
    data <- data %>%
      dplyr::mutate(
        .name_normalized = .data$unified_app_name %>%
          # Remove special characters and symbols
          gsub("\u2122|\u00AE|\u00A9|:|\\*|\u00A4", "", .) %>%
          # Handle specific known patterns
          gsub("NYT Games.*|NYTimes.*", "nyt crossword", ., ignore.case = TRUE) %>%
          gsub("Scrabble.*GO.*", "scrabble go", ., ignore.case = TRUE) %>%
          gsub("Words With Friends.*", "words with friends", ., ignore.case = TRUE) %>%
          gsub("Elevate.*Brain.*", "elevate brain training", ., ignore.case = TRUE) %>%
          gsub("Word Trip.*|WordTrip.*", "word trip", ., ignore.case = TRUE) %>%
          gsub("Heads Up.*|Warner.*Heads Up", "heads up", ., ignore.case = TRUE) %>%
          gsub("Word Connect.*", "word connect", ., ignore.case = TRUE) %>%
          # General cleanup
          gsub("\\s+", " ", .) %>%
          trimws() %>%
          tolower()
      )
  } else {
    # Simple normalization
    data <- data %>%
      dplyr::mutate(
        .name_normalized = tolower(trimws(.data$unified_app_name))
      )
  }

  # Check if there are actually duplicates to consolidate (after normalization)
  if (length(unique(data$.name_normalized)) == nrow(data)) {
    return(dplyr::select(data, -dplyr::any_of(".name_normalized"))) # No duplicates, return as-is
  }

  message(sprintf(
    "Consolidating %d app entries into %d unique apps...",
    nrow(data), length(unique(data$.name_normalized))
  ))

  # Identify numeric columns to sum vs average
  numeric_cols <- names(data)[sapply(data, is.numeric)]

  # Metrics to SUM (additive across platforms)
  # IMPORTANT: DAU/MAU/WAU/users are NOT additive - same user can be on multiple platforms
  sum_metrics <- numeric_cols[grepl(
    "downloads|revenue|units|count|absolute",
    numeric_cols,
    ignore.case = TRUE
  )]
  # Explicitly exclude user metrics from sum
  sum_metrics <- sum_metrics[!grepl("dau|mau|wau|users", sum_metrics, ignore.case = TRUE)]

  # Metrics to TAKE MAX (user counts - same users across platforms)
  max_metrics <- numeric_cols[grepl(
    "dau|mau|wau|users",
    numeric_cols,
    ignore.case = TRUE
  )]
  # Exclude from max if it's a ratio/rate
  max_metrics <- max_metrics[!grepl("rate|ratio|percent", max_metrics, ignore.case = TRUE)]

  # Metrics to AVERAGE (rates, percentages, ratios)
  avg_metrics <- numeric_cols[grepl(
    "retention|rpd|rating|age|share|percent|rate|ratio|transformed",
    numeric_cols,
    ignore.case = TRUE
  )]

  # Everything else (first value)
  other_metrics <- setdiff(numeric_cols, c(sum_metrics, avg_metrics, max_metrics))

  # Group by normalized name
  result <- tryCatch(
    {
      data %>%
        dplyr::group_by(.data$.name_normalized) %>%
        dplyr::summarise(
          # Keep the first unified_app_name
          unified_app_name = dplyr::first(.data$unified_app_name),

          # Keep first unified_app_id
          unified_app_id = dplyr::first(.data$unified_app_id),

          # Sum metrics that should be additive
          dplyr::across(dplyr::all_of(sum_metrics), ~ sum(.x, na.rm = TRUE)),

          # Max for user metrics (same users across platforms)
          dplyr::across(dplyr::all_of(max_metrics), ~ max(.x, na.rm = TRUE)),

          # Average metrics that are rates/percentages
          dplyr::across(dplyr::all_of(avg_metrics), ~ mean(.x, na.rm = TRUE)),

          # First value for other metrics
          dplyr::across(dplyr::all_of(other_metrics), ~ dplyr::first(.x[!is.na(.x)])),

          # First value for character/date columns
          dplyr::across(where(is.character), ~ dplyr::first(.x[!is.na(.x) & .x != ""])),
          dplyr::across(where(lubridate::is.Date), ~ dplyr::first(.x[!is.na(.x)])),
          dplyr::across(where(lubridate::is.POSIXt), ~ dplyr::first(.x[!is.na(.x)])),
          .groups = "drop"
        ) %>%
        dplyr::relocate(unified_app_name, unified_app_id) %>%
        dplyr::select(-dplyr::any_of(".name_normalized"))
    },
    error = function(e) {
      message("Warning: Deduplication failed, returning original data. Error: ", e$message)
      return(dplyr::select(data, -dplyr::any_of(".name_normalized")))
    }
  )

  # Convert 0 values back to NA where appropriate for averaged and max metrics
  for (col in c(avg_metrics, max_metrics)) {
    if (col %in% names(result)) {
      # For max metrics, -Inf means all were NA, convert to NA
      if (col %in% max_metrics) {
        result[[col]][is.infinite(result[[col]])] <- NA
      }
      # For avg metrics, 0 from all NAs should be NA
      result[[col]][result[[col]] == 0] <- NA
    }
  }

  return(result)
}

# Helper function to convert date columns to proper Date class
clean_date_values <- function(data) {
  if (nrow(data) == 0) {
    return(data)
  }

  # Find columns that likely contain date data
  date_cols <- names(data)[sapply(names(data), function(col_name) {
    # Skip non-character columns
    if (!is.character(data[[col_name]])) {
      return(FALSE)
    }

    # Check if column name suggests it contains dates
    is_date_column <- grepl("date|release", col_name, ignore.case = TRUE)

    # Skip columns that are clearly not dates (like frequency, days ago)
    is_non_date <- grepl("frequency|days ago|update|freq", col_name, ignore.case = TRUE)

    if (is_date_column && !is_non_date) {
      # Check if the column actually contains date-like values
      sample_values <- head(data[[col_name]][!is.na(data[[col_name]]) & data[[col_name]] != ""], 5)
      if (length(sample_values) > 0) {
        # Check for various date patterns
        has_iso_dates <- any(grepl("\\d{4}-\\d{2}-\\d{2}", sample_values))
        has_slash_dates <- any(grepl("\\d{4}/\\d{2}/\\d{2}", sample_values))
        has_datetime <- any(grepl("\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}", sample_values))

        return(has_iso_dates || has_slash_dates || has_datetime)
      }
    }
    return(FALSE)
  })]

  if (length(date_cols) == 0) {
    return(data)
  }

  # Convert each date column
  for (col in date_cols) {
    if (is.character(data[[col]])) {
      original_values <- data[[col]]

      # Try to convert to dates
      converted_dates <- tryCatch(
        {
          # Handle different date formats
          parsed_dates <- rep(as.Date(NA), length(original_values))

          # Handle ISO datetime format (e.g., "2025-07-01T00:00:00Z")
          iso_pattern <- grepl("\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}", original_values)
          if (any(iso_pattern, na.rm = TRUE)) {
            parsed_dates[iso_pattern] <- as.Date(lubridate::ymd_hms(original_values[iso_pattern]))
          }

          # Handle YYYY-MM-DD format
          ymd_pattern <- grepl("^\\d{4}-\\d{2}-\\d{2}$", original_values) & !iso_pattern
          if (any(ymd_pattern, na.rm = TRUE)) {
            parsed_dates[ymd_pattern] <- as.Date(original_values[ymd_pattern])
          }

          # Handle YYYY/MM/DD format
          slash_pattern <- grepl("^\\d{4}/\\d{2}/\\d{2}$", original_values)
          if (any(slash_pattern, na.rm = TRUE)) {
            parsed_dates[slash_pattern] <- as.Date(original_values[slash_pattern], format = "%Y/%m/%d")
          }

          parsed_dates
        },
        error = function(e) {
          # If conversion fails, return original
          original_values
        }
      )

      # Only replace if we successfully converted most values
      if (inherits(converted_dates, "Date")) {
        non_na_original <- sum(!is.na(original_values) & original_values != "")
        non_na_converted <- sum(!is.na(converted_dates))

        if (non_na_original > 0) {
          conversion_rate <- non_na_converted / non_na_original
          if (conversion_rate > 0.5) { # If more than 50% converted successfully
            data[[col]] <- converted_dates
            message(sprintf("Converted column '%s' to Date class", col))
          }
        }
      }
    }
  }

  return(data)
}

#' Clear App Name Cache
#'
#' Clears the internal cache of app name lookups. Useful for testing or when
#' you want to refresh app name data.
#'
#' @return No return value, called for side effects (clearing the cache).
#' @export
st_clear_app_cache <- function() {
  rm(list = ls(envir = .app_name_cache), envir = .app_name_cache)
  message("App name cache cleared")
}

#' Fetch and Unified Data from Platforms
#'
#' Fetches data from iOS and/or Android and optionally combines them.
#' Handles missing data from one platform gracefully.
#'
#' @param ios_app_id Character. iOS App ID.
#' @param android_app_id Character. Android App ID.
#' @param start_date Date. Start date.
#' @param end_date Date. End date.
#' @param countries Character vector. Country codes.
#' @param date_granularity Character. Granularity.
#' @param auth_token Character. API token.
#' @param verbose Logical. Verbose output.
#' @param combine_to_unified Logical. Whether to sum metrics into a unified view.
#'
#' @return A tibble with columns date, country, revenue, downloads, and optionally platform/app_id.
#' @keywords internal
fetch_unified_data <- function(
  ios_app_id = NULL,
  android_app_id = NULL,
  start_date,
  end_date,
  countries,
  date_granularity,
  auth_token,
  verbose = FALSE,
  combine_to_unified = TRUE
) {
  all_data <- tibble::tibble()

  # Track what we requested vs what we got
  ios_requested <- !is.null(ios_app_id) && !is.na(ios_app_id)
  android_requested <- !is.null(android_app_id) && !is.na(android_app_id)
  ios_has_data <- FALSE
  android_has_data <- FALSE

  # Fetch iOS data
  if (ios_requested) {
    if (verbose) message("Fetching iOS data for: ", ios_app_id)
    ios_result <- tryCatch(
      {
        st_sales_report_impl(
          os = "ios",
          ios_app_id = ios_app_id,
          countries = countries,
          start_date = start_date,
          end_date = end_date,
          date_granularity = date_granularity,
          auth_token = auth_token,
          verbose = FALSE # Suppress inner verbose to avoid noise
        )
      },
      error = function(e) {
        if (verbose) message("iOS fetch warning: ", e$message)
        NULL
      }
    )

    if (!is.null(ios_result) && nrow(ios_result) > 0) {
      ios_has_data <- TRUE
      # Standardize columns
      ios_result <- ios_result %>%
        dplyr::mutate(
          platform = "ios",
          app_id = as.character(ios_app_id),
          app_id_type = "ios",
          revenue = if ("total_revenue" %in% names(.)) total_revenue else if ("revenue" %in% names(.)) revenue else 0,
          downloads = if ("total_downloads" %in% names(.)) total_downloads else if ("downloads" %in% names(.)) downloads else 0
        ) %>%
        dplyr::select(date, country, revenue, downloads, platform, app_id, app_id_type)

      all_data <- dplyr::bind_rows(all_data, ios_result)
    }
  }

  # Fetch Android data
  if (android_requested) {
    if (verbose) message("Fetching Android data for: ", android_app_id)
    android_result <- tryCatch(
      {
        st_sales_report_impl(
          os = "android",
          android_app_id = android_app_id,
          countries = countries,
          start_date = start_date,
          end_date = end_date,
          date_granularity = date_granularity,
          auth_token = auth_token,
          verbose = FALSE
        )
      },
      error = function(e) {
        if (verbose) message("Android fetch warning: ", e$message)
        NULL
      }
    )

    if (!is.null(android_result) && nrow(android_result) > 0) {
      android_has_data <- TRUE
      # Standardize columns
      android_result <- android_result %>%
        dplyr::mutate(
          platform = "android",
          app_id = as.character(android_app_id),
          app_id_type = "android",
          country = if ("c" %in% names(.)) c else country
        ) %>%
        dplyr::select(date, country, revenue, downloads, platform, app_id, app_id_type)

      all_data <- dplyr::bind_rows(all_data, android_result)
    }
  }

  # Handle missing data scenarios
  if (ios_requested && android_requested) {
    if (!ios_has_data && !android_has_data) {
      if (verbose) message("No data available for either platform.")
      return(tibble::tibble(date = as.Date(character()), country = character(), revenue = numeric(), downloads = numeric()))
    } else if (!ios_has_data && verbose) {
      message("Missing iOS data (returning Android only).")
    } else if (!android_has_data && verbose) {
      message("Missing Android data (returning iOS only).")
    }
  } else if ((ios_requested && !ios_has_data) || (android_requested && !android_has_data)) {
    if (verbose) message("No data returned for requested platform.")
    return(tibble::tibble(date = as.Date(character()), country = character(), revenue = numeric(), downloads = numeric()))
  }

  # Combine or return raw
  if (nrow(all_data) > 0) {
    if (combine_to_unified) {
      combined <- all_data %>%
        dplyr::group_by(date, country) %>%
        dplyr::summarise(
          revenue = sum(revenue, na.rm = TRUE),
          downloads = sum(downloads, na.rm = TRUE),
          .groups = "drop"
        )
      return(combined)
    } else {
      return(all_data)
    }
  }

  return(all_data)
}

# --- v1.0.0 shared parameter normalizers ---
# Used by the unified st_metrics(), st_rankings(), st_app(), st_apps(), st_filter()

normalize_os <- function(os) {
  if (is.null(os) || !length(os)) {
    rlang::abort("`os` must be one of 'ios', 'android', or 'unified'.")
  }
  os <- tolower(as.character(os[1]))
  if (!os %in% c("ios", "android", "unified")) {
    rlang::abort(sprintf("`os` must be one of 'ios', 'android', or 'unified' (got '%s').", os))
  }
  os
}

normalize_country <- function(country, allow_null = FALSE) {
  if (is.null(country) || !length(country)) {
    if (allow_null) return(NULL)
    rlang::abort("`country` must be a non-empty character scalar (e.g., 'US', 'WW').")
  }
  country <- toupper(as.character(country[1]))
  if (!grepl("^[A-Z]{2}$", country)) {
    rlang::abort(sprintf("`country` must be a 2-letter code (got '%s').", country))
  }
  country
}

normalize_countries <- function(countries, allow_null = FALSE) {
  if (is.null(countries) || !length(countries)) {
    if (allow_null) return(NULL)
    rlang::abort("`countries` must be a non-empty character vector (e.g., 'WW' or c('US','JP')).")
  }
  countries <- toupper(as.character(countries))
  bad <- countries[!grepl("^[A-Z]{2}$", countries)]
  if (length(bad)) {
    rlang::abort(sprintf("`countries` entries must be 2-letter codes (bad: %s).",
                         paste(bad, collapse = ", ")))
  }
  unique(countries)
}

normalize_dates <- function(date_from, date_to) {
  df <- tryCatch(as.Date(date_from), error = function(e) NA)
  dt <- tryCatch(as.Date(date_to),   error = function(e) NA)
  if (is.na(df) || is.na(dt)) {
    rlang::abort("`date_from` and `date_to` must be Date or ISO date strings (YYYY-MM-DD).")
  }
  if (df > dt) {
    rlang::abort("`date_from` must be on or before `date_to`.")
  }
  list(date_from = df, date_to = dt)
}

normalize_granularity <- function(granularity) {
  allowed <- c("daily", "weekly", "monthly", "quarterly")
  if (is.null(granularity) || !length(granularity)) {
    rlang::abort(sprintf("`granularity` must be one of %s.",
                         paste(sprintf("'%s'", allowed), collapse = ", ")))
  }
  granularity <- tolower(as.character(granularity[1]))
  if (!granularity %in% allowed) {
    rlang::abort(sprintf("`granularity` must be one of %s (got '%s').",
                         paste(sprintf("'%s'", allowed), collapse = ", "), granularity))
  }
  granularity
}

normalize_metrics <- function(metrics, allowed) {
  if (is.null(metrics) || !length(metrics)) {
    rlang::abort(sprintf("`metrics` must be one or more of %s.",
                         paste(sprintf("'%s'", allowed), collapse = ", ")))
  }
  metrics <- tolower(as.character(metrics))
  bad <- setdiff(metrics, allowed)
  if (length(bad)) {
    rlang::abort(sprintf("Unknown metric(s): %s. Allowed: %s.",
                         paste(sprintf("'%s'", bad), collapse = ", "),
                         paste(sprintf("'%s'", allowed), collapse = ", ")))
  }
  unique(metrics)
}

# Alias kept for the public-facing helper name used across v1.0.0 docs.
get_auth_token <- function(auth_token = NULL, env_var = "SENSORTOWER_AUTH_TOKEN",
                           error_message = NULL) {
  resolve_auth_token(auth_token, env_var = env_var, error_message = error_message)
}

test_that("enrich_response preserves unified_app_id from original app_id", {
  skip("enrich_response is an internal function - testing through st_top_charts instead")

  # Check that unified_app_id was created from the original app_id
  expect_true(
    "unified_app_id" %in% names(enriched),
    "unified_app_id should be created during enrichment"
  )

  # Check that the unified_app_id values are the original hex IDs
  expect_equal(
    unique(enriched$unified_app_id),
    c("5a998ab0ae1f4d57ea38e0f8", "5919fd19ee05f407440011a7"),
    "unified_app_id should preserve the original hex app_id values"
  )

  # Check that entities.app_id contains the platform-specific IDs
  expect_true(
    "entities.app_id" %in% names(enriched) || "platform_app_id" %in% names(enriched),
    "Platform-specific IDs should be preserved separately"
  )
})

test_that("enrich_response handles missing app_id gracefully", {
  skip("enrich_response is an internal function - testing through st_top_charts instead")
})

test_that("deduplication consolidates apps with same unified_app_id", {
  skip_if_no_auth()

  # Create a scenario where we have duplicate unified_app_ids
  # This simulates what happens after unnesting
  mock_duplicated_data <- data.frame(
    unified_app_id = c("5a998ab0ae1f4d57ea38e0f8", "5a998ab0ae1f4d57ea38e0f8"),
    unified_app_name = c("Bible Word Puzzle", "Bible Word Puzzle"),
    platform_app_id = c("1342112505", "bible.wordgame.words.connect.crossword.cookies"),
    dau_30d_us = c(165000, 165000),
    downloads_30d_ww = c(1350000, 1350000),
    revenue_30d_ww = c(19200, 19200),
    stringsAsFactors = FALSE
  )

  # Apply deduplication using internal function with namespace
  deduplicated <- sensortowerR:::deduplicate_by_group_id(mock_duplicated_data, "unified_app_id")

  # Should have only one row
  expect_equal(nrow(deduplicated), 1)

  # DAU should be averaged (not summed) for duplicates
  expect_equal(deduplicated$dau_30d_us, 165000)

  # Downloads should be summed for duplicates
  expect_equal(deduplicated$downloads_30d_ww, 2700000)
})

test_that("st_top_charts handles all scenarios correctly", {
  skip_if_no_auth()

  # Test multiple scenarios to ensure robustness
  scenarios <- list(
    list(
      name = "Word games with custom filter",
      params = list(
        os = "unified",
        category = 0,
        custom_fields_filter_id = "603697f4241bc16eb8570d37",
        custom_tags_mode = "include_unified_apps",
        measure = "DAU",
        regions = "US",
        limit = 5
      )
    ),
    list(
      name = "Regular category without custom filter",
      params = list(
        os = "unified",
        category = 7003, # Puzzle games
        measure = "revenue",
        regions = "US",
        limit = 5
      )
    )
  )

  for (scenario in scenarios) {
    result <- do.call(sensortowerR:::st_top_charts_impl, c(
      scenario$params,
      list(enrich_response = TRUE, deduplicate_apps = TRUE)
    ))

    # Basic checks
    expect_s3_class(result, "data.frame")

    expect_true(
      "unified_app_id" %in% names(result),
      sprintf("Scenario '%s' should have unified_app_id column", scenario$name)
    )

    # Check for duplicates
    dup_count <- sum(duplicated(result$unified_app_id))
    expect_true(dup_count == 0,
      label = sprintf("Scenario '%s' should have no duplicate unified_app_ids", scenario$name)
    )
  }
})

ns <- getNamespace("sensortowerR")

local_clean_cache <- function() {
  env <- ns$.sensortowerR_env
  env$id_cache <- list()
  rm(list = ls(envir = ns$.app_name_cache), envir = ns$.app_name_cache)
}

test_that("extract_custom_metrics handles aggregate and custom tag columns", {
  df <- tibble::tibble(
    `aggregate_tags.Last 30 Days Average DAU (US)` = "1,234",
    `aggregate_tags.Last 30 Days Average DAU (WW)` = "5,678",
    `custom_tags.Day 1 Retention (Last Quarter, US)` = "45%",
    `custom_tags.Day 1 Retention (Last Quarter, WW)` = "55%",
    `entities.custom_tags.Game Genre` = "Strategy"
  )

  processed <- ns$extract_custom_metrics(df)
  processed <- ns$clean_numeric_values(processed)

  expect_equal(processed$dau_30d_us, 1234)
  expect_equal(processed$dau_30d_ww, 5678)
  expect_equal(processed$retention_1d_us, 0.45)
  expect_equal(processed$retention_1d_ww, 0.55)
  expect_equal(processed$game_genre, "Strategy")
})

test_that("extract_custom_metrics maps custom tags when aggregate tags missing", {
  df <- tibble::tibble(
    `custom_tags.Last 30 Days Average DAU (US)` = "9,876",
    `custom_tags.Last 30 Days Average DAU (WW)` = "12,345",
    `custom_tags.Day 7 Retention (Last Quarter, US)` = "32%",
    `custom_tags.Day 7 Retention (Last Quarter, WW)` = "38%"
  )

  processed <- ns$extract_custom_metrics(df)
  processed <- ns$clean_numeric_values(processed)

  expect_equal(processed$dau_30d_us, 9876)
  expect_equal(processed$dau_30d_ww, 12345)
  expect_equal(processed$retention_7d_us, 0.32)
  expect_equal(processed$retention_7d_ww, 0.38)
})

test_that("lookup_app_names_by_id prefers cache metadata", {
  on.exit(local_clean_cache(), add = TRUE)

  ns <- getNamespace("sensortowerR")
  ns$cache_id_mapping(
    input_id = "123",
    ios_id = "123",
    android_id = "com.example.app",
    unified_id = "abc123abc123abc123abc123",
    app_name = "Example App"
  )

  df <- tibble::tibble(
    `entities.app_id` = "123",
    unified_app_name = NA_character_,
    unified_app_id = NA_character_
  )

  result <- ns$lookup_app_names_by_id(df)
  expect_equal(result$unified_app_name, "Example App")
  expect_equal(result$unified_app_id, "abc123abc123abc123abc123")
})

test_that("endpoint helpers build expected paths", {
  expect_equal(
    sensortowerR:::st_endpoint_path("usage_active_users", os = "ios"),
    "v1/ios/usage/active_users"
  )

  expect_equal(
    sensortowerR:::st_endpoint_path("facets_metrics"),
    "v1/facets/metrics"
  )

  expect_equal(
    sensortowerR:::st_endpoint_relative_path("search_entities", app_store = "unified"),
    "unified/search_entities"
  )

  expect_error(
    sensortowerR:::st_endpoint_path("usage_active_users"),
    "Missing required endpoint placeholder"
  )

  expect_error(
    sensortowerR:::st_endpoint_path("not_a_real_endpoint"),
    "Unknown endpoint key"
  )
})

test_that("resolve_auth_token supports direct and env-token resolution", {
  expect_equal(
    sensortowerR:::resolve_auth_token("  token_123  "),
    "token_123"
  )

  previous <- Sys.getenv("SENSORTOWER_AUTH_TOKEN", unset = NA_character_)
  on.exit({
    if (is.na(previous)) {
      Sys.unsetenv("SENSORTOWER_AUTH_TOKEN")
    } else {
      Sys.setenv(SENSORTOWER_AUTH_TOKEN = previous)
    }
  }, add = TRUE)

  Sys.setenv(SENSORTOWER_AUTH_TOKEN = "from_env")
  expect_equal(
    sensortowerR:::resolve_auth_token(NULL),
    "from_env"
  )

  Sys.setenv(SENSORTOWER_AUTH_TOKEN = "")
  expect_error(
    sensortowerR:::resolve_auth_token(NULL),
    "Authentication token not found"
  )
})

test_that("st_facets_metrics validates raw query fragments", {
  expect_error(
    st_facets_metrics(query = 1, auth_token = "token"),
    "`query` must be a character vector"
  )

  expect_error(
    st_facets_metrics(params = "nope", auth_token = "token"),
    "`params` must be a named list"
  )
})

test_that("st_retention_facets validates entity filters and breakdowns", {
  expect_error(
    st_retention_facets(
      start_date = "2025-01-01",
      end_date = "2025-01-31",
      auth_token = "token"
    ),
    "Exactly one of `app_ids` or `unified_app_ids` must be supplied"
  )

  expect_error(
    st_retention_facets(
      app_ids = "553834731",
      unified_app_ids = "5f16a8019f7b275235017614",
      start_date = "2025-01-01",
      end_date = "2025-01-31",
      auth_token = "token"
    ),
    "Exactly one of `app_ids` or `unified_app_ids` must be supplied"
  )

  expect_error(
    st_retention_facets(
      app_ids = "553834731",
      breakdown = c("date", "region"),
      start_date = "2025-01-01",
      end_date = "2025-01-31",
      auth_token = "token"
    ),
    "Unsupported `breakdown` field"
  )

  expect_error(
    st_retention_facets(
      app_ids = "553834731",
      breakdown = c("date", "app_id", "unified_app_id", "extra"),
      start_date = "2025-01-01",
      end_date = "2025-01-31",
      auth_token = "token"
    ),
    "Unsupported `breakdown` field"
  )

  expect_error(
    st_retention_facets(
      app_ids = "553834731",
      start_date = "2025-02-01",
      end_date = "2025-01-31",
      auth_token = "token"
    ),
    "`end_date` must be on or after `start_date`"
  )
})

test_that("st_ratings_facets validates required parameters", {
  expect_error(
    st_ratings_facets(
      start_date = "2024-01-01",
      end_date = "2024-01-07",
      auth_token = "token"
    ),
    "`app_ids` must contain at least one app ID"
  )

  expect_error(
    st_ratings_facets(
      app_ids = "553834731",
      breakdown = c("app_id", "date"),
      start_date = "2024-01-01",
      end_date = "2024-01-07",
      auth_token = "token"
    ),
    "`date_granularity` is required when `breakdown` includes `date`"
  )

  expect_error(
    st_ratings_facets(
      app_ids = "553834731",
      breakdown = c("app_id", "region"),
      start_date = "2024-01-01",
      end_date = "2024-01-07",
      date_granularity = "day",
      auth_token = "token"
    ),
    "Unsupported `breakdown` for ratings facets"
  )
})

test_that("st_reviews_by_rating_facets validates required parameters", {
  expect_error(
    st_reviews_by_rating_facets(
      start_date = "2024-01-01",
      end_date = "2024-01-07",
      auth_token = "token"
    ),
    "`app_id` must be a single non-empty app ID"
  )

  expect_error(
    st_reviews_by_rating_facets(
      app_id = "553834731",
      breakdown = c("date", "review_rating"),
      start_date = "2024-01-01",
      end_date = "2024-01-07",
      auth_token = "token"
    ),
    "`date_granularity` is required when `breakdown` includes `date`"
  )

  expect_error(
    st_reviews_by_rating_facets(
      app_id = "553834731",
      breakdown = c("region", "language"),
      start_date = "2024-01-01",
      end_date = "2024-01-07",
      date_granularity = "day",
      auth_token = "token"
    ),
    "Unsupported `breakdown` for reviews-by-rating facets"
  )
})

mock_st_response <- function(payload) {
  httr2::response(
    status_code = 200,
    body = charToRaw(jsonlite::toJSON(payload, auto_unbox = TRUE))
  )
}

expect_no_title_level_requests <- function(urls) {
  forbidden <- c(
    "ranking",
    "top_and_trending",
    "sales_report",
    "apps/timeseries",
    "app_ids",
    "unified_app_ids"
  )
  pattern <- paste(forbidden, collapse = "|")
  testthat::expect_false(any(grepl(pattern, urls)))
}

test_that("st_market_metrics returns tidy long iOS metrics from one aggregate request", {
  requested_urls <- character()

  testthat::local_mocked_bindings(
    perform_request = function(req) {
      requested_urls <<- c(requested_urls, req$url)
      mock_st_response(list(
        list(ca = 7006, cc = "US", d = "2024-01-01T00:00:00Z", iu = 10, au = 5, ir = 1000, ar = 500)
      ))
    },
    st_batch_metrics_impl = function(...) stop("title-level batch path should not be used"),
    st_top_charts_impl = function(...) stop("top charts path should not be used"),
    .package = "sensortowerR"
  )

  result <- st_market_metrics(
    category = 7006,
    countries = "US",
    os = "ios",
    date_from = "2024-01-01",
    date_to = "2024-01-31",
    granularity = "monthly",
    auth_token = "token"
  )

  expect_equal(length(requested_urls), 1)
  expect_true(grepl("/v1/ios/games_breakdown", requested_urls, fixed = TRUE))
  expect_no_title_level_requests(requested_urls)
  expect_equal(
    names(result),
    c("date", "country", "category_id", "os", "metric", "value")
  )
  expect_equal(nrow(result), 2)
  expect_equal(result$value[result$metric == "downloads"], 15)
  expect_equal(result$value[result$metric == "revenue"], 15)
})

test_that("st_market_metrics returns tidy wide Android metrics with dollar revenue", {
  requested_urls <- character()

  testthat::local_mocked_bindings(
    perform_request = function(req) {
      requested_urls <<- c(requested_urls, req$url)
      mock_st_response(list(
        list(ca = "game_casino", cc = "US", d = "2024-01-01T00:00:00Z", u = 20, r = 2500)
      ))
    },
    .package = "sensortowerR"
  )

  result <- st_market_metrics(
    category = "game_casino",
    countries = "US",
    os = "android",
    date_from = "2024-01-01",
    date_to = "2024-01-31",
    granularity = "monthly",
    shape = "wide",
    auth_token = "token"
  )

  expect_equal(length(requested_urls), 1)
  expect_true(grepl("/v1/android/games_breakdown", requested_urls, fixed = TRUE))
  expect_equal(
    names(result),
    c("date", "country", "category_id", "os", "revenue_usd", "downloads")
  )
  expect_equal(result$revenue_usd, 25)
  expect_equal(result$downloads, 20)
})

test_that("st_market_metrics unified makes exactly two aggregate requests", {
  requested_urls <- character()

  testthat::local_mocked_bindings(
    perform_request = function(req) {
      requested_urls <<- c(requested_urls, req$url)

      if (grepl("/v1/ios/games_breakdown", req$url, fixed = TRUE)) {
        return(mock_st_response(list(
          list(ca = 7006, cc = "US", d = "2024-01-01T00:00:00Z", iu = 10, au = 5, ir = 1000, ar = 500)
        )))
      }

      if (grepl("/v1/android/games_breakdown", req$url, fixed = TRUE)) {
        return(mock_st_response(list(
          list(ca = "game_casino", cc = "US", d = "2024-01-01T00:00:00Z", u = 20, r = 2500)
        )))
      }

      stop("Unexpected request URL: ", req$url)
    },
    st_apps = function(...) stop("app discovery should not be used"),
    st_rankings = function(...) stop("rankings should not be used"),
    st_metrics = function(...) stop("app metrics should not be used"),
    st_batch_metrics_impl = function(...) stop("title-level batch path should not be used"),
    st_top_charts_impl = function(...) stop("top charts path should not be used"),
    .package = "sensortowerR"
  )

  result <- st_market_metrics(
    category = c("7006", "game_casino"),
    countries = c("US", "GB"),
    os = "unified",
    date_from = "2024-01-01",
    date_to = "2024-01-31",
    granularity = "monthly",
    shape = "wide",
    auth_token = "token"
  )

  expect_equal(length(requested_urls), 2)
  expect_true(any(grepl("/v1/ios/games_breakdown", requested_urls, fixed = TRUE)))
  expect_true(any(grepl("/v1/android/games_breakdown", requested_urls, fixed = TRUE)))
  expect_no_title_level_requests(requested_urls)
  expect_equal(nrow(result), 1)
  expect_equal(result$os, "unified")
  expect_equal(result$category_id, "7006,game_casino")
  expect_equal(result$revenue_usd, 40)
  expect_equal(result$downloads, 35)
})

test_that("st_market_metrics preserves stable schemas for empty responses", {
  testthat::local_mocked_bindings(
    perform_request = function(req) mock_st_response(list()),
    .package = "sensortowerR"
  )

  long_result <- st_market_metrics(
    category = 7006,
    countries = "US",
    os = "ios",
    date_from = "2024-01-01",
    date_to = "2024-01-31",
    auth_token = "token"
  )
  wide_result <- st_market_metrics(
    category = 7006,
    countries = "US",
    os = "ios",
    date_from = "2024-01-01",
    date_to = "2024-01-31",
    shape = "wide",
    auth_token = "token"
  )

  expect_equal(nrow(long_result), 0)
  expect_equal(
    names(long_result),
    c("date", "country", "category_id", "os", "metric", "value")
  )
  expect_equal(nrow(wide_result), 0)
  expect_equal(
    names(wide_result),
    c("date", "country", "category_id", "os", "revenue_usd", "downloads")
  )
})

# Process-local cache for standardized st_metrics() results
.st_metrics_cache <- new.env(parent = emptyenv())

#' Fetch Sensor Tower Metrics
#'
#' Thin v1.0.0 facade for sales/download metrics. The function validates a
#' standardized argument set, dispatches to the legacy implementation that best
#' matches the request, and then normalizes the result into a stable schema.
#'
#' @param app_id Character scalar or vector of app identifiers. Each entry can be
#'   a Sensor Tower unified app ID, an iOS numeric app ID, or an Android package
#'   name.
#' @param metrics Character vector of metrics to return. Supported values are
#'   `"revenue"` and `"downloads"`.
#' @param os Character scalar. One of `"ios"`, `"android"`, or `"unified"`.
#' @param countries Character vector of 2-letter country codes. Use `"WW"` for
#'   worldwide aggregates.
#' @param date_from,date_to Date bounds for the query. Accept `Date` objects or
#'   ISO date strings.
#' @param granularity Character scalar. One of `"daily"`, `"weekly"`,
#'   `"monthly"`, or `"quarterly"`.
#' @param revenue_unit Character scalar. `"dollars"` (default) returns revenue in
#'   base currency units. `"cents"` returns revenue in cents for compatibility
#'   with the legacy API surface.
#' @param shape Character scalar. `"long"` returns one row per metric observation.
#'   `"wide"` returns one row per app/date/country with separate metric columns.
#' @param cache Logical. If `TRUE`, use a process-local cache keyed on the
#'   normalized arguments.
#' @param auth_token Optional Sensor Tower API token. If `NULL`, falls back to
#'   `SENSORTOWER_AUTH_TOKEN`.
#'
#' @return
#' If `shape = "long"`, a tibble with columns:
#'   - `app_id`: identifier supplied to `st_metrics()`
#'   - `os`: normalized operating system
#'   - `country`: 2-letter country code
#'   - `date`: observation date
#'   - `metric`: one of `"revenue"` or `"downloads"`
#'   - `value`: metric value; revenue is in dollars by default and cents when
#'     `revenue_unit = "cents"`
#'
#' If `shape = "wide"`, a tibble with columns:
#'   - `app_id`
#'   - `os`
#'   - `country`
#'   - `date`
#'   - one numeric column per requested metric
#'
#' @examples
#' \dontrun{
#' st_metrics(
#'   app_id = "553834731",
#'   os = "ios",
#'   countries = "US",
#'   date_from = Sys.Date() - 30,
#'   date_to = Sys.Date() - 1
#' )
#'
#' st_metrics(
#'   app_id = c("553834731", "com.supercell.clashofclans"),
#'   os = "unified",
#'   shape = "wide",
#'   countries = c("US", "GB"),
#'   revenue_unit = "cents"
#' )
#' }
#'
#' @export
st_metrics <- function(app_id,
                       metrics = c("revenue", "downloads"),
                       os = "unified",
                       countries = "WW",
                       date_from = Sys.Date() - 90,
                       date_to = Sys.Date(),
                       granularity = "daily",
                       revenue_unit = c("dollars", "cents"),
                       shape = c("long", "wide"),
                       cache = TRUE,
                       auth_token = NULL) {
  if (missing(app_id) || is.null(app_id) || !length(app_id)) {
    rlang::abort("`app_id` must be a non-empty character scalar or vector.")
  }

  app_id <- as.character(app_id)
  if (any(is.na(app_id) | !nzchar(app_id))) {
    rlang::abort("`app_id` entries must be non-empty strings.")
  }

  os <- normalize_os(os)
  countries <- normalize_countries(countries)
  dates <- normalize_dates(date_from, date_to)
  granularity <- normalize_granularity(granularity)
  metrics <- normalize_metrics(metrics, allowed = c("revenue", "downloads"))
  revenue_unit <- match.arg(revenue_unit)
  shape <- match.arg(shape)
  cache <- isTRUE(cache)
  auth_token <- get_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required.",
      "Set SENSORTOWER_AUTH_TOKEN or pass `auth_token`."
    )
  )

  cache_key <- .st_metrics_cache_key(
    app_id = app_id,
    metrics = metrics,
    os = os,
    countries = countries,
    date_from = dates$date_from,
    date_to = dates$date_to,
    granularity = granularity,
    revenue_unit = revenue_unit,
    shape = shape
  )

  if (cache && exists(cache_key, envir = .st_metrics_cache, inherits = FALSE)) {
    return(get(cache_key, envir = .st_metrics_cache, inherits = FALSE))
  }

  raw_result <- if (length(app_id) == 1L) {
    .st_metrics_dispatch_single(
      app_id = app_id,
      os = os,
      countries = countries,
      date_from = dates$date_from,
      date_to = dates$date_to,
      granularity = granularity,
      auth_token = auth_token
    )
  } else {
    st_batch_metrics_impl(
      os = os,
      app_list = app_id,
      metrics = metrics,
      date_range = list(
        start_date = dates$date_from,
        end_date = dates$date_to
      ),
      countries = countries,
      granularity = granularity,
      parallel = FALSE,
      verbose = FALSE,
      auth_token = auth_token
    )
  }

  normalized <- .st_metrics_normalize_output(
    data = raw_result,
    requested_ids = app_id,
    requested_countries = countries,
    os = os,
    metrics = metrics,
    revenue_unit = revenue_unit,
    shape = shape
  )

  if (cache) {
    assign(cache_key, normalized, envir = .st_metrics_cache)
  }

  normalized
}

.st_metrics_cache_key <- function(app_id,
                                  metrics,
                                  os,
                                  countries,
                                  date_from,
                                  date_to,
                                  granularity,
                                  revenue_unit,
                                  shape) {
  key_payload <- list(
    app_id = unname(app_id),
    metrics = unname(metrics),
    os = os,
    countries = unname(countries),
    date_from = format(date_from, "%Y-%m-%d"),
    date_to = format(date_to, "%Y-%m-%d"),
    granularity = granularity,
    revenue_unit = revenue_unit,
    shape = shape
  )

  openssl::sha1(jsonlite::toJSON(key_payload, auto_unbox = TRUE, null = "null"))
}

.st_metrics_dispatch_single <- function(app_id,
                                        os,
                                        countries,
                                        date_from,
                                        date_to,
                                        granularity,
                                        auth_token) {
  resolved_id <- .st_metrics_resolve_single_id(
    app_id = app_id,
    os = os,
    auth_token = auth_token
  )

  if (os == "unified") {
    return(
      st_unified_sales_report_impl(
        unified_app_id = resolved_id,
        countries = countries,
        start_date = date_from,
        end_date = date_to,
        date_granularity = granularity,
        auth_token = auth_token,
        verbose = FALSE
      )
    )
  }

  args <- list(
    os = os,
    countries = countries,
    start_date = date_from,
    end_date = date_to,
    date_granularity = granularity,
    auth_token = auth_token,
    auto_segment = TRUE,
    verbose = FALSE
  )

  if (os == "ios") {
    args$ios_app_id <- resolved_id
  } else {
    args$android_app_id <- resolved_id
  }

  do.call(st_sales_report_impl, args)
}

.st_metrics_resolve_single_id <- function(app_id, os, auth_token) {
  app_id <- as.character(app_id[1])

  if (os == "ios" && grepl("^\\d+$", app_id)) {
    return(app_id)
  }
  if (os == "android" && grepl("^(com|net|org|io|app|game)\\.", app_id)) {
    return(app_id)
  }
  if (os == "unified" && grepl("^[a-f0-9]{24}$", app_id)) {
    return(app_id)
  }

  resolved <- tryCatch(
    resolve_app_id(app_id, auth_token = auth_token, use_cache = TRUE, verbose = FALSE),
    error = function(e) NULL
  )

  if (is.null(resolved)) {
    rlang::abort(
      sprintf(
        "Failed to resolve `app_id = '%s'` for `os = '%s'`.",
        app_id,
        os
      )
    )
  }

  resolved_id <- switch(
    os,
    ios = resolved$ios_app_id,
    android = resolved$android_app_id,
    unified = resolved$unified_app_id
  )

  if (is.null(resolved_id) || is.na(resolved_id) || !nzchar(as.character(resolved_id))) {
    rlang::abort(
      sprintf(
        "Could not resolve a %s app ID for input '%s'.",
        os,
        app_id
      )
    )
  }

  as.character(resolved_id)
}

.st_metrics_normalize_output <- function(data,
                                         requested_ids,
                                         requested_countries,
                                         os,
                                         metrics,
                                         revenue_unit,
                                         shape) {
  if (is.null(data) || !nrow(data)) {
    return(
      if (shape == "long") {
        tibble::tibble(
          app_id = character(),
          os = character(),
          country = character(),
          date = as.Date(character()),
          metric = character(),
          value = numeric()
        )
      } else {
        wide <- tibble::tibble(
          app_id = character(),
          os = character(),
          country = character(),
          date = as.Date(character())
        )
        for (metric_name in metrics) {
          wide[[metric_name]] <- numeric()
        }
        wide
      }
    )
  }

  if (all(c("metric", "value") %in% names(data))) {
    data_tbl <- tibble::as_tibble(data)
    if (!"original_id" %in% names(data_tbl)) {
      data_tbl$original_id <- NA_character_
    }
    if (!"app_id" %in% names(data_tbl)) {
      data_tbl$app_id <- NA_character_
    }
    if (!"app_id_type" %in% names(data_tbl)) {
      data_tbl$app_id_type <- NA_character_
    }
    has_canonical_unified <- !is.na(data_tbl$app_id_type) &
      data_tbl$app_id_type == "unified" &
      !is.na(data_tbl$app_id) &
      nzchar(as.character(data_tbl$app_id))
    data_tbl$.st_output_app_id <- ifelse(
      has_canonical_unified,
      as.character(data_tbl$app_id),
      dplyr::coalesce(as.character(data_tbl$original_id), as.character(data_tbl$app_id))
    )

    normalized_long <- data_tbl %>%
      dplyr::transmute(
        app_id = as.character(.data$.st_output_app_id),
        os = os,
        country = as.character(.data$country),
        date = as.Date(.data$date),
        metric = tolower(as.character(.data$metric)),
        value = as.numeric(.data$value)
      ) %>%
      dplyr::filter(.data$metric %in% metrics)
  } else {
    normalized_wide <- .st_metrics_coerce_wide(
      data = tibble::as_tibble(data),
      requested_id = requested_ids[1],
      os = os
    )

    normalized_long <- normalized_wide %>%
      tidyr::pivot_longer(
        cols = dplyr::any_of(metrics),
        names_to = "metric",
        values_to = "value"
      )
  }

  if (!"WW" %in% requested_countries) {
    normalized_long <- normalized_long %>%
      dplyr::filter(.data$country %in% requested_countries)
  }

  conflicts <- normalized_long %>%
    dplyr::distinct(.data$app_id, .data$os, .data$country, .data$date, .data$metric, .data$value) %>%
    dplyr::count(.data$app_id, .data$os, .data$country, .data$date, .data$metric, name = "n_values") %>%
    dplyr::filter(.data$n_values > 1)
  if (nrow(conflicts) > 0) {
    rlang::abort(sprintf(
      "st_metrics() returned conflicting duplicate metric values for app_id '%s', country '%s', date '%s', metric '%s'.",
      conflicts$app_id[[1]],
      conflicts$country[[1]],
      format(conflicts$date[[1]], "%Y-%m-%d"),
      conflicts$metric[[1]]
    ))
  }

  normalized_long <- normalized_long %>%
    dplyr::distinct(.data$app_id, .data$os, .data$country, .data$date, .data$metric, .data$value)

  if (revenue_unit == "cents") {
    revenue_rows <- normalized_long$metric == "revenue"
    normalized_long$value[revenue_rows] <- normalized_long$value[revenue_rows] * 100
  }

  normalized_long <- normalized_long %>%
    dplyr::arrange(.data$app_id, .data$date, .data$country, .data$metric)

  if (shape == "long") {
    return(normalized_long)
  }

  wide <- normalized_long %>%
    tidyr::pivot_wider(names_from = "metric", values_from = "value") %>%
    dplyr::select("app_id", "os", "country", "date", dplyr::any_of(metrics))
  for (metric_name in setdiff(metrics, names(wide))) {
    wide[[metric_name]] <- numeric(nrow(wide))
  }
  wide <- wide %>%
    dplyr::select("app_id", "os", "country", "date", dplyr::all_of(metrics)) %>%
    dplyr::arrange(.data$app_id, .data$date, .data$country)

  wide
}

.st_metrics_coerce_wide <- function(data, requested_id, os) {
  revenue <- if ("revenue" %in% names(data)) {
    data$revenue
  } else if ("total_revenue" %in% names(data)) {
    data$total_revenue
  } else {
    rep(NA_real_, nrow(data))
  }

  downloads <- if ("downloads" %in% names(data)) {
    data$downloads
  } else if ("total_downloads" %in% names(data)) {
    data$total_downloads
  } else {
    rep(NA_real_, nrow(data))
  }

  tibble::tibble(
    app_id = rep(as.character(requested_id), nrow(data)),
    os = rep(os, nrow(data)),
    country = as.character(data$country),
    date = as.Date(data$date),
    revenue = as.numeric(revenue),
    downloads = as.numeric(downloads)
  )
}

# Basic arg-validation coverage for st_metrics().
# Live-API behavior is covered in test-live-smoke.R (gated on SENSORTOWER_RUN_LIVE).

test_that("st_metrics rejects missing app_id", {
  expect_error(
    st_metrics(auth_token = "dummy"),
    "app_id"
  )
})

test_that("st_metrics rejects invalid os", {
  expect_error(
    st_metrics(
      app_id     = "553834731",
      os         = "bogus",
      countries  = "US",
      auth_token = "dummy"
    ),
    "must be one of"
  )
})

test_that("st_metrics rejects invalid granularity", {
  expect_error(
    st_metrics(
      app_id      = "553834731",
      os          = "ios",
      countries   = "US",
      granularity = "hourly",
      auth_token  = "dummy"
    ),
    "must be one of"
  )
})

test_that("st_metrics rejects inverted date range", {
  expect_error(
    st_metrics(
      app_id     = "553834731",
      os         = "ios",
      countries  = "US",
      date_from  = "2024-06-01",
      date_to    = "2024-01-01",
      auth_token = "dummy"
    ),
    "must be on or before"
  )
})

test_that("st_metrics rejects malformed country codes", {
  expect_error(
    st_metrics(
      app_id     = "553834731",
      os         = "ios",
      countries  = c("United States"),
      auth_token = "dummy"
    ),
    "2-letter"
  )
})

test_that("st_metrics rejects unknown metrics", {
  expect_error(
    st_metrics(
      app_id     = "553834731",
      os         = "ios",
      countries  = "US",
      metrics    = c("revenue", "bogus_metric"),
      auth_token = "dummy"
    ),
    "Unknown metric"
  )
})

test_that("st_metrics normalization removes identical duplicate metric rows", {
  raw <- tibble::tibble(
    original_id = c("com.example.game", "com.example.game", "com.example.game"),
    app_id = c("com.example.game", "com.example.game", "com.example.game"),
    country = c("US", "US", "US"),
    date = as.Date(c("2026-04-27", "2026-04-27", "2026-04-27")),
    metric = c("revenue", "revenue", "downloads"),
    value = c(100, 100, 10)
  )

  normalized <- sensortowerR:::.st_metrics_normalize_output(
    data = raw,
    requested_ids = "com.example.game",
    requested_countries = "US",
    os = "android",
    metrics = c("revenue", "downloads"),
    revenue_unit = "dollars",
    shape = "long"
  )

  expect_equal(nrow(normalized), 2)
  expect_equal(
    normalized$value[normalized$metric == "revenue"],
    100
  )
})

test_that("st_metrics normalization rejects conflicting duplicate metric rows", {
  raw <- tibble::tibble(
    original_id = c("com.example.game", "com.example.game"),
    app_id = c("com.example.game", "com.example.game"),
    country = c("US", "US"),
    date = as.Date(c("2026-04-27", "2026-04-27")),
    metric = c("revenue", "revenue"),
    value = c(100, 101)
  )

  expect_error(
    sensortowerR:::.st_metrics_normalize_output(
      data = raw,
      requested_ids = "com.example.game",
      requested_countries = "US",
      os = "android",
      metrics = c("revenue", "downloads"),
      revenue_unit = "dollars",
      shape = "long"
    ),
    "conflicting duplicate metric values"
  )
})

test_that("st_metrics normalization filters rows outside requested countries", {
  raw <- tibble::tibble(
    original_id = c("123", "123", "123", "123"),
    app_id = c("123", "123", "123", "123"),
    country = c("CN", "CN", "US", "US"),
    date = as.Date(rep("2026-04-27", 4)),
    metric = c("revenue", "downloads", "revenue", "downloads"),
    value = c(100, 10, 999, 99)
  )

  normalized <- sensortowerR:::.st_metrics_normalize_output(
    data = raw,
    requested_ids = "123",
    requested_countries = "CN",
    os = "ios",
    metrics = c("revenue", "downloads"),
    revenue_unit = "dollars",
    shape = "wide"
  )

  expect_equal(unique(normalized$country), "CN")
  expect_equal(normalized$revenue, 100)
  expect_equal(normalized$downloads, 10)
})

test_that("st_metrics wide normalization preserves metric columns after country filter removes all rows", {
  raw <- tibble::tibble(
    original_id = "123",
    app_id = "123",
    country = "US",
    date = as.Date("2026-04-27"),
    metric = "revenue",
    value = 999
  )

  normalized <- sensortowerR:::.st_metrics_normalize_output(
    data = raw,
    requested_ids = "123",
    requested_countries = "CN",
    os = "ios",
    metrics = c("revenue", "downloads"),
    revenue_unit = "dollars",
    shape = "wide"
  )

  expect_equal(nrow(normalized), 0)
  expect_true(all(c("revenue", "downloads") %in% names(normalized)))
})

test_that("st_metrics normalization preserves canonical unified app IDs from batch output", {
  raw <- tibble::tibble(
    original_id = c("requested_game", "requested_game"),
    app_id = c("abcdefabcdefabcdefabcdef", "abcdefabcdefabcdefabcdef"),
    app_id_type = c("unified", "unified"),
    country = c("US", "US"),
    date = as.Date(c("2026-04-27", "2026-04-27")),
    metric = c("revenue", "downloads"),
    value = c(100, 10)
  )

  normalized <- sensortowerR:::.st_metrics_normalize_output(
    data = raw,
    requested_ids = "requested_game",
    requested_countries = "US",
    os = "unified",
    metrics = c("revenue", "downloads"),
    revenue_unit = "dollars",
    shape = "long"
  )

  expect_equal(unique(normalized$app_id), "abcdefabcdefabcdefabcdef")
})

.st_filter_new <- function(criteria, combine, filter_id, custom_fields = NULL) {
  structure(
    list(
      criteria = criteria,
      combine = combine,
      filter_id = filter_id,
      custom_fields = custom_fields
    ),
    class = "st_filter"
  )
}

#' Create a Sensor Tower Filter Object
#'
#' Thin v1.0.0 facade for building or wrapping Sensor Tower custom-field filter
#' IDs. When criteria are supplied, the function translates them into the legacy
#' custom-fields request structure and creates a server-side filter ID using the
#' existing implementation.
#'
#' @param date_from,date_to Optional date bounds for a release-date criterion.
#' @param genre Optional character vector of game genres.
#' @param monetization Optional monetization criteria. Accept either a character
#'   vector such as `c("free", "iap")` or a named list like
#'   `list(free = TRUE, iap = TRUE)`.
#' @param publisher Optional character vector of publisher names.
#' @param sdk Optional character vector of SDK names.
#' @param custom_fields Optional raw custom-fields specification. Can be a
#'   list-of-lists already matching the legacy request format or a named list of
#'   field names and values.
#' @param filter_id Optional existing 24-character Sensor Tower filter ID to
#'   wrap.
#' @param combine Logical operator metadata for multi-criterion filters: `"and"`
#'   or `"or"`.
#' @param auth_token Optional Sensor Tower API token.
#' @param x An `st_filter` object (S3 method argument).
#' @param ... Additional arguments. For `c.st_filter()`, further `st_filter`
#'   objects to combine; ignored by `print.st_filter()`, `format.st_filter()`,
#'   and `as.character.st_filter()`.
#' @param recursive Unused; present for S3 method-signature compatibility with
#'   [base::c()].
#'
#' @return An object of class `st_filter`, implemented as a list with elements:
#'   - `criteria`: normalized criteria supplied to `st_filter()`
#'   - `combine`: `"and"` or `"or"`
#'   - `filter_id`: the wrapped or created server-side filter ID
#'
#' @examples
#' \dontrun{
#' st_filter(genre = c("Puzzle", "Strategy"))
#'
#' st_filter(
#'   date_from = "2024-01-01",
#'   date_to = "2024-12-31",
#'   monetization = c("free", "iap")
#' )
#'
#' existing <- st_filter(filter_id = "687df26ac5a19ebcfe817d7f")
#' as.character(existing)
#' }
#'
#' @export
st_filter <- function(date_from = NULL,
                      date_to = NULL,
                      genre = NULL,
                      monetization = NULL,
                      publisher = NULL,
                      sdk = NULL,
                      custom_fields = NULL,
                      filter_id = NULL,
                      combine = c("and", "or"),
                      auth_token = NULL) {
  combine <- match.arg(combine)
  criteria_present <- !all(vapply(
    list(date_from, date_to, genre, monetization, publisher, sdk, custom_fields),
    is.null,
    logical(1)
  ))

  if (!criteria_present) {
    if (is.null(filter_id)) {
      rlang::abort("Supply at least one filter criterion or an existing `filter_id`.")
    }
    if (!st_is_valid_filter_id(filter_id)) {
      rlang::abort("`filter_id` must be a 24-character hexadecimal Sensor Tower filter ID.")
    }

    return(
      .st_filter_new(
        criteria = list(),
        combine = combine,
        filter_id = as.character(filter_id[1]),
        custom_fields = NULL
      )
    )
  }

  if (!is.null(filter_id)) {
    rlang::abort(
      "When criteria are supplied, omit `filter_id`; `st_filter()` will create a new server-side filter."
    )
  }

  auth_token <- get_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required to create a filter.",
      "Set SENSORTOWER_AUTH_TOKEN or pass `auth_token`."
    )
  )

  bounds <- .st_filter_normalize_bounds(date_from, date_to)
  built_custom_fields <- .st_filter_build_custom_fields(
    date_from = bounds$date_from,
    date_to = bounds$date_to,
    genre = genre,
    monetization = monetization,
    publisher = publisher,
    sdk = sdk,
    custom_fields = custom_fields
  )

  created_filter_id <- st_custom_fields_filter_impl(
    custom_fields = built_custom_fields,
    auth_token = auth_token
  )

  criteria <- list(
    date_from = bounds$date_from,
    date_to = bounds$date_to,
    genre = if (!is.null(genre)) unique(as.character(genre)) else NULL,
    monetization = monetization,
    publisher = if (!is.null(publisher)) unique(as.character(publisher)) else NULL,
    sdk = if (!is.null(sdk)) unique(as.character(sdk)) else NULL,
    custom_fields = custom_fields
  )
  criteria <- criteria[!vapply(criteria, is.null, logical(1))]

  .st_filter_new(
    criteria = criteria,
    combine = combine,
    filter_id = created_filter_id,
    custom_fields = built_custom_fields
  )
}

#' @rdname st_filter
#' @export
print.st_filter <- function(x, ...) {
  cat(format(x, ...), "\n", sep = "")
  invisible(x)
}

#' @rdname st_filter
#' @export
format.st_filter <- function(x, ...) {
  criteria_names <- names(x$criteria)
  criteria_text <- if (length(criteria_names)) {
    paste(criteria_names, collapse = ", ")
  } else {
    "none"
  }

  sprintf(
    "<st_filter> id=%s combine=%s criteria=%s",
    x$filter_id %||% "NA",
    x$combine %||% "and",
    criteria_text
  )
}

#' @rdname st_filter
#' @export
as.character.st_filter <- function(x, ...) {
  x$filter_id %||% NA_character_
}

#' @rdname st_filter
#' @export
c.st_filter <- function(..., recursive = FALSE, combine = c("and", "or")) {
  filters <- list(...)
  combine <- match.arg(combine)

  if (!length(filters)) {
    return(.st_filter_new(criteria = list(), combine = combine, filter_id = NULL))
  }
  if (!all(vapply(filters, inherits, logical(1), what = "st_filter"))) {
    rlang::abort("`c.st_filter()` only accepts `st_filter` objects.")
  }

  auth_token <- get_auth_token(
    NULL,
    error_message = paste(
      "Authentication token is required to combine filters.",
      "Set SENSORTOWER_AUTH_TOKEN or pass `auth_token` when creating the filters."
    )
  )

  custom_fields <- unlist(
    lapply(filters, .st_filter_extract_custom_fields, auth_token = auth_token),
    recursive = FALSE
  )

  if (!length(custom_fields)) {
    rlang::abort("Could not extract any criteria to combine from the supplied filters.")
  }

  created_filter_id <- st_custom_fields_filter_impl(
    custom_fields = custom_fields,
    auth_token = auth_token
  )

  .st_filter_new(
    criteria = lapply(filters, `[[`, "criteria"),
    combine = combine,
    filter_id = created_filter_id,
    custom_fields = custom_fields
  )
}

.st_filter_resolve_id <- function(filter) {
  if (is.null(filter)) {
    return(NULL)
  }

  if (inherits(filter, "st_filter")) {
    filter_id <- filter$filter_id %||% NA_character_
    if (is.na(filter_id) || !nzchar(filter_id)) {
      rlang::abort("`st_filter` object does not contain a server-side `filter_id`.")
    }
    return(filter_id)
  }

  if (is.character(filter) && length(filter) == 1L && st_is_valid_filter_id(filter)) {
    return(as.character(filter[1]))
  }

  rlang::abort("`filter` must be `NULL`, an `st_filter` object, or a 24-character filter ID.")
}

.st_filter_normalize_bounds <- function(date_from, date_to) {
  if (is.null(date_from) && is.null(date_to)) {
    return(list(date_from = NULL, date_to = NULL))
  }

  if (is.null(date_from)) {
    date_from <- date_to
  }
  if (is.null(date_to)) {
    date_to <- date_from
  }

  normalize_dates(date_from, date_to)
}

.st_filter_build_custom_fields <- function(date_from,
                                           date_to,
                                           genre,
                                           monetization,
                                           publisher,
                                           sdk,
                                           custom_fields) {
  fields <- list()

  if (!is.null(date_from) || !is.null(date_to)) {
    values <- if (!is.null(date_from) && !is.null(date_to)) {
      list(paste(format(date_from, "%Y-%m-%d"), format(date_to, "%Y-%m-%d"), sep = " to "))
    } else if (!is.null(date_from)) {
      list(paste("after", format(date_from, "%Y-%m-%d")))
    } else {
      list(paste("before", format(date_to, "%Y-%m-%d")))
    }

    fields <- append(fields, list(list(
      name = "Release Date (US)",
      values = values,
      global = TRUE,
      exclude = FALSE
    )))
  }

  if (!is.null(genre)) {
    fields <- append(fields, list(list(
      name = "Game Genre",
      values = as.list(unique(as.character(genre))),
      global = TRUE,
      exclude = FALSE
    )))
  }

  fields <- append(fields, .st_filter_build_monetization_fields(monetization))

  if (!is.null(publisher)) {
    fields <- append(fields, list(list(
      name = "Publisher",
      values = as.list(unique(as.character(publisher))),
      global = TRUE,
      exclude = FALSE
    )))
  }

  if (!is.null(sdk)) {
    sdk <- unique(as.character(sdk))
    fields <- append(fields, lapply(sdk, function(one_sdk) {
      list(
        name = paste0("SDK: ", one_sdk),
        values = list(),
        global = TRUE,
        exclude = FALSE,
        true = TRUE
      )
    }))
  }

  fields <- append(fields, .st_filter_normalize_custom_fields(custom_fields))

  if (!length(fields)) {
    rlang::abort("No valid filter criteria were supplied.")
  }

  fields
}

.st_filter_build_monetization_fields <- function(monetization) {
  if (is.null(monetization)) {
    return(list())
  }

  field_map <- list(
    free = "Free",
    iap = "In-App Purchases",
    ads = "Contains Ads",
    subscription = "In-App Subscription"
  )

  if (is.character(monetization)) {
    keys <- tolower(unique(as.character(monetization)))
    bad <- setdiff(keys, names(field_map))
    if (length(bad)) {
      rlang::abort(
        sprintf(
          "Unknown `monetization` value(s): %s.",
          paste(bad, collapse = ", ")
        )
      )
    }
    values <- stats::setNames(rep(list(TRUE), length(keys)), keys)
  } else if (is.list(monetization)) {
    values <- monetization
  } else {
    rlang::abort("`monetization` must be a character vector or named list.")
  }

  keys <- intersect(names(values), names(field_map))
  if (!length(keys)) {
    rlang::abort(
      "`monetization` must include one or more of: free, iap, ads, subscription."
    )
  }

  lapply(keys, function(key) {
    list(
      name = field_map[[key]],
      global = TRUE,
      true = isTRUE(values[[key]])
    )
  })
}

.st_filter_normalize_custom_fields <- function(custom_fields) {
  if (is.null(custom_fields)) {
    return(list())
  }

  if (!is.list(custom_fields)) {
    rlang::abort("`custom_fields` must be a list.")
  }

  if (length(custom_fields) == 0) {
    return(list())
  }

  if (all(vapply(custom_fields, function(x) is.list(x) && !is.null(x$name), logical(1)))) {
    return(custom_fields)
  }

  if (is.null(names(custom_fields)) || any(!nzchar(names(custom_fields)))) {
    rlang::abort(
      "`custom_fields` must be either a legacy list-of-lists or a named list of field values."
    )
  }

  lapply(names(custom_fields), function(field_name) {
    value <- custom_fields[[field_name]]
    field <- list(
      name = field_name,
      global = TRUE,
      exclude = FALSE
    )

    if (is.logical(value) && length(value) == 1L) {
      field$true <- isTRUE(value)
    } else {
      field$values <- as.list(value)
    }

    field
  })
}

.st_filter_extract_custom_fields <- function(filter, auth_token) {
  if (!is.null(filter$custom_fields)) {
    return(filter$custom_fields)
  }

  if (is.null(filter$filter_id)) {
    return(list())
  }

  details <- st_custom_fields_filter_by_id_impl(
    id = filter$filter_id,
    auth_token = auth_token
  )

  .st_filter_details_to_custom_fields(details)
}

.st_filter_details_to_custom_fields <- function(details) {
  custom_fields <- details$custom_fields
  if (is.null(custom_fields) || !nrow(custom_fields)) {
    return(list())
  }

  lapply(seq_len(nrow(custom_fields)), function(i) {
    field <- list(
      name = custom_fields$name[i],
      global = if ("global" %in% names(custom_fields)) isTRUE(custom_fields$global[i]) else TRUE,
      exclude = if ("exclude" %in% names(custom_fields)) isTRUE(custom_fields$exclude[i]) else FALSE
    )

    if ("values" %in% names(custom_fields)) {
      field$values <- custom_fields$values[[i]]
    }
    if ("true" %in% names(custom_fields) && !is.na(custom_fields$true[i])) {
      field$true <- custom_fields$true[i]
    }

    field
  })
}

test_that("st_filter() constructs an st_filter object from criteria", {
  skip_on_cran()
  skip_if_no_auth()

  f <- st_filter(
    date_from = "2025-01-01",
    date_to   = "2025-06-30"
  )
  expect_s3_class(f, "st_filter")
  expect_true(is.character(as.character(f)))
})

test_that("st_filter() wraps a pre-existing filter_id without an API call", {
  f <- st_filter(filter_id = "5ba4585f539ce75b97db6bcb")
  expect_s3_class(f, "st_filter")
  expect_equal(as.character(f), "5ba4585f539ce75b97db6bcb")
})

test_that("print.st_filter produces output", {
  f <- st_filter(filter_id = "5ba4585f539ce75b97db6bcb")
  expect_output(print(f))
})

test_that("format.st_filter returns a character string", {
  f <- st_filter(filter_id = "5ba4585f539ce75b97db6bcb")
  expect_type(format(f), "character")
})

#' Defunct Functions In sensortowerR
#'
#' Legacy entrypoints retained only as defunct stubs for the v1.0.0 transition.
#'
#' @name sensortowerR-defunct
#' @keywords internal
NULL

#' @rdname sensortowerR-defunct
#' @export
st_sales_report <- function(...) {
  .Defunct(
    "st_metrics",
    package = "sensortowerR",
    msg = paste(
      "`st_sales_report()` is defunct.",
      "Use `st_metrics()` (note: revenue is now in dollars by default;",
      "pass `revenue_unit = \"cents\"` for the old behavior).",
      "See `?st_metrics`."
    )
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_unified_sales_report <- function(...) {
  .Defunct(
    "st_metrics",
    package = "sensortowerR",
    msg = "`st_unified_sales_report()` is defunct. Use `st_metrics()` with the default `os = \"unified\"`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_batch_metrics <- function(...) {
  .Defunct(
    "st_metrics",
    package = "sensortowerR",
    msg = "`st_batch_metrics()` is defunct. Use `st_metrics()` with a vector `app_id`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_smart_metrics <- function(...) {
  .Defunct(
    "st_metrics",
    package = "sensortowerR",
    msg = "`st_smart_metrics()` is defunct. Use `st_metrics()`; caching is enabled by default."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_top_charts <- function(...) {
  .Defunct(
    "st_rankings",
    package = "sensortowerR",
    msg = "`st_top_charts()` is defunct. Use `st_rankings(entity = \"app\")`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_top_publishers <- function(...) {
  .Defunct(
    "st_rankings",
    package = "sensortowerR",
    msg = "`st_top_publishers()` is defunct. Use `st_rankings(entity = \"publisher\")`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_category_rankings <- function(...) {
  .Defunct(
    "st_rankings",
    package = "sensortowerR",
    msg = "`st_category_rankings()` is defunct. Use `st_rankings(entity = \"category\")`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_app_info <- function(...) {
  .Defunct(
    "st_apps",
    package = "sensortowerR",
    msg = "`st_app_info()` is defunct. Use `st_apps(query = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_app_lookup <- function(...) {
  .Defunct(
    "st_app",
    package = "sensortowerR",
    msg = "`st_app_lookup()` is defunct. Use `st_app(app_id = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_app_details <- function(...) {
  .Defunct(
    "st_app",
    package = "sensortowerR",
    msg = "`st_app_details()` is defunct. Use `st_app(app_id = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_filter_by_date <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_filter_by_date()` is defunct. Use `st_filter(date_from = ..., date_to = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_filter_by_genre <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_filter_by_genre()` is defunct. Use `st_filter(genre = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_filter_by_monetization <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_filter_by_monetization()` is defunct. Use `st_filter(monetization = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_filter_by_publisher <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_filter_by_publisher()` is defunct. Use `st_filter(publisher = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_filter_by_sdk <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_filter_by_sdk()` is defunct. Use `st_filter(sdk = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_custom_fields_filter <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_custom_fields_filter()` is defunct. Use `st_filter(custom_fields = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_custom_fields_filter_by_id <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_custom_fields_filter_by_id()` is defunct. Use `st_filter(filter_id = ...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_combine_filters <- function(...) {
  .Defunct(
    package = "sensortowerR",
    msg = "`st_combine_filters()` is defunct. Use `c(st_filter(...), st_filter(...))`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_create_simple_filter <- function(...) {
  .Defunct(
    "st_filter",
    package = "sensortowerR",
    msg = "`st_create_simple_filter()` is defunct. Use `st_filter(...)`."
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_compare_filter_results <- function(...) {
  .Defunct(
    package = "sensortowerR",
    msg = paste(
      "`st_compare_filter_results()` is defunct.",
      "There is no direct replacement; use `st_apps(filter = ...)` and compare results manually."
    )
  )
}

#' @rdname sensortowerR-defunct
#' @export
st_generate_example_filter_ids <- function(...) {
  .Defunct(
    package = "sensortowerR",
    msg = paste(
      "`st_generate_example_filter_ids()` is defunct.",
      "There is no direct replacement; it was a diagnostic helper."
    )
  )
}

#' Fetch Apps by Custom Fields and Tags
#'
#' Retrieves apps filtered by custom fields and tags from Sensor Tower.
#' This function uses the /v1/app_tag/apps endpoint.
#'
#' @param app_id_type Character string. Operating System. Must be one of
#'   "itunes" (iOS) or "unified". Required.
#' @param custom_fields_filter_id Character string. ID of a Sensor Tower
#'   custom field filter. Required. Use the filter ID from relevant endpoint.
#' @param name Optional. Character string. Name of Custom or Global Field.
#'   Defaults to "Stock Ticker".
#' @param value Optional. Character string. Tag value for custom or global 
#'   field provided. Leave blank to fetch all possible apps.
#' @param global Optional. Logical. Filter by global or organization custom 
#'   fields. Defaults to TRUE (false means organization custom fields).
#' @param last_known_id Optional. Character string. Supply last_known_id from 
#'   previous request to get next page. Leave blank to get first page.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param base_url Optional. Character string. The base URL for the API.
#'
#' @return A [tibble][tibble::tibble] with app data including IDs and metadata.
#'
#' @export
#' @importFrom httr2 request req_url_path_append req_url_query req_headers
#' @importFrom httr2 req_perform resp_body_string
#' @importFrom jsonlite fromJSON
#' @importFrom tibble as_tibble
#' @importFrom rlang abort %||%
st_app_tag <- function(
    app_id_type,
    custom_fields_filter_id,
    name = NULL,
    value = NULL,
    global = TRUE,
    last_known_id = NULL,
    auth_token = NULL,
    base_url = "https://api.sensortower.com"
) {
  
  # --- Input Validation ---
  app_id_type <- match.arg(app_id_type, c("itunes", "unified"))
  
  if (missing(custom_fields_filter_id) || is.null(custom_fields_filter_id)) {
    rlang::abort("'custom_fields_filter_id' parameter is required.")
  }
  
  # --- Authentication ---
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set it as an environment variable."
  )
  
  # --- Build Query Parameters ---
  query_params <- list(
    custom_fields_filter_id = custom_fields_filter_id
  )
  
  # Add optional parameters
  if (!is.null(name)) {
    query_params$name <- name
  }
  
  if (!is.null(value)) {
    query_params$value <- value
  }
  
  query_params$global <- tolower(as.character(global))
  
  if (!is.null(last_known_id)) {
    query_params$last_known_id <- last_known_id
  }
  
  # --- Build Request ---
  request_query <- c(list(app_id_type = app_id_type), query_params)
  req <- build_request(
    base_url = base_url,
    path_segments = st_endpoint_segments("app_tag_apps"),
    query_params = request_query
  ) %>%
    httr2::req_headers(
      "Authorization" = paste("Bearer", auth_token_val),
      "Accept" = "application/json"
    ) %>%
    httr2::req_timeout(60)
  
  # --- Perform Request ---
  tryCatch({
    resp <- httr2::req_perform(req)
    
    # Parse response
    content <- httr2::resp_body_string(resp)
    data <- jsonlite::fromJSON(content, flatten = TRUE)
    
    # Convert to tibble
    if (!is.null(data) && length(data) > 0) {
      result <- tibble::as_tibble(data)
    } else {
      result <- tibble::tibble()
    }
    
    return(result)
    
  }, error = function(e) {
    resp_obj <- NULL
    if (!is.null(e$response)) resp_obj <- e$response
    if (is.null(resp_obj) && !is.null(e$resp)) resp_obj <- e$resp
    if (inherits(e, "httr2_http") && !is.null(resp_obj)) {
      status <- httr2::resp_status(resp_obj)
      body <- httr2::resp_body_string(resp_obj)
      
      if (status == 401) {
        rlang::abort("Invalid authentication token.")
      } else if (status == 403) {
        rlang::abort("Your API token is not authorized.")
      } else if (status == 422) {
        rlang::abort(paste("Invalid Query Parameter:", body))
      } else {
        rlang::abort(paste("API request failed with status", status, ":", body))
      }
    } else {
      rlang::abort(paste("Unexpected error:", e$message))
    }
  })
}

#' Fetch Sales Report Estimates
#'
#' Retrieves download and revenue estimates of apps by country and date.
#' Note: All revenues are returned in cents and need to be divided by 100 for dollar amounts.
#'
#' @param ios_app_id Character string. iOS app ID (numeric, e.g., "1234567890").
#' @param android_app_id Character string. Android package name (e.g., "com.example.app").
#' @param unified_app_id Character string. Sensor Tower unified app ID (24-character hex).
#' @param publisher_ids Character vector. Publisher IDs to query. Some Android publisher IDs contain commas.
#' @param custom_fields_filter_id Optional. Character string. ID of a Sensor
#'   Tower custom field filter to apply. Use filter IDs from the web interface
#'   at app.sensortower.com. When provided, this filter will be used instead of
#'   app_ids or publisher_ids.
#' @param custom_tags_mode Optional. Character string. Required if `os` is
#'   'unified' and `custom_fields_filter_id` is provided. Specifies how the
#'   custom filter applies to unified apps. Options: "include", "exclude",
#'   "include_unified_apps". The "include_unified_apps" option includes all
#'   platform versions when any version matches the filter.
#' @param os Character string. Required. Operating system: "ios", "android", or "unified".
#' @param countries Character vector. Country codes (e.g., c("US", "GB", "JP"), or "WW" for worldwide). Required.
#' @param start_date Date or character string. Start date in "YYYY-MM-DD" format. Required.
#' @param end_date Date or character string. End date in "YYYY-MM-DD" format. Required.
#' @param date_granularity Character string. One of "daily", "weekly", "monthly", "quarterly". Required.
#' @param limit Numeric. Number of results to return when using custom_fields_filter_id.
#'   Ignored when using specific app ID parameters or publisher_ids. Defaults to 100.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param auto_segment Logical. If TRUE, automatically segments date ranges to avoid timeouts.
#' @param verbose Logical. If TRUE, prints progress messages.
#'
#' @return A tibble with download and revenue estimates.
#'
#' @details
#' **App ID Parameters**: Provide one of the following:
#' - `ios_app_id`: Specifically for iOS app IDs (numeric)
#' - `android_app_id`: Specifically for Android package names
#' - `unified_app_id`: Specifically for Sensor Tower unified IDs
#'
#' The function will automatically resolve IDs if needed. For example, if you provide
#' a `unified_app_id` but set `os="ios"`, it will look up the iOS app ID.
#'
#' The API has timeout limitations based on date granularity:
#' - daily: limit to 1 week segments
#' - weekly: limit to 3 month segments  
#' - monthly: limit to 1 year segments
#' - quarterly: limit to 2 year segments
#' 
#' When auto_segment = TRUE, the function automatically breaks up the date range
#' into appropriate segments and combines the results.
#' 
#'
#' @examples
#' \dontrun{
#' # Get daily sales for a single app using specific parameter
#' sales <- st_sales_report(
#'   os = "ios",
#'   ios_app_id = "553834731",  # Candy Crush iOS
#'   countries = c("US", "GB"),
#'   start_date = "2024-01-01",
#'   end_date = "2024-01-07",
#'   date_granularity = "daily"
#' )
#' 
#' # Get Android data using specific parameter
#' android_sales <- st_sales_report(
#'   os = "android",
#'   android_app_id = "com.king.candycrushsaga",
#'   countries = "US",
#'   start_date = "2024-01-01",
#'   end_date = "2024-01-07",
#'   date_granularity = "daily"
#' )
#' 
#' # Get iOS data from unified ID (automatic lookup)
#' unified_sales <- st_sales_report(
#'   os = "ios",
#'   unified_app_id = "5ba4585f539ce75b97db6bcb",
#'   countries = "US",
#'   start_date = "2024-01-01",
#'   end_date = "2024-01-07",
#'   date_granularity = "daily"
#' )
#' 
#' }
#'
#' @keywords internal
#' @noRd
st_sales_report_impl <- function(os,
                           countries,
                           start_date,
                           end_date,
                           date_granularity,
                           ios_app_id = NULL,
                           android_app_id = NULL,
                           unified_app_id = NULL,
                            publisher_ids = NULL,
                           custom_fields_filter_id = NULL,
                           custom_tags_mode = NULL,
                           limit = 100,
                            auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                           auto_segment = TRUE,
                           verbose = TRUE) {
  
  # Validate OS parameter
  if (missing(os) || is.null(os) || !os %in% c("ios", "android", "unified")) {
    rlang::abort("'os' parameter is required and must be one of: 'ios', 'android', or 'unified'")
  }
  
  
  # Validate required parameters
  if (missing(countries) || is.null(countries) || length(countries) == 0) {
    rlang::abort("'countries' parameter is required. Specify country codes (e.g., 'US', 'GB', 'JP', or 'WW' for worldwide).")
  }
  
  if (missing(start_date) || is.null(start_date)) {
    rlang::abort("'start_date' parameter is required. Specify in YYYY-MM-DD format.")
  }
  
  if (missing(end_date) || is.null(end_date)) {
    rlang::abort("'end_date' parameter is required. Specify in YYYY-MM-DD format.")
  }
  
  if (missing(date_granularity) || is.null(date_granularity)) {
    rlang::abort("'date_granularity' parameter is required. Specify one of: 'daily', 'weekly', 'monthly', 'quarterly'.")
  }
  
  # Load custom filter utilities if available
  if (exists("validate_custom_filter_params", mode = "function")) {
    validate_custom_filter_params(
      custom_fields_filter_id = custom_fields_filter_id,
      custom_tags_mode = custom_tags_mode,
      os = os,
      require_category = FALSE
    )
  }
  
  # Input validation
  if (is.null(ios_app_id) && is.null(android_app_id) && is.null(unified_app_id) && is.null(publisher_ids) && is.null(custom_fields_filter_id)) {
    rlang::abort("At least one of ios_app_id, android_app_id, unified_app_id, publisher_ids, or custom_fields_filter_id is required")
  }
  date_granularity <- match.arg(date_granularity, c("daily", "weekly", "monthly", "quarterly"))
  
  # Convert dates to Date objects
  start_date <- as.Date(start_date)
  end_date <- as.Date(end_date)
  
  # Authentication
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set SENSORTOWER_AUTH_TOKEN environment variable."
  )
  
  # For unified OS, we don't support st_sales_report
  if (os == "unified") {
    rlang::abort(paste0(
      "st_sales_report_impl does not support os='unified'.\n",
      "Please use platform-specific calls with os='ios' or os='android'.\n",
      "For unified data, consider using st_metrics() instead."
    ))
  }
  
  # Resolve app IDs only when using app-based requests (not publisher or custom filter)
  app_ids <- NULL
  if (is.null(publisher_ids) && is.null(custom_fields_filter_id)) {
    id_resolution <- resolve_ids_for_os(
      unified_app_id = unified_app_id,
      ios_app_id = ios_app_id,
      android_app_id = android_app_id,
      os = os,
      auth_token = auth_token_val,
      verbose = verbose
    )
    # Extract the resolved app ID for the API call
    resolved <- id_resolution$resolved_ids
    if (os == "ios" && !is.null(resolved$ios_app_id)) {
      app_ids <- resolved$ios_app_id
    } else if (os == "android" && !is.null(resolved$android_app_id)) {
      app_ids <- resolved$android_app_id
    }
  }
  
  # Determine date segments if auto_segment is TRUE
  if (auto_segment) {
    segments <- get_date_segments(start_date, end_date, date_granularity)
    if (verbose && nrow(segments) > 1) {
      message(sprintf("Breaking date range into %d segments to avoid timeouts", nrow(segments)))
    }
  } else {
    segments <- tibble(start = start_date, end = end_date)
  }
  
  # Collect results from all segments
  all_results <- list()
  
  for (i in seq_len(nrow(segments))) {
    segment_start <- segments$start[i]
    segment_end <- segments$end[i]
    
    if (verbose && nrow(segments) > 1) {
      message(sprintf("Fetching segment %d/%d: %s to %s", 
                     i, nrow(segments), segment_start, segment_end))
    }
    
    # Build query parameters
    query_params <- list(
      auth_token = auth_token_val,
      start_date = format(segment_start, "%Y-%m-%d"),
      end_date = format(segment_end, "%Y-%m-%d"),
      date_granularity = date_granularity,
      countries = paste(countries, collapse = ",")
    )
    
    # Add app IDs, publisher IDs, or custom filter
    if (!is.null(custom_fields_filter_id)) {
      query_params$limit <- limit
      # Add custom filter parameters if available
      if (exists("add_custom_filter_params", mode = "function")) {
        query_params <- add_custom_filter_params(
          query_params,
          custom_fields_filter_id = custom_fields_filter_id,
          custom_tags_mode = custom_tags_mode,
          os = os
        )
      } else {
        query_params$custom_fields_filter_id <- custom_fields_filter_id
        if (!is.null(custom_tags_mode)) {
          query_params$custom_tags_mode <- custom_tags_mode
        }
      }
    } else {
      if (!is.null(app_ids)) {
        query_params$app_ids <- paste(app_ids, collapse = ",")
      }
      
      if (!is.null(publisher_ids)) {
        # Handle publisher IDs with commas using array format
        if (any(grepl(",", publisher_ids))) {
          query_params$`publisher_ids[]` <- publisher_ids
        } else {
          query_params$publisher_ids <- paste(publisher_ids, collapse = ",")
        }
      }
    }
    
    # Build and perform request
    path <- st_endpoint_relative_path("sales_report_estimates", os = os)
    
    # Create a wrapper for process_sales_response that binds the 'os' argument
    sales_processor <- function(resp, ...) {
      process_sales_response(resp, os = os)
    }
    
    result <- fetch_data_core(
      endpoint = path,
      params = query_params,
      auth_token = auth_token_val,
      verbose = verbose,
      processor = sales_processor
    )
    
    if (!is.null(result) && nrow(result) > 0) {
      all_results[[i]] <- result
    }
  }
  
  # Combine all results
  if (length(all_results) > 0) {
    final_result <- bind_rows(all_results)
    
    if (verbose) {
      message(sprintf("Retrieved %d records", nrow(final_result)))
    }
    
    # Add platform information
    final_result$platform <- os
    
    return(final_result)
  } else {
    return(tibble())
  }
}

#' Get date segments based on granularity
#' @noRd
get_date_segments <- function(start_date, end_date, granularity) {
  # Define segment lengths based on API recommendations
  segment_days <- switch(granularity,
    daily = 7,      # 1 week
    weekly = 90,    # ~3 months
    monthly = 365,  # 1 year
    quarterly = 730 # 2 years
  )
  
  # Create segments
  segments <- tibble()
  current_start <- start_date
  
  while (current_start <= end_date) {
    current_end <- min(current_start + segment_days - 1, end_date)
    segments <- bind_rows(segments, 
                         tibble(start = current_start, end = current_end))
    current_start <- current_end + 1
  }
  
  return(segments)
}

#' Process sales report response
#' @noRd
process_sales_response <- function(resp, os) {
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(NULL)
  }
  
  body_text <- rawToChar(body_raw)
  result <- jsonlite::fromJSON(body_text, flatten = TRUE)
  
  if (length(result) == 0 || nrow(result) == 0) {
    return(NULL)
  }
  
  # Convert to tibble
  result_tbl <- as_tibble(result)
  
  # Process based on OS
  if (os == "ios") {
    # Rename columns for clarity
    result_tbl <- result_tbl %>%
      rename_with(~ case_when(
        . == "aid" ~ "app_id",
        . == "cc" ~ "country",
        . == "d" ~ "date",
        . == "iu" ~ "iphone_downloads",
        . == "ir" ~ "iphone_revenue_cents",
        . == "au" ~ "ipad_downloads",
        . == "ar" ~ "ipad_revenue_cents",
        TRUE ~ .
      ))
    
    # Add calculated fields
    result_tbl <- result_tbl %>%
      mutate(
        date = as.Date(date),
        iphone_revenue = iphone_revenue_cents / 100,
        ipad_revenue = ipad_revenue_cents / 100,
        total_downloads = iphone_downloads + ipad_downloads,
        total_revenue = iphone_revenue + ipad_revenue,
        .after = date
      )
  } else {
    # Android
    result_tbl <- result_tbl %>%
      rename_with(~ case_when(
        . == "aid" ~ "app_id",
        . == "c" ~ "country",    # Android uses 'c' for country
        . == "d" ~ "date",
        . == "u" ~ "downloads",
        . == "r" ~ "revenue_cents",
        TRUE ~ .
      ))
    
    # Add calculated fields
    result_tbl <- result_tbl %>%
      mutate(
        date = as.Date(date),
        revenue = revenue_cents / 100,
        .after = date
      )
  }
  
  return(result_tbl)
}

#' Helper function to look up category names
#'
#' @param category_ids Character vector of category IDs
#' @param platform Character string. "ios" or "android"
#' @return Character vector of category names
#' @importFrom purrr map_chr
#' @export
lookup_category_names <- function(category_ids, platform = "ios") {
  # Load internal data
  st_category_data <- NULL
  data("st_category_data", envir = environment())
  
  # Filter for platform and match IDs
  categories <- st_category_data %>%
    filter(platform == !!platform) %>%
    filter(category_id %in% category_ids)
  
  # Return matched names in same order as input
  category_ids %>%
    map_chr(~ {
      matched <- categories %>%
        filter(category_id == .x) %>%
        pull(category_name)
      
      if (length(matched) > 0) {
        matched[1]
      } else {
        as.character(.x)  # Return ID if no match found
      }
    })
}

#' Get Top Publishers by Revenue or Downloads
#'
#' Retrieves top app publishers ranked by revenue or downloads for a specified
#' category, time period, and country. This function uses the
#' `/v1/\{os\}/top_and_trending/publishers` endpoint.
#'
#' @param measure Character. Metric to rank by: "revenue" or "units" (downloads).
#'   Defaults to "revenue".
#' @param os Character. Operating system: "ios", "android", or "unified". Required.
#' @param category Integer or character. Category ID to filter publishers.
#'   For iOS use numeric IDs (e.g., 6014 for Games), for Android use strings
#'   (e.g., "game"). Use 0 or "all" for all categories.
#' @param time_range Character. Time period: "day", "week", "month", "quarter",
#'   or "year". Defaults to "month".
#' @param comparison_attribute Character. Data type to return: "absolute"
#'   (total values), "delta" (growth), or "transformed_delta" (growth rate).
#'   Defaults to "absolute".
#' @param date Date or character. Start date in "YYYY-MM-DD" format. Required.
#'   **Important**: Must align with time_range boundaries:
#'   - `month`: Must be first day of month (e.g., 2025-06-01)
#'   - `week`: Must be Monday
#'   - `quarter`: Must be quarter start (Jan 1, Apr 1, Jul 1, Oct 1)
#'   - `year`: Must be January 1st
#'   - `day`: Any date allowed
#'   Function will error if date doesn't align. Defaults to 30 days ago.
#' @param end_date Date or character. Optional end date for aggregating multiple
#'   periods. If not provided with `time_range = "month"`, "quarter", or "year",
#'   it will be automatically set to the last day of the period. 
#'   **Important**: If provided, must align with time_range boundaries:
#'   - `month`: Must be last day of month (e.g., 2025-06-30, 2025-07-31)
#'   - `week`: Must be Sunday  
#'   - `quarter`: Must be quarter end (Mar 31, Jun 30, Sep 30, Dec 31)
#'   - `year`: Must be December 31st
#'   - `day`: Any date allowed
#'   Function will error if date doesn't align. Use `time_range = "day"`
#'   for custom date ranges.
#' @param country Character. Country or region code (e.g., "US", "GB", "WW" for worldwide). Required.
#' @param limit Integer. Number of publishers to return (1-100). Defaults to 25.
#' @param offset Integer. Number of publishers to skip for pagination. Defaults to 0.
#' @param device_type Character. For iOS: "iphone", "ipad", or "total".
#'   For unified: "total". Ignored for Android. Defaults to "total".
#' @param include_apps Logical. Whether to include each publisher's top apps
#'   in the response. Defaults to TRUE.
#' @param auth_token Character. Your Sensor Tower API authentication token.
#'   Defaults to the value stored in the `SENSORTOWER_AUTH_TOKEN` environment
#'   variable.
#'
#' @return A [tibble][tibble::tibble] containing top publishers with columns:
#'   - `publisher_id`: Unique publisher identifier
#'   - `publisher_name`: Publisher display name
#'   - `date`: Date of the metrics (as provided by API)
#'   - `date_start`: Actual start date of the period covered
#'   - `date_end`: Actual end date of the period covered
#'   - `units_absolute`: Total downloads for the period
#'   - `units_delta`: Download growth vs previous period
#'   - `units_transformed_delta`: Download growth rate
#'   - `revenue_absolute`: Total revenue in cents for the period
#'   - `revenue_delta`: Revenue growth vs previous period
#'   - `revenue_transformed_delta`: Revenue growth rate
#'   - `revenue_usd`: Total revenue in USD (revenue_absolute / 100)
#'   - `rank`: Publisher rank based on selected measure
#'   - `apps`: Nested tibble with top apps (if include_apps = TRUE)
#'
#' @section API Notes:
#'   - All revenue values are returned in cents from the API
#'   - The function adds a `revenue_usd` column for convenience
#'   - Growth metrics compare to the previous equivalent period
#'   - Worldwide data may have a 2-3 day lag vs country-specific data
#'   
#' @section Date Handling:
#'   When using `time_range = "month"`, "quarter", or "year":
#'   - Dates MUST align with period boundaries or the function will error
#'   - Example: For `time_range = "month"`, use `date = "2025-06-01"`, not `"2025-06-27"`
#'   - This prevents unexpected data ranges from the API
#'   - For custom date ranges (e.g., June 27 - July 26), use `time_range = "day"` and aggregate
#'
#' @examples
#' \dontrun{
#' # Get top 10 game publishers by revenue for last month
#' top_publishers <- st_top_publishers(
#'   measure = "revenue",
#'   category = 6014,  # Games category for iOS
#'   limit = 10
#' )
#'
#' # Get top publishers by downloads with growth metrics
#' growth_publishers <- st_top_publishers(
#'   measure = "units",
#'   comparison_attribute = "delta",
#'   time_range = "week",
#'   limit = 20
#' )
#'
#' # This will ERROR - dates don't align with month boundaries:
#' # publishers_custom <- st_top_publishers(
#' #   date = "2025-06-27",        # ERROR: Not start of month!
#' #   end_date = "2025-07-26",    # ERROR: Not end of month!
#' #   time_range = "month"
#' # )
#' 
#' # Correct way for full months (end_date auto-calculated):
#' publishers_month <- st_top_publishers(
#'   date = "2025-06-01",      # First day of June
#'   time_range = "month"      # end_date auto-set to 2025-06-30
#' )
#' 
#' # Or specify multiple months:
#' publishers_months <- st_top_publishers(
#'   date = "2025-06-01",      # First day of June
#'   end_date = "2025-07-31",  # Last day of July
#'   time_range = "month"
#' )
#' 
#' # For custom date ranges (e.g., June 27 - July 26), use daily:
#' daily_publishers <- purrr::map_df(
#'   seq(as.Date("2025-06-27"), as.Date("2025-07-26"), by = "day"),
#'   ~ st_top_publishers(date = .x, time_range = "day", limit = 50)
#' ) %>%
#'   group_by(publisher_id, publisher_name) %>%
#'   summarise(total_revenue = sum(revenue_usd))
#' }
#'
#' @import dplyr
#' @importFrom httr GET add_headers stop_for_status content http_status
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble as_tibble
#' @importFrom tidyr nest
#' @importFrom rlang abort warn %||% .data sym
#' @importFrom lubridate floor_date ceiling_date
#' @importFrom purrr map map_int
#' @keywords internal
#' @noRd
st_top_publishers_impl <- function(measure = "revenue",
                             os,
                             category = 0,
                             time_range = "month",
                             comparison_attribute = "absolute",
                             date,
                             end_date = NULL,
                             country,
                             limit = 25,
                             offset = 0,
                             device_type = "total",
                             include_apps = TRUE,
                             auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN")) {
  
  auth_token <- resolve_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required.",
      "Set SENSORTOWER_AUTH_TOKEN environment variable",
      "or pass via auth_token argument."
    )
  )

  # Validate required parameters
  if (missing(os) || is.null(os)) {
    rlang::abort("'os' parameter is required. Specify one of: 'ios', 'android', 'unified'.")
  }
  
  if (missing(date) || is.null(date)) {
    rlang::abort("'date' parameter is required. Specify in YYYY-MM-DD format.")
  }
  
  if (missing(country) || is.null(country)) {
    rlang::abort("'country' parameter is required. Specify country code (e.g., 'US', 'GB', or 'WW' for worldwide).")
  }
  
  # Input validation
  measure <- match.arg(measure, c("revenue", "units"))
  os <- match.arg(os, c("ios", "android", "unified"))
  time_range <- match.arg(time_range, c("day", "week", "month", "quarter", "year"))
  comparison_attribute <- match.arg(comparison_attribute, 
                                   c("absolute", "delta", "transformed_delta"))
  
  # Display parameters being used
  message("\n=== Sensor Tower API Request ===")
  message(sprintf("  Endpoint: Top Publishers"))
  message(sprintf("  Measure: %s", measure))
  message(sprintf("  OS: %s", os))
  message(sprintf("  Country: %s", country))
  message(sprintf("  Time Range: %s", time_range))
  message(sprintf("  Date: %s", date))
  if (!is.null(end_date)) {
    message(sprintf("  End Date: %s", end_date))
  }
  message(sprintf("  Category: %s", category))
  message(sprintf("  Limit: %d", limit))
  message(sprintf("  Include Apps: %s", include_apps))
  message("================================\n")
  
  # Validate numeric inputs
  if (!is.numeric(limit) || limit < 1 || limit > 10) {
    rlang::abort("limit must be between 1 and 10")
  }
  
  if (!is.numeric(offset) || offset < 0) {
    rlang::abort("offset must be non-negative")
  }
  
  # Convert dates to proper format
  date <- as.Date(date)
  
  # Check if date aligns with time_range boundaries
  expected_date <- switch(time_range,
    day = date,
    week = floor_date(date, "week", week_start = 1),
    month = floor_date(date, "month"),
    quarter = floor_date(date, "quarter"),
    year = floor_date(date, "year")
  )
  
  # Error if date doesn't align with period boundaries
  if (time_range != "day" && date != expected_date) {
    rlang::abort(
      sprintf(
        "Date %s does not align with %s boundaries. For time_range='%s', date must be %s (start of %s). Use time_range='day' for custom date ranges.",
        format(date, "%Y-%m-%d"),
        time_range,
        time_range,
        format(expected_date, "%Y-%m-%d"),
        time_range
      )
    )
  }
  
  # Build API URL early for potential data availability check
  url <- paste0(st_api_base_url(), "/", st_endpoint_path("top_and_trending_publishers", os = os))
  
  # Auto-set end_date for period-based time_ranges if not provided
  if (is.null(end_date) && time_range != "day") {
    # Calculate theoretical end date
    theoretical_end <- switch(time_range,
      week = date + 6,  # Sunday of the same week
      month = ceiling_date(date, "month") - 1,  # Last day of the month
      quarter = ceiling_date(date, "quarter") - 1,  # Last day of the quarter
      year = ceiling_date(date, "year") - 1  # December 31st
    )
    
    # For current periods, we need to check actual data availability
    # The API typically has a 1-2 day lag
    if (theoretical_end >= Sys.Date() - 3) {
      # We're dealing with very recent data, need to check availability
      # Quick check: try yesterday first (most common case)
      test_date <- Sys.Date() - 1
      
      test_response <- tryCatch({
        httr::GET(
          url = url,
          query = list(
            auth_token = auth_token,
            comparison_attribute = "absolute",
            time_range = "day",
            measure = measure,
            category = as.character(category),
            date = format(test_date, "%Y-%m-%d"),
            limit = 1
          ),
          httr::add_headers("Accept" = "application/json")
        )
      }, error = function(e) NULL)
      
      # Check if yesterday has data
      if (!is.null(test_response) && httr::status_code(test_response) == 200) {
        test_content <- httr::content(test_response, "text", encoding = "UTF-8")
        if (nzchar(test_content) && test_content != "[]") {
          # Yesterday has data, use it
          latest_available <- test_date
        } else {
          # Yesterday is empty, try day before
          latest_available <- Sys.Date() - 2
        }
      } else {
        # API call failed, assume 2-day lag
        latest_available <- Sys.Date() - 2
      }
      
      # Use the minimum of theoretical end and latest available
      if (theoretical_end > latest_available) {
        end_date <- latest_available
        message(sprintf(
          "Auto-setting end_date to %s (latest available data)",
          format(end_date, "%Y-%m-%d")
        ))
      } else {
        end_date <- theoretical_end
        message(sprintf(
          "Auto-setting end_date to %s (last day of %s starting %s)",
          format(end_date, "%Y-%m-%d"),
          time_range,
          format(date, "%Y-%m-%d")
        ))
      }
    } else {
      # Historical data, use theoretical end
      end_date <- theoretical_end
      message(sprintf(
        "Auto-setting end_date to %s (last day of %s starting %s)",
        format(end_date, "%Y-%m-%d"),
        time_range,
        format(date, "%Y-%m-%d")
      ))
    }
  } else if (!is.null(end_date)) {
    # If end_date is provided, validate it
    end_date <- as.Date(end_date)
    
    # For non-day time ranges, check that end_date aligns with period boundaries
    if (time_range != "day") {
      expected_end_date <- switch(time_range,
        week = ceiling_date(end_date, "week", week_start = 1) - 1,
        month = ceiling_date(floor_date(end_date, "month"), "month") - 1,
        quarter = ceiling_date(floor_date(end_date, "quarter"), "quarter") - 1,
        year = ceiling_date(floor_date(end_date, "year"), "year") - 1
      )
      
      # Check if end_date is at period boundary
      if (end_date != expected_end_date) {
        rlang::abort(
          sprintf(
            "End date %s does not align with %s boundaries. For time_range='%s', end_date must be the last day of a %s. Use time_range='day' for custom date ranges.",
            format(end_date, "%Y-%m-%d"),
            time_range,
            time_range,
            time_range
          )
        )
      }
    }
  }
  
  # URL was already built earlier for data availability check
  
  # Build query parameters
  query_params <- list(
    auth_token = auth_token,
    comparison_attribute = comparison_attribute,
    time_range = time_range,
    measure = measure,
    category = as.character(category),
    date = format(date, "%Y-%m-%d"),
    limit = limit,
    offset = offset
  )
  
  # Add optional parameters
  if (!is.null(end_date)) {
    query_params$end_date <- format(end_date, "%Y-%m-%d")
  }
  
  if (!is.null(country) && country != "" && country != "WW") {
    query_params$country <- country
  }
  
  if (os %in% c("ios", "unified")) {
    query_params$device_type <- device_type
  }
  
  # Make API request
  response <- tryCatch(
    httr::GET(
      url = url,
      query = query_params,
      httr::add_headers("Accept" = "application/json")
    ),
    error = function(e) {
      rlang::abort(paste("HTTP request failed:", e$message))
    }
  )
  
  # Handle response
  parsed_data <- tryCatch(
    {
      httr::stop_for_status(
        response,
        task = paste("fetch top publishers for", measure, "in", os)
      )
      content_text <- httr::content(response, "text", encoding = "UTF-8")
      if (content_text == "") {
        rlang::warn("API returned an empty response body.")
        return(tibble::tibble())
      }
      jsonlite::fromJSON(content_text, flatten = TRUE)
    },
    error = function(e) {
      if (httr::http_error(response)) {
        status <- httr::http_status(response)
        error_msg <- paste0(
          "API request failed (HTTP ", status$status_code, "): ",
          status$reason
        )
        
        # Try to parse error message from response
        error_content <- httr::content(response, "text", encoding = "UTF-8")
        if (nzchar(error_content)) {
          tryCatch({
            error_json <- jsonlite::fromJSON(error_content)
            if (!is.null(error_json$error)) {
              error_msg <- paste0(error_msg, " - ", error_json$error)
            }
          }, error = function(e) {})
        }
        
        rlang::abort(error_msg)
      } else {
        rlang::abort(paste("Error processing API response:", e$message))
      }
    }
  )
  
  # Convert to tibble
  if (is.null(parsed_data) || length(parsed_data) == 0) {
    rlang::warn("No publishers found for the specified criteria.")
    return(tibble::tibble())
  }
  
  # Process the response
  publishers_df <- tibble::as_tibble(parsed_data)
  
  # Add date_start and date_end columns to clarify the actual period covered
  # This is crucial for understanding what data we're actually looking at
  # Use the validated date objects from earlier in the function
  actual_start_date <- date  # This was converted to Date earlier
  actual_end_date <- end_date %||% 
                     switch(time_range,
                       day = date,
                       week = date + 6,
                       month = ceiling_date(date, "month") - 1,
                       quarter = ceiling_date(date, "quarter") - 1,
                       year = ceiling_date(date, "year") - 1
                     )
  
  # For current periods, API might return partial data
  if (time_range != "day" && actual_end_date > Sys.Date()) {
    actual_end_date <- Sys.Date() - 1  # Yesterday
    message(sprintf("Note: Data covers %s to %s (partial %s)\n", 
                    actual_start_date, actual_end_date, time_range))
  }
  
  publishers_df <- publishers_df %>%
    mutate(
      date_start = format(actual_start_date, "%Y-%m-%d"),
      date_end = format(actual_end_date, "%Y-%m-%d"),
      # Add a warning column for YTD metrics
      ytd_warning = if (time_range == "year" && 
                       comparison_attribute %in% c("delta", "transformed_delta")) {
        "Use st_ytd_metrics() for accurate YTD metrics"
      } else {
        NA_character_
      }
    )
  
  # Add rank based on the measure used
  rank_column <- paste0(measure, "_absolute")
  if (rank_column %in% names(publishers_df)) {
    publishers_df <- publishers_df %>%
      arrange(desc(!!sym(rank_column))) %>%
      mutate(rank = row_number())
  }
  
  # Convert revenue from cents to dollars
  if ("revenue_absolute" %in% names(publishers_df)) {
    publishers_df <- publishers_df %>%
      mutate(revenue_usd = revenue_absolute / 100)
  }
  
  # Process nested apps data if present
  if ("apps" %in% names(publishers_df) && include_apps) {
    publishers_df <- publishers_df %>%
      mutate(apps = purrr::map(apps, ~{
        if (!is.null(.x) && length(.x) > 0) {
          app_df <- as_tibble(.x)
          if ("revenue_absolute" %in% names(app_df)) {
            app_df <- app_df %>%
              mutate(revenue_usd = revenue_absolute / 100)
          }
          app_df
        } else {
          tibble()
        }
      }))
  } else if (!include_apps && "apps" %in% names(publishers_df)) {
    publishers_df <- publishers_df %>%
      select(-apps)
  }
  
  # Reorder columns for better readability
  col_order <- c("rank", "publisher_id", "publisher_name", "date", 
                 "date_start", "date_end",
                 "revenue_absolute", "revenue_usd", "revenue_delta", 
                 "revenue_transformed_delta", "units_absolute", "units_delta",
                 "units_transformed_delta")
  
  if (include_apps) {
    col_order <- c(col_order, "apps")
  }
  
  # Keep only columns that exist
  col_order <- col_order[col_order %in% names(publishers_df)]
  
  # Add any remaining columns
  remaining_cols <- setdiff(names(publishers_df), col_order)
  col_order <- c(col_order, remaining_cols)
  
  publishers_df <- publishers_df %>%
    select(all_of(col_order))
  
  return(publishers_df)
}

# Test helper functions

skip_if_no_auth <- function() {
  if (!identical(tolower(Sys.getenv("SENSORTOWER_RUN_LIVE")), "true")) {
    skip("Set SENSORTOWER_RUN_LIVE=true to run live Sensor Tower API tests")
  }
  auth_token <- Sys.getenv("SENSORTOWER_AUTH_TOKEN")
  if (auth_token == "") {
    skip("Sensor Tower authentication token not found")
  }
}

test_that("live st_market_metrics smoke test uses aggregate market parameters", {
  skip_on_cran()
  skip_if_no_auth()

  result <- st_market_metrics(
    category = 7001,
    countries = "US",
    os = "ios",
    date_from = "2024-01-01",
    date_to = "2024-01-31",
    granularity = "monthly",
    shape = "wide"
  )

  expect_s3_class(result, "tbl_df")
  expect_true(all(c("date", "country", "category_id", "os", "revenue_usd", "downloads") %in% names(result)))
  if (nrow(result) > 0) {
    expect_true(all(result$revenue_usd >= 0, na.rm = TRUE))
    expect_true(all(result$downloads >= 0, na.rm = TRUE))
    expect_equal(unique(result$os), "ios")
    expect_equal(unique(result$country), "US")
  }
})

# --- Constants for sensortowerR ---

#' Metric Mapping for Custom Tags
#' 
#' Maps raw API field names (from entities.custom_tags or aggregate_tags)
#' to cleaner, user-friendly column names.
#' 
#' @keywords internal
METRIC_MAPPING <- c(
  # Performance metrics
  "aggregate_tags.Last 180 Days Downloads (WW)" = "downloads_180d_ww",
  "aggregate_tags.Last 180 Days Revenue (WW)" = "revenue_180d_ww", 
  "aggregate_tags.Last 30 Days Average DAU (US)" = "dau_30d_us",
  "aggregate_tags.Last 30 Days Average DAU (WW)" = "dau_30d_ww",
  "aggregate_tags.Last 30 Days Downloads (WW)" = "downloads_30d_ww",
  "aggregate_tags.Last 30 Days Revenue (WW)" = "revenue_30d_ww",
  "aggregate_tags.Last 4 Weeks Average WAU (US)" = "wau_4w_us",
  "aggregate_tags.Last 4 Weeks Average WAU (WW)" = "wau_4w_ww",
  "aggregate_tags.Last Month Average MAU (US)" = "mau_month_us",
  "aggregate_tags.Last Month Average MAU (WW)" = "mau_month_ww",
  
  # Historical metrics
  "aggregate_tags.All Time Downloads (US)" = "downloads_alltime_us",
  "aggregate_tags.All Time Downloads (WW)" = "downloads_alltime_ww",
  "aggregate_tags.All Time Revenue (US)" = "revenue_alltime_us", 
  "aggregate_tags.All Time Revenue (WW)" = "revenue_alltime_ww",
  "aggregate_tags.All Time Publisher Downloads (WW)" = "publisher_downloads_alltime_ww",
  "aggregate_tags.All Time Publisher Revenue (WW)" = "publisher_revenue_alltime_ww",
  
  # RPD metrics
  "aggregate_tags.RPD (All Time, US)" = "rpd_alltime_us",
  "aggregate_tags.RPD (All Time, WW)" = "rpd_alltime_ww",
  
  # Launch metrics
  "aggregate_tags.Release Date (US)" = "release_date_us",
  "aggregate_tags.Release Date (WW)" = "release_date_ww",
  "aggregate_tags.Release Date (JP)" = "release_date_jp",
  "aggregate_tags.Earliest Release Date" = "earliest_release_date",
  "aggregate_tags.Revenue First 30 Days (WW)" = "revenue_first30d_ww",
  "aggregate_tags.Downloads First 30 Days (WW)" = "downloads_first30d_ww",
  
  # Retention metrics
  "aggregate_tags.Day 1 Retention (Last Quarter, US)" = "retention_1d_us",
  "aggregate_tags.Day 1 Retention (Last Quarter, WW)" = "retention_1d_ww",
  "aggregate_tags.Day 7 Retention (Last Quarter, US)" = "retention_7d_us",
  "aggregate_tags.Day 7 Retention (Last Quarter, WW)" = "retention_7d_ww",
  "aggregate_tags.Day 14 Retention (Last Quarter, US)" = "retention_14d_us",
  "aggregate_tags.Day 14 Retention (Last Quarter, WW)" = "retention_14d_ww",
  "aggregate_tags.Day 30 Retention (Last Quarter, US)" = "retention_30d_us",
  "aggregate_tags.Day 30 Retention (Last Quarter, WW)" = "retention_30d_ww",
  "aggregate_tags.Day 60 Retention (Last Quarter, US)" = "retention_60d_us",
  "aggregate_tags.Day 60 Retention (Last Quarter, WW)" = "retention_60d_ww",
  
  # Demographics
  "aggregate_tags.Age (Last Quarter, US)" = "age_us",
  "aggregate_tags.Age (Last Quarter, WW)" = "age_ww",
  "aggregate_tags.Genders (Last Quarter, US)" = "genders_us",
  "aggregate_tags.Genders (Last Quarter, WW)" = "genders_ww",
  "aggregate_tags.Gender (Last Quarter, US)" = "gender_us",
  "aggregate_tags.Gender (Last Quarter, WW)" = "gender_ww",
  
  # Additional revenue metrics
  "aggregate_tags.Most Popular Country by Revenue" = "most_popular_country_revenue",
  "aggregate_tags.Last 90 Days Downloads (US)" = "downloads_90d_us",
  "aggregate_tags.Last 90 Days Downloads (WW)" = "downloads_90d_ww",
  "aggregate_tags.Last 90 Days Revenue (US)" = "revenue_90d_us",
  "aggregate_tags.Last 90 Days Revenue (WW)" = "revenue_90d_ww",
  "aggregate_tags.ARPU (90 Days, US)" = "arpu_90d_us",
  "aggregate_tags.ARPU (90 Days, WW)" = "arpu_90d_ww",
  
  # Platform shares
  "aggregate_tags.Android Share (WW)" = "android_share_ww",
  "aggregate_tags.iOS Share (WW)" = "ios_share_ww",
  "aggregate_tags.Female Share (US)" = "female_share_us",
  "aggregate_tags.Male Share (US)" = "male_share_us",
  
  # Additional useful metrics from entities.custom_tags (when available)
  "entities.custom_tags.Game Sub-genre" = "game_subgenre",
  "entities.custom_tags.Game Genre" = "game_genre",
  "entities.custom_tags.Game Art Style" = "game_art_style",
  "entities.custom_tags.Primary Category" = "primary_category",
  "entities.custom_tags.Overall US Rating" = "us_rating",
  "entities.custom_tags.Current US Rating" = "current_us_rating",
  "entities.custom_tags.Free" = "is_free",
  "entities.custom_tags.In-App Purchases" = "has_iap",
  "entities.custom_tags.Contains Ads" = "has_ads"
)

#' Numeric Metric Patterns
#' 
#' Patterns to identify columns that should be converted to numeric.
#' 
#' @keywords internal
NUMERIC_METRIC_PATTERNS <- c(
  "downloads", "revenue", "users", "dau", "wau", "mau", "rpd", 
  "retention", "age", "rating", "count", "days", "size", "absolute",
  "delta", "share", "percent", "time", "last", "average", "total",
  "first", "all time", "180", "30", "90", "7d", "14d", "month"
)

#' Fetch Aggregate Sensor Tower Market Metrics
#'
#' Tidy facade for market/category revenue and download denominators. This
#' function routes to Sensor Tower's aggregate `games_breakdown` endpoint through
#' [st_game_summary()] and never approximates market totals by enumerating apps,
#' rankings, top charts, or title cohorts.
#'
#' @param category Character or numeric vector of Sensor Tower game category IDs.
#'   iOS and Android use different category identifiers; pass the appropriate
#'   platform category for platform-specific calls, or the full platform basket
#'   for `os = "unified"`.
#' @param countries Character vector of 2-letter country codes. Use `"WW"` for
#'   worldwide.
#' @param os Character scalar. One of `"ios"`, `"android"`, or `"unified"`.
#' @param date_from,date_to Date bounds for the query. Accept `Date` objects or
#'   ISO date strings.
#' @param granularity Character scalar. One of `"daily"`, `"weekly"`,
#'   `"monthly"`, or `"quarterly"`.
#' @param metrics Character vector. One or both of `"revenue"` and `"downloads"`.
#' @param revenue_unit Character scalar. `"dollars"` returns revenue in base
#'   currency units. `"cents"` returns revenue in cents.
#' @param shape Character scalar. `"long"` returns one row per metric
#'   observation. `"wide"` returns metric columns.
#' @param auth_token Optional Sensor Tower API token. If `NULL`, falls back to
#'   `SENSORTOWER_AUTH_TOKEN`.
#'
#' @return
#' If `shape = "long"`, a tibble with columns `date`, `country`, `category_id`,
#' `os`, `metric`, and `value`. If `shape = "wide"`, a tibble with columns
#' `date`, `country`, `category_id`, `os`, and requested metric columns.
#'
#' @section Denominator Safety:
#' Use `st_market_metrics()` for game category market denominators. Use
#' [st_rankings()], [st_apps()], and custom-field top-chart workflows for
#' discovery, leaderboards, and title cohorts only. True Game IQ subgenre market
#' denominators require a Sensor Tower market-size export or another confirmed
#' aggregate market endpoint.
#'
#' @examples
#' \dontrun{
#' casino_market <- st_market_metrics(
#'   category = c("7006", "game_casino"),
#'   countries = "WW",
#'   os = "unified",
#'   date_from = "2025-01-01",
#'   date_to = "2025-12-31",
#'   granularity = "monthly"
#' )
#' }
#'
#' @export
st_market_metrics <- function(category,
                              countries,
                              os,
                              date_from,
                              date_to,
                              granularity = "monthly",
                              metrics = c("revenue", "downloads"),
                              revenue_unit = c("dollars", "cents"),
                              shape = c("long", "wide"),
                              auth_token = NULL) {
  if (missing(category) || is.null(category) || !length(category)) {
    rlang::abort("`category` must be a non-empty character or numeric vector.")
  }

  category <- unique(as.character(category))
  if (any(is.na(category) | !nzchar(category))) {
    rlang::abort("`category` entries must be non-empty strings.")
  }

  os <- normalize_os(os)
  countries <- normalize_countries(countries)
  dates <- normalize_dates(date_from, date_to)
  granularity <- normalize_granularity(granularity)
  metrics <- normalize_metrics(metrics, allowed = c("revenue", "downloads"))
  revenue_unit <- match.arg(revenue_unit)
  shape <- match.arg(shape)
  auth_token <- get_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required.",
      "Set SENSORTOWER_AUTH_TOKEN or pass `auth_token`."
    )
  )

  raw <- st_game_summary(
    categories = category,
    countries = countries,
    os = os,
    date_granularity = granularity,
    start_date = dates$date_from,
    end_date = dates$date_to,
    auth_token = auth_token,
    enrich_response = TRUE
  )

  .st_market_metrics_normalize(
    data = raw,
    category = category,
    os = os,
    metrics = metrics,
    revenue_unit = revenue_unit,
    shape = shape
  )
}

.st_market_metrics_normalize <- function(data,
                                         category,
                                         os,
                                         metrics,
                                         revenue_unit,
                                         shape) {
  if (is.null(data) || !nrow(data)) {
    return(.st_market_metrics_empty(shape = shape, metrics = metrics, revenue_unit = revenue_unit))
  }

  data <- tibble::as_tibble(data)
  n <- nrow(data)
  category_label <- paste(category, collapse = ",")

  date <- .st_market_col(data, "Date", default = as.Date(rep(NA_character_, n)))
  country <- .st_market_col(data, "Country Code", default = rep(NA_character_, n))
  category_id <- if ("Category" %in% names(data)) {
    as.character(data$Category)
  } else {
    rep(category_label, n)
  }

  revenue_col <- dplyr::case_when(
    os == "ios" ~ "iOS Revenue",
    os == "android" ~ "Android Revenue",
    TRUE ~ "Total Revenue"
  )
  downloads_col <- dplyr::case_when(
    os == "ios" ~ "iOS Downloads",
    os == "android" ~ "Android Downloads",
    TRUE ~ "Total Downloads"
  )

  wide <- tibble::tibble(
    date = as.Date(date),
    country = as.character(country),
    category_id = as.character(category_id),
    os = rep(os, n),
    revenue_usd = .st_market_numeric_col(data, revenue_col),
    downloads = .st_market_numeric_col(data, downloads_col)
  )

  if (identical(revenue_unit, "cents")) {
    wide$revenue_usd <- wide$revenue_usd * 100
    names(wide)[names(wide) == "revenue_usd"] <- "revenue_cents"
  }

  metric_cols <- .st_market_metric_columns(metrics, revenue_unit)
  wide <- dplyr::select(wide, dplyr::all_of(c("date", "country", "category_id", "os", metric_cols)))

  if (identical(shape, "wide")) {
    return(wide)
  }

  long <- tidyr::pivot_longer(
    wide,
    cols = dplyr::all_of(metric_cols),
    names_to = "metric",
    values_to = "value"
  )
  long$metric <- dplyr::recode(
    long$metric,
    revenue_usd = "revenue",
    revenue_cents = "revenue",
    downloads = "downloads"
  )
  dplyr::select(long, date, country, category_id, os, metric, value)
}

.st_market_metrics_empty <- function(shape, metrics, revenue_unit) {
  if (identical(shape, "wide")) {
    metric_cols <- .st_market_metric_columns(metrics, revenue_unit)
    out <- tibble::tibble(
      date = as.Date(character()),
      country = character(),
      category_id = character(),
      os = character()
    )
    for (col in metric_cols) {
      out[[col]] <- numeric()
    }
    return(out)
  }

  tibble::tibble(
    date = as.Date(character()),
    country = character(),
    category_id = character(),
    os = character(),
    metric = character(),
    value = numeric()
  )
}

.st_market_metric_columns <- function(metrics, revenue_unit) {
  cols <- character()
  if ("revenue" %in% metrics) {
    cols <- c(cols, if (identical(revenue_unit, "cents")) "revenue_cents" else "revenue_usd")
  }
  if ("downloads" %in% metrics) {
    cols <- c(cols, "downloads")
  }
  cols
}

.st_market_col <- function(data, column, default) {
  if (column %in% names(data)) {
    return(data[[column]])
  }
  default
}

.st_market_numeric_col <- function(data, column) {
  if (!column %in% names(data)) {
    return(rep(NA_real_, nrow(data)))
  }
  as.numeric(data[[column]])
}

#' Year-over-Year Metrics Comparison
#'
#' Fetches metrics for the same date range across multiple years for 
#' year-over-year comparisons. Allows flexible date ranges and supports
#' all available metrics including revenue, downloads, and active users.
#'
#' @param os Character. Required. Operating system: "ios", "android", or "unified".
#'   This determines which platform's data is returned.
#' @param unified_app_id Character vector. Sensor Tower unified app ID(s). 
#'   Must be 24-character hex format (e.g., "5ba4585f539ce75b97db6bcb").
#' @param ios_app_id Character vector. iOS app ID(s) (e.g., "1234567890").
#' @param android_app_id Character vector. Android package name(s) (e.g., "com.example.app").
#' @param publisher_id Character vector. Publisher ID(s) (alternative to app IDs).
#' @param years Integer vector. Years to compare (e.g., c(2022, 2023, 2024)).
#'   If NULL, uses current year and previous year.
#' @param period_start Character string or Date. Start of the comparison period.
#'   Can be "MM-DD" format (e.g., "01-01" for Jan 1) or a full date.
#'   If a full date is provided, only the month and day are used.
#' @param period_end Character string or Date. End of the comparison period.  
#'   Can be "MM-DD" format (e.g., "03-31" for Mar 31) or a full date.
#'   If a full date is provided, only the month and day are used.
#' @param metrics Character vector. Metrics to fetch. Supports "revenue", "downloads", 
#'   "dau", "wau", and "mau". Default is both revenue and downloads.
#' @param countries Character vector. Country codes (e.g., "US", "GB", "JP"). Required.
#' @param cache_dir Character. Directory for caching API responses (optional).
#' @param auth_token Character string. Sensor Tower API token.
#' @param verbose Logical. Print progress messages.
#' @param granularity Character. Date granularity for the data (e.g., "daily", "monthly").
#' @param use_single_fetch Logical. If TRUE, uses a single API call to fetch all data.
#'   Defaults to TRUE for efficiency.
#'
#' @return A tibble in tidy/long format with columns:
#'   - `app_id`: The app ID used for fetching data
#'   - `app_id_type`: Type of app ID ("ios", "android", or "unified")
#'   - `entity_id`: App or publisher ID
#'   - `entity_name`: App or publisher name
#'   - `entity_type`: "app" or "publisher"
#'   - `year`: Year of the data
#'   - `date_start`: Start date of the period (YYYY-MM-DD)
#'   - `date_end`: End date of the period (YYYY-MM-DD)
#'   - `country`: Country code
#'   - `metric`: The metric name (e.g., "revenue", "downloads", "dau")
#'   - `value`: Metric value (units depend on metric type)
#'   - `yoy_change`: Year-over-year change (percentage)
#'   - `yoy_change_absolute`: Year-over-year change (absolute value)
#'
#' @details
#' This function is designed for year-over-year comparisons:
#' 
#' - **Flexible date ranges**: Compare any period (e.g., Q1, specific months, custom ranges)
#' - **Multiple years**: Compare across 2+ years in a single call
#' - **Smart date handling**: Automatically handles leap years and invalid dates
#' - **YoY calculations**: Includes both percentage and absolute change
#' - **Caching**: Reuses cached data to minimize API calls
#'
#' The function will apply the same calendar period (month/day range) to each
#' specified year, making it easy to compare seasonal trends, campaign periods,
#' or any custom date range across years.
#'
#' @examples
#' \dontrun{
#' # Compare Q1 performance across years
#' q1_comparison <- st_yoy_metrics(
#'   os = "ios",
#'   ios_app_id = "553834731",  # Candy Crush iOS
#'   years = c(2022, 2023, 2024),
#'   period_start = "01-01",
#'   period_end = "03-31",
#'   countries = "US",
#'   metrics = c("revenue", "downloads")
#' )
#'
#' # Compare holiday season (Nov-Dec) across years
#' holiday_comparison <- st_yoy_metrics(
#'   os = "unified",
#'   unified_app_id = "5ba4585f539ce75b97db6bcb",
#'   years = c(2021, 2022, 2023),
#'   period_start = "11-01",
#'   period_end = "12-31",
#'   countries = c("US", "GB", "JP"),
#'   metrics = c("revenue", "downloads", "dau")
#' )
#'
#' # Compare specific campaign period using full dates
#' campaign_comparison <- st_yoy_metrics(
#'   os = "android", 
#'   android_app_id = "com.king.candycrushsaga",
#'   years = NULL,  # Uses current and previous year
#'   period_start = as.Date("2024-02-14"),  # Valentine's campaign
#'   period_end = as.Date("2024-02-28"),
#'   countries = c("US", "GB", "JP"),
#'   metrics = c("revenue", "downloads", "dau", "wau")
#' )
#' }
#'
#' @importFrom dplyr %>% group_by summarise mutate select filter bind_rows arrange lag
#' @importFrom lubridate year month day
#' @importFrom tibble tibble
#' @export
st_yoy_metrics <- function(
  os,
  unified_app_id = NULL,
  ios_app_id = NULL,
  android_app_id = NULL,
  publisher_id = NULL,
  years = NULL,
  period_start,
  period_end,
  metrics = c("revenue", "downloads"),
  countries,
  cache_dir = NULL,
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
  verbose = TRUE,
  granularity,
  use_single_fetch = TRUE
) {
  
  # Validate OS parameter
  if (missing(os) || is.null(os) || !os %in% c("ios", "android", "unified")) {
    rlang::abort("'os' parameter is required and must be one of: 'ios', 'android', or 'unified'")
  }
  
  # Validate required parameters
  if (missing(period_start) || is.null(period_start)) {
    rlang::abort("'period_start' is required")
  }
  
  if (missing(period_end) || is.null(period_end)) {
    rlang::abort("'period_end' is required")
  }
  
  if (missing(countries) || is.null(countries) || length(countries) == 0) {
    rlang::abort("'countries' parameter is required")
  }
  
  # Convert dates to MM-DD format if full dates provided
  if (inherits(period_start, "Date") || nchar(as.character(period_start)) == 10) {
    start_date <- as.Date(as.character(period_start))
    period_start <- format(start_date, "%m-%d")
  }
  
  if (inherits(period_end, "Date") || nchar(as.character(period_end)) == 10) {
    end_date <- as.Date(as.character(period_end))  
    period_end <- format(end_date, "%m-%d")
  }
  
  # Validate date formats
  if (!grepl("^\\d{2}-\\d{2}$", period_start)) {
    rlang::abort("period_start must be in MM-DD format (e.g., '02-01') or a Date object")
  }
  if (!grepl("^\\d{2}-\\d{2}$", period_end)) {
    rlang::abort("period_end must be in MM-DD format (e.g., '02-28') or a Date object")
  }
  
  # Default years to current and previous if not specified
  if (is.null(years)) {
    current_year <- as.integer(format(Sys.Date(), "%Y"))
    years <- c(current_year - 1, current_year)
    if (verbose) {
      message("No years specified. Using: ", paste(years, collapse = ", "))
    }
  }
  
  # Ensure years are integers and sorted
  years <- sort(as.integer(years))
  
  # Create end_dates for each year
  end_dates <- character()
  for (year in years) {
    # Construct the end date for this year
    end_date_str <- paste(year, period_end, sep = "-")
    
    # Try to create the date, handling invalid dates like Feb 29 in non-leap years
    end_date <- tryCatch(
      as.Date(end_date_str),
      error = function(e) {
        # If date is invalid, use the last valid day of the month
        month <- as.integer(substr(period_end, 1, 2))
        if (month == 2 && substr(period_end, 4, 5) == "29") {
          # Special case for Feb 29 - use Feb 28 in non-leap years
          as.Date(paste(year, "02-28", sep = "-"))
        } else {
          # General case - use last day of month
          last_day <- lubridate::ceiling_date(
            as.Date(paste(year, month, "01", sep = "-")), 
            "month"
          ) - 1
          last_day
        }
      }
    )
    
    end_dates <- c(end_dates, as.character(end_date))
  }
  
  if (verbose) {
    message("\nFetching year-over-year metrics:")
    message("  Years: ", paste(years, collapse = ", "))
    message("  Period: ", period_start, " to ", period_end, " (each year)")
    message("  Metrics: ", paste(metrics, collapse = ", "))
    message("  Countries: ", paste(countries, collapse = ", "))
  }
  
  # Validate granularity explicitly
  if (missing(granularity) || is.null(granularity)) {
    rlang::abort("'granularity' parameter is required. Specify one of: 'daily', 'weekly', 'monthly', 'quarterly'.")
  }

  # Build unified app_list once (or set publisher mode)
  app_list <- NULL
  publisher_mode <- FALSE
  if (!is.null(unified_app_id)) {
    app_list <- unified_app_id
  } else if (!is.null(ios_app_id) && !is.null(android_app_id)) {
    # Prefer passing a unified ID to ensure both-platform pairing downstream
    unified_map <- tryCatch(
      st_get_unified_mapping(ios_app_id, os = "ios", auth_token = auth_token),
      error = function(e) NULL
    )
    if (!is.null(unified_map) && !is.null(unified_map$unified_app_id[1])) {
      app_list <- as.character(unified_map$unified_app_id[1])
    } else {
      # Fallback: pass both IDs; downstream will fetch per-platform separately
      app_list <- c(as.character(ios_app_id), as.character(android_app_id))
    }
  } else if (!is.null(ios_app_id) || !is.null(android_app_id)) {
    # If unified OS requested but platform IDs supplied, resolve to unified to capture all regional SKUs
    if (os == 'unified') {
      unified_try <- NULL
      if (!is.null(ios_app_id)) {
        unified_try <- tryCatch(st_get_unified_mapping(ios_app_id, os = 'ios', auth_token = auth_token), error = function(e) NULL)
      }
      if ((is.null(unified_try) || is.null(unified_try$unified_app_id[1])) && !is.null(android_app_id)) {
        unified_try <- tryCatch(st_get_unified_mapping(android_app_id, os = 'android', auth_token = auth_token), error = function(e) NULL)
      }
      if (!is.null(unified_try) && !is.null(unified_try$unified_app_id[1])) {
        app_list <- as.character(unified_try$unified_app_id[1])
      } else {
        # Fallback to passing both platform IDs
        ids <- c(as.character(ios_app_id %||% character()), as.character(android_app_id %||% character()))
        app_list <- ids[ids != '']
      }
    } else {
      # Non-unified OS: pass platform IDs through
      app_list <- if (!is.null(ios_app_id)) ios_app_id else android_app_id
    }
  } else if (!is.null(publisher_id)) {
    publisher_mode <- TRUE
    # Resolve publisher apps once and create a cross-platform app list
    apps_tbl <- tryCatch(st_publisher_apps(publisher_id, auth_token = auth_token), error = function(e) NULL)
    if (!is.null(apps_tbl) && nrow(apps_tbl) > 0) {
      # Prefer unified IDs; pass simple vector compatible with st_batch_metrics
      unified_ids <- if ("unified_app_id" %in% names(apps_tbl)) unique(na.omit(as.character(apps_tbl$unified_app_id))) else character(0)
      app_list <- unified_ids
    } else {
      app_list <- character(0)
    }
  }

  # Use the user-specified granularity; no auto-granularity
  chosen_granularity <- granularity

  # If we can fetch once, do a single spanning request; otherwise loop per year
  all_results <- list()
  if (use_single_fetch) {
    # Determine overall start/end
    min_year <- min(years)
    max_year <- max(years)
    all_start <- as.Date(paste(min_year, period_start, sep = "-"))
    all_end <- as.Date(paste(max_year, period_end, sep = "-"))
    # Adjust invalid dates (e.g., 02-29) to valid
    if (is.na(all_start)) {
      m <- as.integer(substr(period_start, 1, 2))
      all_start <- lubridate::floor_date(as.Date(paste(min_year, m, "01", sep = "-")), "month")
    }
    if (is.na(all_end)) {
      m <- as.integer(substr(period_end, 1, 2))
      all_end <- lubridate::ceiling_date(as.Date(paste(max_year, m, "01", sep = "-")), "month") - 1
    }

    fetched <- tryCatch({
      st_batch_metrics_impl(
        os = os,
        app_list = app_list,
        metrics = metrics,
        date_range = list(start_date = all_start, end_date = all_end),
        countries = countries,
        granularity = chosen_granularity,
        parallel = FALSE,
        cache_dir = cache_dir,
        verbose = FALSE,
        auth_token = auth_token
      )
    }, error = function(e) {
      if (verbose) message("  Error fetching data: ", e$message)
      NULL
    })

    if (!is.null(fetched) && nrow(fetched) > 0) {
      fetched$date <- as.Date(fetched$date)
      # Cut per year windows and sum per app/country/metric
      per_year <- lapply(seq_along(years), function(i) {
        y <- years[i]
        ys <- tryCatch(as.Date(paste(y, period_start, sep = "-")), error = function(e) NA)
        ye <- tryCatch(as.Date(paste(y, period_end, sep = "-")), error = function(e) NA)
        if (is.na(ys)) {
          m <- as.integer(substr(period_start, 1, 2))
          ys <- lubridate::floor_date(as.Date(paste(y, m, "01", sep = "-")), "month")
        }
        if (is.na(ye)) {
          m <- as.integer(substr(period_end, 1, 2))
          ye <- lubridate::ceiling_date(as.Date(paste(y, m, "01", sep = "-")), "month") - 1
        }
        sub <- fetched %>% dplyr::filter(.data$date >= ys, .data$date <= ye)
        if (nrow(sub) == 0) return(NULL)
        sub %>%
          dplyr::group_by(.data$original_id, .data$app_name, .data$country, .data$metric) %>%
          dplyr::summarise(value = sum(.data$value, na.rm = TRUE), .groups = "drop") %>%
          dplyr::mutate(
            year = y,
            date_start = as.character(ys),
            date_end = as.character(ye),
            entity_id = .data$original_id,
            entity_name = .data$app_name,
            entity_type = if (publisher_mode) "publisher" else "app",
            app_id = .data$original_id,
            app_id_type = if (publisher_mode) "publisher" else os
          )
      })
      per_year <- per_year[!vapply(per_year, is.null, logical(1))]
      if (length(per_year) > 0) all_results[[1]] <- dplyr::bind_rows(per_year)
    }
  } else {
    # Per-year loop (with chosen granularity)
    for (i in seq_along(years)) {
      year <- years[i]
      start_date <- tryCatch(as.Date(paste(year, period_start, sep = "-")), error = function(e) NA)
      end_date <- tryCatch(as.Date(paste(year, period_end, sep = "-")), error = function(e) NA)
      if (is.na(start_date)) {
        m <- as.integer(substr(period_start, 1, 2))
        start_date <- lubridate::floor_date(as.Date(paste(year, m, "01", sep = "-")), "month")
      }
      if (is.na(end_date)) {
        m <- as.integer(substr(period_end, 1, 2))
        end_date <- lubridate::ceiling_date(as.Date(paste(year, m, "01", sep = "-")), "month") - 1
      }
      if (verbose) message(sprintf("Fetching data for %d: %s to %s", year, start_date, end_date))
      year_data <- tryCatch({
        st_batch_metrics_impl(
          os = os,
          app_list = app_list,
          metrics = metrics,
          date_range = list(start_date = start_date, end_date = end_date),
          countries = countries,
          granularity = chosen_granularity,
          parallel = FALSE,
          cache_dir = cache_dir,
          verbose = FALSE,
          auth_token = auth_token
        )
      }, error = function(e) NULL)
      if (!is.null(year_data) && nrow(year_data) > 0) {
        year_summary <- year_data %>%
          dplyr::group_by(.data$original_id, .data$app_name, .data$country, .data$metric) %>%
          dplyr::summarise(value = sum(.data$value, na.rm = TRUE), .groups = "drop") %>%
          dplyr::mutate(year = year, date_start = as.character(start_date), date_end = as.character(end_date), entity_id = .data$original_id, entity_name = .data$app_name, entity_type = if (publisher_mode) "publisher" else "app", app_id = .data$original_id, app_id_type = if (publisher_mode) "publisher" else os)
        all_results[[length(all_results) + 1]] <- year_summary
      }
    }
  }

  # Combine all years
  ytd_data <- if (length(all_results) > 0) dplyr::bind_rows(all_results) else tibble::tibble()
  
  # If no data returned, return empty tibble with expected columns
  if (length(all_results) == 0 || nrow(ytd_data) == 0) {
    return(tibble::tibble(
      app_id = character(),
      app_id_type = character(),
      entity_id = character(),
      entity_name = character(),
      entity_type = character(),
      year = integer(),
      date_start = character(),
      date_end = character(),
      country = character(),
      metric = character(),
      value = numeric(),
      yoy_change = numeric(),
      yoy_change_absolute = numeric()
    ))
  }
  
  # Calculate year-over-year changes
  yoy_data <- ytd_data %>%
    # Ensure data is sorted by entity, country, metric, and year
    arrange(entity_id, country, metric, year) %>%
    # Group by entity, country, and metric to calculate YoY within each group
    group_by(entity_id, country, metric) %>%
    # Calculate YoY changes
    mutate(
      # Previous year's value
      prev_value = lag(.data$value, n = 1),
      # Year-over-year absolute change
      yoy_change_absolute = .data$value - .data$prev_value,
      # Year-over-year percentage change
      yoy_change = ifelse(
        is.na(.data$prev_value) | .data$prev_value == 0,
        NA_real_,
        ((.data$value - .data$prev_value) / .data$prev_value) * 100
      )
    ) %>%
    # Remove the temporary prev_value column
    select(-prev_value) %>%
    ungroup()
  
  if (verbose) {
    # Print summary of YoY changes
    summary_data <- yoy_data %>%
      filter(!is.na(.data$yoy_change)) %>%
      group_by(metric, year) %>%
      summarise(
        avg_yoy_change = mean(.data$yoy_change, na.rm = TRUE),
        .groups = "drop"
      )
    
    if (nrow(summary_data) > 0) {
      message("\nYear-over-Year Summary:")
      for (i in seq_len(nrow(summary_data))) {
        message(sprintf("  %s %d: %+.1f%% avg change", 
                       summary_data$metric[i], 
                       summary_data$year[i],
                       summary_data$avg_yoy_change[i]))
      }
    }
  }
  
  return(yoy_data)
}

#' Calculate year-over-year growth rates
#'
#' Helper function to calculate YoY growth rates from the output of st_yoy_metrics
#'
#' @param yoy_data Output from st_yoy_metrics
#' @param baseline_year The year to use as baseline (default: earliest year)
#'
#' @return A tibble with growth rates relative to baseline year
#' @export
calculate_yoy_growth <- function(yoy_data, baseline_year = NULL) {
  
  if (is.null(baseline_year)) {
    baseline_year <- min(yoy_data$year)
  }
  
  # Get baseline values
  baseline_data <- yoy_data %>%
    filter(.data$year == baseline_year) %>%
    select(entity_id, country, metric, baseline_value = value)
  
  # Calculate growth rates
  growth_data <- yoy_data %>%
    left_join(baseline_data, by = c("entity_id", "country", "metric")) %>%
    mutate(
      growth_from_baseline = ifelse(
        is.na(.data$baseline_value) | .data$baseline_value == 0,
        NA_real_,
        ((.data$value - .data$baseline_value) / .data$baseline_value) * 100
      ),
      growth_index = ifelse(
        is.na(.data$baseline_value) | .data$baseline_value == 0,
        NA_real_,
        (.data$value / .data$baseline_value) * 100
      )
    )
  
  return(growth_data)
}

test_that("st_yoy_metrics validates required parameters", {
  expect_error(
    st_yoy_metrics(
      ios_app_id = "123456789",
      period_start = "01-01",
      period_end = "03-31",
      countries = "US"
    ),
    "'os' parameter is required"
  )
  
  expect_error(
    st_yoy_metrics(
      os = "ios",
      ios_app_id = "123456789",
      period_end = "03-31",
      countries = "US"
    ),
    "'period_start' is required"
  )
  
  expect_error(
    st_yoy_metrics(
      os = "ios",
      ios_app_id = "123456789",
      period_start = "01-01",
      countries = "US"
    ),
    "'period_end' is required"
  )
  
  expect_error(
    st_yoy_metrics(
      os = "ios",
      ios_app_id = "123456789",
      period_start = "01-01",
      period_end = "03-31"
    ),
    "'countries' parameter is required"
  )
})

test_that("st_yoy_metrics works with basic inputs", {
  skip_if_no_auth()
  
  # Test with a short date range to minimize API calls
  result <- st_yoy_metrics(
    os = "ios",
    ios_app_id = "553834731",  # Candy Crush
    years = c(2023, 2024),
    period_start = "01-01",
    period_end = "01-07",
    countries = "US",
    metrics = "revenue",
    granularity = "daily"  )
  
  # Check structure
  expect_s3_class(result, "tbl_df")
  
  # Check expected columns
  expected_cols <- c("app_id", "app_id_type", "entity_id", "entity_name", 
                    "entity_type", "year", "date_start", "date_end", 
                    "country", "metric", "value", "yoy_change", "yoy_change_absolute")
  expect_true(all(expected_cols %in% names(result)))
  
  # Check that we have data for both years if data was returned
  if (nrow(result) > 0) {
    years_in_data <- unique(result$year)
    expect_true(2023 %in% years_in_data || 2024 %in% years_in_data)
  }
})

test_that("st_yoy_metrics calculates YoY changes correctly", {
  # Create mock data to test calculations
  test_data <- tibble::tibble(
    original_id = "test_app",
    app_name = "Test App",
    country = c("US", "US"),
    metric = c("revenue", "revenue"),
    value = c(1000, 1200),
    year = c(2023, 2024),
    date_start = c("2023-01-01", "2024-01-01"),
    date_end = c("2023-01-31", "2024-01-31"),
    entity_id = "test_app",
    entity_name = "Test App",
    entity_type = "app",
    app_id = "test_app",
    app_id_type = "ios"
  )
  
  # Apply the YoY calculation logic from st_yoy_metrics
  yoy_data <- test_data %>%
    dplyr::arrange(entity_id, country, metric, year) %>%
    dplyr::group_by(entity_id, country, metric) %>%
    dplyr::mutate(
      prev_value = dplyr::lag(value, n = 1),
      yoy_change_absolute = value - prev_value,
      yoy_change = ifelse(
        is.na(prev_value) | prev_value == 0,
        NA_real_,
        ((value - prev_value) / prev_value) * 100
      )
    ) %>%
    dplyr::select(-prev_value) %>%
    dplyr::ungroup()
  
  # Check calculations
  expect_equal(yoy_data$yoy_change[2], 20)  # 20% increase
  expect_equal(yoy_data$yoy_change_absolute[2], 200)  # $200 increase
  expect_true(is.na(yoy_data$yoy_change[1]))  # First year has no previous
})

test_that("st_yoy_metrics handles leap years correctly", {
  skip_if_no_auth()
  
  # Test Feb 29 handling across leap and non-leap years
  result <- st_yoy_metrics(
    os = "ios",
    ios_app_id = "553834731",
    years = c(2023, 2024),  # 2024 is a leap year
    period_start = "02-01",
    period_end = "02-29",  # Feb 29
    countries = "US",
    metrics = "revenue",
    granularity = "daily"  )
  
  # Check that dates were adjusted appropriately
  if (nrow(result) > 0) {
    feb_2023 <- result[result$year == 2023, ]$date_end[1]
    feb_2024 <- result[result$year == 2024, ]$date_end[1]
    
    if (!is.na(feb_2023)) expect_equal(feb_2023, "2023-02-28")  # Non-leap year
    if (!is.na(feb_2024)) expect_equal(feb_2024, "2024-02-29")  # Leap year
  }
})

test_that("st_yoy_metrics uses default years when not specified", {
  skip_if_no_auth()
  
  result <- st_yoy_metrics(
    os = "ios",
    ios_app_id = "553834731",
    years = NULL,  # Should default to current and previous year
    period_start = "01-01",
    period_end = "01-07",
    countries = "US",
    metrics = "revenue",
    granularity = "daily"  )
  
  current_year <- as.integer(format(Sys.Date(), "%Y"))
  
  # Check structure even if no data
  expect_s3_class(result, "tbl_df")
  
  if (nrow(result) > 0) {
    years_in_data <- unique(result$year)
    # At least one of the default years should be present
    expect_true(current_year %in% years_in_data || (current_year - 1) %in% years_in_data)
  }
})

test_that("calculate_yoy_growth works correctly", {
  # Create test data
  test_data <- tibble::tibble(
    entity_id = "app1",
    country = "US",
    metric = "revenue",
    year = c(2021, 2022, 2023),
    value = c(1000, 1500, 2000)
  )
  
  # Test with default baseline (earliest year)
  growth <- calculate_yoy_growth(test_data)
  
  expect_equal(growth$growth_from_baseline[1], 0)  # 2021 is baseline
  expect_equal(growth$growth_from_baseline[2], 50)  # 50% growth from 2021
  expect_equal(growth$growth_from_baseline[3], 100)  # 100% growth from 2021
  
  expect_equal(growth$growth_index[1], 100)  # Base year = 100
  expect_equal(growth$growth_index[2], 150)  # 150% of base
  expect_equal(growth$growth_index[3], 200)  # 200% of base
  
  # Test with specific baseline year
  growth_2022 <- calculate_yoy_growth(test_data, baseline_year = 2022)
  
  # For year 2021, it's -33.33% from the 2022 baseline (1000 vs 1500)
  expect_equal(growth_2022$growth_from_baseline[growth_2022$year == 2021], -33.33333, tolerance = 0.001)
  # For year 2022 (baseline year), growth should be 0
  expect_equal(growth_2022$growth_from_baseline[growth_2022$year == 2022], 0)
  # For year 2023, growth from 2022 baseline should be 33.33%
  expect_equal(growth_2022$growth_from_baseline[growth_2022$year == 2023], 33.33333, tolerance = 0.001)
})

test_that("st_yoy_metrics handles multiple countries", {
  skip_if_no_auth()
  
  result <- st_yoy_metrics(
    os = "ios",
    ios_app_id = "553834731",
    years = c(2024),  # Just one year to minimize API calls
    period_start = "01-01",
    period_end = "01-07",
    countries = c("US", "GB"),
    metrics = "revenue",
    granularity = "daily"  )
  
  # Check structure
  expect_s3_class(result, "tbl_df")
  
  if (nrow(result) > 0) {
    countries_in_data <- unique(result$country)
    # Check that we got data for at least one country
    expect_true(length(countries_in_data) >= 1)
  }
})

#' Batch Look up app information
#'
#' This function takes a vector of unified app IDs and returns a single
#' data frame containing the corresponding unified_app_id, app_name,
#' platform-specific IDs (ios_app_id, android_app_id), and icon_url.
#' It handles chunking requests to the st_app_details endpoint (limit 100).
#'
#' @param app_ids Character vector of unified app IDs.
#' @param auth_token Character string. Your Sensor Tower API authentication token.
#' @param verbose Logical. Whether to show progress messages. Default is FALSE.
#'
#' @return A \code{tibble} with columns:
#'   \itemize{
#'     \item \code{unified_app_id}: The Sensor Tower unified app ID
#'     \item \code{app_name}: The app's display name
#'     \item \code{ios_app_id}: iOS app ID if found
#'     \item \code{android_app_id}: Android app ID if found
#'     \item \code{icon_url}: URL of the app icon
#'   }
#'   Returns empty tibble if no apps can be found or details cannot be fetched.
#'
#' @examples
#' \dontrun{
#' # Look up multiple games
#' apps <- st_batch_app_lookup(c("5ba4585f539ce75b97db6bcb", "5d10d6bfecb4db5f93902f23"))
#' }
#'
#' @importFrom dplyr select mutate left_join bind_rows distinct coalesce rename transmute
#' @importFrom purrr map_df
#' @importFrom tibble tibble
#' @export
st_batch_app_lookup <- function(app_ids,
                                auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                                verbose = FALSE) {
    if (length(app_ids) == 0) {
        return(tibble::tibble(
            unified_app_id = character(),
            app_name = character(),
            ios_app_id = character(),
            android_app_id = character(),
            icon_url = character()
        ))
    }

    app_ids <- unique(app_ids[!is.na(app_ids) & nzchar(app_ids)])

    if (length(app_ids) == 0) {
        return(tibble::tibble(
            unified_app_id = character(),
            app_name = character(),
            ios_app_id = character(),
            android_app_id = character(),
            icon_url = character()
        ))
    }

    auth_token <- resolve_auth_token(
        auth_token,
        error_message = "Authentication token is required. Set SENSORTOWER_AUTH_TOKEN environment variable."
    )

    # Helper to chunk requests to st_app_details
    chunk_vector <- function(x, size) {
        split(x, ceiling(seq_along(x) / size))
    }

    id_chunks <- chunk_vector(app_ids, 100)

    if (verbose) message("Fetching details for ", length(app_ids), " unified apps in ", length(id_chunks), " batches.")

    details_list <- lapply(id_chunks, function(chunk) {
        tryCatch(
            {
                st_app_details_impl(
                    app_ids = chunk,
                    os = "unified",
                    include_developer_contacts = FALSE,
                    auth_token = auth_token
                )
            },
            error = function(e) {
                if (verbose) message("Batch detail fetch failed: ", e$message)
                tibble::tibble()
            }
        )
    })

    combined_details <- dplyr::bind_rows(details_list)

    if (nrow(combined_details) == 0) {
        return(tibble::tibble(
            unified_app_id = character(),
            app_name = character(),
            ios_app_id = character(),
            android_app_id = character(),
            icon_url = character()
        ))
    }

    # Ensure all desired columns exist
    res <- combined_details
    if (!"app_id" %in% names(res)) res$app_id <- NA_character_
    if (!"app_name" %in% names(res)) res$app_name <- NA_character_
    if (!"ios_app_id" %in% names(res)) res$ios_app_id <- NA_character_
    if (!"android_app_id" %in% names(res)) res$android_app_id <- NA_character_
    if (!"icon_url" %in% names(res)) res$icon_url <- NA_character_

    res <- res %>%
        dplyr::transmute(
            unified_app_id = as.character(.data$app_id),
            app_name = as.character(.data$app_name),
            ios_app_id = as.character(.data$ios_app_id),
            android_app_id = as.character(.data$android_app_id),
            icon_url = as.character(.data$icon_url)
        )

    return(res)
}

#' Intelligent App ID Cache and Resolution System
#'
#' Internal functions to manage a persistent cache of app ID mappings,
#' reducing API calls and improving performance.
#'
#' @keywords internal
#' @name id_cache

# Package environment for storing cache
.sensortowerR_env <- new.env(parent = emptyenv())

#' Initialize or Get ID Cache
#' @noRd
get_id_cache <- function() {
  if (!exists("id_cache", envir = .sensortowerR_env)) {
    .sensortowerR_env$id_cache <- list()
  }
  .sensortowerR_env$id_cache
}

#' Save ID Mapping to Cache
#' @noRd
cache_id_mapping <- function(input_id, ios_id = NULL, android_id = NULL,
                           unified_id = NULL, app_name = NULL, publisher_id = NULL) {
  cache <- get_id_cache()

  # Create cache entry with consistent naming (use _app_id suffix to match st_app_lookup output)
  entry <- list(
    ios_app_id = ios_id,
    android_app_id = android_id,
    unified_app_id = unified_id,
    app_name = app_name,
    publisher_id = publisher_id,
    cached_at = Sys.time()
  )

  # Store under all provided IDs for quick lookup
  if (!is.null(input_id)) cache[[as.character(input_id)]] <- entry
  if (!is.null(ios_id)) cache[[as.character(ios_id)]] <- entry
  if (!is.null(android_id)) cache[[as.character(android_id)]] <- entry
  if (!is.null(unified_id)) cache[[as.character(unified_id)]] <- entry

  .sensortowerR_env$id_cache <- cache
  invisible(entry)
}

#' Look Up ID in Cache
#' @noRd
lookup_cached_id <- function(id) {
  if (is.null(id)) return(NULL)
  cache <- get_id_cache()
  cache[[as.character(id)]]
}

#' Get Default Cache Directory
#'
#' Returns the CRAN-compliant cache directory using tools::R_user_dir()
#' @noRd
get_cache_dir <- function() {
  new_cache_dir <- tools::R_user_dir("sensortowerR", "cache")
  old_cache_dir <- tools::R_user_dir("SensorTowerR", "cache")

  if (!dir.exists(new_cache_dir) && dir.exists(old_cache_dir)) {
    return(old_cache_dir)
  }

  new_cache_dir
}

#' Save Cache to Disk
#'
#' Saves the ID cache to disk. Only creates the cache directory when
#' this function is explicitly called.
#' @noRd
save_id_cache <- function(path = NULL) {
  if (is.null(path)) {
    cache_dir <- get_cache_dir()
    # Only create directory when user explicitly saves cache
    if (!dir.exists(cache_dir)) {
      dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
    }
    path <- file.path(cache_dir, "id_cache.rds")
  }

  cache <- get_id_cache()
  if (length(cache) > 0) {
    saveRDS(cache, path)
    if (getOption("sensortowerR.verbose", getOption("SensorTowerR.verbose", FALSE))) {
      message("Saved ", length(cache), " ID mappings to cache")
    }
  }
}

#' Load Cache from Disk
#'
#' Loads the ID cache from disk if it exists. Does NOT create any directories.
#' @noRd
load_id_cache <- function(path = NULL) {
  if (is.null(path)) {
    path <- file.path(get_cache_dir(), "id_cache.rds")
  }

  if (file.exists(path)) {
    cache <- readRDS(path)
    .sensortowerR_env$id_cache <- cache
    if (getOption("sensortowerR.verbose", getOption("SensorTowerR.verbose", FALSE))) {
      message("Loaded ", length(cache), " ID mappings from cache")
    }
  }
}

#' Smart ID Resolution with Caching
#' 
#' Resolves app IDs using cache first, then API if needed
#' @noRd
resolve_app_id <- function(id, auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"), 
                          use_cache = TRUE, verbose = FALSE) {
  
  # Check cache first
  if (use_cache) {
    cached <- lookup_cached_id(id)
    if (!is.null(cached)) {
      # Check if cache is fresh (within 30 days)
      age_days <- as.numeric(difftime(Sys.time(), cached$cached_at, units = "days"))
      if (age_days < 30) {
        if (verbose) message("  Found in cache (", round(age_days, 1), " days old)")
        return(cached)
      }
    }
  }
  
  # Not in cache or stale - make API call
  if (verbose) message("  Looking up ID via API...")
  
  result <- tryCatch({
    st_app_lookup_impl(id, auth_token = auth_token, verbose = FALSE)
  }, error = function(e) NULL)
  
  # Cache the result
  if (!is.null(result) && use_cache) {
    cache_id_mapping(
      input_id = id,
      ios_id = result$ios_app_id,
      android_id = result$android_app_id,
      unified_id = result$unified_app_id,
      app_name = result$app_name,
      publisher_id = result$publisher_id
    )
  }
  
  result
}

#' Batch Resolve Multiple IDs
#' 
#' Efficiently resolves multiple app IDs with minimal API calls
#' @noRd
batch_resolve_ids <- function(ids, auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                            use_cache = TRUE, verbose = FALSE) {
  
  ids <- as.character(unique(ids))
  ids <- ids[!is.na(ids) & nzchar(ids)]
  if (length(ids) == 0) {
    return(list())
  }
  
  if (verbose) message("Resolving ", length(ids), " app IDs...")
  
  # First pass: check cache
  results <- list()
  needs_api <- character()
  
  for (id in ids) {
    if (use_cache) {
      cached <- lookup_cached_id(id)
      if (!is.null(cached)) {
        age_days <- as.numeric(difftime(Sys.time(), cached$cached_at, units = "days"))
        if (age_days < 30) {
          results[[id]] <- cached
          next
        }
      }
    }
    needs_api <- c(needs_api, id)
  }
  
  if (verbose && length(results) > 0) {
    message("  Found ", length(results), " IDs in cache")
  }
  
  # Second pass: batch API lookup for remaining IDs
  if (length(needs_api) > 0) {
    if (verbose) message("  Looking up ", length(needs_api), " IDs via API...")
    
    # Try to use comparison attributes endpoint for batch lookup
    if (length(needs_api) > 1) {
      batch_result <- tryCatch({
        st_get_unified_mapping(needs_api, auth_token = auth_token)
      }, error = function(e) NULL)
      
      if (!is.null(batch_result)) {
        # Process batch results
        for (i in seq_len(nrow(batch_result))) {
          id <- batch_result$input_id[i]
          results[[id]] <- list(
            ios_app_id = batch_result$ios_app_id[i],
            android_app_id = batch_result$android_app_id[i],
            unified_app_id = batch_result$unified_app_id[i],
            app_name = batch_result$app_name[i],
            publisher_id = batch_result$publisher_id[i]
          )
          
          # Cache the result
          if (use_cache) {
            cache_id_mapping(
              input_id = id,
              ios_id = batch_result$ios_app_id[i],
              android_id = batch_result$android_app_id[i],
              unified_id = batch_result$unified_app_id[i],
              app_name = batch_result$app_name[i],
              publisher_id = batch_result$publisher_id[i]
            )
          }
        }
        needs_api <- character() # All resolved
      }
    }
    
    # Fall back to individual lookups if batch failed
    for (id in needs_api) {
      result <- resolve_app_id(id, auth_token, use_cache, verbose = FALSE)
      if (!is.null(result)) {
        results[[id]] <- result
      }
    }
  }
  
  # NOTE: Auto-save removed to comply with CRAN policy.
  # Users must explicitly call save_id_cache() to persist the cache.
  
  results
}

#' Smart ID Type Detection
#' 
#' Intelligently detects the type of ID and optimal resolution strategy
#' @noRd
detect_id_type <- function(id) {
  if (is.null(id) || is.na(id)) return("unknown")
  
  id_str <- as.character(id)
  
  # iOS numeric ID (e.g., "553834731")
  if (grepl("^\\d{9,10}$", id_str)) {
    return("ios")
  }
  
  # Android package name (e.g., "com.king.candycrushsaga")
  if (grepl("^(com|net|org|io|app|game)\\.", id_str)) {
    return("android")
  }
  
  # Sensor Tower hex ID (24 chars, e.g., "5ba4585f539ce75b97db6bcb")
  if (grepl("^[a-f0-9]{24}$", id_str)) {
    return("unified_hex")
  }
  
  # Publisher ID patterns
  if (grepl("^pub", id_str) || grepl("^[a-f0-9]{24}pub", id_str)) {
    return("publisher")
  }
  
  # Combined iOS_Android format
  if (grepl("^\\d+_com\\.", id_str)) {
    return("combined")
  }
  
  return("unknown")
}

#' Preload Common App IDs
#' 
#' Preloads cache with commonly used app IDs to minimize API calls
#' @noRd
preload_common_apps <- function(auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN")) {
  
  # Common game IDs that are frequently used
  common_apps <- c(
    "553834731",           # Candy Crush iOS
    "1195621598",          # Homescapes iOS  
    "1094591345",          # Pokemon GO iOS
    "529479190",           # Clash of Clans iOS
    "1053012308",          # MONOPOLY GO! iOS
    "com.king.candycrushsaga",
    "com.playrix.homescapes",
    "com.nianticlabs.pokemongo",
    "com.supercell.clashofclans",
    "com.scopely.monopolygo"
  )
  
  # Check which ones aren't cached
  needs_lookup <- character()
  for (id in common_apps) {
    if (is.null(lookup_cached_id(id))) {
      needs_lookup <- c(needs_lookup, id)
    }
  }
  
  if (length(needs_lookup) > 0) {
    message("Preloading ", length(needs_lookup), " common app IDs...")
    batch_resolve_ids(needs_lookup, auth_token, use_cache = TRUE, verbose = FALSE)
    save_id_cache()
  }
}

# NOTE: Removed .onAttach and .onDetach hooks that automatically created
# the package cache directory to comply with CRAN policy.
# Cache is now only loaded/saved when explicitly requested by user functions.
# Users can call save_id_cache() explicitly to persist the cache.

# Test internal caching functions
library(testthat)
library(sensortowerR)

test_that("ID type detection works correctly", {
  detect_id_type <- sensortowerR:::detect_id_type
  
  expect_equal(detect_id_type("553834731"), "ios")
  expect_equal(detect_id_type("1234567890"), "ios")
  expect_equal(detect_id_type("com.king.candycrushsaga"), "android")
  expect_equal(detect_id_type("net.example.app"), "android")
  expect_equal(detect_id_type("5ba4585f539ce75b97db6bcb"), "unified_hex")
  expect_equal(detect_id_type("pub123456"), "publisher")
  expect_equal(detect_id_type("123456_com.example"), "combined")
  expect_equal(detect_id_type("random_string"), "unknown")
  expect_equal(detect_id_type(NULL), "unknown")
  expect_equal(detect_id_type(NA), "unknown")
})

test_that("cache operations work correctly", {
  # Get internal functions
  get_id_cache <- sensortowerR:::get_id_cache
  cache_id_mapping <- sensortowerR:::cache_id_mapping
  lookup_cached_id <- sensortowerR:::lookup_cached_id
  save_id_cache <- sensortowerR:::save_id_cache
  load_id_cache <- sensortowerR:::load_id_cache

  # Use tempdir for CRAN compliance
  test_cache_path <- file.path(tempdir(), "sensortowerR_test_cache.rds")

  # Clear cache first (memory only for this test)
  st_clear_id_cache(disk = FALSE)

  # Verify cache is empty
  cache <- get_id_cache()
  expect_equal(length(cache), 0)

  # Add an entry to cache
  cache_id_mapping(
    input_id = "test123",
    ios_id = "123456789",
    android_id = "com.test.app",
    unified_id = "abcdef123456789012345678",
    app_name = "Test App"
  )

  # Verify entry was added
  cache <- get_id_cache()
  expect_gt(length(cache), 0)

  # Lookup by different IDs (cache uses _app_id suffix)
  result1 <- lookup_cached_id("test123")
  expect_equal(result1$ios_app_id, "123456789")
  expect_equal(result1$app_name, "Test App")

  result2 <- lookup_cached_id("123456789")
  expect_equal(result2$android_app_id, "com.test.app")

  result3 <- lookup_cached_id("com.test.app")
  expect_equal(result3$unified_app_id, "abcdef123456789012345678")

  # Test cache persistence using tempdir
  save_id_cache(path = test_cache_path)

  # Clear memory cache
  st_clear_id_cache(disk = FALSE)

  # Load from tempdir
  load_id_cache(path = test_cache_path)

  # Verify data persisted
  result4 <- lookup_cached_id("test123")
  expect_equal(result4$app_name, "Test App")

  # Clean up - remove temp file and clear memory
  if (file.exists(test_cache_path)) {
    file.remove(test_cache_path)
  }
  st_clear_id_cache(disk = FALSE)
})

test_that("st_cache_info displays correct information", {
  # Get internal function
  cache_id_mapping <- sensortowerR:::cache_id_mapping

  # Clear cache first (memory only for CRAN compliance)
  st_clear_id_cache(disk = FALSE)

  # Add some test entries
  cache_id_mapping("app1", ios_id = "111", app_name = "App 1")
  cache_id_mapping("app2", android_id = "com.app2", app_name = "App 2")
  cache_id_mapping("app3", ios_id = "333", android_id = "com.app3", app_name = "App 3")

  # Capture output
  cache_output <- capture.output(st_cache_info())

  # Verify the cache contains expected entries (don't check exact counts due to test isolation)
  expect_true(any(grepl("Total entries:", cache_output)))
  expect_true(any(grepl("Apps with iOS ID:", cache_output)))
  expect_true(any(grepl("Apps with Android ID:", cache_output)))
  expect_true(any(grepl("Apps with both platforms:", cache_output)))

  # Clean up (memory only)
  st_clear_id_cache(disk = FALSE)
})

test_that("st_smart_metrics wrapper exports exist", {
  expect_true(exists("st_smart_metrics"))
  expect_true(exists("st_clear_id_cache"))
  expect_true(exists("st_cache_info"))
})

#' Get Unified ID Mapping for Apps
#'
#' Retrieves the mapping between platform-specific app IDs and unified app IDs.
#' This function handles cases where platform IDs from st_top_charts may not be
#' directly searchable, using app names as a fallback resolution method.
#'
#' @param app_ids Character vector of app IDs (can be iOS, Android, or unified hex IDs)
#' @param app_names Character vector of app names (optional, helps with resolution)
#' @param os Character string. Operating system: "ios", "android", or "unified"
#' @param auth_token Character string. Sensor Tower API authentication token.
#'   Defaults to environment variable SENSORTOWER_AUTH_TOKEN.
#'
#' @return A data frame with columns:
#'   - `input_id`: The original ID provided
#'   - `unified_app_id`: The unified app ID (hex format)
#'   - `unified_app_name`: The unified app name
#'   - `ios_app_id`: iOS app ID (if available)
#'   - `android_app_id`: Android app ID (if available)
#'   - `publisher_id`: Publisher ID
#'   - `publisher_name`: Publisher name
#'
#' @details
#' This function uses an ID-first approach (no name-based resolution):
#' 1. For hex IDs (24-char), uses st_app_lookup to get platform IDs
#' 2. For platform IDs, first tries to look them up via st_app_lookup
#' 3. If direct lookup fails, searches the unified index using the platform ID
#'    as the term and matches exact IDs within nested ios_apps/android_apps
#' 4. Returns the best available mapping for each app using IDs only
#'
#' Note: Platform IDs from st_top_charts may be regional or legacy IDs that
#' aren't directly searchable. In these cases, name-based search provides
#' the most reliable resolution to unified IDs.
#'
#' @examples
#' \dontrun{
#' # Get mapping with app names for better resolution
#' mapping <- st_get_unified_mapping(
#'   app_ids = c("943599237", "com.bandainamcogames.dbzdokkan"),
#'   app_names = c("Dragon Ball Z Dokkan Battle", "Dragon Ball Z Dokkan Battle"),
#'   os = "unified"
#' )
#' }
#'
#' @importFrom httr2 request req_url_path_append req_url_query req_headers req_timeout req_perform resp_body_string
#' @importFrom jsonlite fromJSON
#' @importFrom dplyr %>%
#' @export
st_get_unified_mapping <- function(app_ids, 
                                   app_names = NULL,
                                   os = "unified",
                                   auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN")) {
  
  # Validate inputs
  if (missing(app_ids) || length(app_ids) == 0) {
    rlang::abort("At least one app_id is required")
  }
  
  # Get auth token
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token is required. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )
  
  # Initialize result data frame
  result_df <- tibble::tibble(
    input_id = app_ids,
    unified_app_id = NA_character_,
    unified_app_name = NA_character_,
    ios_app_id = NA_character_,
    android_app_id = NA_character_,
    publisher_id = NA_character_,
    publisher_name = NA_character_,

  )
  
  # Process each app ID
  for (i in seq_along(app_ids)) {
    app_id <- app_ids[i]
    id_type <- if (grepl("^[a-f0-9]{24}$", app_id)) "unified" else if (grepl("^\\d+$", app_id)) "ios" else if (grepl("^(com|net|org|io)\\.", app_id)) "android" else NA
    
    # Skip if already a hex ID (unified ID)
    if (grepl("^[a-f0-9]{24}$", app_id)) {
      # Use st_app_lookup to get platform IDs for unified ID
      lookup_result <- tryCatch({
        st_app_lookup_impl(app_id, auth_token = auth_token_val, verbose = FALSE)
      }, error = function(e) NULL)
      
      if (!is.null(lookup_result)) {
        result_df$unified_app_id[i] <- app_id
        result_df$unified_app_name[i] <- lookup_result$app_name
        result_df$ios_app_id[i] <- lookup_result$ios_app_id
        result_df$android_app_id[i] <- lookup_result$android_app_id
        result_df$publisher_name[i] <- lookup_result$publisher_name
      }
      next
    }
    # First attempt: use st_app_lookup on platform IDs as well
    if (!grepl("^[a-f0-9]{24}$", app_id)) {
      lk <- tryCatch({ st_app_lookup_impl(app_id, auth_token = auth_token_val, verbose = FALSE) }, error = function(e) NULL)
      if (!is.null(lk) && !is.null(lk$unified_app_id)) {
        result_df$unified_app_id[i] <- lk$unified_app_id
        result_df$unified_app_name[i] <- lk$app_name %||% NA_character_
        result_df$ios_app_id[i] <- lk$ios_app_id %||% NA_character_
        result_df$android_app_id[i] <- lk$android_app_id %||% NA_character_
        result_df$publisher_name[i] <- lk$publisher_name %||% NA_character_
        next
      }
    }

    # Second attempt: search unified index by the platform ID as term, then match nested platforms exactly
    unified_search <- tryCatch({
      st_app_info_impl(
        term = app_id,
        app_store = "unified",
        return_all_fields = TRUE,
        limit = 50,
        auth_token = auth_token_val
      )
    }, error = function(e) NULL)
    if (!is.null(unified_search) && nrow(unified_search) > 0) {
      match_idx <- NA_integer_
      if (!is.na(id_type) && id_type == "ios" && "ios_apps" %in% names(unified_search)) {
        for (ri in seq_len(nrow(unified_search))) {
          ios_tbl <- unified_search$ios_apps[[ri]]
          if (is.data.frame(ios_tbl) && nrow(ios_tbl) > 0 && app_id %in% ios_tbl$app_id) { match_idx <- ri; break }
        }
      } else if (!is.na(id_type) && id_type == "android" && "android_apps" %in% names(unified_search)) {
        for (ri in seq_len(nrow(unified_search))) {
          and_tbl <- unified_search$android_apps[[ri]]
          if (is.data.frame(and_tbl) && nrow(and_tbl) > 0 && app_id %in% and_tbl$app_id) { match_idx <- ri; break }
        }
      }
      if (!is.na(match_idx)) {
        result_df$unified_app_id[i] <- unified_search$app_id[match_idx]
        result_df$unified_app_name[i] <- unified_search$name[match_idx]
        result_df$publisher_name[i] <- unified_search$publisher_name[match_idx] %||% NA_character_
        # Populate platform IDs from the matched row
        if ("ios_apps" %in% names(unified_search) && !is.null(unified_search$ios_apps[[match_idx]]) && nrow(unified_search$ios_apps[[match_idx]]) > 0) {
          result_df$ios_app_id[i] <- unified_search$ios_apps[[match_idx]]$app_id[1]
        }
        if ("android_apps" %in% names(unified_search) && !is.null(unified_search$android_apps[[match_idx]]) && nrow(unified_search$android_apps[[match_idx]]) > 0) {
          result_df$android_app_id[i] <- unified_search$android_apps[[match_idx]]$app_id[1]
        }
        next
      }
    }

    # If we still don't have a unified ID, use the input as a fallback for platform fields
    if (is.na(result_df$unified_app_id[i])) {
      # Just populate the appropriate platform field
      if (grepl("^\\d+$", app_id)) {
        # iOS ID
        result_df$ios_app_id[i] <- app_id
      } else if (grepl("^(com|net|org|io)\\.", app_id)) {
        # Android ID
        result_df$android_app_id[i] <- app_id
      }
      
      # Use the provided name if available
      if (!is.null(app_names) && length(app_names) >= i && !is.na(app_names[i])) {
        result_df$unified_app_name[i] <- app_names[i]
      }
    }
  }
  
  return(result_df)
}

# Helper operator
`%||%` <- function(x, y) if (is.null(x)) y else x

#' Custom Fields Filter Workflow Helper Functions
#'
#' Helper functions to streamline working with custom fields filters in Sensor Tower.
#' These functions combine the custom fields endpoints with data retrieval functions
#' to provide a complete workflow.
#'
#' @name st_custom_fields_workflow
#' @rdname st_custom_fields_workflow
NULL

#' Create or Get Filter ID for Custom Criteria
#'
#' Creates a custom fields filter or retrieves existing filter ID if the same
#' criteria already exists. This is a convenience wrapper around st_custom_fields_filter.
#'
#' @param field_name Character. Name of the custom field to filter by
#' @param field_values Character vector. Values to filter for
#' @param global Logical. Whether this is a global field (TRUE) or organization field (FALSE)
#' @param exclude Logical. Whether to exclude apps matching criteria (FALSE = include)
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return Character string containing the filter ID
#'
#' @examples
#' \dontrun{
#' # Get filter ID for Word games
#' filter_id <- st_create_simple_filter(
#'   field_name = "Game Sub-genre",
#'   field_values = "Word"
#' )
#' 
#' # Get filter ID for multiple genres
#' filter_id <- st_create_simple_filter(
#'   field_name = "Game Genre",
#'   field_values = c("Puzzle", "Word")
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_create_simple_filter_impl <- function(
    field_name,
    field_values,
    global = TRUE,
    exclude = FALSE,
    auth_token = NULL
) {
  
  # Input validation
  if (missing(field_name) || is.null(field_name) || field_name == "") {
    rlang::abort("field_name is required and cannot be empty")
  }
  
  # Convert single value to list for consistency
  if (!is.list(field_values)) {
    if (is.null(field_values)) {
      field_values <- list()
    } else {
      field_values <- as.list(field_values)
    }
  }
  
  # If possible, validate values for the given field and drop invalid entries
  valid_values <- NULL
  try({
    fields_df <- st_custom_fields_values(auth_token = auth_token)
    if (nrow(fields_df) > 0 && "name" %in% names(fields_df) && "values" %in% names(fields_df)) {
      row <- fields_df[fields_df$name == field_name, , drop = FALSE]
      if (nrow(row) == 1 && length(row$values[[1]]) > 0) {
        valid_values <- as.character(unlist(row$values[[1]]))
      }
    }
  }, silent = TRUE)
  if (!is.null(valid_values)) {
    # Keep only values present for this field
    field_values <- Filter(function(v) v %in% valid_values, field_values)
  }
  
  # Create filter criteria
  custom_fields <- list(
    list(
      exclude = exclude,
      global = global,
      name = field_name,
      values = field_values
    )
  )
  
  # Create or get filter ID
  filter_id <- tryCatch({
    st_custom_fields_filter_impl(
      custom_fields = custom_fields,
      auth_token = auth_token
    )
  }, error = function(e) {
    # If server error occurs, return a deterministic placeholder ObjectId so callers can proceed
    if (inherits(e, c("httr2_http", "httr2_http_500"))) {
      # Create 24-char hex id from hash of field and values
      id_src <- paste0(field_name, "|", paste(unlist(field_values), collapse=","))
      hex <- substr(openssl::sha1(id_src), 1, 24)
      message("Warning: filter creation failed with server error; returning placeholder id ", hex)
      return(hex)
    }
    rlang::abort(e)
  })
  
  return(filter_id)
}

#' Discover Available Custom Fields
#'
#' Searches and displays available custom fields that can be used for filtering.
#' This is helpful for discovering what fields are available before creating filters.
#'
#' @param search_term Optional. Character string to search for in field names
#' @param show_values Logical. Whether to show possible values for each field
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return A tibble with custom fields information
#'
#' @examples
#' \dontrun{
#' # Find all game-related fields
#' game_fields <- st_discover_fields("game")
#' 
#' # Find all date fields
#' date_fields <- st_discover_fields("date")
#' 
#' # Show all fields with their values
#' all_fields <- st_discover_fields(show_values = TRUE)
#' }
#'
#' @export
#' @importFrom dplyr filter mutate select
#' @importFrom stringr str_detect str_to_lower
st_discover_fields <- function(
    search_term = NULL,
    show_values = FALSE,
    auth_token = NULL
) {
  
  # Get all custom fields
  fields <- st_custom_fields_values(term = search_term, auth_token = auth_token)
  
  if (nrow(fields) == 0) {
    message("No custom fields found")
    return(fields)
  }
  
  # Filter by search term if provided
  if (!is.null(search_term)) {
    fields <- fields %>%
      dplyr::filter(stringr::str_detect(
        stringr::str_to_lower(name), 
        stringr::str_to_lower(search_term)
      ))
  }
  
  # Format output
  if (!show_values && "values" %in% names(fields)) {
    # Show count of possible values instead of the values themselves
    fields <- fields %>%
      dplyr::mutate(
        value_count = sapply(values, function(x) {
          if (is.null(x)) return(0)
          if (is.list(x)) return(length(x))
          return(1)
        })
      ) %>%
      dplyr::select(-values)
  }
  
  # Sort by name for easier reading
  fields <- fields %>%
    dplyr::arrange(name)
  
  message(sprintf("Found %d custom fields%s", 
                  nrow(fields),
                  ifelse(!is.null(search_term), 
                         sprintf(" matching '%s'", search_term), 
                         "")))
  
  return(fields)
}

#' Get Top Apps with Custom Filter
#'
#' Retrieves top apps using a custom fields filter. This combines filter creation
#' with data retrieval in a single workflow.
#'
#' @param field_name Character. Name of the custom field to filter by (or NULL to use filter_id)
#' @param field_values Character vector. Values to filter for (or NULL to use filter_id)
#' @param filter_id Character. Existing filter ID to use (alternative to field_name/values)
#' @param measure Character. Metric to measure: "DAU", "WAU", "MAU", "revenue", or "units"
#' @param regions Character vector. Region codes (e.g., "US", "WW")
#' @param date Character or Date. Start date for the query
#' @param end_date Optional. Character or Date. End date for the query
#' @param limit Integer. Maximum number of apps to return (default 100)
#' @param enrich_response Logical. Whether to enrich with additional metrics
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param ... Additional parameters passed to st_top_charts
#'
#' @return A tibble with top apps data
#'
#' @examples
#' \dontrun{
#' # Get top Word games by DAU
#' word_games <- st_get_filtered_apps(
#'   field_name = "Game Sub-genre",
#'   field_values = "Word",
#'   measure = "DAU",
#'   regions = "US",
#'   limit = 20
#' )
#' 
#' # Use existing filter ID
#' apps <- st_get_filtered_apps(
#'   filter_id = "603697f4241bc16eb8570d37",
#'   measure = "revenue",
#'   regions = "US"
#' )
#' }
#'
#' @export
st_get_filtered_apps <- function(
    field_name = NULL,
    field_values = NULL,
    filter_id = NULL,
    measure = "DAU",
    regions = "US",
    date = NULL,
    end_date = NULL,
    limit = 100,
    enrich_response = TRUE,
    auth_token = NULL,
    ...
) {
  
  # Determine filter ID
  if (is.null(filter_id)) {
    if (is.null(field_name) || is.null(field_values)) {
      rlang::abort("Either provide filter_id or both field_name and field_values")
    }
    
    # Create filter
    filter_id <- st_create_simple_filter_impl(
      field_name = field_name,
      field_values = field_values,
      auth_token = auth_token
    )
    
    message(sprintf("Using filter ID: %s", filter_id))
  }
  
  # Get filter details for logging
  filter_details <- tryCatch({
    st_custom_fields_filter_by_id_impl(id = filter_id, auth_token = auth_token)
  }, error = function(e) NULL)
  
  if (!is.null(filter_details) && !is.null(filter_details$custom_fields)) {
    fields_desc <- paste(
      sprintf("%s=%s", 
              filter_details$custom_fields$name,
              sapply(filter_details$custom_fields$values, 
                     function(x) paste(x, collapse = ","))),
      collapse = "; "
    )
    message(sprintf("Filter criteria: %s", fields_desc))
  }
  
  # Retrieve data using st_top_charts
  result <- st_top_charts_impl(
    os = "unified",
    category = 0,  # Use 0 when using custom filter
    custom_fields_filter_id = filter_id,
    custom_tags_mode = "include_unified_apps",
    measure = measure,
    regions = regions,
    date = date,
    end_date = end_date,
    limit = limit,
    device_type = "total",
    enrich_response = enrich_response,
    auth_token = auth_token,
    ...
  )
  
  return(result)
}

#' Analyze Custom Filter Performance
#'
#' Provides a summary analysis of apps matching a custom filter, including
#' top performers, growth metrics, and category breakdown.
#'
#' @param filter_id Character. The custom fields filter ID to analyze
#' @param measure Character. Metric to analyze: "DAU", "revenue", or "units"
#' @param regions Character vector. Region codes (default "US")
#' @param top_n Integer. Number of top apps to show (default 10)
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'
#' @return A list containing summary statistics and top apps
#'
#' @export
st_analyze_filter <- function(
    filter_id,
    measure = "DAU",
    regions = "US",
    top_n = 10,
    auth_token = NULL
) {
  
  # Get filter details
  filter_details <- st_custom_fields_filter_by_id_impl(
    id = filter_id, 
    auth_token = auth_token
  )
  
  # Get current period data
  current_data <- st_top_charts_impl(
    os = "unified",
    category = 0,
    custom_fields_filter_id = filter_id,
    custom_tags_mode = "include_unified_apps",
    measure = measure,
    regions = regions,
    limit = 100,
    device_type = "total",
    enrich_response = TRUE,
    auth_token = auth_token
  )
  
  if (nrow(current_data) == 0) {
    message("No apps found for this filter")
    return(NULL)
  }
  
  # Prepare summary
  summary <- list(
    filter_id = filter_id,
    filter_criteria = filter_details$custom_fields,
    total_apps = nrow(current_data),
    measure = measure,
    regions = regions,
    top_apps = head(current_data, top_n)
  )
  
  # Add measure-specific summaries
  if (measure == "DAU" && "aggregate_tags.Last 30 Days Average DAU (US)" %in% names(current_data)) {
    summary$total_dau <- sum(current_data$`aggregate_tags.Last 30 Days Average DAU (US)`, na.rm = TRUE)
    summary$avg_dau_per_app <- mean(current_data$`aggregate_tags.Last 30 Days Average DAU (US)`, na.rm = TRUE)
  }
  
  message(sprintf("\nAnalysis Summary for Filter: %s", filter_id))
  message(sprintf("Total apps found: %d", summary$total_apps))
  message(sprintf("Measure: %s | Region: %s", measure, regions))
  
  if (measure == "DAU" && !is.null(summary$total_dau)) {
    message(sprintf("Total DAU: %s", format(summary$total_dau, big.mark = ",")))
    message(sprintf("Average DAU per app: %s", format(round(summary$avg_dau_per_app), big.mark = ",")))
  }
  
  return(summary)
}

#' Fetch Game Market Summary Data
#'
#' Retrieves aggregate download and revenue estimates by game categories,
#' countries, and date ranges from Sensor Tower's `games_breakdown` endpoint.
#' Use this for market/category denominator series; do not approximate market
#' totals by batching top charts, rankings, or large app rosters.
#'
#' @param categories Character string or numeric vector. Game category IDs to 
#'   analyze. Defaults to 7001 (a popular game category). Use `st_categories()` 
#'   to find valid category IDs.
#' @param countries Character vector or comma-separated string. Country codes 
#'   (e.g., `"US"`, `c("US", "GB")`, `"WW"` for worldwide) to analyze. Required.
#' @param os Character string. Operating System. Must be one of "ios", "android", 
#'   or "unified". Required. Note: The underlying API only supports "ios" and "android";
#'   when `os = "unified"` this function automatically fetches both platforms and
#'   combines them into a single table with total columns.
#' @param date_granularity Character string. Time granularity for aggregation. 
#'   Must be one of "daily", "weekly", "monthly", or "quarterly". Required.
#' @param start_date Character string or Date object. Start date for the query 
#'   in "YYYY-MM-DD" format. Required.
#' @param end_date Character string or Date object. End date for the query 
#'   in "YYYY-MM-DD" format, inclusive. Required.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param base_url Optional. Character string. The base URL for the API.
#' @param enrich_response Optional. Logical. If `TRUE` (default), enriches
#'   the response with readable column names and processes the data.
#'
#' @return A [tibble][tibble::tibble] with aggregate game market summary data including:
#'   - **Category information**: Game category details
#'   - **Geographic data**: Country-level breakdowns
#'   - **Downloads**: iOS (iPhone + iPad combined) and Android download estimates
#'   - **Revenue**: iOS (iPhone + iPad combined) and Android revenue estimates,
#'     expressed in dollars when `enrich_response = TRUE`
#'   - **Totals (unified only)**: `Total Downloads`, `Total Revenue`
#'   - **Time series**: Data broken down by specified granularity
#'   
#'   **Automatic Data Combination**: For iOS and unified platforms, iPhone and iPad
#'   data are automatically combined into single "iOS Downloads" and "iOS Revenue" 
#'   columns for simplified analysis.
#'
#' @section API Endpoint Used:
#'   - **Game Summary**: `GET /v1/\{os\}/games_breakdown`
#'     (API only supports `os = "ios"` or `os = "android"`; unified is synthesized)
#'
#' @section Revenue Units:
#'   The raw `games_breakdown` endpoint returns revenue in cents. With
#'   `enrich_response = TRUE`, this function converts revenue fields to dollars
#'   before combining devices or platforms. With `enrich_response = FALSE`, raw
#'   endpoint fields are returned unchanged.
#'
#' @section Field Mappings and Processing:
#'   The API returns abbreviated field names which are automatically mapped to 
#'   descriptive names and processed:
#'   - **iOS**: `iu` + `au` = iOS Downloads (combined), `ir` + `ar` = iOS Revenue (combined)
#'   - **Android**: `u` = Android Downloads, `r` = Android Revenue
#'   - **Common**: `ca` = Category, `cc` = Country Code, `d` = Date
#'   
#'   iPhone and iPad data are automatically combined for simplified analysis.
#'   For `os = "unified"`, iOS and Android aggregate rows are summed by date and
#'   country across the requested category basket.
#'
#' @examples
#' \dontrun{
#' # Specific categories and countries
#' rpg_summary <- st_game_summary(
#'   categories = c(7001, 7002),
#'   countries = c("US", "GB", "DE"),
#'   os = "ios",
#'   date_granularity = "weekly",
#'   start_date = "2024-01-01",
#'   end_date = "2024-03-31"
#' )
#'
#' # Monthly summary for iOS games in the US
#' ios_monthly <- st_game_summary(
#'   os = "ios",
#'   countries = "US", 
#'   date_granularity = "monthly",
#'   start_date = "2024-01-01",
#'   end_date = "2024-06-30"
#' )
#' }
#'
#' @seealso [st_categories()], [st_rankings()], [st_metrics()]
#' @export
st_game_summary <- function(categories = 7001,
                            countries,
                            os,
                            date_granularity,
                            start_date,
                            end_date,
                            auth_token = NULL,
                            base_url = "https://api.sensortower.com",
                            enrich_response = TRUE) {
  
  # Validate required parameters
  if (missing(countries) || is.null(countries) || length(countries) == 0) {
    rlang::abort("'countries' parameter is required. Specify country codes (e.g., 'US', 'GB', 'JP', or 'WW' for worldwide).")
  }
  
  if (missing(os) || is.null(os)) {
    rlang::abort("'os' parameter is required. Specify one of: 'ios', 'android', 'unified'.")
  }
  
  if (missing(date_granularity) || is.null(date_granularity)) {
    rlang::abort("'date_granularity' parameter is required. Specify one of: 'daily', 'weekly', 'monthly', 'quarterly'.")
  }
  
  if (missing(start_date) || is.null(start_date)) {
    rlang::abort("'start_date' parameter is required. Specify in YYYY-MM-DD format.")
  }
  
  if (missing(end_date) || is.null(end_date)) {
    rlang::abort("'end_date' parameter is required. Specify in YYYY-MM-DD format.")
  }
  
  # Validate inputs
  if (!date_granularity %in% c("daily", "weekly", "monthly", "quarterly")) {
    rlang::abort("date_granularity must be one of: daily, weekly, monthly, quarterly")
  }
  
  if (!os %in% c("ios", "android", "unified")) {
    rlang::abort("os must be one of: ios, android, unified")
  }
  
  # --- Authentication ---
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set it as an environment variable."
  )
  
  # Convert dates to proper format
  start_date <- as.character(as.Date(start_date))
  end_date <- as.character(as.Date(end_date))
  
  # Convert categories to comma-separated string if vector
  if (is.numeric(categories) || length(categories) > 1) {
    categories <- paste(categories, collapse = ",")
  }
  
  # Convert countries to comma-separated string if vector
  if (length(countries) > 1) {
    countries <- paste(countries, collapse = ",")
  }
  
  # If unified requested, fetch iOS + Android and combine
  if (os == "unified") {
    ios_res <- tryCatch(
      st_game_summary(
        categories = categories,
        countries = countries,
        os = "ios",
        date_granularity = date_granularity,
        start_date = start_date,
        end_date = end_date,
        auth_token = auth_token_val,
        base_url = base_url,
        enrich_response = enrich_response
      ),
      error = function(e) tibble::tibble()
    )
    android_res <- tryCatch(
      st_game_summary(
        categories = categories,
        countries = countries,
        os = "android",
        date_granularity = date_granularity,
        start_date = start_date,
        end_date = end_date,
        auth_token = auth_token_val,
        base_url = base_url,
        enrich_response = enrich_response
      ),
      error = function(e) tibble::tibble()
    )

    if (nrow(ios_res) == 0 && nrow(android_res) == 0) {
      return(tibble::tibble())
    }

    if (!enrich_response) {
      return(dplyr::bind_rows(
        dplyr::mutate(ios_res, os = "ios"),
        dplyr::mutate(android_res, os = "android")
      ))
    }

    unified_keys <- game_summary_unified_keys(ios_res, android_res)
    ios_res <- summarize_game_summary_platform(
      ios_res,
      metric_cols = c("iOS Downloads", "iOS Revenue"),
      keys = unified_keys
    )
    android_res <- summarize_game_summary_platform(
      android_res,
      metric_cols = c("Android Downloads", "Android Revenue"),
      keys = unified_keys
    )

    joined <- dplyr::full_join(ios_res, android_res, by = unified_keys)

    # Add total columns when platform-specific metrics exist
    if ("iOS Downloads" %in% names(joined) || "Android Downloads" %in% names(joined)) {
      joined$`Total Downloads` <- row_sum_metric_columns(
        joined,
        intersect(c("iOS Downloads", "Android Downloads"), names(joined))
      )
    }
    if ("iOS Revenue" %in% names(joined) || "Android Revenue" %in% names(joined)) {
      joined$`Total Revenue` <- row_sum_metric_columns(
        joined,
        intersect(c("iOS Revenue", "Android Revenue"), names(joined))
      )
    }

    return(joined)
  }

  # Build API request (platform-specific)
  path_segments <- st_endpoint_segments("games_breakdown", os = os)
  
  query_params <- list(
    auth_token = auth_token_val,
    categories = categories,
    date_granularity = date_granularity,
    start_date = start_date,
    end_date = end_date
  )
  
  # Add countries parameter (API uses "WW" for worldwide by default)
  if (!is.null(countries) && countries != "WW") {
    query_params$countries <- countries
  }
  
  # Build and perform request
  req <- build_request(base_url, path_segments, query_params)
  resp <- perform_request(req)
  
  # Process response
  if (enrich_response) {
    result <- process_game_summary_response(resp, os)
  } else {
    result <- process_response(resp, FALSE)
  }
  
  return(result)
}

#' Process Game Summary API Response
#'
#' Internal function to process and enrich game summary API responses.
#'
#' @param resp List. Raw API response from game summary endpoint.
#' @param os Character string. Operating system to determine field mappings.
#'
#' @return A processed tibble with descriptive column names.
#' @keywords internal
process_game_summary_response <- function(resp, os) {
  
  # First get the raw response like process_response does
  body_raw <- httr2::resp_body_raw(resp)
  if (length(body_raw) == 0) {
    return(tibble::tibble())
  }

  body_text <- rawToChar(body_raw)
  result <- jsonlite::fromJSON(body_text, flatten = TRUE)

  if (length(result) == 0) {
    return(tibble::tibble())
  }

  # Convert to tibble
  result_tbl <- tibble::as_tibble(result)
  
  if (nrow(result_tbl) == 0) {
    return(result_tbl)
  }
  
  # Apply field name mappings using the games_breakdown_key data
  result_tbl <- map_game_summary_fields(result_tbl, os)
  
  # Convert date fields to proper Date format
  if ("Date" %in% names(result_tbl)) {
    result_tbl$Date <- as.Date(substr(result_tbl$Date, 1, 10))
  }
  
  download_cols <- intersect(
    c("iPhone Downloads", "iPad Downloads", "Android Downloads"),
    names(result_tbl)
  )
  revenue_cols <- intersect(
    c("iPhone Revenue", "iPad Revenue", "Android Revenue"),
    names(result_tbl)
  )

  for (col in download_cols) {
    result_tbl[[col]] <- as.numeric(result_tbl[[col]])
  }

  for (col in revenue_cols) {
    result_tbl[[col]] <- as.numeric(result_tbl[[col]]) / 100
  }
  
  # Automatically combine iPad and iPhone data into unified iOS totals
  if (os %in% c("ios", "unified")) {
    # Check if we have both iPhone and iPad columns
    has_iphone_downloads <- "iPhone Downloads" %in% names(result_tbl)
    has_ipad_downloads <- "iPad Downloads" %in% names(result_tbl)
    has_iphone_revenue <- "iPhone Revenue" %in% names(result_tbl)
    has_ipad_revenue <- "iPad Revenue" %in% names(result_tbl)
    
    if (has_iphone_downloads && has_ipad_downloads) {
      result_tbl$`iOS Downloads` <- row_sum_metric_columns(
        result_tbl,
        c("iPhone Downloads", "iPad Downloads")
      )
      # Remove individual device columns
      result_tbl$`iPhone Downloads` <- NULL
      result_tbl$`iPad Downloads` <- NULL
    }
    
    if (has_iphone_revenue && has_ipad_revenue) {
      result_tbl$`iOS Revenue` <- row_sum_metric_columns(
        result_tbl,
        c("iPhone Revenue", "iPad Revenue")
      )
      # Remove individual device columns
      result_tbl$`iPhone Revenue` <- NULL
      result_tbl$`iPad Revenue` <- NULL
    }
  }
  
  return(result_tbl)
}

#' Map Game Summary Field Names
#'
#' Internal function to map abbreviated API field names to descriptive names.
#'
#' @param data Tibble. Data with abbreviated field names.
#' @param os Character string. Operating system to determine mappings.
#'
#' @return Tibble with descriptive field names.
#' @keywords internal
map_game_summary_fields <- function(data, os) {
  
  # Get the appropriate field mappings from internal data
  if (os == "unified") {
    # For unified, we need to handle both iOS and Android fields
    ios_mappings <- games_breakdown_key$ios
    android_mappings <- games_breakdown_key$android
    
    # Combine mappings, with iOS taking precedence for common fields
    field_mappings <- c(ios_mappings, android_mappings[!names(android_mappings) %in% names(ios_mappings)])
  } else {
    field_mappings <- games_breakdown_key[[os]]
  }

  field_mappings <- c(
    list(ca = "Category"),
    field_mappings
  )
  
  if (is.null(field_mappings)) {
    warning("Unknown OS: ", os, ". Field names will not be mapped.")
    return(data)
  }
  
  # Apply mappings
  current_names <- names(data)
  new_names <- current_names
  
  for (i in seq_along(current_names)) {
    old_name <- current_names[i]
    if (old_name %in% names(field_mappings)) {
      new_names[i] <- field_mappings[[old_name]]
    }
  }
  
  names(data) <- new_names
  return(data)
} 

game_summary_unified_keys <- function(ios_data, android_data) {
  candidate_keys <- c("Country Code", "Date")
  intersect(candidate_keys, intersect(names(ios_data), names(android_data)))
}

summarize_game_summary_platform <- function(data, metric_cols, keys) {
  if (nrow(data) == 0) {
    return(data)
  }

  metric_cols <- intersect(metric_cols, names(data))
  keep_cols <- c(keys, metric_cols)
  data <- data[, keep_cols, drop = FALSE]

  if (length(metric_cols) == 0) {
    return(data)
  }

  if (length(keys) == 0) {
    return(tibble::as_tibble(stats::setNames(
      lapply(metric_cols, function(col) sum_metric_column(data[[col]])),
      metric_cols
    )))
  }

  data %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(keys))) %>%
    dplyr::summarise(
      dplyr::across(
        dplyr::all_of(metric_cols),
        sum_metric_column
      ),
      .groups = "drop"
    )
}

sum_metric_column <- function(x) {
  x <- as.numeric(x)
  if (all(is.na(x))) {
    return(NA_real_)
  }
  sum(x, na.rm = TRUE)
}

row_sum_metric_columns <- function(data, cols) {
  cols <- intersect(cols, names(data))
  if (length(cols) == 0) {
    return(rep(NA_real_, nrow(data)))
  }

  values <- as.data.frame(lapply(data[, cols, drop = FALSE], as.numeric))
  out <- rowSums(values, na.rm = TRUE)
  out[rowSums(!is.na(values)) == 0] <- NA_real_
  out
}

game_summary_response <- function(payload) {
  httr2::response(
    status_code = 200,
    body = charToRaw(jsonlite::toJSON(payload, auto_unbox = TRUE))
  )
}

test_that("iOS game summary converts cents to dollars before combining revenue", {
  payload <- list(
    list(
      ca = 7001,
      cc = "US",
      d = "2024-01-01T00:00:00Z",
      iu = 10,
      au = 5,
      ir = 2147483647,
      ar = 2147483647
    )
  )

  result <- sensortowerR:::process_game_summary_response(
    game_summary_response(payload),
    os = "ios"
  )

  expect_equal(result$`iOS Downloads`, 15)
  expect_type(result$`iOS Revenue`, "double")
  expect_equal(result$`iOS Revenue`, 42949672.94)
  expect_false(any(is.na(result$`iOS Revenue`)))
})

test_that("Android game summary maps downloads and dollar revenue", {
  payload <- list(
    list(
      ca = "game_action",
      cc = "US",
      d = "2024-01-01T00:00:00Z",
      u = 123,
      r = 456789
    )
  )

  result <- sensortowerR:::process_game_summary_response(
    game_summary_response(payload),
    os = "android"
  )

  expect_equal(result$Category, "game_action")
  expect_equal(result$`Country Code`, "US")
  expect_equal(result$Date, as.Date("2024-01-01"))
  expect_equal(result$`Android Downloads`, 123)
  expect_equal(result$`Android Revenue`, 4567.89)
})

test_that("unified game summary aggregates iOS and Android endpoint rows", {
  requested_urls <- character()

  testthat::local_mocked_bindings(
    perform_request = function(req) {
      requested_urls <<- c(requested_urls, req$url)

      if (grepl("/v1/ios/games_breakdown", req$url, fixed = TRUE)) {
        return(game_summary_response(list(
          list(ca = 7001, cc = "US", d = "2024-01-01T00:00:00Z", iu = 10, au = 5, ir = 1000, ar = 500)
        )))
      }

      if (grepl("/v1/android/games_breakdown", req$url, fixed = TRUE)) {
        return(game_summary_response(list(
          list(ca = "game_action", cc = "US", d = "2024-01-01T00:00:00Z", u = 20, r = 2500)
        )))
      }

      stop("Unexpected request URL: ", req$url)
    },
    .package = "sensortowerR"
  )

  result <- st_game_summary(
    categories = c("7001", "game_action"),
    countries = "US",
    os = "unified",
    date_granularity = "daily",
    start_date = "2024-01-01",
    end_date = "2024-01-01",
    auth_token = "token"
  )

  expect_equal(length(requested_urls), 2)
  expect_true(any(grepl("/v1/ios/games_breakdown", requested_urls, fixed = TRUE)))
  expect_true(any(grepl("/v1/android/games_breakdown", requested_urls, fixed = TRUE)))
  expect_equal(nrow(result), 1)
  expect_equal(result$`Total Downloads`, 35)
  expect_equal(result$`Total Revenue`, 40)
})

test_that("game summary uses aggregate breakdown endpoint, not title-batch paths", {
  requested_urls <- character()
  title_batch_error <- function(...) {
    stop("title-batch path should not be used for market totals")
  }

  testthat::local_mocked_bindings(
    perform_request = function(req) {
      requested_urls <<- c(requested_urls, req$url)
      game_summary_response(list(
        list(ca = 7001, cc = "US", d = "2024-01-01T00:00:00Z", iu = 1, au = 1, ir = 100, ar = 100)
      ))
    },
    st_batch_metrics_impl = title_batch_error,
    st_top_charts_impl = title_batch_error,
    .package = "sensortowerR"
  )

  result <- st_game_summary(
    categories = 7001,
    countries = "US",
    os = "ios",
    date_granularity = "daily",
    start_date = "2024-01-01",
    end_date = "2024-01-01",
    auth_token = "token"
  )

  expect_equal(nrow(result), 1)
  expect_equal(result$`iOS Revenue`, 2)
  expect_equal(length(requested_urls), 1)
  expect_true(grepl("/v1/ios/games_breakdown", requested_urls, fixed = TRUE))
  expect_false(grepl("ranking|top_and_trending|timeseries|sales_report", requested_urls))
})

#' Fetch App Details
#'
#' Thin v1.0.0 facade for app metadata lookups. By default it dispatches to the
#' legacy detailed app metadata endpoint. When `fields` is supplied, it dispatches
#' to `st_app_enriched()` and optionally subsets to the requested fields.
#'
#' @param app_id Character scalar or vector of app identifiers.
#' @param os Operating system context. One of `"ios"`, `"android"`, or
#'   `"unified"`.
#' @param fields Optional character vector of enriched fields to keep. When
#'   supplied, the request is routed through `st_app_enriched()`.
#' @param auth_token Optional Sensor Tower API token.
#'
#' @return A tibble returned by the dispatched legacy implementation.
#'
#' @examples
#' \dontrun{
#' st_app("553834731", os = "ios")
#'
#' st_app(
#'   c("553834731", "com.supercell.clashofclans"),
#'   fields = c("revenue_90d_ww", "downloads_30d_ww")
#' )
#' }
#'
#' @export
st_app <- function(app_id,
                   os = "unified",
                   fields = NULL,
                   auth_token = NULL) {
  if (missing(app_id) || is.null(app_id) || !length(app_id)) {
    rlang::abort("`app_id` must be a non-empty character scalar or vector.")
  }

  app_id <- as.character(app_id)
  if (any(is.na(app_id) | !nzchar(app_id))) {
    rlang::abort("`app_id` entries must be non-empty strings.")
  }

  os <- normalize_os(os)
  auth_token <- get_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required.",
      "Set SENSORTOWER_AUTH_TOKEN or pass `auth_token`."
    )
  )

  if (is.null(fields)) {
    resolved_ids <- .st_app_resolve_ids_for_os(
      app_id = app_id,
      os = os,
      auth_token = auth_token
    )

    return(
      st_app_details_impl(
        app_ids = resolved_ids,
        os = os,
        auth_token = auth_token
      )
    )
  }

  fields <- unique(as.character(fields))
  unified_ids <- .st_app_resolve_ids_for_os(
    app_id = app_id,
    os = "unified",
    auth_token = auth_token
  )

  enriched <- st_app_enriched(
    unified_app_ids = unified_ids,
    os = "unified",
    auth_token = auth_token
  )

  keep <- unique(c(
    "app_id", "unified_app_id", "app_name", "unified_app_name", "os", fields
  ))

  dplyr::select(enriched, dplyr::any_of(keep))
}

#' Search or Filter Apps
#'
#' Thin v1.0.0 facade for app discovery by query string or server-side filter.
#'
#' @param query Optional search string for app-name search.
#' @param filter Optional `st_filter` object or 24-character filter ID string.
#' @param os Operating system context for query searches.
#' @param country Two-letter country code used by the filtered-app workflow.
#' @param limit Positive integer row limit.
#' @param auth_token Optional Sensor Tower API token.
#'
#' @return A tibble returned by the dispatched legacy implementation.
#'
#' @examples
#' \dontrun{
#' st_apps(query = "Royal Match", os = "unified", limit = 25)
#'
#' filt <- st_filter(genre = "Puzzle")
#' st_apps(filter = filt, country = "US", limit = 50)
#' }
#'
#' @export
st_apps <- function(query = NULL,
                    filter = NULL,
                    os = "ios",
                    country = "US",
                    limit = 100,
                    auth_token = NULL) {
  if (is.null(query) && is.null(filter)) {
    rlang::abort("Provide either `query` or `filter`.")
  }
  if (!is.null(query) && !is.null(filter)) {
    rlang::abort("Provide only one of `query` or `filter`, not both.")
  }

  os <- normalize_os(os)
  country <- normalize_country(country)
  limit <- suppressWarnings(as.integer(limit[1]))
  if (is.na(limit) || limit < 1L) {
    rlang::abort("`limit` must be a positive integer.")
  }

  auth_token <- get_auth_token(
    auth_token,
    error_message = paste(
      "Authentication token is required.",
      "Set SENSORTOWER_AUTH_TOKEN or pass `auth_token`."
    )
  )

  if (!is.null(query)) {
    query <- as.character(query[1])
    if (!nzchar(query)) {
      rlang::abort("`query` must be a non-empty string when supplied.")
    }

    return(
      st_app_info_impl(
        term = query,
        app_store = os,
        entity_type = "app",
        limit = limit,
        auth_token = auth_token
      )
    )
  }

  filter_id <- .st_filter_resolve_id(filter)
  if (is.null(filter_id)) {
    rlang::abort("`filter` must be an `st_filter` object or a 24-character filter ID.")
  }

  if (os != "unified") {
    rlang::warn(
      "`st_apps(filter = ...)` uses the unified filtered-app workflow and ignores `os`."
    )
  }

  st_get_filtered_apps(
    filter_id = filter_id,
    regions = country,
    limit = limit,
    auth_token = auth_token
  )
}

.st_app_resolve_ids_for_os <- function(app_id, os, auth_token) {
  vapply(
    app_id,
    function(one_id) .st_app_resolve_single_id(one_id, os, auth_token),
    character(1)
  )
}

.st_app_resolve_single_id <- function(app_id, os, auth_token) {
  app_id <- as.character(app_id[1])

  if (os == "ios" && grepl("^\\d+$", app_id)) {
    return(app_id)
  }
  if (os == "android" && grepl("^(com|net|org|io|app|game)\\.", app_id)) {
    return(app_id)
  }
  if (os == "unified" && grepl("^[a-f0-9]{24}$", app_id)) {
    return(app_id)
  }

  resolved <- tryCatch(
    resolve_app_id(app_id, auth_token = auth_token, use_cache = TRUE, verbose = FALSE),
    error = function(e) NULL
  )

  if (is.null(resolved)) {
    rlang::abort(sprintf("Failed to resolve app ID '%s' for `os = '%s'`.", app_id, os))
  }

  resolved_id <- switch(
    os,
    ios = resolved$ios_app_id,
    android = resolved$android_app_id,
    unified = resolved$unified_app_id
  )

  if (is.null(resolved_id) || is.na(resolved_id) || !nzchar(as.character(resolved_id))) {
    rlang::abort(
      sprintf("Could not resolve a %s app ID for input '%s'.", os, app_id)
    )
  }

  as.character(resolved_id)
}

#' Formatting Helper Functions
#'
#' @description
#' Functions for formatting numeric values in a human-readable way,
#' particularly useful for revenue, download counts, and percentages.
#' 
#' @name formatting_helpers
NULL

#' Format large numbers with K/M/B suffixes
#'
#' @param val Numeric value to format
#' @param digits Number of digits after decimal for millions/billions (default: 1)
#' @param prefix Optional prefix (e.g., "$" for currency)
#' @return Formatted string with appropriate suffix
#' @export
#' @examples
#' format_large_number(1234567)      # "1.2M"
#' format_large_number(1234567890)   # "1.2B"
#' format_large_number(1234567, prefix = "$") # "$1.2M"
format_large_number <- function(val, digits = 1, prefix = "") {
  if (is.na(val) || abs(val) < 0.5) return("\u2014")  # em-dash for NA or ~0

  if (abs(val) >= 1e9) {
    paste0(prefix, formatC(val/1e9, format = "f", digits = digits), "B")
  } else if (abs(val) >= 1e6) {
    paste0(prefix, formatC(val/1e6, format = "f", digits = digits), "M")
  } else if (abs(val) >= 1e3) {
    paste0(prefix, formatC(val/1e3, format = "f", digits = 0), "K")
  } else {
    paste0(prefix, formatC(val, format = "f", digits = 0))
  }
}

#' Format currency values with appropriate suffixes
#'
#' @param val Numeric value to format as currency
#' @param digits Number of digits after decimal for millions/billions (default: 2)
#' @return Formatted currency string
#' @export
#' @examples
#' format_currency(1234567)      # "$1.23M"
#' format_currency(1234567890)   # "$1.23B"
#' format_currency(123)          # "$123"
format_currency <- function(val, digits = 2) {
  format_large_number(val, digits = digits, prefix = "$")
}

#' Format download counts with appropriate suffixes
#'
#' @param val Numeric value to format as downloads
#' @return Formatted download count string
#' @export
#' @examples
#' format_downloads(1234567)      # "1.2M"
#' format_downloads(1234567890)   # "1.2B"
format_downloads <- function(val) {
  format_large_number(val, digits = 1, prefix = "")
}

#' Format percentages
#'
#' @param val Numeric value to format as percentage (0-100 scale)
#' @param digits Number of decimal places (default: 1)
#' @return Formatted percentage string
#' @export
#' @examples
#' format_percent(23.456)    # "23.5%"
#' format_percent(0.234, digits = 2)    # "0.23%"
format_percent <- function(val, digits = 1) {
  if (is.na(val)) return("-")
  paste0(formatC(val, format = "f", digits = digits), "%")
}

#' Format retention rates
#'
#' @param val Numeric value as decimal (0-1 scale)
#' @param digits Number of decimal places (default: 1)
#' @return Formatted retention percentage string
#' @export
#' @examples
#' format_retention(0.234)    # "23.4%"
#' format_retention(0.85)     # "85.0%"
format_retention <- function(val, digits = 1) {
  if (is.na(val)) return("-")
  paste0(formatC(val * 100, format = "f", digits = digits), "%")
}

#' Format user counts (DAU/MAU/WAU)
#'
#' @param val Numeric value to format as user count
#' @return Formatted user count string
#' @export
#' @examples
#' format_users(1234567)      # "1.2M"
#' format_users(1234)         # "1.2K"
format_users <- function(val) {
  format_large_number(val, digits = 1, prefix = "")
}

#' Format ARPU (Average Revenue Per User)
#'
#' @param val Numeric ARPU value
#' @param digits Number of decimal places (default: 2)
#' @return Formatted ARPU string
#' @export
#' @examples
#' format_arpu(5.234)    # "$5.23"
#' format_arpu(0.99)     # "$0.99"
format_arpu <- function(val, digits = 2) {
  if (is.na(val)) return("-")
  paste0("$", formatC(val, format = "f", digits = digits))
}

#' Format market share as percentage
#'
#' @param val Numeric value as decimal (0-1 scale)
#' @param digits Number of decimal places (default: 1)
#' @return Formatted market share percentage string
#' @export
#' @examples
#' format_market_share(0.234)    # "23.4%"
#' format_market_share(0.05)     # "5.0%"
format_market_share <- function(val, digits = 1) {
  format_retention(val, digits = digits)
}

#' Create a vector of formatted values
#'
#' @keywords internal
#' @noRd
format_vector <- function(values, type = c("currency", "downloads", "percent", "users", "retention", "arpu"), ...) {
  type <- match.arg(type)
  
  format_fn <- switch(type,
    currency = format_currency,
    downloads = format_downloads,
    percent = format_percent,
    users = format_users,
    retention = format_retention,
    arpu = format_arpu
  )
  
  sapply(values, format_fn, ...)
}

# Global variables used in the package
# These are variables used in NSE contexts (dplyr, tidyselect) that
# R CMD check would otherwise flag as undefined global variables
utils::globalVariables(c(
  # Common tidyverse variables
  ".",

  # Variables from API responses
  "category_id",
  "name",
  "user_rating",
  "user_rating_count",
  "score",
  "developer",
  "developer_id",
  "developer_email",
  "developer_address",
  "iphone_revenue_cents",
  "ipad_revenue_cents",
  "iphone_downloads",
  "ipad_downloads",
  "iphone_revenue",
  "ipad_revenue",
  "revenue_cents",

  # Variables from custom tags
  "entities.custom_tags.Genders (Last Quarter, WW)",
  "revenue_30d_ww",
  "revenue_alltime_ww",
  "genders_ww",
  "unified_app_name",

  # Variables used in dplyr/tidyr operations
  "android_id",
  "android_only",
  "app_id",
  "app_id_type",
  "both_id",
  "custom_fields",
  "ios_id",
  "ios_only",
  "original_id",
  "prev_value",
  "true_unified_id",
  "unified_app_id",
  "unified_id",
  "unified_revenue",
  "users_absolute",
  "values",

  # gt function names that are used without namespace prefix in some contexts
  "cell_text",
  "cells_body",
  "px",
  "text_transform",

  # Variables used in st_publisher_portfolio
  "aggregate_tags.Game Sub-genre",
  "global_rank",
  "subgenre_rank",
  "Month",
  "Year",
  "subgenre",
  "revenue",
  "downloads",
  "mau",
  "value"
))

# Import na.omit from stats
#' @importFrom stats na.omit
NULL

#' Custom Filter Helper Functions
#' 
#' Functions to validate, test, and manage Sensor Tower custom field filters
#' 
#' @name filter_helpers
NULL

#' Validate Custom Field Filter ID Format
#' 
#' Checks if a filter ID matches the expected 24-character hexadecimal format
#' used by Sensor Tower.
#' 
#' @param filter_id Character string. The filter ID to validate
#' @return Logical. TRUE if valid format, FALSE otherwise
#' @export
#' @examples
#' \dontrun{
#' # Valid filter ID
#' st_is_valid_filter_id("687df26ac5a19ebcfe817d7f")  # TRUE
#' 
#' # Invalid filter IDs
#' st_is_valid_filter_id("invalid")  # FALSE
#' st_is_valid_filter_id("687df26ac5a19ebcfe817d7")  # FALSE (too short)
#' }
st_is_valid_filter_id <- function(filter_id) {
  if (is.null(filter_id) || !is.character(filter_id) || length(filter_id) != 1) {
    return(FALSE)
  }
  grepl("^[a-f0-9]{24}$", filter_id, ignore.case = TRUE)
}

#' Test a Custom Filter ID
#' 
#' Tests whether a custom filter ID works with the Sensor Tower API by making
#' a minimal test request. This helps verify that the filter exists and is
#' accessible with your authentication.
#' 
#' @param filter_id Character string. The filter ID to test
#' @param os Character string. Operating system to test with. One of "ios", 
#'   "android", or "unified". Defaults to "ios".
#' @param verbose Logical. Whether to print detailed test results. Defaults to TRUE.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'   Defaults to environment variable SENSORTOWER_AUTH_TOKEN.
#' @return List with test results including success status and any error messages
#' @export
#' @examples
#' \dontrun{
#' # Test a filter ID
#' result <- st_test_filter("687df26ac5a19ebcfe817d7f")
#' 
#' # Test silently
#' result <- st_test_filter("687df26ac5a19ebcfe817d7f", verbose = FALSE)
#' 
#' # Test with different OS
#' result <- st_test_filter("687df26ac5a19ebcfe817d7f", os = "unified")
#' }
st_test_filter <- function(filter_id, 
                          os = "ios", 
                          verbose = TRUE,
                          auth_token = NULL) {
  
  # Validate filter format first
  if (!st_is_valid_filter_id(filter_id)) {
    result <- list(
      success = FALSE,
      filter_id = filter_id,
      error = "Invalid filter ID format. Expected 24-character hexadecimal string.",
      os_tested = os
    )
    
    if (verbose) {
      cat("X Invalid filter ID format:", filter_id, "\n")
    }
    
    return(invisible(result))
  }
  
  if (verbose) {
    cat("Testing filter ID:", filter_id, "\n")
    cat("Operating System:", os, "\n")
  }
  
  # Prepare test parameters
  test_params <- list(
    os = os,
    measure = "revenue",
    custom_fields_filter_id = filter_id,
    regions = "US",
    limit = 1,
    auth_token = auth_token
  )
  
  # Add custom_tags_mode if unified
  if (os == "unified") {
    test_params$custom_tags_mode <- "include"
  }
  
  # Make test API call
  test_result <- tryCatch(
    {
      data <- do.call(st_top_charts_impl, test_params)
      list(
        success = TRUE,
        filter_id = filter_id,
        os_tested = os,
        rows_returned = nrow(data),
        data_sample = if (nrow(data) > 0) data[1, ] else NULL
      )
    },
    error = function(e) {
      error_details <- list(
        success = FALSE,
        filter_id = filter_id,
        os_tested = os,
        error = e$message
      )
      
      # Categorize error type
      if (grepl("422", e$message)) {
        error_details$error_type <- "invalid_filter"
        error_details$suggestion <- "Filter doesn't exist or belongs to different account"
      } else if (grepl("403", e$message)) {
        error_details$error_type <- "access_denied"
        error_details$suggestion <- "Check your API authentication"
      } else if (grepl("401", e$message)) {
        error_details$error_type <- "authentication"
        error_details$suggestion <- "Invalid or missing API token"
      } else {
        error_details$error_type <- "unknown"
        error_details$suggestion <- "Check API status and parameters"
      }
      
      error_details
    }
  )
  
  if (verbose) {
    if (test_result$success) {
      cat("v Filter test successful!\n")
      if (!is.null(test_result$rows_returned)) {
        cat("   Returned", test_result$rows_returned, "rows\n")
      }
    } else {
      cat("X Filter test failed\n")
      cat("   Error:", test_result$error, "\n")
      if (!is.null(test_result$suggestion)) {
        cat("   Suggestion:", test_result$suggestion, "\n")
      }
    }
  }
  
  invisible(test_result)
}

#' Generate Example Filter IDs for Testing
#' 
#' Generates example filter IDs in the correct format for testing purposes.
#' Note: These are randomly generated and won't work with the actual API unless
#' they happen to match a real filter in your account.
#' 
#' @param n Integer. Number of example IDs to generate. Defaults to 5.
#' @param seed Integer. Random seed for reproducibility. Optional.
#' @return Character vector of example filter IDs
#' @keywords internal
#' @noRd
#' @examples
#' # Generate example IDs
#' st_generate_example_filter_ids(3)
#' 
#' # Generate with seed for reproducibility
#' st_generate_example_filter_ids(3, seed = 123)
st_generate_example_filter_ids_impl <- function(n = 5, seed = NULL) {
  if (!is.null(seed)) {
    set.seed(seed)
  }
  
  # Generate random hex strings
  ids <- character(n)
  for (i in seq_len(n)) {
    ids[i] <- paste0(
      sprintf("%x", sample(0:15, 24, replace = TRUE)),
      collapse = ""
    )
  }
  
  ids
}

#' Build Sensor Tower Filter URL
#' 
#' Constructs a URL for the Sensor Tower web interface where you can create
#' custom filters. Visit this URL, configure your filters, and then copy the
#' custom_fields_filter_id from the resulting URL.
#' 
#' @param os Character string. Operating system filter. One of "ios", "android", 
#'   or "unified". Defaults to "unified".
#' @param category Optional. Category ID to pre-select.
#' @param countries Optional. Character vector of country codes to pre-select.
#' @param base_url Character string. Base URL for Sensor Tower. Defaults to 
#'   "https://app.sensortower.com/top-charts".
#' @return Character string. The constructed URL.
#' @export
#' @examples
#' \dontrun{
#' # Build URL for iOS games in US
#' url <- st_build_filter_url(os = "ios", category = 6014, countries = "US")
#' 
#' # Open in browser
#' browseURL(url)
#' }
st_build_filter_url <- function(os = "unified",
                               category = NULL,
                               countries = NULL,
                               base_url = "https://app.sensortower.com/top-charts") {
  
  # Validate OS
  os <- match.arg(os, c("ios", "android", "unified"))
  
  # Build query parameters
  params <- list(os = os)
  
  if (!is.null(category)) {
    params$category <- as.character(category)
  }
  
  if (!is.null(countries)) {
    if (is.character(countries)) {
      params$countries <- paste(countries, collapse = ",")
    }
  }
  
  # Construct URL
  url <- httr::modify_url(base_url, query = params)
  
  message("Filter creation URL generated:")
  message(url)
  message("\nSteps to create a custom filter:")
  message("1. Visit the URL above")
  message("2. Configure your desired filters")
  message("3. Copy the custom_fields_filter_id from the URL")
  message("4. Use it in sensortowerR functions")
  
  invisible(url)
}

#' Extract Filter ID from Sensor Tower URL
#' 
#' Extracts the custom_fields_filter_id parameter from a Sensor Tower URL.
#' This is helpful when copying URLs from the web interface.
#' 
#' @param url Character string. A Sensor Tower URL containing custom_fields_filter_id
#' @return Character string. The extracted filter ID, or NULL if not found.
#' @export
#' @examples
#' \dontrun{
#' url <- "https://app.sensortower.com/top-charts?custom_fields_filter_id=687df26ac5a19ebcfe817d7f"
#' filter_id <- st_extract_filter_id(url)
#' }
st_extract_filter_id <- function(url) {
  if (is.null(url) || !is.character(url)) {
    return(NULL)
  }
  
  # Parse URL
  parsed <- httr::parse_url(url)
  
  # Extract from query parameters
  filter_id <- parsed$query$custom_fields_filter_id
  
  if (is.null(filter_id)) {
    # Try regex as fallback
    match <- regmatches(url, regexpr("custom_fields_filter_id=([a-f0-9]{24})", url))
    if (length(match) > 0) {
      filter_id <- sub("custom_fields_filter_id=", "", match[1])
    }
  }
  
  # Validate format
  if (!is.null(filter_id) && st_is_valid_filter_id(filter_id)) {
    message("Extracted filter ID: ", filter_id)
    return(filter_id)
  }
  
  message("No valid filter ID found in URL")
  return(NULL)
}

#' Compare Results With and Without Custom Filter
#' 
#' Compares the results of API calls with a custom filter versus a standard
#' category filter to understand what the custom filter is doing.
#' 
#' @param filter_id Character string. The custom filter ID to test
#' @param category Category ID to use for comparison
#' @param os Character string. Operating system. Defaults to "ios".
#' @param regions Character string. Region code. Defaults to "US".
#' @param limit Integer. Number of results to fetch. Defaults to 20.
#' @return List containing both result sets and comparison statistics
#' @keywords internal
#' @noRd
st_compare_filter_results_impl <- function(filter_id,
                                     category = 6014,  # Games
                                     os = "ios",
                                     regions = "US",
                                     limit = 20) {
  
  if (!st_is_valid_filter_id(filter_id)) {
    rlang::abort("Invalid filter ID format")
  }
  
  message("Fetching results with custom filter...")
  filtered_results <- tryCatch(
    st_top_charts_impl(
      os = os,
      custom_fields_filter_id = filter_id,
      regions = regions,
      limit = limit
    ),
    error = function(e) {
      rlang::abort("Failed to fetch filtered results: ", e$message)
    }
  )
  
  message("Fetching results with category filter...")
  category_results <- tryCatch(
    st_top_charts_impl(
      os = os,
      category = category,
      regions = regions,
      limit = limit
    ),
    error = function(e) {
      rlang::abort("Failed to fetch category results: ", e$message)
    }
  )
  
  # Compare results
  comparison <- list(
    filtered = filtered_results,
    category = category_results,
    stats = list(
      filtered_count = nrow(filtered_results),
      category_count = nrow(category_results)
    )
  )
  
  # Find apps in filtered but not in category
  if ("unified_app_id" %in% names(filtered_results) && 
      "unified_app_id" %in% names(category_results)) {
    
    comparison$stats$unique_to_filtered <- setdiff(
      filtered_results$unified_app_id,
      category_results$unified_app_id
    )
    
    comparison$stats$unique_to_category <- setdiff(
      category_results$unified_app_id,
      filtered_results$unified_app_id
    )
    
    comparison$stats$common_apps <- intersect(
      filtered_results$unified_app_id,
      category_results$unified_app_id
    )
  }
  
  message("\nComparison Results:")
  message("Filtered results: ", comparison$stats$filtered_count, " apps")
  message("Category results: ", comparison$stats$category_count, " apps")
  
  if (!is.null(comparison$stats$common_apps)) {
    message("Common apps: ", length(comparison$stats$common_apps))
    message("Unique to filter: ", length(comparison$stats$unique_to_filtered))
    message("Unique to category: ", length(comparison$stats$unique_to_category))
  }
  
  invisible(comparison)
}

#' Fetch Active User Metrics with a Tidy Long Output
#'
#' Lightweight wrapper around [st_batch_metrics()] focused only on active-user
#' metrics (`dau`, `wau`, `mau`). Returns a tidy long-format tibble that is
#' designed for straightforward tidyverse chaining.
#'
#' @param os Character. Required. Operating system: `"ios"`, `"android"`, or
#'   `"unified"`.
#' @param app_list List or data frame containing app information. Same formats
#'   accepted by [st_batch_metrics()].
#' @param metrics Character vector. Active-user metrics to return. Must be one
#'   or more of `"dau"`, `"wau"`, `"mau"`. Default is all three.
#' @param date_range List with `start_date` and `end_date`.
#' @param countries Character vector of country codes.
#' @param granularity Character granularity (`"daily"`, `"weekly"`,
#'   `"monthly"`, `"quarterly"`). Default `"monthly"`.
#' @param parallel Logical. Whether to use parallel processing.
#' @param verbose Logical. Whether to print progress messages.
#' @param auth_token Character string. Sensor Tower API token. Defaults to
#'   `SENSORTOWER_AUTH_TOKEN` environment variable.
#' @param max_cores Integer. Maximum cores for parallel processing.
#'
#' @return A tibble with columns including `original_id`, `app_name`, `app_id`,
#'   `app_id_type`, `date`, `country`, `metric`, and `value`.
#' @export
st_active_users <- function(os,
                            app_list,
                            metrics = c("dau", "wau", "mau"),
                            date_range = list(
                              start_date = Sys.Date() - 90,
                              end_date = Sys.Date() - 1
                            ),
                            countries,
                            granularity = "monthly",
                            parallel = FALSE,
                            verbose = TRUE,
                            auth_token = NULL,
                            max_cores = 2) {
  if (missing(os) || is.null(os)) {
    rlang::abort("'os' parameter is required. Specify one of: 'ios', 'android', 'unified'.")
  }

  if (missing(app_list) || is.null(app_list)) {
    rlang::abort("'app_list' parameter is required.")
  }

  if (missing(countries) || is.null(countries) || length(countries) == 0) {
    rlang::abort("'countries' parameter is required.")
  }

  metrics <- tolower(as.character(metrics))
  valid_metrics <- c("dau", "wau", "mau")
  invalid_metrics <- setdiff(metrics, valid_metrics)
  if (length(invalid_metrics) > 0) {
    rlang::abort(
      sprintf(
        "Invalid metrics: %s. `metrics` must be one or more of: dau, wau, mau.",
        paste(invalid_metrics, collapse = ", ")
      )
    )
  }

  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token is required. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  results <- st_batch_metrics_impl(
    os = os,
    app_list = app_list,
    metrics = metrics,
    date_range = date_range,
    countries = countries,
    granularity = granularity,
    parallel = parallel,
    verbose = verbose,
    auth_token = auth_token_val,
    max_cores = max_cores
  )

  if (nrow(results) == 0) {
    return(results)
  }

  results %>%
    dplyr::filter(.data$metric %in% metrics) %>%
    dplyr::arrange(.data$original_id, .data$metric, .data$date, .data$country)
}

test_that("st_active_users validates metric choices", {
  expect_error(
    st_active_users(
      os = "ios",
      app_list = "553834731",
      metrics = c("dau", "revenue"),
      countries = "US",
      auth_token = "token"
    ),
    "must be one or more of: dau, wau, mau"
  )
})

test_that("st_active_users delegates to st_batch_metrics and returns active metrics only", {
  testthat::local_mocked_bindings(
    resolve_auth_token = function(auth_token, env_var = "SENSORTOWER_AUTH_TOKEN", error_message = NULL) {
      "token"
    },
    st_batch_metrics_impl = function(...) {
      tibble::tibble(
        original_id = c("app_b", "app_a", "app_a"),
        app_name = c("B", "A", "A"),
        app_id = c("2", "1", "1"),
        app_id_type = c("ios", "ios", "ios"),
        date = as.Date(c("2024-01-02", "2024-01-01", "2024-01-01")),
        country = c("US", "US", "US"),
        metric = c("wau", "revenue", "dau"),
        value = c(20, 999, 10)
      )
    },
    .package = "sensortowerR"
  )

  result <- st_active_users(
    os = "ios",
    app_list = c("1", "2"),
    metrics = c("dau", "wau"),
    countries = "US",
    auth_token = "token",
    verbose = FALSE
  )

  expect_equal(nrow(result), 2)
  expect_true(all(result$metric %in% c("dau", "wau")))
  expect_false(any(result$metric == "revenue"))
  expect_equal(result$original_id, c("app_a", "app_b"))
})

#' List Available Sensor Tower Categories
#'
#' Returns a tibble of app categories recognized by the Sensor Tower API,
#' mapping category IDs to category names for different platforms (iOS/Android).
#' Useful for finding valid inputs for the `category` parameter in other
#' functions.
#'
#' @param platform Optional character string. Filter results for a specific
#'   platform ("ios" or "android"). If NULL (default), returns categories for
#'   both platforms.
#'
#' @return A tibble with columns `platform` (character, "ios" or "android"),
#'   `category_id` (character, e.g., "6014"), and `category_name`
#'   (character, e.g., "Games").
#' @export
#' @examples
#' # Get all categories
#' all_cats <- st_categories()
#' head(all_cats)
#'
#' # Get only iOS categories
#' ios_cats <- st_categories(platform = "ios")
#' head(ios_cats)
#'
#' # Find game categories on iOS
#' ios_games <- subset(st_categories("ios"), grepl("Game", category_name))
#' head(ios_games)
st_categories <- function(platform = NULL) { # Renamed function
  # Access the internal data object st_category_data
  data <- st_category_data

  if (!is.null(platform)) {
    platform <- tolower(platform)
    if (!platform %in% c("ios", "android")) {
      warning(
        "Invalid platform specified. Choose 'ios' or 'android'. ",
        "Returning all categories.",
        call. = FALSE
      )
    } else {
      data <- data[data$platform == platform, ]
    }
  }
  return(data)
} 

#' Fetch Enriched Metrics for Specific Apps
#'
#' Retrieves comprehensive metrics including retention, MAU, DAU, demographics,
#' and other aggregate tags for specific apps by their unified app IDs.
#'
#' This function is designed for the common workflow of:
#' 1. Search for apps by name using `st_app_info()`
#' 2. Get their unified IDs
#' 3. Fetch enriched metrics for those specific apps using this function
#'
#' @param unified_app_ids Character vector. One or more unified app IDs
#'   (24-character hex strings). Required. Use `st_app_info()` to find these.
#' @param os Character string. Operating system context for the request.
#'   Must be "unified" (default), "ios", or "android".
#' @param regions Character vector. Region codes for data filtering.
#'   Defaults to "WW" (worldwide).
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#'   Defaults to environment variable SENSORTOWER_AUTH_TOKEN.
#'
#' @return A [tibble][tibble::tibble] with enriched metrics including:
#'   - **Identification**: `unified_app_id`, `unified_app_name`
#'   - **Active Users**: `dau_30d_us`, `dau_30d_ww`, `wau_4w_us`, `wau_4w_ww`,
#'     `mau_month_us`, `mau_month_ww`
#'   - **Retention**: `retention_1d_us/ww`, `retention_7d_us/ww`,
#'     `retention_14d_us/ww`, `retention_30d_us/ww`, `retention_60d_us/ww`
#'   - **Demographics**: `genders_us`, `genders_ww`, `age_us`, `age_ww`,
#'     `male_share_us`, `female_share_us`
#'   - **Revenue/Downloads**: `revenue_30d_ww`, `revenue_90d_ww`, `revenue_alltime_us/ww`,
#'     `downloads_30d_ww`, `downloads_alltime_us/ww`
#'   - **Monetization**: `rpd_alltime_us/ww`, `arpu_90d_us/ww`
#'   - **Launch**: `release_date_us/ww`, `earliest_release_date`
#'
#' @section Recommended Workflow:
#' ```
#' # Step 1: Search for apps by name
#' apps <- st_app_info("Royal Match")
#'
#' # Step 2: Get unified IDs
#' app_ids <- apps$unified_app_id
#'
#' # Step 3: Fetch enriched metrics
#' metrics <- st_app_enriched(app_ids)
#' ```
#'
#' @section Data Availability Notes:
#' - **IMPORTANT: Geographic Limitations** - All enriched metrics are only available
#'   for **US market (`_us` suffix)** and **Worldwide aggregates (`_ww` suffix)**.
#'   Per-country data (e.g., GB, DE, FR, JP) is NOT available through this endpoint.
#'   For per-country data, use [st_sales_report()] for revenue/downloads or
#'   [st_batch_metrics()] for MAU/DAU time-series.
#' - Retention data (D1, D7, D14, D30, D60) is aggregated for the "last quarter" -
#'   not time-series data. D90 retention is NOT available through the API.
#' - Demographics (age/gender) are primarily available for US market only.
#' - Not all metrics are available for all apps - smaller apps may have NA values.
#' - This returns **snapshot data**, not time-series. For historical trends, use
#'   [st_batch_metrics()] or [st_sales_report()].
#'
#' @examples
#' \dontrun{
#' # Get enriched data for specific apps
#' royal_match <- st_app_info("Royal Match")
#' enriched <- st_app_enriched(royal_match$unified_app_id)
#'
#' # Get data for multiple apps at once
#' game_ids <- c("5f16a8019f7b275235017614", "660af7c66237390ce7c829fc")
#' multi_enriched <- st_app_enriched(game_ids)
#'
#' # View retention metrics
#' multi_enriched %>%
#'   select(unified_app_name, starts_with("retention"))
#' }
#'
#' @seealso [st_app_info()] for searching apps by name,
#'   [st_app_lookup()] for resolving app IDs,
#'   [st_sales_report()] for time-series revenue/download data,
#'   [st_batch_metrics()] for time-series DAU/WAU/MAU data
#'
#' @export
#' @importFrom rlang %||% abort
#' @importFrom httr2 request req_url_path_append req_url_query req_headers
#' @importFrom httr2 req_perform resp_body_raw
#' @importFrom jsonlite fromJSON
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr filter bind_rows mutate select any_of
st_app_enriched <- function(unified_app_ids,
                            os = "unified",
                            regions = "WW",
                            auth_token = NULL) {


  if (missing(unified_app_ids) || is.null(unified_app_ids) ||
      length(unified_app_ids) == 0) {
    rlang::abort(c(
      "'unified_app_ids' parameter is required.",
      "i" = "Use st_app_info_impl('app name') to search for apps and get their IDs.",
      "x" = "Example: st_app_enriched(c('5f16a8019f7b275235017614'))"
    ))
  }


  unified_app_ids <- as.character(unified_app_ids)
  invalid_ids <- unified_app_ids[!grepl("^[a-f0-9]{24}$", unified_app_ids)]
  if (length(invalid_ids) > 0) {
    rlang::abort(c(
      "Invalid unified app ID format detected.",
      "i" = "Unified IDs must be 24-character hexadecimal strings.",
      "x" = paste("Invalid IDs:", paste(invalid_ids[1:min(3, length(invalid_ids))],
                                         collapse = ", ")),
      "i" = "Use st_app_info_impl('app name') to find valid unified IDs."
    ))
  }

  # Validate OS
  os <- match.arg(os, c("unified", "ios", "android"))

  # Authentication
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  # Build filter for specific app IDs
  # The API requires a custom_fields_filter to get aggregate tags
  # We'll create one targeting these specific unified app IDs

  # For apps without a filter, we fetch from top_charts with a large limit
  # and filter client-side. This is the only reliable way to get aggregate_tags.

  # Strategy: Use st_top_charts with category=0 (all) and filter results
  # This is not ideal but is the only way given API limitations

  message(sprintf("Fetching enriched data for %d app(s)...", length(unified_app_ids)))

  # Use the sales_report_estimates_comparison_attributes endpoint
  # which returns aggregate_tags when queried properly

  # Build query parameters (auth via header, not query param)
  query_params <- list(
    measure = "revenue",  # Required parameter
    comparison_attribute = "absolute",
    time_range = "month",
    date = format(Sys.Date() - 30, "%Y-%m-%d"),
    end_date = format(Sys.Date() - 1, "%Y-%m-%d"),
    category = "0",  # All categories
    regions = paste(regions, collapse = ","),
    limit = 2000,  # Fetch enough to hopefully include our apps
    device_type = if (os %in% c("ios", "unified")) "total" else NULL
  )

  # Remove NULL parameters
 query_params <- query_params[!sapply(query_params, is.null)]

  # Build request
  req <- build_request(
    base_url = st_api_base_url(),
    path_segments = st_endpoint_segments("sales_report_estimates_comparison_attributes", os = os),
    query_params = query_params
  ) %>%
    httr2::req_headers(
      "Authorization" = paste("Bearer", auth_token_val),
      "Accept" = "application/json"
    ) %>%
    httr2::req_timeout(60)

  # Perform request
  resp <- tryCatch({
    httr2::req_perform(req)
  }, error = function(e) {
    if (inherits(e, "httr2_http")) {
      # httr2 stores response in e$resp, not e$response
      resp_obj <- e$resp
      if (!is.null(resp_obj)) {
        status <- httr2::resp_status(resp_obj)
        body <- tryCatch(httr2::resp_body_string(resp_obj), error = function(.) "")

        if (status == 401) {
          rlang::abort("Invalid authentication token.")
        } else if (status == 403) {
          rlang::abort("API token not authorized for this endpoint.")
        } else {
          rlang::abort(sprintf("API request failed [%d]: %s", status, body))
        }
      }
    }
    rlang::abort(sprintf("Request failed: %s", conditionMessage(e)))
  })

  # Process response with enrichment (reuse existing utilities)
  result_tbl <- process_response(resp, enrich_response = TRUE)

  if (nrow(result_tbl) == 0) {
    message("No results in API response")
    return(tibble::tibble())
  }

  # Filter to our requested app IDs
  if ("unified_app_id" %in% names(result_tbl)) {
    filtered <- result_tbl %>%
      dplyr::filter(unified_app_id %in% unified_app_ids)
  } else if ("app_id" %in% names(result_tbl)) {
    filtered <- result_tbl %>%
      dplyr::filter(app_id %in% unified_app_ids)
  } else {
    message("Warning: Could not find app ID column in response")
    filtered <- tibble::tibble()
  }

  # Report on what we found
  found_ids <- if ("unified_app_id" %in% names(filtered)) {
    unique(filtered$unified_app_id)
  } else if ("app_id" %in% names(filtered)) {
    unique(filtered$app_id)
  } else {
    character(0)
  }

  missing_ids <- setdiff(unified_app_ids, found_ids)

  if (length(found_ids) > 0) {
    message(sprintf("Found enriched data for %d of %d requested apps",
                   length(found_ids), length(unified_app_ids)))
  }

  if (length(missing_ids) > 0) {
    message(sprintf(
      "Note: %d app(s) not found in top charts. These may be smaller apps without aggregate metrics: %s",
      length(missing_ids),
      paste(missing_ids[1:min(3, length(missing_ids))], collapse = ", ")
    ))
  }

  # Select and order key columns
  key_cols <- c(
    "unified_app_id", "unified_app_name",
    # Active users
    "dau_30d_us", "dau_30d_ww",
    "wau_4w_us", "wau_4w_ww",
    "mau_month_us", "mau_month_ww",
    # Retention
    "retention_1d_us", "retention_1d_ww",
    "retention_7d_us", "retention_7d_ww",
    "retention_14d_us", "retention_14d_ww",
    "retention_30d_us", "retention_30d_ww",
    "retention_60d_us", "retention_60d_ww",
    # Demographics
    "genders_us", "genders_ww",
    "age_us", "age_ww",
    "male_share_us", "female_share_us",
    # Revenue
    "revenue_30d_ww", "revenue_90d_ww", "revenue_180d_ww",
    "revenue_alltime_us", "revenue_alltime_ww",
    # Downloads
    "downloads_30d_ww", "downloads_90d_ww", "downloads_180d_ww",
    "downloads_alltime_us", "downloads_alltime_ww",
    # Monetization
    "rpd_alltime_us", "rpd_alltime_ww",
    "arpu_90d_us", "arpu_90d_ww",
    # Launch
    "release_date_us", "release_date_ww", "earliest_release_date"
  )

  # Get columns that exist in our data
  available_key_cols <- intersect(key_cols, names(filtered))
  other_cols <- setdiff(names(filtered), key_cols)

  # Reorder with key columns first
  if (nrow(filtered) > 0 && length(available_key_cols) > 0) {
    filtered <- filtered %>%
      dplyr::select(dplyr::any_of(available_key_cols), dplyr::any_of(other_cols))
  }

  return(filtered)
}

#' Fetch Top Apps by Various Metrics
#'
#' Retrieves top apps from Sensor Tower based on revenue, downloads ("units"), 
#' or active user metrics (DAU, WAU, MAU). This unified function automatically 
#' selects the appropriate API endpoint based on the measure specified.
#'
#' @param measure Character string. Metric to measure. Must be one of:
#'   - **Revenue/Downloads**: "revenue" (default), "units" 
#'   - **Active Users**: "DAU", "WAU", "MAU"
#' @param os Character string. Operating System. Must be one of
#'   "ios", "android", or "unified". Required.
#' @param comparison_attribute Character string. Comparison attribute
#'   type. Must be one of "absolute", "delta", or "transformed_delta". 
#'   Defaults to "absolute".
#' @param time_range Character string. Time granularity. Must be one of
#'   "day", "week", "month", or "quarter". Defaults to "month".
#'   Note: "week" is not available when `measure` is "MAU".
#' @param date Character string or Date object. Start date for the
#'   query in "YYYY-MM-DD" format. Defaults to the start of the current month.
#' @param category Character string or numeric. The ID of the category to 
#'   filter by. **Required unless `custom_fields_filter_id` is provided**.
#'   Use `st_categories()` to find valid IDs.
#' @param regions Character vector or comma-separated string. Region
#'   codes (e.g., `"US"`, `c("US", "GB")`, `"WW"` for worldwide) to filter results. 
#'   Required.
#' @param end_date Optional. Character string or Date object. End date for the
#'   query in "YYYY-MM-DD" format, inclusive. Only used for revenue/downloads.
#' @param limit Optional. Integer. Maximum number of apps to return per call.
#'   Defaults to 20.
#' @param offset Optional. Integer. Number of apps to skip for pagination.
#' @param device_type Optional. Character string. For `os = "ios"` or
#'   `os = "unified"`: "iphone", "ipad", or "total". Defaults to `"total"`.
#' @param custom_fields_filter_id Optional. Character string. ID of a Sensor
#'   Tower custom field filter to apply.
#' @param custom_tags_mode Optional. Character string. Required if `os` is
#'   'unified' and `custom_fields_filter_id` is provided.
#' @param data_model Optional. Character string. The data model to use.
#'   Defaults to "DM_2025_Q2". Only used for active user metrics.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param base_url Optional. Character string. The base URL for the API.
#' @param enrich_response Optional. Logical. If `TRUE` (default), enriches
#'   the response with app metadata and custom metrics.
#' @param deduplicate_apps Optional. Logical. If `TRUE` (default), consolidates
#'   apps with the same name but different platform/regional SKUs into single rows
#'   with aggregated metrics. If `FALSE`, returns separate rows for each SKU.
#'
#' @return A [tibble][tibble::tibble] with top app data including enhanced
#'   custom metrics like downloads, revenue, retention rates, and more.
#'   For sales data (revenue/downloads), app names are automatically looked up
#'   using the app IDs since the sales endpoint doesn't provide app names natively.
#'   
#'   **Revenue Units**: Revenue values are standardized to base currency units (dollars,
#'   euros, etc.) for consistency across all sensortowerR functions. The function provides
#'   a `revenue` column in base units alongside the original `revenue_absolute` (in cents).
#'   
#'   **Data Cleaning**: Numeric metric values are automatically cleaned of special 
#'   characters (%, $, commas) and converted to proper numeric format for analysis.
#'   
#'   **App Deduplication**: By default, apps with the same name but different 
#'   platform/regional SKUs are consolidated into single rows with aggregated metrics
#'   (downloads/revenue summed, rates/percentages averaged).
#'
#' @section API Endpoints Used:
#'   - **All Measures**: `GET /v1/\{os\}/sales_report_estimates_comparison_attributes`
#'   - Note: DAU/WAU/MAU measures now use the sales endpoint with custom filters for correct sorting
#'
#' @section Enhanced Custom Metrics:
#'   The function extracts comprehensive custom metrics including:
#'   - Downloads: `downloads_180d_ww`, `downloads_90d_us`
#'   - Revenue: `revenue_180d_ww`, `revenue_90d_us`
#'   - Retention: `retention_1d_us`, `retention_7d_us`, `retention_30d_us`
#'   - Monetization: `rpd_alltime_us`, `arpu_90d_us`
#'   - Demographics: `male_share_us`, `female_share_us`
#'   - Platform: `ios_share_ww`, `android_share_ww`
#'
#' @examples
#' \dontrun{
#' # Top apps by revenue (default)
#' top_revenue <- st_top_charts(
#'   os = "ios",
#'   category = 6000,  # iOS Games
#'   regions = "WW"
#' )
#' 
#' # Top apps by downloads
#' top_downloads <- st_top_charts(
#'   os = "android",
#'   measure = "units", 
#'   category = 6000,
#'   regions = "US"
#' )
#' 
#' # Top apps by Daily Active Users with custom filter
#' # Custom filter URLs from Sensor Tower web interface can be used directly
#' # Extract the custom_fields_filter_id from the URL parameter 'uai'
#' top_word_puzzles <- st_top_charts(
#'   os = "unified",
#'   measure = "revenue",  # Use revenue but custom filter handles DAU sorting
#'   custom_fields_filter_id = "5a39e9681454d22f5a5e75ca",  # Word puzzle filter
#'   custom_tags_mode = "include_unified_apps",
#'   category = 7019,  # Puzzle category
#'   regions = "US",
#'   date = "2025-07-20",
#'   end_date = "2025-08-18"
#' )
#' 
#' # Custom time range and region
#' top_quarter <- st_top_charts(
#'   os = "ios",
#'   measure = "revenue",
#'   time_range = "quarter", 
#'   regions = "US",
#'   category = 6000
#' )
#' }
#'
#' @keywords internal
#' @noRd
st_top_charts_impl <- function(measure = "revenue",
                          os,
                          comparison_attribute = "absolute",
                          time_range = "month",
                          date = NULL,
                          category = NULL,
                          regions,
                          end_date = NULL,
                          limit = 20,
                          offset = NULL,
                          device_type = NULL,
                          custom_fields_filter_id = NULL,
                          custom_tags_mode = NULL,
                          data_model = "DM_2025_Q2",
                          auth_token = NULL,
                          base_url = "https://api.sensortower.com",
                          enrich_response = TRUE,
                          deduplicate_apps = TRUE) {
  
  # Validate required parameters
  if (missing(os) || is.null(os)) {
    rlang::abort("'os' parameter is required. Specify one of: 'ios', 'android', 'unified'.")
  }
  
  if (missing(regions) || is.null(regions) || length(regions) == 0) {
    rlang::abort("'regions' parameter is required. Specify country codes (e.g., 'US', 'GB', 'JP', or 'WW' for worldwide).")
  }
  
  # --- Input Validation ---
  measure <- match.arg(measure, c("revenue", "units", "DAU", "WAU", "MAU"))
  
  # Always use sales endpoint - it handles all measures including DAU
  # Determine which endpoint to use based on measure
  is_active_users <- toupper(measure) %in% c("DAU", "WAU", "MAU")
  is_sales <- !is_active_users
  
  # Validate category requirement - either category or custom_fields_filter_id must be provided
  if (is.null(category) && is.null(custom_fields_filter_id)) {
    rlang::abort("Either 'category' or 'custom_fields_filter_id' parameter is required. Use st_categories() to find valid IDs or provide a custom filter ID from the Sensor Tower web interface.")
  }
  
  # --- Handle Default Date ---
  if (is.null(date)) {
    date <- lubridate::floor_date(Sys.Date(), "month")
  }
  
  # --- Authentication ---
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set it as an environment variable."
  )
  
  # Handle default device_type
  if (is.null(device_type) && os %in% c("ios", "unified")) {
    device_type <- "total"
    if (os == "unified") {
      message("`device_type` is not specified for `os = 'unified'`. Defaulting to 'total'.")
    }
  }
  
  # If using a custom filter, enforce category = 0 and message when overridden
  if (!is.null(custom_fields_filter_id)) {
    if (!is.null(category) && !(as.character(category) %in% c("0"))) {
      message("Using category = 0 when a custom filter is provided (ignoring non-zero category)")
      category <- 0
    }
    if (is.null(category)) {
      # Default to 0 when custom filter is set and category omitted
      category <- 0
    }
  }
  
  # Route to appropriate API endpoint based on measure type
  if (is_active_users) {
    # --- Input Validation for Active Users ---
    validate_inputs(
      os = os,
      comparison_attribute = comparison_attribute,
      time_range = time_range,
      measure = measure,
      date = date,
      category = category,
      regions = regions,
      limit = limit,
      offset = offset,
      device_type = device_type,
      custom_fields_filter_id = custom_fields_filter_id,
      custom_tags_mode = custom_tags_mode,
      data_model = data_model
    )

    # --- Prepare Query Parameters for Active Users ---
    query_params <- prepare_query_params_active_users(
      auth_token = auth_token_val,
      comparison_attribute = comparison_attribute,
      time_range = time_range,
      measure = measure,
      date = date,
      category = category,
      regions = regions,
      limit = limit,
      offset = offset,
      device_type = device_type,
      custom_fields_filter_id = custom_fields_filter_id,
      custom_tags_mode = custom_tags_mode,
      data_model = data_model,
      os = os
    )

    # --- Build and Perform Request for Active Users ---
    path <- st_endpoint_segments("top_and_trending_active_users", os = os)
    
  } else {
    # --- Input Validation for Sales ---
    validate_inputs(
      os = os,
      comparison_attribute = comparison_attribute,
      time_range = time_range,
      measure = measure,
      date = date,
      category = category,
      regions = regions,
      end_date = end_date,
      limit = limit,
      offset = offset,
      device_type = device_type,
      custom_fields_filter_id = custom_fields_filter_id,
      custom_tags_mode = custom_tags_mode
    )

    # --- Prepare Query Parameters for Sales ---
    query_params <- prepare_query_params_sales(
      auth_token = auth_token_val,
      comparison_attribute = comparison_attribute,
      time_range = time_range,
      measure = measure,
      date = date,
      category = category,
      end_date = end_date,
      regions = regions,
      limit = limit,
      offset = offset,
      device_type = device_type,
      custom_fields_filter_id = custom_fields_filter_id,
      custom_tags_mode = custom_tags_mode,
      os = os
    )

    # --- Build and Perform Request for Sales ---
    path <- st_endpoint_segments("sales_report_estimates_comparison_attributes", os = os)
  }
  
  # --- Common Request Building and Processing ---
  req <- build_request(base_url, path, query_params)
  resp <- perform_request(req)
  
  # --- Process Response ---
  result <- process_response(resp, enrich_response)
  
  # --- Validate Data ---
  if (nrow(result) > 0) {
    result <- validate_top_charts_data(result, measure, regions)
  }
  
  # --- Deduplicate Apps ---
  # For unified OS, resolve true unified IDs and consolidate
  if (os == "unified" && "unified_app_id" %in% names(result) && nrow(result) > 1 && deduplicate_apps) {
    original_count <- nrow(result)
    
    # First, check if we have duplicate unified_app_ids (can happen after unnesting)
    if (any(duplicated(result$unified_app_id))) {
      message(sprintf("Consolidating %d duplicate entries...", sum(duplicated(result$unified_app_id))))
      
      # Group by unified_app_id and aggregate metrics
      result <- deduplicate_by_group_id(result, "unified_app_id")
      
      # Update count after initial deduplication
      new_count <- nrow(result)
      if (original_count != new_count) {
        message(sprintf("Consolidated %d app entries into %d unique apps", 
                       original_count, new_count))
      }
      original_count <- new_count
    }
    
    # Get unique platform IDs and their corresponding names
    unique_ids <- unique(result$unified_app_id)
    # Create a mapping that handles duplicate keys by taking the first occurrence
    id_name_df <- result %>%
      dplyr::select(unified_app_id, unified_app_name) %>%
      dplyr::distinct(unified_app_id, .keep_all = TRUE)
    id_to_name <- stats::setNames(id_name_df$unified_app_name, id_name_df$unified_app_id)
    
    # Check if we need to resolve platform IDs to true unified IDs
    # Filter out NA values first
    non_na_ids <- unique_ids[!is.na(unique_ids)]
    non_hex_ids <- non_na_ids[!grepl("^[a-f0-9]{24}$", non_na_ids)]
    
    if (length(non_hex_ids) > 0) {
      message(sprintf("Resolving unified IDs for %d apps...", length(non_hex_ids)))
      
      # Get corresponding names for these IDs
      # Handle potential NA values in the mapping
      app_names <- character(length(non_hex_ids))
      for (i in seq_along(non_hex_ids)) {
        name <- id_to_name[non_hex_ids[i]]
        app_names[i] <- if (!is.null(name) && !is.na(name)) as.character(name) else NA_character_
      }
      
      # Use st_get_unified_mapping with app names for better resolution
      mapping_result <- tryCatch({
        st_get_unified_mapping(
          app_ids = non_hex_ids,
          app_names = app_names,
          os = "unified",
          auth_token = auth_token_val
        )
      }, error = function(e) {
        message("Warning: Error resolving unified IDs: ", e$message)
        NULL
      })
      
      if (!is.null(mapping_result) && nrow(mapping_result) > 0) {
        # Create mapping of platform IDs to true unified IDs
        id_mapping <- mapping_result %>%
          dplyr::filter(!is.na(.data$unified_app_id)) %>%
          dplyr::select(platform_id = .data$input_id, 
                       true_unified_id = .data$unified_app_id)
        
        if (nrow(id_mapping) > 0) {
          # Apply true unified ID mapping
          result <- result %>%
            dplyr::left_join(id_mapping, by = c("unified_app_id" = "platform_id")) %>%
            dplyr::mutate(
              .group_id = dplyr::coalesce(.data$true_unified_id, .data$unified_app_id)
            )
          
          # Group by true unified ID and aggregate metrics
          result <- deduplicate_by_group_id(result, ".group_id")
          
          # Update the unified_app_id to be the true unified ID where available
          if ("true_unified_id" %in% names(result)) {
            result <- result %>%
              dplyr::mutate(
                unified_app_id = dplyr::coalesce(.data$true_unified_id, .data$unified_app_id)
              ) %>%
              dplyr::select(-true_unified_id)
          }
          
          new_count <- nrow(result)
          if (original_count != new_count) {
            message(sprintf("Consolidated %d app entries into %d unique apps using unified IDs", 
                           original_count, new_count))
          }
        } else {
          message("Warning: Could not resolve unified IDs. Using name-based consolidation.")
          result <- deduplicate_apps_by_name(result)
        }
      } else {
        message("Using name-based consolidation as fallback.")
        result <- tryCatch({
          deduplicate_apps_by_name(result)
        }, error = function(e) {
          message("Error in deduplication: ", e$message)
          result  # Return original on error
        })
      }
    }
    # Note: We already handled hex format duplicates above
  } else if (deduplicate_apps && "unified_app_name" %in% names(result) && os != "unified") {
    # For non-unified OS, use simple name deduplication if requested
    result <- deduplicate_apps_by_name(result)
  }
  
  # --- Standardize Revenue Units ---
  result <- standardize_revenue_units(result, source = "top_charts")
  
  # --- Add platform information ---
  if (nrow(result) > 0) {
    result$platform <- os
    
    # Add app_id and app_id_type columns based on OS parameter
    if ("unified_app_id" %in% names(result)) {
      result$app_id <- result$unified_app_id
      result$app_id_type <- os  # The OS parameter determines the type
    }
    
    # --- Sort by the requested measure ---
    # After enrichment, the relevant columns are renamed
    if (is_active_users) {
      # For DAU/WAU/MAU, sort by the appropriate active user metric
      sort_col <- switch(
        toupper(measure),
        "DAU" = "dau_30d_us",
        "WAU" = "wau_4w_us", 
        "MAU" = "mau_month_us",
        NULL
      )
      # Fallback to users_absolute if enrichment didn't happen
      if (!is.null(sort_col) && sort_col %in% names(result)) {
        result <- result %>% dplyr::arrange(dplyr::desc(!!rlang::sym(sort_col)))
      } else if ("users_absolute" %in% names(result)) {
        result <- result %>% dplyr::arrange(dplyr::desc(users_absolute))
      }
    } else {
      # For sales measures, sort by the appropriate column
      sort_col <- switch(
        toupper(measure),
        "REVENUE" = if ("revenue_30d_ww" %in% names(result)) "revenue_30d_ww" else "revenue_absolute",
        "UNITS" = if ("downloads_30d_ww" %in% names(result)) "downloads_30d_ww" else "units_absolute",
        NULL
      )
      if (!is.null(sort_col) && sort_col %in% names(result)) {
        result <- result %>% dplyr::arrange(dplyr::desc(!!rlang::sym(sort_col)))
      }
    }
  }
  
  return(result)
} 

#' Create FiveThirtyEight-styled GT Dashboard from Top Charts Data
#'
#' @description
#' Creates a professional, FiveThirtyEight-themed GT table dashboard from 
#' Sensor Tower top charts data with customizable styling and metric options.
#'
#' @param data Data frame from st_top_charts() or similar Sensor Tower function
#' @param title Character string for the table title (default: "Top Mobile Games")
#' @param subtitle Character string for subtitle. If NULL, auto-generates based on data
#' @param ranking_metric Character string specifying which metric to use for ranking.
#'   Options: "revenue_180d_ww", "revenue_30d_ww", "downloads_180d_ww", "downloads_30d_ww", etc.
#'   (default: "revenue_180d_ww")
#' @param show_demographics Logical, whether to show demographic columns (age, gender)
#'   (default: TRUE)
#' @param show_engagement Logical, whether to show engagement metrics (DAU, WAU, MAU)
#'   (default: TRUE)
#' @param show_retention Logical, whether to show retention metrics
#'   (default: TRUE)
#' @param retention_region Character string for retention region ("us", "ww", etc.)
#'   (default: "us")
#' @param show_rpd Logical, whether to show Revenue Per Download
#'   (default: TRUE)
#' @param bar_charts Logical, whether to show bar chart visualizations
#'   (default: TRUE)
#' @param bar_chart_columns Character vector of column patterns to add bar charts to.
#'   If NULL, applies to all numeric columns except RPD and retention.
#' @param heatmap_retention Logical, whether to apply heatmap to retention columns
#'   (default: TRUE)
#' @param compact_mode Logical, whether to use compact row heights
#'   (default: TRUE)
#' @param width Numeric, table width in pixels (default: 1800)
#' @param height Numeric, table height in pixels (default: 700)
#' @param save_path Character string, path to save the table image. If NULL, returns GT object
#' @param icon_cache_dir Character string, directory to cache app icons
#'   (default: "inst/images/app_icons")
#' @param raw Logical, whether to return a minimally styled table without 
#'   custom formatting, bar charts, or heatmaps (default: FALSE)
#' @param color_scheme List with color codes for different metric types:
#'   - revenue: Revenue metrics color (default: "#FF6600")
#'   - downloads: Downloads metrics color (default: "#008FD5")
#'   - engagement: Engagement metrics color (default: "#9C27B0")
#'   - rpd: RPD metrics color (default: "#4CAF50")
#'   - retention_low: Low retention color (default: "#FFCDD2")
#'   - retention_mid: Mid retention color (default: "#C8E6C9")
#'   - retention_high: High retention color (default: "#4CAF50")
#'
#' @return GT object (if save_path is NULL) or saves image and returns path
#'
#' @importFrom purrr map2_chr map_chr
#' @importFrom utils data
#' @importFrom stats setNames
#' @importFrom graphics title
#' @export
#'
#' @examples
#' \dontrun{
#' # Basic usage - one line after st_top_charts()
#' top_rpgs <- st_top_charts(category = 7014, measure = "revenue")
#' st_gt_dashboard(top_rpgs)
#' 
#' # Raw mode for minimal styling
#' st_gt_dashboard(top_rpgs, raw = TRUE)
#' 
#' # Customize the dashboard
#' st_gt_dashboard(
#'   top_rpgs,
#'   title = "Top Role-Playing Games Q4 2024",
#'   ranking_metric = "revenue_30d_ww",
#'   show_retention = FALSE,
#'   save_path = "dashboard.png"
#' )
#' 
#' # Change color scheme
#' st_gt_dashboard(
#'   top_rpgs,
#'   color_scheme = list(
#'     revenue = "#E74C3C",
#'     downloads = "#3498DB",
#'     engagement = "#9B59B6"
#'   )
#' )
#' }
st_gt_dashboard <- function(
  data,
  title = "Top Mobile Games",
  subtitle = NULL,
  ranking_metric = "revenue_180d_ww",
  show_demographics = TRUE,
  show_engagement = TRUE,
  show_retention = TRUE,
  retention_region = "us",
  show_rpd = TRUE,
  bar_charts = TRUE,
  bar_chart_columns = NULL,
  heatmap_retention = TRUE,
  compact_mode = TRUE,
  width = 1800,
  height = 700,
  save_path = NULL,
  icon_cache_dir = "inst/images/app_icons",
  raw = FALSE,
  color_scheme = list(
    revenue = "#FF6600",
    downloads = "#008FD5", 
    engagement = "#9C27B0",
    rpd = "#4CAF50",
    retention_low = "#FFCDD2",
    retention_mid = "#C8E6C9",
    retention_high = "#4CAF50"
  )
) {
  
  # Check for required packages
  if (!requireNamespace("gt", quietly = TRUE)) {
    rlang::abort("Package 'gt' is required for this function. Please install it with: install.packages('gt')")
  }
  if (!requireNamespace("gtExtras", quietly = TRUE)) {
    rlang::abort("Package 'gtExtras' is required for this function. Please install it with: install.packages('gtExtras')")
  }
  if (!requireNamespace("stringr", quietly = TRUE)) {
    rlang::abort("Package 'stringr' is required for this function. Please install it with: install.packages('stringr')")
  }
  
  # Override styling options if raw mode
  if (raw) {
    bar_charts <- FALSE
    heatmap_retention <- FALSE
  }
  
  # Helper function to format billions with 2 decimals
  format_billions <- function(val) {
    if (is.na(val)) return("-")
    if (val >= 1e9) {
      paste0(formatC(val/1e9, format = "f", digits = 2), "B")
    } else if (val >= 1e6) {
      paste0(round(val/1e6), "M")
    } else if (val >= 1e3) {
      paste0(round(val/1e3), "K")
    } else {
      as.character(round(val))
    }
  }
  
  # Helper function to format currency with billions
  format_currency_billions <- function(val) {
    if (is.na(val)) return("-")
    if (val >= 1e9) {
      paste0("$", formatC(val/1e9, format = "f", digits = 2), "B")
    } else if (val >= 1e6) {
      paste0("$", round(val/1e6), "M")
    } else if (val >= 1e3) {
      paste0("$", round(val/1e3), "K")
    } else {
      paste0("$", round(val))
    }
  }
  
  # Helper function to create bar chart HTML
  create_bar_chart <- function(values, labels, color, format_fn = NULL, use_labels = FALSE) {
    max_val <- max(values, na.rm = TRUE)
    
    purrr::map2_chr(values, labels, function(val, label) {
      if (is.na(val)) return("-")
      bar_width <- (val / max_val) * 100
      
      # Use provided labels or format the value
      if (use_labels) {
        formatted_label <- label
      } else if (!is.null(format_fn)) {
        formatted_label <- format_fn(val)
      } else {
        formatted_label <- as.character(val)
      }
      
      sprintf(
        '<div style="display: flex; align-items: center; gap: 8px;">
          <div style="flex: 0 0 60px; height: 15px; background: #E1E1E1; position: relative; border-radius: 2px;">
            <div style="height: 100%%; width: %s%%; background: %s; border-radius: 2px;"></div>
          </div>
          <span style="font-size: 12px; color: #333; font-weight: 500;">%s</span>
        </div>',
        round(bar_width),
        color,
        formatted_label
      )
    })
  }
  
  # Country to flag emoji mapping
  country_to_flag <- function(country_code) {
    if (is.na(country_code) || country_code == "") return("")
    
    country_code <- toupper(country_code)
    
    # Using ASCII-only country codes instead of flag emojis
    flags <- c(
      "US" = "[US]", "GB" = "[GB]", "CA" = "[CA]", "AU" = "[AU]", "DE" = "[DE]",
      "FR" = "[FR]", "IT" = "[IT]", "ES" = "[ES]", "NL" = "[NL]", "SE" = "[SE]",
      "NO" = "[NO]", "DK" = "[DK]", "FI" = "[FI]", "PL" = "[PL]", "RU" = "[RU]",
      "BR" = "[BR]", "MX" = "[MX]", "AR" = "[AR]", "CL" = "[CL]", "CO" = "[CO]",
      "JP" = "[JP]", "KR" = "[KR]", "CN" = "[CN]", "TW" = "[TW]", "HK" = "[HK]",
      "SG" = "[SG]", "MY" = "[MY]", "TH" = "[TH]", "ID" = "[ID]", "PH" = "[PH]",
      "VN" = "[VN]", "IN" = "[IN]", "PK" = "[PK]", "BD" = "[BD]", "LK" = "[LK]",
      "AE" = "[AE]", "SA" = "[SA]", "EG" = "[EG]", "IL" = "[IL]", "TR" = "[TR]",
      "ZA" = "[ZA]", "NG" = "[NG]", "KE" = "[KE]", "GH" = "[GH]", "MA" = "[MA]",
      "PT" = "[PT]", "GR" = "[GR]", "CH" = "[CH]", "AT" = "[AT]", "BE" = "[BE]",
      "CZ" = "[CZ]", "HU" = "[HU]", "RO" = "[RO]", "BG" = "[BG]", "SK" = "[SK]",
      "HR" = "[HR]", "SI" = "[SI]", "LT" = "[LT]", "LV" = "[LV]", "EE" = "[EE]",
      "IE" = "[IE]", "NZ" = "[NZ]", "PE" = "[PE]", "VE" = "[VE]", "EC" = "[EC]",
      "UY" = "[UY]", "PY" = "[PY]", "BO" = "[BO]", "CR" = "[CR]", "PA" = "[PA]",
      "GT" = "[GT]", "HN" = "[HN]", "SV" = "[SV]", "NI" = "[NI]", "DO" = "[DO]",
      "JM" = "[JM]", "TT" = "[TT]", "CU" = "[CU]", "PR" = "[PR]", "LU" = "[LU]",
      "IS" = "[IS]", "MT" = "[MT]", "CY" = "[CY]", "AL" = "[AL]", "MK" = "[MK]",
      "RS" = "[RS]", "BA" = "[BA]", "ME" = "[ME]", "XK" = "[XK]", "MD" = "[MD]",
      "UA" = "[UA]", "BY" = "[BY]", "AM" = "[AM]", "GE" = "[GE]", "AZ" = "[AZ]",
      "KZ" = "[KZ]", "UZ" = "[UZ]", "TM" = "[TM]", "KG" = "[KG]", "TJ" = "[TJ]"
    )
    
    if (country_code %in% names(flags)) {
      return(paste0(flags[country_code], " ", country_code))
    } else {
      return(country_code)
    }
  }
  
  # Prepare data
  table_data <- data %>%
    arrange(desc(!!sym(ranking_metric))) %>%
    mutate(
      rank = row_number(),
      # Map gender column if it exists
      genders_ww = if ("entities.custom_tags.Genders (Last Quarter, WW)" %in% names(.)) {
        `entities.custom_tags.Genders (Last Quarter, WW)`
      } else {
        NA_character_
      },
      # Calculate years since launch
      years_since_launch = case_when(
        !is.na(release_date_ww) ~ {
          days_diff <- as.numeric(difftime(Sys.Date(), release_date_ww, units = "days"))
          years <- floor(days_diff / 365.25)
          months <- floor((days_diff %% 365.25) / 30.44)
          case_when(
            years > 0 & months > 0 ~ paste0(years, "y ", months, "m"),
            years > 0 ~ paste0(years, "y"),
            months > 0 ~ paste0(months, "m"),
            TRUE ~ "<1m"
          )
        },
        TRUE ~ "-"
      )
    )
  
  # Build column selection dynamically
  selected_cols <- c("rank", "unified_app_name")
  
  # Add ranking metric
  selected_cols <- c(selected_cols, ranking_metric)
  
  # Add demographics if requested
  if (show_demographics) {
    demo_cols <- c("age_ww", "genders_ww", "years_since_launch")
    
    # Check for country column variants
    country_col <- if ("entities.custom_tags.Most Popular Country by Revenue" %in% names(table_data)) {
      "entities.custom_tags.Most Popular Country by Revenue"
    } else if ("most_popular_country_revenue" %in% names(table_data)) {
      "most_popular_country_revenue"
    } else {
      NULL
    }
    
    if (!is.null(country_col)) demo_cols <- c(demo_cols, country_col)
    
    selected_cols <- c(selected_cols, intersect(names(table_data), demo_cols))
  }
  
  # Add revenue columns
  revenue_cols <- c("revenue_30d_ww", "revenue_alltime_ww")
  if (show_rpd) revenue_cols <- c(revenue_cols, "rpd_alltime_ww")
  selected_cols <- c(selected_cols, intersect(names(table_data), revenue_cols))
  
  # Add engagement if requested
  if (show_engagement) {
    engagement_cols <- c("dau_30d_ww", "wau_4w_ww", "mau_month_ww")
    selected_cols <- c(selected_cols, intersect(names(table_data), engagement_cols))
  }
  
  # Add downloads
  downloads_cols <- c("downloads_30d_ww", "downloads_180d_ww", "downloads_alltime_ww")
  selected_cols <- c(selected_cols, intersect(names(table_data), downloads_cols))
  
  # Add retention if requested
  available_retention <- character(0)  # Initialize for later use
  if (show_retention) {
    retention_cols <- paste0("retention_", c("1d", "7d", "30d", "60d"), "_", retention_region)
    available_retention <- intersect(names(table_data), retention_cols)
    selected_cols <- c(selected_cols, available_retention)
    
    # Retention values should already be in percentage format (0-100)
    # No conversion needed
  }
  
  # Select final columns
  table_data <- table_data %>%
    select(all_of(selected_cols))
  
  # Auto-generate subtitle if not provided
  if (is.null(subtitle)) {
    date_range <- paste0(
      format(Sys.Date() - 180, "%b %Y"), "-",
      format(Sys.Date(), "%b %Y")
    )
    
    ranking_label <- case_when(
      grepl("revenue_180d", ranking_metric) ~ "180-day revenue",
      grepl("revenue_30d", ranking_metric) ~ "30-day revenue", 
      grepl("revenue_alltime", ranking_metric) ~ "lifetime revenue",
      grepl("downloads_180d", ranking_metric) ~ "180-day downloads",
      grepl("downloads_30d", ranking_metric) ~ "30-day downloads",
      grepl("downloads_alltime", ranking_metric) ~ "lifetime downloads",
      TRUE ~ ranking_metric
    )
    
    subtitle <- paste0("Ranked by ", ranking_label, " (", date_range, ")")
  }
  
  # Create GT table
  gt_table <- table_data %>%
    gt::gt()
  
  # Apply theme based on raw parameter
  if (!raw) {
    gt_table <- gt_table %>%
      gtExtras::gt_theme_538()
  }
  
  # Add header
  gt_table <- gt_table %>%
    gt::tab_header(
      title = title,
      subtitle = subtitle
    )
  
  # Apply compact mode if requested
  if (compact_mode && !raw) {
    gt_table <- gt_table %>%
      gt::tab_options(
        data_row.padding = gt::px(2),
        row.striping.include_table_body = FALSE
      )
  }
  
  # Center all column headers
  gt_table <- gt_table %>%
    gt::tab_style(
      style = gt::cell_text(align = "center"),
      locations = gt::cells_column_labels(everything())
    )
  
  # Build column labels dynamically
  col_labels <- list(
    rank = "",
    unified_app_name = "Game",
    age_ww = "Avg Age",
    genders_ww = "Gender",
    years_since_launch = "Years Live",
    revenue_30d_ww = "30-Day Revenue",
    revenue_alltime_ww = "Lifetime Revenue", 
    rpd_alltime_ww = "Revenue per Download",
    downloads_30d_ww = "30-Day Downloads",
    downloads_180d_ww = "180-Day Downloads",
    downloads_alltime_ww = "Lifetime Downloads",
    dau_30d_ww = "DAU (30d avg)",
    wau_4w_ww = "WAU (4w avg)",
    mau_month_ww = "MAU (monthly avg)"
  )
  
  # Add retention labels
  if (show_retention) {
    retention_labels <- list(
      "retention_1d" = "Day 1",
      "retention_7d" = "Day 7",
      "retention_30d" = "Day 30",
      "retention_60d" = "Day 60"
    )
    
    for (ret_type in names(retention_labels)) {
      col_name <- paste0(ret_type, "_", retention_region)
      if (col_name %in% names(table_data)) {
        col_labels[[col_name]] <- retention_labels[[ret_type]]
      }
    }
  }
  
  # Add ranking metric label
  col_labels[[ranking_metric]] <- if (raw) {
    "Ranking Metric"
  } else {
    paste0(gsub("_", " ", gsub("_ww$", "", ranking_metric)), " (Ranking KPI)")
  }
  
  # Add country column label
  if ("entities.custom_tags.Most Popular Country by Revenue" %in% names(table_data)) {
    col_labels[["entities.custom_tags.Most Popular Country by Revenue"]] <- "Top Country (Rev)"
  }
  if ("most_popular_country_revenue" %in% names(table_data)) {
    col_labels[["most_popular_country_revenue"]] <- "Top Country (Rev)"
  }
  
  # Filter col_labels to only include columns that exist in the table
  existing_cols <- names(table_data)
  col_labels <- col_labels[names(col_labels) %in% existing_cols]
  
  # Apply column labels
  gt_table <- gt_table %>%
    gt::cols_label(!!!col_labels)
  
  # Add column groups (spanners)
  if (!raw || ranking_metric %in% names(table_data)) {
    gt_table <- gt_table %>%
      gt::tab_spanner(
        label = "RANKING KPI",
        columns = all_of(ranking_metric)
      )
  }
  
  if (show_demographics && any(c("age_ww", "genders_ww", "years_since_launch") %in% names(table_data))) {
    demo_cols_present <- intersect(names(table_data), 
                                   c("age_ww", "genders_ww", "years_since_launch",
                                     "entities.custom_tags.Most Popular Country by Revenue",
                                     "most_popular_country_revenue"))
    gt_table <- gt_table %>%
      gt::tab_spanner(
        label = "DEMOGRAPHICS", 
        columns = all_of(demo_cols_present)
      )
  }
  
  revenue_cols_present <- intersect(names(table_data), 
                                    c("revenue_30d_ww", "revenue_alltime_ww", "rpd_alltime_ww"))
  if (length(revenue_cols_present) > 0) {
    gt_table <- gt_table %>%
      gt::tab_spanner(
        label = "REVENUE",
        columns = all_of(revenue_cols_present)
      )
  }
  
  if (show_engagement) {
    engagement_cols_present <- intersect(names(table_data),
                                         c("dau_30d_ww", "wau_4w_ww", "mau_month_ww"))
    if (length(engagement_cols_present) > 0) {
      gt_table <- gt_table %>%
        gt::tab_spanner(
          label = "ENGAGEMENT",
          columns = all_of(engagement_cols_present)
        )
    }
  }
  
  downloads_cols_present <- intersect(names(table_data),
                                      c("downloads_30d_ww", "downloads_180d_ww", "downloads_alltime_ww"))
  if (length(downloads_cols_present) > 0) {
    gt_table <- gt_table %>%
      gt::tab_spanner(
        label = "DOWNLOADS",
        columns = all_of(downloads_cols_present)
      )
  }
  
  if (show_retention && length(available_retention) > 0) {
    gt_table <- gt_table %>%
      gt::tab_spanner(
        label = paste0("RETENTION (", toupper(retention_region), ")"),
        columns = all_of(available_retention)
      )
  }
  
  # Apply number formatting
  gt_table <- gt_table %>%
    gt::fmt_number(
      columns = rank,
      decimals = 0
    ) %>%
    gt::sub_missing(
      columns = everything(),
      missing_text = "-"
    )
  
  # Format numeric columns based on raw parameter
  if (raw) {
    # Simple formatting for raw mode
    gt_table <- gt_table %>%
      gt::fmt_number(
        columns = any_of(c("age_ww", "dau_30d_ww", "wau_4w_ww", "mau_month_ww",
                          "downloads_30d_ww", "downloads_180d_ww", "downloads_alltime_ww")),
        decimals = 0,
        suffixing = TRUE
      ) %>%
      gt::fmt_currency(
        columns = any_of(c(ranking_metric, "revenue_30d_ww", "revenue_alltime_ww")),
        currency = "USD",
        decimals = 0,
        suffixing = TRUE
      ) %>%
      gt::fmt_currency(
        columns = any_of("rpd_alltime_ww"),
        currency = "USD",
        decimals = 2
      ) %>%
      gt::fmt_percent(
        columns = any_of(available_retention),
        decimals = 0,
        scale_values = FALSE
      )
  } else {
    # Full formatting with bar charts and custom styling
    
    # Format engagement metrics
    gt_table <- gt_table %>%
      gt::fmt_number(
        columns = any_of(c("age_ww", "dau_30d_ww", "wau_4w_ww", "mau_month_ww")),
        decimals = 0,
        suffixing = TRUE
      ) %>%
      gt::fmt_currency(
        columns = any_of("rpd_alltime_ww"),
        currency = "USD",
        decimals = 2
      ) %>%
      gt::fmt_percent(
        columns = any_of(available_retention),
        decimals = 0,
        scale_values = FALSE
      )
    
    # Apply bar charts if requested
    if (bar_charts) {
      # Revenue columns with bar charts
      if (ranking_metric %in% names(table_data)) {
        # Get original numeric values
        original_ranking_values <- table_data[[ranking_metric]]
        
        gt_table <- gt_table %>%
          gt::text_transform(
            locations = gt::cells_body(columns = all_of(ranking_metric)),
            fn = function(x) {
              create_bar_chart(
                original_ranking_values, 
                x, 
                color_scheme$revenue,
                format_currency_billions
              )
            }
          )
      }
      
      if ("revenue_30d_ww" %in% names(table_data)) {
        gt_table <- gt_table %>%
          gt::text_transform(
            locations = gt::cells_body(columns = revenue_30d_ww),
            fn = function(x) {
              create_bar_chart(
                table_data$revenue_30d_ww,
                x,
                color_scheme$revenue,
                format_currency_billions
              )
            }
          )
      }
      
      if ("revenue_alltime_ww" %in% names(table_data)) {
        gt_table <- gt_table %>%
          gt::text_transform(
            locations = gt::cells_body(columns = revenue_alltime_ww),
            fn = function(x) {
              create_bar_chart(
                table_data$revenue_alltime_ww,
                x,
                color_scheme$revenue,
                format_currency_billions
              )
            }
          )
      }
      
      # Downloads columns with bar charts
      for (dl_col in c("downloads_30d_ww", "downloads_180d_ww", "downloads_alltime_ww")) {
        if (dl_col %in% names(table_data)) {
          gt_table <- gt_table %>%
            gt::text_transform(
              locations = gt::cells_body(columns = all_of(dl_col)),
              fn = function(x) {
                create_bar_chart(
                  table_data[[dl_col]],
                  x,
                  color_scheme$downloads,
                  format_billions
                )
              }
            )
        }
      }
      
      # Engagement columns with bar charts
      for (eng_col in c("dau_30d_ww", "wau_4w_ww", "mau_month_ww")) {
        if (eng_col %in% names(table_data)) {
          # Get the original numeric values before formatting
          original_values <- table_data[[eng_col]]
          
          gt_table <- gt_table %>%
            gt::text_transform(
              locations = gt::cells_body(columns = all_of(eng_col)),
              fn = function(x) {
                # x contains the formatted values (e.g., "779M", "2.1B")
                # Use the formatted values as labels
                create_bar_chart(
                  original_values, 
                  x,  # Use the pre-formatted values as labels
                  color_scheme$engagement,
                  use_labels = TRUE  # Use the labels from x
                )
              }
            )
        }
      }
    }
    
    # Apply retention heatmaps if requested
    if (heatmap_retention && show_retention && length(available_retention) > 0) {
      for (ret_col in available_retention) {
        gt_table <- gt_table %>%
          gt::data_color(
            columns = all_of(ret_col),
            method = "numeric",
            palette = c(color_scheme$retention_low, 
                       color_scheme$retention_mid,
                       color_scheme$retention_high),
            domain = c(0, 100),
            na_color = "#F5F5F5"
          )
      }
    }
    
    # Format gender column with custom HTML
    if ("genders_ww" %in% names(table_data)) {
      gt_table <- gt_table %>%
        text_transform(
          locations = cells_body(columns = genders_ww),
          fn = function(x) {
            purrr::map_chr(x, function(gender) {
              if (is.na(gender) || gender == "") return("-")
              
              # Parse the gender percentage
              male_match <- regexpr("[0-9]+(?=% Male)", gender, perl = TRUE)
              if (male_match > 0) {
                male_pct <- as.numeric(regmatches(gender, male_match))
              } else {
                male_pct <- as.numeric(gsub("[^0-9]", "", gender))
              }
              
              female_pct <- 100 - male_pct
              
              sprintf(
                '<div style="font-size: 12px; line-height: 1.3;">
                  <div style="color: #4A90E2; font-weight: 500;">\u2642 %d%%</div>
                  <div style="color: #E91E63; font-weight: 500;">\u2640 %d%%</div>
                </div>',
                male_pct,
                female_pct
              )
            })
          }
        )
    }
    
    # Add country flags
    country_cols <- c("entities.custom_tags.Most Popular Country by Revenue", 
                      "most_popular_country_revenue")
    country_col <- intersect(names(table_data), country_cols)
    
    if (length(country_col) > 0) {
      gt_table <- gt_table %>%
        text_transform(
          locations = cells_body(columns = all_of(country_col)),
          fn = function(x) {
            purrr::map_chr(x, function(code) {
              flag_text <- country_to_flag(code)
              if (flag_text != "") {
                sprintf('<span style="font-size: 14px;">%s</span>', flag_text)
              } else {
                "-"
              }
            })
          }
        )
    }
    
    # Style the rank column
    gt_table <- gt_table %>%
      gt::tab_style(
        style = list(
          cell_text(
            size = px(20),
            weight = "bold",
            color = "#666666"
          )
        ),
        locations = cells_body(columns = rank)
      )
    
    # Add medal emojis to top 3
    gt_table <- gt_table %>%
      text_transform(
        locations = cells_body(columns = rank, rows = 1),
        fn = function(x) paste0(x, " \U0001f947")
      ) %>%
      text_transform(
        locations = cells_body(columns = rank, rows = 2),
        fn = function(x) paste0(x, " \U0001f948")
      ) %>%
      text_transform(
        locations = cells_body(columns = rank, rows = 3),
        fn = function(x) paste0(x, " \U0001f949")
      )
    
    # Highlight top 3 games
    gt_table <- gt_table %>%
      gt::tab_style(
        style = list(
          gt::cell_fill(color = "#F2F2F2"),
          cell_text(weight = "bold")
        ),
        locations = cells_body(
          rows = 1:3,
          columns = unified_app_name
        )
      )
    
    # Style game names
    gt_table <- gt_table %>%
      gt::tab_style(
        style = list(
          cell_text(
            size = px(14),
            weight = "600"
          )
        ),
        locations = cells_body(columns = unified_app_name)
      )
  }
  
  # Add footnotes
  if (!raw) {
    gt_table <- gt_table %>%
      gt::tab_footnote(
        footnote = "Data represents worldwide metrics unless otherwise specified",
        locations = gt::cells_column_labels(columns = all_of(ranking_metric))
      )
    
    # Check if we have engagement columns to add footnote
    engagement_cols_in_table <- intersect(names(table_data), 
                                         c("dau_30d_ww", "wau_4w_ww", "mau_month_ww"))
    if (show_engagement && length(engagement_cols_in_table) > 0) {
      gt_table <- gt_table %>%
        gt::tab_footnote(
          footnote = "All engagement metrics show average values over their respective time periods",
          locations = gt::cells_column_spanners(spanners = "ENGAGEMENT")
        )
    }
    
    gt_table <- gt_table %>%
      gt::tab_source_note(
        source_note = gt::md("**Source:** Sensor Tower Store Intelligence")
      )
  }
  
  # Save or return
  if (!is.null(save_path)) {
    gt_table %>%
      gt::gtsave(
        filename = save_path,
        vwidth = width,
        vheight = height
      )
    message("Dashboard saved to: ", save_path)
    return(invisible(save_path))
  } else {
    return(gt_table)
  }
}

#' Look up app information by any ID type
#'
#' This function looks up app information using any type of app ID - unified,
#' iOS, or Android. It returns the unified ID and platform-specific IDs that 
#' can be used with other API functions.
#' 
#' The function automatically detects the ID type:
#' - 24-character hex strings are treated as unified IDs
#' - Numeric strings are treated as iOS app IDs
#' - Strings starting with com/net/org/io are treated as Android package names
#'
#' @param app_id Character string. Can be:
#'   - Sensor Tower unified app ID (24-char hex like "5ba4585f539ce75b97db6bcb")
#'   - iOS app ID (numeric like "943599237")
#'   - Android package name (like "com.bandainamcogames.dbzdokkanww")
#' @param auth_token Character string. Your Sensor Tower API authentication token.
#' @param verbose Logical. Whether to show progress messages. Default is FALSE.
#'
#' @return A list with components:
#'   - `unified_app_id`: The Sensor Tower unified app ID
#'   - `ios_app_id`: iOS app ID if found
#'   - `android_app_id`: Android app ID if found
#'   - `app_name`: App name if found
#'   - `publisher_name`: Publisher name if found
#'   Returns NULL if app cannot be found.
#'
#' @examples
#' \dontrun{
#' # Look up Star Trek Fleet Command
#' app_ids <- st_app_lookup("5ba4585f539ce75b97db6bcb")
#' 
#' # Use the IDs with st_ytd_metrics
#' if (!is.null(app_ids)) {
#'   metrics <- st_ytd_metrics(
#'     ios_app_id = app_ids$ios_app_id,
#'     android_app_id = app_ids$android_app_id,
#'     years = 2025,
#'     metrics = "revenue",
#'     countries = "WW"
#'   )
#' }
#' }
#' 
#' @keywords internal
#' @noRd
st_app_lookup_impl <- function(app_id, 
                         auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                         verbose = FALSE) {
  
  auth_token <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token is required. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )
  
  # Detect the type of ID
  id_type <- NULL
  if (grepl("^[a-f0-9]{24}$", app_id)) {
    id_type <- "unified"
  } else if (grepl("^\\d+$", app_id)) {
    id_type <- "ios"
  } else if (grepl("^(com|net|org|io)\\.", app_id)) {
    id_type <- "android"
  } else {
    rlang::abort(paste0(
      "Invalid app ID format. Expected one of:\n",
      "- 24-character hex ID (unified): '5ba4585f539ce75b97db6bcb'\n",
      "- iOS numeric ID: '1234567890'\n",
      "- Android package name: 'com.example.app'\n",
      "Got: '", app_id, "'"
    ))
  }
  
  if (verbose) message("Looking up ", id_type, " ID: ", app_id)
  
  # For unified IDs, use direct unified apps lookup first.
  if (id_type == "unified") {
    direct_details <- tryCatch({
      st_app_details_impl(
        app_ids = app_id,
        os = "unified",
        include_developer_contacts = FALSE,
        auth_token = auth_token
      )
    }, error = function(e) {
      if (verbose) message("Direct unified lookup failed: ", e$message)
      NULL
    })

    if (!is.null(direct_details) && nrow(direct_details) > 0) {
      ios_id <- if ("ios_app_id" %in% names(direct_details)) as.character(direct_details$ios_app_id[1]) else NULL
      android_id <- if ("android_app_id" %in% names(direct_details)) as.character(direct_details$android_app_id[1]) else NULL
      app_name <- if ("app_name" %in% names(direct_details)) direct_details$app_name[1] else NA_character_
      publisher_name <- if ("publisher_name" %in% names(direct_details)) direct_details$publisher_name[1] else NA_character_

      if (verbose) {
        message("Found app via unified apps endpoint: ", app_name)
        message("iOS ID: ", ios_id %||% "none")
        message("Android ID: ", android_id %||% "none")
      }

      return(list(
        unified_app_id = app_id,
        ios_app_id = ios_id,
        android_app_id = android_id,
        app_name = app_name,
        publisher_name = publisher_name
      ))
    }

    # Fallback to search-based lookup for backward compatibility.
    search_results <- tryCatch({
      st_app_info_impl(
        term = app_id,
        return_all_fields = TRUE,
        limit = 1,
        auth_token = auth_token
      )
    }, error = function(e) {
      if (verbose) message("Search by ID failed: ", e$message)
      NULL
    })
    
    if (!is.null(search_results) && nrow(search_results) > 0 && 
        search_results$app_id[1] == app_id) {
      # Found the app by its unified ID
      app_name <- search_results$name[1]
      unified_id <- app_id
      
      # Extract platform IDs
      ios_id <- NULL
      android_id <- NULL
      
      if ("ios_apps" %in% names(search_results) && length(search_results$ios_apps[[1]]) > 0) {
        ios_id <- search_results$ios_apps[[1]]$app_id[1]
      }
      
      if ("android_apps" %in% names(search_results) && length(search_results$android_apps[[1]]) > 0) {
        android_id <- search_results$android_apps[[1]]$app_id[1]
      }
      
      if (verbose) {
        message("Found app: ", app_name)
        message("iOS ID: ", ios_id %||% "none")
        message("Android ID: ", android_id %||% "none")
      }
      
      return(list(
        unified_app_id = unified_id,
        ios_app_id = ios_id,
        android_app_id = android_id,
        app_name = app_name,
        publisher_name = search_results$publisher_name[1] %||% NA
      ))
    }
  } else {
    # For platform IDs, search by app name to find the unified ID
    # First, try to find apps that contain this platform ID
    
    # Determine search app store
    search_store <- if (id_type == "ios") "unified" else "unified"
    
    # Search more broadly - the ID might not be searchable
    # So we'll need to look through results manually
    search_results <- NULL
    
    # Try different search strategies
    search_terms <- c()
    
    # For iOS IDs, we can't reliably search by the ID itself
    # For Android IDs, try the package name parts
    if (id_type == "android") {
      # Extract app name from package
      parts <- strsplit(app_id, "\\.")[[1]]
      if (length(parts) > 2) {
        # Try the last part (often the app name)
        search_terms <- c(parts[length(parts)], paste(parts[(length(parts)-1):length(parts)], collapse = " "))
      }
    }
    
    # If we have search terms, try them
    for (term in search_terms) {
      if (verbose) message("Searching for: ", term)
      
      search_results <- tryCatch({
        st_app_info_impl(
          term = term,
          app_store = search_store,
          return_all_fields = TRUE,
          limit = 50,
          auth_token = auth_token
        )
      }, error = function(e) {
        if (verbose) message("Search failed: ", e$message)
        NULL
      })
      
      if (!is.null(search_results) && nrow(search_results) > 0) {
        # Look through results for our platform ID
        for (i in 1:nrow(search_results)) {
          found <- FALSE
          
          if (id_type == "ios" && "ios_apps" %in% names(search_results)) {
            ios_apps <- search_results$ios_apps[[i]]
            if (!is.null(ios_apps) && is.data.frame(ios_apps) && app_id %in% ios_apps$app_id) {
              found <- TRUE
            }
          } else if (id_type == "android" && "android_apps" %in% names(search_results)) {
            android_apps <- search_results$android_apps[[i]]
            if (!is.null(android_apps) && is.data.frame(android_apps) && app_id %in% android_apps$app_id) {
              found <- TRUE
            }
          }
          
          if (found) {
            # Extract all the information
            unified_id <- search_results$app_id[i]
            app_name <- search_results$name[i]
            publisher_name <- search_results$publisher_name[i]
            
            # Get platform IDs
            ios_id <- NULL
            android_id <- NULL
            
            if ("ios_apps" %in% names(search_results) && !is.null(search_results$ios_apps[[i]])) {
              ios_apps <- search_results$ios_apps[[i]]
              if (is.data.frame(ios_apps) && nrow(ios_apps) > 0) {
                ios_id <- ios_apps$app_id[1]
              }
            }
            
            if ("android_apps" %in% names(search_results) && !is.null(search_results$android_apps[[i]])) {
              android_apps <- search_results$android_apps[[i]]
              if (is.data.frame(android_apps) && nrow(android_apps) > 0) {
                android_id <- android_apps$app_id[1]
              }
            }
            
            if (verbose) {
              message("Found app: ", app_name)
              message("Unified ID: ", unified_id)
              message("iOS ID: ", ios_id %||% "none")
              message("Android ID: ", android_id %||% "none")
            }
            
            return(list(
              unified_app_id = unified_id,
              ios_app_id = ios_id,
              android_app_id = android_id,
              app_name = app_name,
              publisher_name = publisher_name %||% NA
            ))
          }
        }
      }
    }
  }
  
  # If we couldn't find the app
  if (verbose) message("Could not find app with ", id_type, " ID: ", app_id)
  
  return(NULL)
}

# Helper operator
`%||%` <- function(x, y) if (is.null(x)) y else x

#' Batch Fetch Metrics for Multiple Apps
#'
#' Efficiently fetch metrics for multiple apps by batching API calls and
#' automatically handling platform-specific requirements. The OS parameter
#' controls which platform's data is returned for all apps.
#'
#' @param os Character. Required. Operating system: "ios", "android", or "unified".
#'   This determines which platform's data is returned for all apps.
#' @param app_list List or data frame containing app information. Can be:
#'   - Character vector of app IDs
#'   - Data frame with columns: app_id, app_name (optional), platform (optional)
#'   - List of lists with app_id and optional metadata
#' @param metrics Character vector. Metrics to fetch. Supported values:
#'   - "revenue" - App revenue estimates
#'   - "downloads" - App download estimates
#'   - "dau" - Daily Active Users
#'   - "wau" - Weekly Active Users
#'   - "mau" - Monthly Active Users
#' @param date_range List with start_date and end_date, or "ytd" for year-to-date
#' @param countries Character vector. Country codes. Required.
#' @param granularity Character. Date granularity (default "monthly")
#' @param parallel Logical. Use parallel processing (default FALSE)
#' @param cache_dir Character. Directory for caching results (optional)
#' @param verbose Logical. Show progress messages (default TRUE)
#' @param auth_token Character string. Your Sensor Tower API authentication token.
#' @param max_cores Integer. Maximum number of cores to use for parallel processing.
#' @param max_concurrent_requests Integer. Max concurrent requests (deprecated/unused).
#' @param retry Logical. Whether to retry failed requests.
#' @param max_retries Integer. Max retries.
#' @param publisher_ids Character vector. Publisher IDs to fetch data for.
#'
#' @return A tibble with all metrics for all apps.
#' @keywords internal
#' @noRd
st_batch_metrics_impl <- function(os,
                             app_list,
                             metrics = c("revenue", "downloads"),
                             date_range = list(
                               start_date = Sys.Date() - 90,
                               end_date = Sys.Date() - 1
                             ),
                             countries,
                             granularity,
                             parallel = FALSE,
                             cache_dir = NULL,
                             verbose = TRUE,
                             auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
                             max_cores = 2,
                             max_concurrent_requests = 2,
                             retry = TRUE,
                             max_retries = 3,
                             publisher_ids = NULL) {
  # Validate OS parameter
  if (missing(os) || is.null(os) || !os %in% c("ios", "android", "unified")) {
    rlang::abort("'os' parameter is required and must be one of: 'ios', 'android', or 'unified'")
  }

  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token is required. Set SENSORTOWER_AUTH_TOKEN environment variable."
  )

  # Validate countries parameter
  if (missing(countries) || is.null(countries)) {
    rlang::abort("'countries' parameter is required")
  }

  # Validate granularity explicitly
  if (missing(granularity) || is.null(granularity)) {
    rlang::abort("'granularity' parameter is required. Specify one of: 'daily', 'weekly', 'monthly', 'quarterly'.")
  }

  # Validate date range
  if (identical(date_range, "ytd")) {
    rlang::abort("YTD mode is no longer supported in st_batch_metrics_impl. Please specify explicit start_date and end_date.")
  }

  # Publisher mode
  if (!is.null(publisher_ids)) {
    return(fetch_publisher_batch(publisher_ids, os, metrics, countries, date_range, granularity, auth_token_val, verbose))
  }

  # Step 1: Normalize app list
  app_count <- if (is.data.frame(app_list)) nrow(app_list) else length(app_list)
  if (verbose) message("Processing ", app_count, " apps...")
  apps_df <- normalize_app_list(app_list, os, auth_token_val, verbose)

  if (verbose) {
    message("Resolved apps:")
    message("  iOS apps: ", sum(!is.na(apps_df$ios_id)))
    message("  Android apps: ", sum(!is.na(apps_df$android_id)))
    message("  Unified only: ", sum(is.na(apps_df$ios_id) & is.na(apps_df$android_id)))
  }

  # Step 2: Group apps by fetch strategy
  fetch_groups <- group_apps_by_platform(apps_df)

  # Step 3: Define worker function
  fetch_worker <- function(group, group_name) {
    if (verbose) message("\nFetching metrics for ", group_name, " group (", nrow(group), " apps)...")

    # Fetch Active Users
    active_res <- fetch_active_users_batch(group, group_name, metrics, countries, date_range, granularity, auth_token_val, verbose)

    # Fetch Sales (Revenue/Downloads)
    sales_res <- fetch_sales_batch(group, group_name, metrics, countries, date_range, granularity, auth_token_val, verbose, os)

    # Combine results
    if (verbose) {
      if (!is.null(active_res) && nrow(active_res) > 0) {
        message("Active res app_id type: ", typeof(active_res$app_id), " class: ", class(active_res$app_id))
      }
      if (!is.null(sales_res) && nrow(sales_res) > 0) {
        message("Sales res app_id type: ", typeof(sales_res$app_id), " class: ", class(sales_res$app_id))
      }
    }

    # Ensure app_id is character in both
    if (!is.null(active_res) && nrow(active_res) > 0 && "app_id" %in% names(active_res)) {
      active_res$app_id <- as.character(active_res$app_id)
    }
    if (!is.null(sales_res) && nrow(sales_res) > 0 && "app_id" %in% names(sales_res)) {
      sales_res$app_id <- as.character(sales_res$app_id)
    }

    dplyr::bind_rows(active_res, sales_res)
  }

  # Step 4: Execute fetches
  results <- execute_batch_fetch(fetch_groups, fetch_worker, parallel, max_cores, verbose)

  # Step 5: Finalize results
  finalize_batch_results_internal(results, apps_df, os = os)
}

# --- Helper Functions ---

normalize_app_list <- function(app_list, os, auth_token, verbose) {
  # Convert input to data frame
  if (is.vector(app_list) && is.character(app_list)) {
    apps_df <- tibble::tibble(app_id = app_list)
  } else if (is.data.frame(app_list)) {
    apps_df <- app_list
  } else if (is.list(app_list)) {
    # Handle list of lists or list of named vectors
    apps_df <- dplyr::bind_rows(app_list)
  } else {
    rlang::abort("Invalid app_list format. Must be character vector, data frame, or list.")
  }

  # Ensure app_id column exists
  if (!"app_id" %in% names(apps_df)) {
    rlang::abort("app_list must contain an 'app_id' column or be a vector of IDs")
  }

  # Add metadata columns if missing
  if (!"app_name" %in% names(apps_df)) apps_df$app_name <- NA_character_
  if (!"unified_id" %in% names(apps_df)) apps_df$unified_id <- NA_character_
  if (!"ios_id" %in% names(apps_df)) apps_df$ios_id <- NA_character_
  if (!"android_id" %in% names(apps_df)) apps_df$android_id <- NA_character_

  # Resolve IDs if needed
  # We need to ensure we have platform-specific IDs for fetching

  # Identify which apps need resolution
  # If we are in unified mode, we need both IDs
  # If we are in ios mode, we need ios_id
  # If we are in android mode, we need android_id

  # First, try to resolve unified IDs for all apps if not present
  missing_unified <- is.na(apps_df$unified_id) & !is.na(apps_df$app_id)
  if (any(missing_unified)) {
    # If app_id looks like a unified ID (no dots, no plain numbers usually, but let's assume input is unified if not specified)
    # Actually, let's use resolve_app_id to get the mapping

    if (verbose) message("Resolving ", sum(missing_unified), " app IDs...")

    # We can use batch resolution if available, or loop
    # For now, let's loop or use lapply with resolve_app_id which is cached

    resolved_list <- lapply(apps_df$app_id[missing_unified], function(id) {
      resolve_app_id(id, auth_token = auth_token)
    })

    # Update the dataframe
    for (i in seq_along(resolved_list)) {
      res <- resolved_list[[i]]
      idx <- which(missing_unified)[i]

      if (!is.null(res)) {
        if (!is.null(res$unified_app_id)) apps_df$unified_id[idx] <- res$unified_app_id
        if (!is.null(res$ios_app_id)) apps_df$ios_id[idx] <- res$ios_app_id
        if (!is.null(res$android_app_id)) apps_df$android_id[idx] <- res$android_app_id
        if (is.na(apps_df$app_name[idx]) && !is.null(res$name)) apps_df$app_name[idx] <- res$name
      }
    }
  }

  # Handle OS-specific requirements
  if (os == "ios") {
    # If ios_id is missing but we have app_id and it looks like an iOS ID (all digits), use it
    missing_ios <- is.na(apps_df$ios_id) & grepl("^\\d+$", apps_df$app_id)
    apps_df$ios_id[missing_ios] <- apps_df$app_id[missing_ios]
  } else if (os == "android") {
    # If android_id is missing but app_id looks like android (contains dots), use it
    missing_android <- is.na(apps_df$android_id) & grepl("\\.", apps_df$app_id)
    apps_df$android_id[missing_android] <- apps_df$app_id[missing_android]
  }

  apps_df
}

fetch_publisher_batch <- function(publisher_ids, os, metrics, countries, date_range, granularity, auth_token, verbose) {
  active_user_metrics <- intersect(metrics, c("dau", "wau", "mau"))
  if (length(active_user_metrics) > 0 && verbose) {
    message("Active user metrics are not supported for publisher IDs. Skipping: ", paste(active_user_metrics, collapse = ", "))
  }
  revenue_download_metrics <- intersect(metrics, c("revenue", "downloads"))
  if (length(revenue_download_metrics) == 0) {
    return(tibble::tibble())
  }

  # Helper to fetch one publisher for one platform
  fetch_publisher_platform <- function(platform_os, pid) {
    tryCatch(
      {
        st_sales_report_impl(
          os = platform_os,
          countries = countries,
          start_date = as.character(date_range$start_date),
          end_date = as.character(date_range$end_date),
          date_granularity = granularity,
          publisher_ids = pid,
          auth_token = auth_token,
          auto_segment = TRUE,
          verbose = FALSE
        )
      },
      error = function(e) NULL
    )
  }

  results <- lapply(seq_along(publisher_ids), function(i) {
    pid <- publisher_ids[i]
    df <- NULL

    if (os == "ios") {
      df <- fetch_publisher_platform("ios", pid)
    } else if (os == "android") {
      df <- fetch_publisher_platform("android", pid)
    } else {
      ios_res <- fetch_publisher_platform("ios", pid)
      if (!is.null(ios_res) && "app_id" %in% names(ios_res)) ios_res$app_id <- as.character(ios_res$app_id)

      and_res <- fetch_publisher_platform("android", pid)
      if (!is.null(and_res) && "app_id" %in% names(and_res)) and_res$app_id <- as.character(and_res$app_id)

      df <- dplyr::bind_rows(ios_res %||% tibble::tibble(), and_res %||% tibble::tibble())
      if (nrow(df) > 0) {
        df <- df %>%
          dplyr::group_by(.data$date, .data$country) %>%
          dplyr::summarise(
            revenue = sum(.data$revenue, na.rm = TRUE),
            downloads = sum(.data$downloads, na.rm = TRUE),
            .groups = "drop"
          )
      }
    }

    if (is.null(df) || nrow(df) == 0) {
      return(NULL)
    }

    df %>%
      tidyr::pivot_longer(
        cols = dplyr::all_of(intersect(revenue_download_metrics, c("revenue", "downloads"))),
        names_to = "metric",
        values_to = "value"
      ) %>%
      dplyr::mutate(
        original_id = pid,
        app_id = pid,
        app_id_type = "publisher",
        entity_id = pid,
        entity_type = "publisher"
      )
  })

  results <- results[!vapply(results, is.null, logical(1))]
  final <- if (length(results) > 0) dplyr::bind_rows(results) else tibble::tibble()

  if (verbose) {
    message("\nBatch fetch complete (publisher mode)!")
    message("Total records: ", nrow(final))
  }
  return(final)
}

group_apps_by_platform <- function(apps_df) {
  fetch_groups <- list()

  # Group 1: Apps with both iOS and Android IDs
  both_platforms <- apps_df[!is.na(apps_df$ios_id) & !is.na(apps_df$android_id), ]
  if (nrow(both_platforms) > 0) {
    fetch_groups$both <- both_platforms
  }

  # Group 2: iOS only
  ios_only <- apps_df[!is.na(apps_df$ios_id) & is.na(apps_df$android_id), ]
  if (nrow(ios_only) > 0) {
    fetch_groups$ios <- ios_only
  }

  # Group 3: Android only
  android_only <- apps_df[is.na(apps_df$ios_id) & !is.na(apps_df$android_id), ]
  if (nrow(android_only) > 0) {
    fetch_groups$android <- android_only
  }

  # Group 4: Unified only (fallback)
  unified_only <- apps_df[is.na(apps_df$ios_id) & is.na(apps_df$android_id), ]
  if (nrow(unified_only) > 0) {
    fetch_groups$unified <- unified_only
  }

  fetch_groups
}

execute_batch_fetch <- function(fetch_groups, fetch_worker, parallel, max_cores, verbose) {
  can_parallel <- parallel && length(fetch_groups) > 1 &&
    .Platform$OS.type != "windows" && max(1L, as.integer(max_cores)) > 1

  if (can_parallel) {
    if (verbose) message("\nUsing parallel processing...")

    results <- parallel::mclapply(
      names(fetch_groups),
      function(group_name) {
        result <- fetch_worker(fetch_groups[[group_name]], group_name)
        if (!is.null(result) && "app_id" %in% names(result)) {
          result$app_id <- as.character(result$app_id)
        }
        result
      },
      mc.cores = min(length(fetch_groups), max(1L, as.integer(max_cores)))
    )
  } else {
    if (parallel && .Platform$OS.type == "windows" && verbose) {
      message("Parallel processing is not supported on Windows; defaulting to sequential execution.")
    }
    results <- lapply(
      names(fetch_groups),
      function(group_name) {
        result <- fetch_worker(fetch_groups[[group_name]], group_name)
        if (!is.null(result) && "app_id" %in% names(result)) {
          result$app_id <- as.character(result$app_id)
        }
        result
      }
    )
  }

  if (verbose) message("Batch fetch complete. Combining results...")
  res <- Filter(Negate(is.null), results)
  if (verbose) message("Filtered results length: ", length(res))
  res
}

finalize_batch_results <- function(results, apps_df, os = NULL) {
  finalize_batch_results_internal(results, apps_df, os = os)
}

active_users_default_time_period <- function(granularity) {
  period <- dplyr::recode(
    granularity,
    "daily" = "day",
    "weekly" = "week",
    "monthly" = "month",
    "quarterly" = "quarter",
    .default = "month"
  )
  as.character(period)
}

active_users_time_period_for_metric <- function(metric_name, default_time_period) {
  period <- dplyr::recode(
    tolower(metric_name),
    "dau" = "day",
    "wau" = "week",
    "mau" = "month",
    .default = default_time_period
  )
  as.character(period)
}

build_active_users_request_plan <- function(group, group_name, active_user_metrics) {
  if (group_name == "ios") {
    base_plan <- tibble::tibble(
      platform = "ios",
      app_ids = list(unique(as.character(stats::na.omit(group$ios_id))))
    )
  } else if (group_name == "android") {
    base_plan <- tibble::tibble(
      platform = "android",
      app_ids = list(unique(as.character(stats::na.omit(group$android_id))))
    )
  } else if (group_name == "both") {
    base_plan <- tibble::tibble(
      platform = c("ios", "android"),
      app_ids = list(
        unique(as.character(stats::na.omit(group$ios_id))),
        unique(as.character(stats::na.omit(group$android_id)))
      )
    ) %>%
      dplyr::filter(lengths(.data$app_ids) > 0)
  } else {
    base_plan <- tibble::tibble(platform = character(), app_ids = list())
  }

  if (nrow(base_plan) == 0 || length(active_user_metrics) == 0) {
    return(tibble::tibble(platform = character(), app_ids = list(), metric = character()))
  }

  tidyr::crossing(
    base_plan,
    metric = active_user_metrics
  ) %>%
    dplyr::mutate(metric = tolower(.data$metric))
}

fetch_active_users_request <- function(platform_os,
                                       app_ids,
                                       metric_name,
                                       countries,
                                       date_range,
                                       default_time_period,
                                       auth_token,
                                       verbose = FALSE) {
  cleaned_ids <- app_ids[!is.na(app_ids) & nzchar(app_ids)]
  if (length(cleaned_ids) == 0) {
    return(tibble::tibble())
  }

  metric_time_period <- active_users_time_period_for_metric(metric_name, default_time_period)

  params <- list(
    app_ids = paste(cleaned_ids, collapse = ","),
    countries = countries,
    start_date = as.character(date_range$start_date),
    end_date = as.character(date_range$end_date),
    time_period = metric_time_period
  )

  active_processor <- function(resp, ...) {
    body_raw <- httr2::resp_body_raw(resp)
    if (length(body_raw) == 0) {
      return(tibble::tibble())
    }

    body_text <- rawToChar(body_raw)
    if (!nzchar(body_text) || body_text %in% c("[]", "{}")) {
      return(tibble::tibble())
    }

    data <- tryCatch(jsonlite::fromJSON(body_text, flatten = TRUE), error = function(e) NULL)
    if (is.null(data) || length(data) == 0) {
      return(tibble::tibble())
    }

    df <- tryCatch(dplyr::bind_rows(data), error = function(e) NULL)
    if (is.null(df) || nrow(df) == 0) {
      return(tibble::tibble())
    }

    if (!"app_id" %in% names(df) || !"date" %in% names(df)) {
      return(tibble::tibble())
    }

    if (platform_os == "ios") {
      has_split_ios_cols <- "iphone_users" %in% names(df) || "ipad_users" %in% names(df)
      if (has_split_ios_cols) {
        if (!"iphone_users" %in% names(df)) df$iphone_users <- 0
        if (!"ipad_users" %in% names(df)) df$ipad_users <- 0
        df <- df %>%
          dplyr::mutate(
            users = dplyr::coalesce(.data$iphone_users, 0) + dplyr::coalesce(.data$ipad_users, 0),
            .keep = "all"
          )
      } else if ("users" %in% names(df)) {
        df <- df %>% dplyr::mutate(users = dplyr::coalesce(.data$users, 0))
      } else {
        return(tibble::tibble())
      }
    } else {
      if (!"users" %in% names(df) && "android_users" %in% names(df)) {
        df <- df %>% dplyr::rename(users = .data$android_users)
      }
      if (!"users" %in% names(df)) {
        return(tibble::tibble())
      }
    }

    if (!"country" %in% names(df)) {
      df$country <- NA_character_
    }

    df %>%
      dplyr::transmute(
        app_id = as.character(.data$app_id),
        date = as.Date(.data$date),
        country = as.character(.data$country),
        metric = metric_name,
        value = as.numeric(.data$users),
        platform = platform_os
      )
  }

  result <- fetch_data_core(
    endpoint = st_endpoint_relative_path("usage_active_users", os = platform_os),
    params = params,
    auth_token = auth_token,
    processor = active_processor,
    verbose = verbose
  )

  if (is.null(result) || nrow(result) == 0) {
    return(tibble::tibble())
  }

  result
}

fetch_active_users_batch <- function(group, group_name, metrics, countries, date_range, granularity, auth_token, verbose) {
  active_user_metrics <- intersect(metrics, c("dau", "wau", "mau"))
  if (length(active_user_metrics) == 0) {
    return(tibble::tibble())
  }

  request_plan <- build_active_users_request_plan(group, group_name, active_user_metrics)
  if (nrow(request_plan) == 0) {
    return(tibble::tibble())
  }

  if (verbose) {
    message("  Fetching active user metrics (platform endpoints)...")
  }

  default_time_period <- active_users_default_time_period(granularity)

  request_results <- purrr::pmap(
    .l = list(
      platform_os = request_plan$platform,
      app_ids = request_plan$app_ids,
      metric_name = request_plan$metric
    ),
    .f = function(platform_os, app_ids, metric_name) {
      fetch_active_users_request(
        platform_os = platform_os,
        app_ids = app_ids,
        metric_name = metric_name,
        countries = countries,
        date_range = date_range,
        default_time_period = default_time_period,
        auth_token = auth_token,
        verbose = verbose
      )
    }
  )

  platform_results <- dplyr::bind_rows(request_results)
  if (nrow(platform_results) == 0) {
    return(tibble::tibble())
  }

  if (group_name %in% c("ios", "android")) {
    return(platform_results %>%
      dplyr::mutate(entity_id = .data$app_id) %>%
      dplyr::select(-"platform"))
  }

  if (group_name == "both") {
    pair_lookup <- dplyr::bind_rows(
      group %>%
        dplyr::transmute(
          platform = "ios",
          app_id = as.character(.data$ios_id),
          entity_id = paste0(.data$ios_id, "_", .data$android_id)
        ),
      group %>%
        dplyr::transmute(
          platform = "android",
          app_id = as.character(.data$android_id),
          entity_id = paste0(.data$ios_id, "_", .data$android_id)
        )
    ) %>%
      dplyr::filter(!is.na(.data$app_id) & nzchar(.data$app_id)) %>%
      dplyr::distinct(.data$platform, .data$app_id, .data$entity_id)

    return(platform_results %>%
      dplyr::inner_join(pair_lookup, by = c("platform", "app_id")) %>%
      dplyr::group_by(.data$entity_id, .data$date, .data$country, .data$metric) %>%
      dplyr::summarise(value = sum(.data$value, na.rm = TRUE), .groups = "drop") %>%
      dplyr::mutate(app_id = .data$entity_id) %>%
      dplyr::select("app_id", "date", "country", "metric", "value", "entity_id"))
  }

  tibble::tibble()
}

fetch_sales_batch <- function(group, group_name, metrics, countries, date_range, granularity, auth_token, verbose, os) {
  revenue_download_metrics <- intersect(metrics, c("revenue", "downloads"))
  if (length(revenue_download_metrics) == 0) {
    return(tibble::tibble())
  }

  # Helper to process individual results
  process_individual <- function(idx, fetch_func) {
    result <- fetch_func(idx)
    if (is.null(result) || nrow(result) == 0) {
      return(NULL)
    }

    # Standardize columns
    if ("total_revenue" %in% names(result)) result$revenue <- result$total_revenue
    if ("total_downloads" %in% names(result)) result$downloads <- result$total_downloads
    if ("iphone_revenue" %in% names(result) && "ipad_revenue" %in% names(result)) result$revenue <- result$iphone_revenue + result$ipad_revenue
    if ("iphone_downloads" %in% names(result) && "ipad_downloads" %in% names(result)) result$downloads <- result$iphone_downloads + result$ipad_downloads

    if (!"revenue" %in% names(result)) result$revenue <- 0
    if (!"downloads" %in% names(result)) result$downloads <- 0L
    if (!"country" %in% names(result) && "c" %in% names(result)) result$country <- result$c

    result %>%
      dplyr::select(date, country, revenue, downloads) %>%
      tidyr::pivot_longer(
        cols = dplyr::all_of(intersect(revenue_download_metrics, c("revenue", "downloads"))),
        names_to = "metric",
        values_to = "value"
      )
  }

  if (group_name == "both") {
    # Fetch revenue/downloads for both platforms per app and sum
    results_list <- lapply(seq_len(nrow(group)), function(i) {
      ios_res <- tryCatch(
        {
          res <- st_sales_report_impl(
            os = "ios",
            ios_app_id = group$ios_id[i],
            countries = countries,
            start_date = as.character(date_range$start_date),
            end_date = as.character(date_range$end_date),
            date_granularity = granularity,
            auth_token = auth_token,
            auto_segment = TRUE,
            verbose = FALSE
          )
          if (!is.null(res) && "app_id" %in% names(res)) res$app_id <- as.character(res$app_id)
          res
        },
        error = function(e) NULL
      )

      android_res <- tryCatch(
        {
          res <- st_sales_report_impl(
            os = "android",
            android_app_id = group$android_id[i],
            countries = countries,
            start_date = as.character(date_range$start_date),
            end_date = as.character(date_range$end_date),
            date_granularity = granularity,
            auth_token = auth_token,
            auto_segment = TRUE,
            verbose = FALSE
          )
          if (!is.null(res) && "app_id" %in% names(res)) res$app_id <- as.character(res$app_id)
          res
        },
        error = function(e) NULL
      )

      combined <- dplyr::bind_rows(ios_res %||% tibble::tibble(), android_res %||% tibble::tibble())
      if (nrow(combined) == 0) {
        return(NULL)
      }

      combined %>%
        dplyr::group_by(.data$date, .data$country) %>%
        dplyr::summarise(
          revenue = sum(.data$revenue, na.rm = TRUE),
          downloads = sum(.data$downloads, na.rm = TRUE),
          .groups = "drop"
        ) %>%
        tidyr::pivot_longer(cols = dplyr::all_of(intersect(revenue_download_metrics, c("revenue", "downloads"))), names_to = "metric", values_to = "value") %>%
        dplyr::mutate(entity_id = as.character(paste0(group$ios_id[i], "_", group$android_id[i])), app_id = entity_id)
    })
    return(dplyr::bind_rows(results_list))
  } else if (group_name == "ios") {
    results_list <- lapply(seq_len(nrow(group)), function(i) {
      res <- process_individual(i, function(idx) {
        tryCatch(
          {
            st_sales_report_impl(
              os = "ios",
              ios_app_id = group$ios_id[idx],
              countries = countries,
              start_date = as.character(date_range$start_date),
              end_date = as.character(date_range$end_date),
              date_granularity = granularity,
              auth_token = auth_token,
              auto_segment = TRUE,
              verbose = FALSE
            )
          },
          error = function(e) NULL
        )
      })
      if (!is.null(res)) {
        res$entity_id <- as.character(group$ios_id[i])
        res$app_id <- as.character(group$ios_id[i])
      }
      res
    })
    return(dplyr::bind_rows(results_list))
  } else if (group_name == "android") {
    results_list <- lapply(seq_len(nrow(group)), function(i) {
      res <- process_individual(i, function(idx) {
        tryCatch(
          {
            st_sales_report_impl(
              os = "android",
              android_app_id = group$android_id[idx],
              countries = countries,
              start_date = as.character(date_range$start_date),
              end_date = as.character(date_range$end_date),
              date_granularity = granularity,
              auth_token = auth_token,
              auto_segment = TRUE,
              verbose = FALSE
            )
          },
          error = function(e) NULL
        )
      })
      if (!is.null(res)) {
        res$entity_id <- as.character(group$android_id[i])
        res$app_id <- as.character(group$android_id[i])
      }
      res
    })
    return(dplyr::bind_rows(results_list))
  } else if (group_name == "unified") {
    # Fallback to st_metrics for unified IDs
    results_list <- lapply(seq_len(nrow(group)), function(i) {
      # Try to use fetch_unified_data if we have platform IDs resolved
      ios_id <- group$ios_id[i]
      android_id <- group$android_id[i]

      if (!is.na(ios_id) || !is.na(android_id)) {
        res <- fetch_unified_data(
          ios_app_id = if (!is.na(ios_id)) ios_id else NULL,
          android_app_id = if (!is.na(android_id)) android_id else NULL,
          start_date = date_range$start_date,
          end_date = date_range$end_date,
          countries = countries,
          date_granularity = granularity,
          auth_token = auth_token,
          verbose = FALSE,
          combine_to_unified = TRUE
        )
        if (nrow(res) > 0) {
          return(res %>%
            tidyr::pivot_longer(cols = dplyr::all_of(intersect(revenue_download_metrics, c("revenue", "downloads"))), names_to = "metric", values_to = "value") %>%
            dplyr::mutate(entity_id = as.character(group$unified_id[i]), app_id = entity_id))
        }
      }

      # Last resort: st_unified_sales_report_impl with unified ID
      tryCatch(
        {
          res <- st_unified_sales_report_impl(
            unified_app_id   = group$unified_id[i],
            countries        = countries,
            start_date       = as.character(date_range$start_date),
            end_date         = as.character(date_range$end_date),
            date_granularity = granularity,
            auth_token       = auth_token,
            verbose          = FALSE
          )
          if (nrow(res) > 0) {
            res %>%
              tidyr::pivot_longer(cols = dplyr::all_of(intersect(revenue_download_metrics, c("revenue", "downloads"))), names_to = "metric", values_to = "value") %>%
              dplyr::mutate(entity_id = as.character(group$unified_id[i]), app_id = entity_id)
          } else {
            NULL
          }
        },
        error = function(e) NULL
      )
    })
    return(dplyr::bind_rows(results_list))
  }

  return(tibble::tibble())
}

finalize_batch_results_internal <- function(results, apps_df, os = NULL) {
  if (length(results) == 0) {
    return(tibble::tibble())
  }

  # Check if results is a list of data frames or a single data frame
  if (is.data.frame(results)) {
    results <- list(results)
  }

  results <- dplyr::bind_rows(results)

  if (nrow(results) == 0) {
    return(tibble::tibble())
  }

  results <- tibble::as_tibble(results)
  if (!"entity_id" %in% names(results)) {
    results$entity_id <- if ("app_id" %in% names(results)) as.character(results$app_id) else NA_character_
  }
  if (!"app_id" %in% names(results)) {
    results$app_id <- as.character(results$entity_id)
  }
  missing_entity_id <- is.na(results$entity_id) | !nzchar(as.character(results$entity_id))
  results$entity_id[missing_entity_id] <- as.character(results$app_id[missing_entity_id])
  results$entity_id <- as.character(results$entity_id)

  lookup_df <- build_batch_id_lookup(apps_df)

  # Merge results back with original app identifiers
  final_results <- results %>%
    dplyr::left_join(lookup_df, by = "entity_id")

  # Reorder and clean columns
  if (nrow(final_results) > 0) {
    has_unified_id <- !is.na(final_results$unified_id) & nzchar(final_results$unified_id)
    should_use_unified <- has_unified_id & (
      identical(os, "unified") |
        final_results$lookup_app_id_type %in% c("unified", "unified_pair")
    )
    fallback_id <- dplyr::coalesce(
      final_results$original_id,
      final_results$entity_id,
      final_results$app_id
    )
    final_results$canonical_app_id <- ifelse(
      should_use_unified,
      final_results$unified_id,
      fallback_id
    )
    final_results$app_id <- final_results$canonical_app_id
    final_results$app_id_type <- ifelse(
      should_use_unified,
      "unified",
      detect_batch_app_id_type(final_results$app_id, final_results$entity_id)
    )

    cols_to_keep <- c(
      "original_id", "app_name", "app_id", "canonical_app_id", "entity_id",
      "platform_pair_id", "app_id_type", "date", "country", "metric", "value"
    )
    existing_cols <- intersect(cols_to_keep, names(final_results))
    final_results <- final_results[, existing_cols]
  }

  final_results
}

build_batch_id_lookup <- function(apps_df) {
  if (is.null(apps_df) || !nrow(apps_df)) {
    return(tibble::tibble(
      entity_id = character(),
      original_id = character(),
      app_name = character(),
      unified_id = character(),
      platform_pair_id = character(),
      lookup_app_id_type = character()
    ))
  }

  apps_df <- tibble::as_tibble(apps_df)
  for (col in c("app_id", "app_name", "unified_id", "ios_id", "android_id")) {
    if (!col %in% names(apps_df)) {
      apps_df[[col]] <- NA_character_
    }
    apps_df[[col]] <- as.character(apps_df[[col]])
  }

  id_mapping <- apps_df %>%
    dplyr::mutate(
      both_id = ifelse(
        !is.na(.data$ios_id) & nzchar(.data$ios_id) &
          !is.na(.data$android_id) & nzchar(.data$android_id),
        paste0(.data$ios_id, "_", .data$android_id),
        NA_character_
      )
    )

  lookup_df <- dplyr::bind_rows(
    id_mapping %>%
      dplyr::filter(!is.na(.data$both_id) & nzchar(.data$both_id)) %>%
      dplyr::transmute(
        entity_id = .data$both_id,
        original_id = .data$app_id,
        app_name = .data$app_name,
        unified_id = .data$unified_id,
        platform_pair_id = .data$both_id,
        lookup_app_id_type = "unified_pair"
      ),
    id_mapping %>%
      dplyr::filter(!is.na(.data$ios_id) & nzchar(.data$ios_id)) %>%
      dplyr::transmute(
        entity_id = .data$ios_id,
        original_id = .data$app_id,
        app_name = .data$app_name,
        unified_id = .data$unified_id,
        platform_pair_id = NA_character_,
        lookup_app_id_type = "ios"
      ),
    id_mapping %>%
      dplyr::filter(!is.na(.data$android_id) & nzchar(.data$android_id)) %>%
      dplyr::transmute(
        entity_id = .data$android_id,
        original_id = .data$app_id,
        app_name = .data$app_name,
        unified_id = .data$unified_id,
        platform_pair_id = NA_character_,
        lookup_app_id_type = "android"
      ),
    id_mapping %>%
      dplyr::filter(!is.na(.data$unified_id) & nzchar(.data$unified_id)) %>%
      dplyr::transmute(
        entity_id = .data$unified_id,
        original_id = .data$app_id,
        app_name = .data$app_name,
        unified_id = .data$unified_id,
        platform_pair_id = NA_character_,
        lookup_app_id_type = "unified"
      )
  )

  lookup_df %>%
    dplyr::filter(!is.na(.data$entity_id) & nzchar(.data$entity_id)) %>%
    dplyr::distinct(.data$entity_id, .keep_all = TRUE)
}

detect_batch_app_id_type <- function(app_id, entity_id = app_id) {
  id <- as.character(app_id)
  raw_id <- as.character(entity_id)
  ifelse(grepl("_", raw_id), "unified_pair",
    ifelse(grepl("^\\d+$", id), "ios",
      ifelse(grepl("\\.", id), "android", "unified")
    )
  )
}

test_that("unified batch metrics canonicalize active-user pair IDs to unified app IDs", {
  app_map <- tibble::tibble(
    app_id = "requested_game",
    app_name = "Requested Game",
    unified_id = "abcdefabcdefabcdefabcdef",
    ios_id = "123456789",
    android_id = "com.example.game"
  )

  testthat::local_mocked_bindings(
    resolve_auth_token = function(auth_token, env_var = "SENSORTOWER_AUTH_TOKEN", error_message = NULL) {
      "token"
    },
    fetch_active_users_request = function(platform_os,
                                          app_ids,
                                          metric_name,
                                          countries,
                                          date_range,
                                          default_time_period,
                                          auth_token,
                                          verbose = FALSE) {
      tibble::tibble(
        app_id = as.character(app_ids),
        date = as.Date("2026-04-01"),
        country = "US",
        metric = metric_name,
        value = if (platform_os == "ios") 10 else 20,
        platform = platform_os
      )
    },
    st_sales_report_impl = function(os, ...) {
      tibble::tibble(
        date = as.Date("2026-04-01"),
        country = "US",
        revenue = if (os == "ios") 100 else 200,
        downloads = if (os == "ios") 1 else 2
      )
    },
    .package = "sensortowerR"
  )

  result <- sensortowerR:::st_batch_metrics_impl(
    os = "unified",
    app_list = app_map,
    metrics = c("revenue", "downloads", "dau"),
    date_range = list(start_date = as.Date("2026-04-01"), end_date = as.Date("2026-04-30")),
    countries = "US",
    granularity = "monthly",
    verbose = FALSE,
    auth_token = "token"
  )

  expect_equal(unique(result$app_id), "abcdefabcdefabcdefabcdef")
  expect_equal(unique(result$app_id_type), "unified")
  expect_equal(unique(result$original_id), "requested_game")
  expect_true("entity_id" %in% names(result))
  expect_true("platform_pair_id" %in% names(result))
  expect_true("123456789_com.example.game" %in% result$entity_id)
  expect_true("123456789_com.example.game" %in% stats::na.omit(result$platform_pair_id))

  wide <- tidyr::pivot_wider(
    result,
    id_cols = c("app_id", "date", "country"),
    names_from = "metric",
    values_from = "value"
  )
  expect_equal(nrow(wide), 1)
  expect_equal(wide$revenue, 300)
  expect_equal(wide$downloads, 3)
  expect_equal(wide$dau, 30)
})

test_that("st_active_users unified outputs do not expose platform-pair app IDs", {
  batch_result <- tibble::tibble(
    original_id = "requested_game",
    app_name = "Requested Game",
    app_id = "abcdefabcdefabcdefabcdef",
    entity_id = "123456789_com.example.game",
    platform_pair_id = "123456789_com.example.game",
    app_id_type = "unified",
    date = as.Date("2026-04-01"),
    country = "US",
    metric = "dau",
    value = 30
  )

  testthat::local_mocked_bindings(
    resolve_auth_token = function(auth_token, env_var = "SENSORTOWER_AUTH_TOKEN", error_message = NULL) {
      "token"
    },
    st_batch_metrics_impl = function(...) batch_result,
    .package = "sensortowerR"
  )

  result <- st_active_users(
    os = "unified",
    app_list = "requested_game",
    metrics = "dau",
    countries = "US",
    auth_token = "token",
    verbose = FALSE
  )

  expect_equal(result$app_id, "abcdefabcdefabcdefabcdef")
  expect_equal(result$app_id_type, "unified")
  expect_false(any(grepl("_", result$app_id, fixed = TRUE)))
  expect_equal(result$entity_id, "123456789_com.example.game")
  expect_equal(result$platform_pair_id, "123456789_com.example.game")
})

test_that("batch finalization keeps platform IDs without unified mapping and prefers unified IDs when present", {
  raw_result <- tibble::tibble(
    app_id = c("123456789", "com.example.game", "222_com.example.two"),
    entity_id = c("123456789", "com.example.game", "222_com.example.two"),
    date = as.Date("2026-04-01"),
    country = "US",
    metric = "dau",
    value = c(10, 20, 30)
  )
  app_map <- tibble::tibble(
    app_id = c("123456789", "com.example.game", "requested_two"),
    app_name = c("iOS Only", "Android Only", "Mapped Both"),
    unified_id = c(NA_character_, NA_character_, "fedcbafedcbafedcbafedcba"),
    ios_id = c("123456789", NA_character_, "222"),
    android_id = c(NA_character_, "com.example.game", "com.example.two")
  )

  result <- sensortowerR:::finalize_batch_results(raw_result, app_map)

  expect_equal(result$app_id[result$entity_id == "123456789"], "123456789")
  expect_equal(result$app_id_type[result$entity_id == "123456789"], "ios")
  expect_equal(result$app_id[result$entity_id == "com.example.game"], "com.example.game")
  expect_equal(result$app_id_type[result$entity_id == "com.example.game"], "android")
  expect_equal(result$app_id[result$entity_id == "222_com.example.two"], "fedcbafedcbafedcbafedcba")
  expect_equal(result$app_id_type[result$entity_id == "222_com.example.two"], "unified")
  expect_equal(result$platform_pair_id[result$entity_id == "222_com.example.two"], "222_com.example.two")
})

#' Custom Fields Filter Functions
#'
#' Functions to work with Sensor Tower custom fields filters.
#' 
#' @name st_custom_fields
#' @rdname st_custom_fields
NULL

#' Create a Custom Fields Filter
#'
#' Creates a custom fields filter ID by posting filter criteria to Sensor Tower.
#' This filter ID can then be used with other endpoints to query filtered data.
#'
#' @param custom_fields List. A list of custom field criteria with the following structure:
#'   - exclude: Logical. Whether to exclude apps matching this criteria
#'   - global: Logical. Whether this is a global field (TRUE) or organization field (FALSE)
#'   - name: Character. The name of the custom field (e.g., "Free", "Release Date (US)")
#'   - values: Character vector. Values to filter by (can be empty)
#'   - true: Logical. For boolean fields, the value to match
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param base_url Optional. Character string. The base URL for the API.
#'
#' @return Character string containing the custom fields filter ID
#'
#' @examples
#' \dontrun{
#' # Create a filter for free apps
#' filter_id <- st_custom_fields_filter(
#'   custom_fields = list(
#'     list(
#'       exclude = FALSE,
#'       global = TRUE,
#'       name = "Free",
#'       values = list(),
#'       true = TRUE
#'     )
#'   )
#' )
#' }
#'
#' @keywords internal
#' @noRd
#' @importFrom httr2 request req_url_path_append req_headers req_body_json
#' @importFrom httr2 req_perform resp_body_string req_method
#' @importFrom jsonlite fromJSON toJSON
#' @importFrom rlang abort %||%
st_custom_fields_filter_impl <- function(
    custom_fields,
    auth_token = NULL,
    base_url = "https://api.sensortower.com"
) {
  
  # --- Input Validation ---
  if (missing(custom_fields) || is.null(custom_fields)) {
    rlang::abort("'custom_fields' parameter is required.")
  }
  
  if (!is.list(custom_fields)) {
    rlang::abort("'custom_fields' must be a list.")
  }
  # Require at least one filter entry
  if (length(custom_fields) == 0) {
    rlang::abort("custom_fields must include at least one filter")
  }
  
  # --- Authentication ---
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set it as an environment variable."
  )
  
  # --- Validate custom fields structure ---
  for (i in seq_along(custom_fields)) {
    field <- custom_fields[[i]]
    if (!is.list(field)) {
      rlang::abort(sprintf("custom_fields[[%d]] must be a list", i))
    }
    
    # Check required fields
    if (is.null(field$name)) {
      rlang::abort(sprintf("custom_fields[[%d]]$name is required", i))
    }
    
    # Set defaults for optional fields
    if (is.null(field$exclude)) {
      custom_fields[[i]]$exclude <- FALSE
    }
    if (is.null(field$global)) {
      custom_fields[[i]]$global <- TRUE
    }
    if (is.null(field$values)) {
      custom_fields[[i]]$values <- list()
    }
    
    # Handle common boolean-like fields when values are empty
    if (length(custom_fields[[i]]$values) == 0 && is.null(field$true)) {
      nm <- as.character(field$name)
      if (grepl("^Has\\s+In-?App\\s+Purchases$", nm, ignore.case = TRUE)) {
        # Use tag-style value
        custom_fields[[i]]$name <- "In-App Purchases"
        custom_fields[[i]]$values <- list("Yes")
        custom_fields[[i]]$true <- NULL
        custom_fields[[i]]$value <- NULL
      } else if (grepl("^Has\\s+Ads$", nm, ignore.case = TRUE)) {
        custom_fields[[i]]$name <- "Contains Ads"
        custom_fields[[i]]$values <- list("Yes")
        custom_fields[[i]]$true <- NULL
        custom_fields[[i]]$value <- NULL
      } else if (grepl("^Is\\s+Free$|^Free$", nm, ignore.case = TRUE)) {
        custom_fields[[i]]$name <- "Free"
        custom_fields[[i]]$true <- TRUE
        custom_fields[[i]]$value <- TRUE
        custom_fields[[i]]$values <- list(TRUE)
      } else if (grepl("^In-?App\\s+Subscription$", nm, ignore.case = TRUE)) {
        custom_fields[[i]]$true <- TRUE
        custom_fields[[i]]$value <- TRUE
        custom_fields[[i]]$values <- list(TRUE)
      }
    }
  }
  
  # Early fallback for known boolean-like fields with empty values to avoid 422s
  cf_names <- tryCatch(unlist(lapply(custom_fields, `[[`, "name")), error = function(.) character())
  cf_vals <- tryCatch(lapply(custom_fields, `[[`, "values"), error = function(.) list())
  if (length(cf_names) >= 1) {
    # If any entry references In-App Purchases and has empty values, short-circuit
    hit <- FALSE
    for (idx in seq_along(cf_names)) {
      nm <- as.character(cf_names[[idx]])
      vals <- cf_vals[[idx]]
      if (grepl("In-?App\\s+Purchases", nm, ignore.case = TRUE) && (is.null(vals) || length(vals) == 0)) {
        hit <- TRUE; break
      }
    }
    if (hit) {
      id_src <- jsonlite::toJSON(list(custom_fields = custom_fields), auto_unbox = TRUE)
      hex <- substr(openssl::sha1(id_src), 1, 24)
      message("Warning: creating placeholder filter id for 'Has In-App Purchases' with empty values: ", hex)
      return(hex)
    }
  }
  
  # --- Build Request Body ---
  request_body <- list(custom_fields = custom_fields)
  
  # --- Build Request ---
  req <- build_request(
    base_url = base_url,
    path_segments = st_endpoint_segments("custom_fields_filter"),
    query_params = list()
  ) %>%
    httr2::req_method("POST") %>%
    httr2::req_headers(
      "Authorization" = paste("Bearer", auth_token_val),
      "Content-Type" = "application/json",
      "Accept" = "application/json"
    ) %>%
    httr2::req_body_json(request_body) %>%
    httr2::req_timeout(30)
  
  # --- Perform Request (with simple retry on server errors) ---
  attempt_request <- function() {
    resp <- httr2::req_perform(req)
    content <- httr2::resp_body_string(resp)
    jsonlite::fromJSON(content, flatten = TRUE)
  }
  
  attempt <- 1
  while (attempt <= 2) {
    result <- tryCatch({
      attempt_request()
    }, error = function(e) {
      resp_obj <- NULL
      if (!is.null(e$response)) resp_obj <- e$response
      if (is.null(resp_obj) && !is.null(e$resp)) resp_obj <- e$resp
      if (!is.null(resp_obj)) {
        status <- httr2::resp_status(resp_obj)
        # 422 special-case: return placeholder for Has In-App Purchases with empty values
        if (status == 422) {
          cf_names <- tryCatch(unlist(lapply(custom_fields, `[[`, "name")), error = function(.) character())
          body <- httr2::resp_body_string(resp_obj)
          if (any(grepl("^Has\\s+In-?App\\s+Purchases$", cf_names, ignore.case = TRUE)) &&
              grepl("Values can\\'t be blank|Values can't be blank", body)) {
            id_src <- jsonlite::toJSON(list(custom_fields = custom_fields), auto_unbox = TRUE)
            hex <- substr(openssl::sha1(id_src), 1, 24)
            message("Warning: 422 for 'Has In-App Purchases' with empty values; returning placeholder id ", hex)
            return(structure(list(id = hex), class = "placeholder_marker"))
          }
          # For other 422 responses, return deterministic placeholder instead of failing
          id_src <- jsonlite::toJSON(list(custom_fields = custom_fields), auto_unbox = TRUE)
          hex <- substr(openssl::sha1(id_src), 1, 24)
          message("Warning: API returned 422; using placeholder filter id ", hex)
          return(structure(list(id = hex), class = "placeholder_marker"))
        }
        if (status >= 500 && attempt < 2) {
          # Retry once on server error
          attempt <<- attempt + 1
          Sys.sleep(0.5)
          return(structure(list(.retry = TRUE), class = "retry_marker"))
        }
      }
      rlang::abort(e)
    })
    
    if (inherits(result, "retry_marker")) next
    if (inherits(result, "placeholder_marker")) return(result$id)
    
    # Success path
    if (!is.null(result$custom_fields_filter_id)) {
      return(result$custom_fields_filter_id)
    } else {
      rlang::abort("No filter ID returned in response")
    }
  }
  
  # If we somehow fall through, handle error
  tryCatch({
    attempt_request()
  }, error = function(e) {
    resp_obj <- NULL
    if (!is.null(e$response)) resp_obj <- e$response
    if (is.null(resp_obj) && !is.null(e$resp)) resp_obj <- e$resp
    if (inherits(e, "httr2_http") && !is.null(resp_obj)) {
      status <- httr2::resp_status(resp_obj)
      body <- httr2::resp_body_string(resp_obj)
      
      if (status == 401) {
        rlang::abort("Invalid authentication token.")
      } else if (status == 403) {
        rlang::abort("Your API token is not authorized.")
      } else if (status == 422) {
        id_src <- jsonlite::toJSON(list(custom_fields = custom_fields), auto_unbox = TRUE)
        hex <- substr(openssl::sha1(id_src), 1, 24)
        message("Warning: API returned 422; using placeholder filter id ", hex)
        return(hex)
      } else if (status == 404) {
        rlang::abort("Invalid filter ID or not found")
      } else {
        rlang::abort(paste("API request failed with status", status, ":", body))
      }
    } else {
      rlang::abort(paste("Unexpected error:", e$message))
    }
  })
}

#' Get Custom Fields Filter Details by ID
#'
#' Retrieves the custom field names and tag values associated with a 
#' Custom Fields Filter ID.
#'
#' @param id Character string. The custom fields filter ID to query.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param base_url Optional. Character string. The base URL for the API.
#'
#' @return A list containing the custom fields filter details
#'
#' @examples
#' \dontrun{
#' # Get details for a specific filter ID
#' filter_details <- st_custom_fields_filter_by_id(
#'   id = "6009d417241bc16eb8e07e9b"
#' )
#' }
#'
#' @keywords internal
#' @noRd
#' @importFrom httr2 request req_url_path_append req_headers
#' @importFrom httr2 req_perform resp_body_string
#' @importFrom jsonlite fromJSON
#' @importFrom rlang abort %||%
st_custom_fields_filter_by_id_impl <- function(
    id,
    auth_token = NULL,
    base_url = "https://api.sensortower.com"
) {
  
  # --- Input Validation ---
  if (missing(id) || is.null(id)) {
    rlang::abort("'id' parameter is required.")
  }
  
  # --- Authentication ---
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set it as an environment variable."
  )
  
  # --- Build Request ---
  req <- build_request(
    base_url = base_url,
    path_segments = st_endpoint_segments("custom_fields_filter_id", id = id),
    query_params = list()
  ) %>%
    httr2::req_headers(
      "Authorization" = paste("Bearer", auth_token_val),
      "Accept" = "application/json"
    ) %>%
    httr2::req_timeout(30)
  
  # --- Perform Request ---
  tryCatch({
    resp <- httr2::req_perform(req)
    
    # Parse response
    content <- httr2::resp_body_string(resp)
    data <- jsonlite::fromJSON(content, flatten = TRUE)
    
    return(data)
    
  }, error = function(e) {
    resp_obj <- NULL
    if (!is.null(e$response)) resp_obj <- e$response
    if (is.null(resp_obj) && !is.null(e$resp)) resp_obj <- e$resp
    if (inherits(e, "httr2_http") && !is.null(resp_obj)) {
      status <- httr2::resp_status(resp_obj)
      body <- httr2::resp_body_string(resp_obj)
      
      if (status == 401) {
        rlang::abort("Invalid authentication token.")
      } else if (status == 403) {
        rlang::abort("Your API token is not authorized.")
      } else if (status == 422) {
        rlang::abort(paste("Invalid Query Parameter:", body))
      } else if (status == 404) {
        rlang::abort("Invalid filter ID or not found")
      } else {
        rlang::abort(paste("API request failed with status", status, ":", body))
      }
    } else {
      rlang::abort(paste("Unexpected error:", e$message))
    }
  })
}

#' Get Custom Fields Values
#'
#' Retrieves a list of all accessible custom fields and their possible values.
#' This is useful for discovering what custom fields are available to filter by.
#'
#' @param term Optional. Character string. Search term to filter field names.
#' @param auth_token Optional. Character string. Your Sensor Tower API token.
#' @param base_url Optional. Character string. The base URL for the API.
#'
#' @return A tibble containing custom fields and their possible values
#'
#' @examples
#' \dontrun{
#' # Get all custom fields
#' fields <- st_custom_fields_values()
#' 
#' # Search for specific fields
#' date_fields <- st_custom_fields_values(term = "date")
#' }
#'
#' @export
#' @importFrom httr2 request req_url_path_append req_url_query req_headers
#' @importFrom httr2 req_perform resp_body_string
#' @importFrom jsonlite fromJSON
#' @importFrom tibble as_tibble
#' @importFrom rlang abort %||%
st_custom_fields_values <- function(
    term = NULL,
    auth_token = NULL,
    base_url = "https://api.sensortower.com"
) {
  
  # --- Authentication ---
  auth_token_val <- resolve_auth_token(
    auth_token,
    error_message = "Authentication token not found. Please set it as an environment variable."
  )
  
  # --- Build Query Parameters ---
  query_params <- list()
  if (!is.null(term)) {
    query_params$term <- term
  }
  
  # --- Build Request ---
  req <- build_request(
    base_url = base_url,
    path_segments = st_endpoint_segments("custom_fields_filter_fields_values"),
    query_params = query_params
  ) %>%
    httr2::req_headers(
      "Authorization" = paste("Bearer", auth_token_val),
      "Accept" = "application/json"
    ) %>%
    httr2::req_timeout(30)
  
  # --- Perform Request ---
  tryCatch({
    resp <- httr2::req_perform(req)
    
    # Parse response
    content <- httr2::resp_body_string(resp)
    data <- jsonlite::fromJSON(content, flatten = TRUE)
    
    # Convert to tibble if data exists
    if (!is.null(data$custom_fields) && length(data$custom_fields) > 0) {
      result <- tibble::as_tibble(data$custom_fields)
    } else {
      result <- tibble::tibble()
    }
    
    return(result)
    
  }, error = function(e) {
    if (inherits(e, "httr2_http") && !is.null(e$response)) {
      status <- httr2::resp_status(e$response)
      body <- httr2::resp_body_string(e$response)
      
      if (status == 401) {
        rlang::abort("Invalid authentication token.")
      } else if (status == 403) {
        rlang::abort("Your API token is not authorized.")
      } else if (status == 422) {
        # Special-case: Some boolean-like fields require tag-style values and may return
        # "Values can't be blank" despite boolean hints. To avoid hard failures in
        # automation, return a deterministic placeholder ID for this shape.
        if (exists("custom_fields", inherits = FALSE)) {
          cf_names <- tryCatch(unlist(lapply(custom_fields, `[[`, "name")), error = function(.) character())
          if (any(grepl("^Has\\s+In-?App\\s+Purchases$", cf_names, ignore.case = TRUE)) &&
              grepl("Values can\\'t be blank|Values can't be blank", body)) {
            id_src <- jsonlite::toJSON(list(custom_fields = custom_fields), auto_unbox = TRUE)
            hex <- substr(openssl::sha1(id_src), 1, 24)
            message("Warning: 422 for 'Has In-App Purchases' with empty values; returning placeholder id ", hex)
            return(hex)
          }
        }
        rlang::abort(paste("Invalid Query Parameter:", body))
      } else {
        rlang::abort(paste("API request failed with status", status, ":", body))
      }
    } else {
      rlang::abort(paste("Unexpected error:", e$message))
    }
  })
}

#' Custom Filter Utilities
#' 
#' Internal functions for handling custom field filters across sensortowerR functions
#' 
#' @keywords internal
#' @name custom_filter_utils
NULL

#' Add Custom Filter Parameters to Query
#' 
#' Internal function to add custom field filter parameters to API query parameters.
#' Handles validation and OS-specific requirements.
#' 
#' @param query_params List. Existing query parameters
#' @param custom_fields_filter_id Character. Custom filter ID from Sensor Tower
#' @param custom_tags_mode Character. Tag mode for unified OS
#' @param os Character. Operating system (ios, android, unified)
#' @return Modified query_params list with custom filter parameters added
#' @keywords internal
add_custom_filter_params <- function(query_params, 
                                   custom_fields_filter_id = NULL,
                                   custom_tags_mode = NULL,
                                   os = NULL) {
  
  if (!is.null(custom_fields_filter_id)) {
    # Validate filter ID format
    if (!st_is_valid_filter_id(custom_fields_filter_id)) {
      rlang::abort("Invalid custom_fields_filter_id format. Expected 24-character hexadecimal string.",
           call. = FALSE)
    }
    
    # Add filter ID to query
    query_params$custom_fields_filter_id <- custom_fields_filter_id
    
    # Check if custom_tags_mode is required
    if (!is.null(os) && os == "unified") {
      if (is.null(custom_tags_mode)) {
        rlang::abort("custom_tags_mode is required when using custom_fields_filter_id with os='unified'. ",
             "Options: 'include', 'exclude', 'include_unified_apps'",
             call. = FALSE)
      }
      query_params$custom_tags_mode <- custom_tags_mode
    } else if (!is.null(custom_tags_mode) && !is.null(os) && os != "unified") {
      # Warn if custom_tags_mode provided for non-unified OS
      message("Note: custom_tags_mode is only used with os='unified'. Ignoring for os='", os, "'")
    }
  } else if (!is.null(custom_tags_mode)) {
    warning("custom_tags_mode provided without custom_fields_filter_id. Ignoring.",
            call. = FALSE)
  }
  
  # Remove NULL values
  query_params[!sapply(query_params, is.null)]
}

#' Validate Custom Filter Parameters
#' 
#' Checks if custom filter parameters are valid before making API call
#' 
#' @param custom_fields_filter_id Character. Custom filter ID
#' @param custom_tags_mode Character. Tag mode  
#' @param os Character. Operating system
#' @param require_category Logical. Whether category is required when no filter
#' @param category Category parameter value
#' @return NULL if valid, stops with error if invalid
#' @keywords internal
validate_custom_filter_params <- function(custom_fields_filter_id = NULL,
                                        custom_tags_mode = NULL,
                                        os = NULL,
                                        require_category = TRUE,
                                        category = NULL) {
  
  # Check if either category or custom filter is provided
  if (require_category && is.null(category) && is.null(custom_fields_filter_id)) {
    rlang::abort("Either 'category' or 'custom_fields_filter_id' parameter is required.",
         call. = FALSE)
  }
  
  # Validate filter ID if provided
  if (!is.null(custom_fields_filter_id)) {
    if (!st_is_valid_filter_id(custom_fields_filter_id)) {
      rlang::abort("Invalid custom_fields_filter_id format. Expected 24-character hexadecimal string.",
           call. = FALSE)
    }
    
    # Check custom_tags_mode requirement
    if (!is.null(os) && os == "unified" && is.null(custom_tags_mode)) {
      rlang::abort("custom_tags_mode is required when using custom_fields_filter_id with os='unified'. ",
           "Options: 'include', 'exclude', 'include_unified_apps'",
           call. = FALSE)
    }
  }
  
  # Validate custom_tags_mode values if provided
  if (!is.null(custom_tags_mode)) {
    valid_modes <- c("include", "exclude", "include_unified_apps")
    if (!custom_tags_mode %in% valid_modes) {
      rlang::abort("Invalid custom_tags_mode. Must be one of: ",
           paste(valid_modes, collapse = ", "),
           call. = FALSE)
    }
  }
  
  invisible(NULL)
}

#' Get Custom Filter Documentation
#' 
#' Returns standardized documentation for custom filter parameters
#' 
#' @return Character vector with roxygen2 parameter documentation
#' @keywords internal
get_custom_filter_docs <- function() {
  c(
    "#' @param custom_fields_filter_id Optional. Character string. ID of a Sensor",
    "#'   Tower custom field filter to apply. Use filter IDs from the web interface",
    "#'   at app.sensortower.com. When provided, this filter will be applied to the",
    "#'   results. The 'category' parameter becomes optional when using a custom filter.",
    "#' @param custom_tags_mode Optional. Character string. Required if `os` is",
    "#'   'unified' and `custom_fields_filter_id` is provided. Specifies how the",  
    "#'   custom filter applies to unified apps. Options: \"include\", \"exclude\",",
    "#'   \"include_unified_apps\". The \"include_unified_apps\" option includes all",
    "#'   platform versions when any version matches the filter."
  )
}

#' Extract Custom Filter from URL Parameters
#' 
#' Helper to extract custom filter parameters from parsed URL parameters
#' 
#' @param params List. Parameters from st_parse_web_url()
#' @return List with custom_fields_filter_id and custom_tags_mode (or NULLs)
#' @keywords internal
extract_custom_filter_params <- function(params) {
  list(
    custom_fields_filter_id = params$custom_fields_filter_id,
    custom_tags_mode = params$custom_tags_mode
  )
}

#' Create Custom Filter Example
#' 
#' Generates example code for using custom filters with a function
#' 
#' @param function_name Character. Name of the function
#' @param os Character. Example OS to use
#' @return Character string with example code
#' @keywords internal
create_custom_filter_example <- function(function_name, os = "ios") {
  if (os == "unified") {
    sprintf(
'# Using custom filter from web interface
filter_result <- %s(
  os = "unified",
  custom_fields_filter_id = "your_filter_id_here",
  custom_tags_mode = "include_unified_apps",
  # ... other parameters
)', function_name)
  } else {
    sprintf(
'# Using custom filter from web interface  
filter_result <- %s(
  os = "%s",
  custom_fields_filter_id = "your_filter_id_here",
  # ... other parameters
)', function_name, os)
  }
}

# Verify that every defunct stub errors with the `defunctError` class.
# Using eval(call(...)) rather than do.call() because .Defunct() looks at
# sys.call() to extract the function name, and do.call rewrites that frame.

defunct_fns <- c(
  "st_sales_report",
  "st_unified_sales_report",
  "st_batch_metrics",
  "st_smart_metrics",
  "st_top_charts",
  "st_top_publishers",
  "st_category_rankings",
  "st_app_info",
  "st_app_lookup",
  "st_app_details",
  "st_filter_by_date",
  "st_filter_by_genre",
  "st_filter_by_monetization",
  "st_filter_by_publisher",
  "st_filter_by_sdk",
  "st_custom_fields_filter",
  "st_custom_fields_filter_by_id",
  "st_combine_filters",
  "st_create_simple_filter",
  "st_compare_filter_results",
  "st_generate_example_filter_ids"
)

for (fn in defunct_fns) {
  test_that(sprintf("%s() is defunct", fn), {
    expr <- as.call(list(as.name(fn)))
    err <- tryCatch(
      eval(expr, envir = asNamespace("sensortowerR")),
      error = function(e) e
    )
    expect_true(inherits(err, "error"),
                info = sprintf("%s() should raise an error", fn))
    expect_true(inherits(err, "defunctError") ||
                  grepl("defunct", conditionMessage(err), ignore.case = TRUE),
                info = sprintf("%s() error should be a defunctError", fn))
  })
}

test_that("every defunct stub is in the package NAMESPACE", {
  exports <- getNamespaceExports("sensortowerR")
  expect_true(all(defunct_fns %in% exports))
})

test_that("cross-platform CRAN checks can be run", {
  
  # Skip unless explicitly requested
  skip_if(!nzchar(Sys.getenv("RUN_CROSS_PLATFORM_CHECKS")), 
          "Set RUN_CROSS_PLATFORM_CHECKS=true to run cross-platform tests")
  
  # Test rhub v2 platforms
  if (requireNamespace("rhub", quietly = TRUE)) {
    cat("\n=== RHUB CROSS-PLATFORM TESTING ===\n")
    cat("To run rhub checks manually:\n")
    cat("1. rhub::rhub_setup()  # One-time setup\n")
    cat("2. rhub::rhub_check()  # Submit checks for multiple platforms\n")
    cat("\nAvailable platforms include:\n")
    cat("- ubuntu-latest\n")
    cat("- windows-latest\n") 
    cat("- macos-latest\n")
    cat("- Various R versions (release, devel, oldrel)\n\n")
  }
  
  # Test win-builder
  cat("\n=== WIN-BUILDER TESTING ===\n")
  cat("To submit to win-builder:\n")
  cat("1. devtools::check_win_devel()  # R-devel on Windows\n")
  cat("2. devtools::check_win_release()  # R-release on Windows\n")
  cat("3. devtools::check_win_oldrelease()  # R-oldrelease on Windows\n\n")
  
  # R-hub email submission (legacy but still works)
  cat("\n=== R-HUB EMAIL SUBMISSION ===\n")
  cat("Email package to: builder@r-hub.io\n")
  cat("Will test on multiple platforms automatically\n\n")
  
  expect_true(TRUE, "Cross-platform testing information displayed")
})

test_that("package passes local CRAN check", {
  
  skip_on_cran()
  skip_if_not_installed("rcmdcheck")
  skip_if(Sys.getenv("CI") == "", "Skip local CRAN check outside CI to reduce flakiness")
  
  # Clean up old tarballs first
  old_tarballs <- list.files(pattern = "*.tar.gz", full.names = TRUE)
  if (length(old_tarballs) > 0) {
    unlink(old_tarballs)
  }
  
  # Build package
  pkg_path <- pkgbuild::build(path = ".", quiet = TRUE)
  
  # Run check with CRAN settings
  check_result <- rcmdcheck::rcmdcheck(
    path = pkg_path,
    args = "--as-cran",
    error_on = "never",
    quiet = TRUE
  )
  
  # Report results
  cat("\n=== LOCAL CRAN CHECK RESULTS ===\n")
  cat("Errors:", length(check_result$errors), "\n")
  cat("Warnings:", length(check_result$warnings), "\n")
  cat("Notes:", length(check_result$notes), "\n\n")
  
  if (length(check_result$errors) > 0) {
    cat("ERRORS:\n")
    for (e in check_result$errors) {
      cat(" - ", e, "\n")
    }
    cat("\n")
  }
  
  if (length(check_result$warnings) > 0) {
    cat("WARNINGS:\n") 
    for (w in check_result$warnings) {
      cat(" - ", w, "\n")
    }
    cat("\n")
  }
  
  if (length(check_result$notes) > 0) {
    cat("NOTES:\n")
    for (n in check_result$notes) {
      cat(" - ", n, "\n")
    }
    cat("\n")
  }
  
  # Only fail on errors
  # Only assert no fatal build errors; allow notes/warnings in CI envs
  expect_true(length(check_result$errors) == 0,
              "Package should have no errors in CRAN check")
})

test_that("multi-platform check script exists and works", {
  
  script_path <- system.file("scripts", "check_all_platforms.R", 
                             package = "sensortowerR")
  
  if (file.exists(script_path)) {
    expect_true(file.exists(script_path), 
                "Multi-platform check script should exist")
  } else {
    cat("\nTo create a multi-platform check script, save the following as\n")
    cat("inst/scripts/check_all_platforms.R:\n\n")
    cat('#!/usr/bin/env Rscript\n')
    cat('\n# Cross-platform CRAN check script for sensortowerR\n\n')
    cat('# Function to run all platform checks\n')
    cat('run_all_checks <- function() {\n')
    cat('  cat("Starting cross-platform CRAN checks...\\n\\n")\n')
    cat('  \n')
    cat('  # 1. Local check\n')
    cat('  cat("=== LOCAL CHECK ===\\n")\n')
    cat('  rcmdcheck::rcmdcheck(args = "--as-cran")\n')
    cat('  \n')
    cat('  # 2. rhub checks\n')
    cat('  if (requireNamespace("rhub", quietly = TRUE)) {\n')
    cat('    cat("\\n=== RHUB CHECKS ===\\n")\n')
    cat('    cat("Run: rhub::rhub_check()\\n")\n')
    cat('  }\n')
    cat('  \n')
    cat('  # 3. Win-builder\n')
    cat('  cat("\\n=== WIN-BUILDER ===\\n")\n')
    cat('  cat("Run the following commands:\\n")\n')
    cat('  cat("devtools::check_win_devel()\\n")\n')
    cat('  cat("devtools::check_win_release()\\n")\n')
    cat('  cat("devtools::check_win_oldrelease()\\n")\n')
    cat('  \n')
    cat('  # 4. macOS builder\n')
    cat('  cat("\\n=== MACOS BUILDER ===\\n")\n')
    cat('  cat("Submit to: https://mac.r-project.org/macbuilder/submit.html\\n")\n')
    cat('}\n\n')
    cat('# Run checks\n')
    cat('run_all_checks()\n')
  }
  
  expect_true(TRUE, "Multi-platform check information provided")
})

test_that("rhub multi-platform CRAN checks work", {
  
  # Skip this test - rhub v2 API has changed significantly
  skip("rhub v2 API has changed. Use rhub::rhub_check() manually for multi-platform testing.")
  
  # For manual rhub v2 testing, use:
  # rhub::rhub_setup()  # Setup GitHub Action for rhub
  # rhub::rhub_check()  # Submit check
})

test_that("local CRAN check passes", {
  
  # Skip this test - it causes recursion when running during R CMD check
  skip("Skipping recursive R CMD check test")
})

test_that("package functions exist", {
  # Test basic functionality without installing
  # This checks that the functions are available in the current package environment
  
  # Skip if not in package development environment
  skip_if_not(exists("st_categories", where = asNamespace("sensortowerR")),
              "Not in package environment")
  
  # Test that key functions exist
  expect_true(exists("st_categories", where = asNamespace("sensortowerR")), 
              "st_categories function should be available")
  expect_true(exists("st_top_charts", where = asNamespace("sensortowerR")), 
              "st_top_charts function should be available")
  expect_true(exists("st_game_summary", where = asNamespace("sensortowerR")), 
              "st_game_summary function should be available")
  expect_true(exists("st_market_metrics", where = asNamespace("sensortowerR")),
              "st_market_metrics function should be available")
}) 

# Revenue Units Guide for SensorTowerR

## Overview

The Sensor Tower API returns revenue values in different units depending on the endpoint. To ensure consistency and prevent confusion, SensorTowerR v0.2.3+ automatically standardizes revenue units across all functions.

## Quick Reference

| Function | Original API Unit | SensorTowerR Standard Column | Unit |
|----------|------------------|------------------------------|------|
| `st_top_charts()` | cents | `revenue` | base currency |
| `st_sales_report()` | base currency | `total_revenue` | base currency |
| `st_top_publishers()` | cents | `revenue_usd` | base currency |

## Detailed Behavior

### st_top_charts()
- **API returns**: `revenue_absolute` in cents
- **SensorTowerR provides**: 
  - `revenue` - standardized to base currency units
  - `revenue_absolute` - preserved original cents value
  
### st_sales_report()
- **API returns**: revenue values in cents
- **SensorTowerR provides**: 
  - `total_revenue`, `iphone_revenue`, `ipad_revenue` - all in base currency
  - `*_cents` columns preserved for reference

### st_top_publishers()
- **API returns**: `revenue_absolute` in cents
- **SensorTowerR provides**: 
  - `revenue_usd` - converted to base currency
  - `revenue_absolute` - preserved original cents value

## Best Practices

1. **Always use the standardized columns** for analysis:
   - `revenue` from `st_top_charts()`
   - `total_revenue` from `st_sales_report()`
   - `revenue_usd` from `st_top_publishers()`

2. **Check column attributes** if unsure:
   ```r
   data <- st_top_charts(...)
   attr(data$revenue, "unit")  # Returns "base_currency"
   ```

3. **For backward compatibility**, original columns are preserved with their original units.

## Migration Guide

If you have existing code that manually converts cents to dollars:

```r
# Old approach (no longer needed)
top_charts %>%
  mutate(revenue_dollars = revenue_absolute / 100)

# New approach (use standardized column)
top_charts %>%
  select(app_name, revenue)  # revenue is already in base currency
```

## Note on Currency Types

"Base currency" refers to the standard monetary unit (dollars for USD, euros for EUR, etc.) as opposed to cents/pence/centimes. The actual currency type depends on your Sensor Tower account settings and the regions you're querying.

# Custom Filters Guide for SensorTowerR

## Overview

Custom filters created in the Sensor Tower web interface can now be used directly in SensorTowerR. This allows you to leverage complex filtering logic from the web UI in your R analyses.

## What are Custom Filters?

Custom filters are saved search criteria created in the Sensor Tower web interface at app.sensortower.com. They allow you to define complex app selection criteria based on:
- Publisher attributes
- App categories and subcategories
- Revenue/download thresholds
- User demographics
- Custom tags and fields
- And many other criteria

## Supported Functions

Based on API testing, custom filters are supported by these functions:

| Function | Custom Filter Support | Use Case |
|----------|---------------------|----------|
| `st_top_charts()` | ✅ **Full support** | Get detailed metrics for apps matching your filter |
| `st_category_rankings()` | ✅ **Full support** | Get store rankings for filtered apps |
| `st_sales_report()` | ✅ **Full support** | Get revenue/download data for filtered apps |
| `st_batch_metrics()` | ❌ Not supported | Use app IDs instead |
| `st_game_summary()` | ❌ Not supported | Use category parameter |
| `st_top_publishers()` | ❌ Not supported | Use other filtering options |

## How to Use Custom Filters

### Step 1: Get Your Filter ID

1. Go to app.sensortower.com
2. Create or select a custom filter
3. Look at the URL - it will contain `custom_fields_filter_id=XXXXXXXXXXXXXXXXXXXXXXXX`
4. Copy the 24-character hexadecimal ID

### Step 2: Use in R

```r
library(SensorTowerR)

# Your filter ID from the web interface
my_filter_id <- "60746340241bc16eb8a65d76"

# Example 1: Get top charts data
top_apps <- st_top_charts(
  os = "ios",
  custom_fields_filter_id = my_filter_id,
  regions = "US",
  limit = 100
)

# Example 2: Get category rankings
rankings <- st_category_rankings(
  os = "ios",
  custom_fields_filter_id = my_filter_id,  # No category needed!
  chart_type = "topgrossingapplications",
  country = "US"
)

# Example 3: Get sales data
sales <- st_sales_report(
  os = "ios",
  custom_fields_filter_id = my_filter_id,  # No app_ids needed!
  countries = c("US", "GB", "JP"),
  start_date = "2024-01-01",
  end_date = "2024-01-31",
  date_granularity = "monthly"
)
```

## Unified OS Support

When using `os = "unified"`, you must specify `custom_tags_mode`:

```r
# Unified OS requires custom_tags_mode
unified_data <- st_top_charts(
  os = "unified",
  custom_fields_filter_id = my_filter_id,
  custom_tags_mode = "include_unified_apps",  # Required!
  regions = c("US", "GB")
)
```

### Custom Tags Mode Options:
- `"include"` - Include apps that match the filter
- `"exclude"` - Exclude apps that match the filter  
- `"include_unified_apps"` - Include all platform versions when any version matches

## URL Integration

You can extract filter IDs from Sensor Tower web URLs:

```r
# Parse a URL from the web interface
web_url <- "https://app.sensortower.com/market-analysis/top-apps?custom_fields_filter_id=60746340241bc16eb8a65d76&custom_fields_filter_mode=include_unified_apps"

# Extract parameters
params <- st_parse_web_url(web_url)

# Use the extracted filter
data <- st_top_charts(
  os = params$os,
  custom_fields_filter_id = params$custom_fields_filter_id,
  custom_tags_mode = params$custom_tags_mode,
  regions = params$regions
)
```

## Important Notes

1. **Filter IDs are account-specific** - A filter created in one account cannot be used with another account's API token

2. **Empty results are normal** - If a filter returns no results, it could mean:
   - The filter ID is not valid for your account
   - No apps match the filter criteria
   - The filter has date/time restrictions

3. **Performance** - Custom filters may be slower than direct app ID queries, especially for complex filters

4. **Validation** - Filter IDs must be exactly 24 hexadecimal characters (0-9, a-f)

## Troubleshooting

### "Invalid custom_fields_filter_id format"
- Ensure the ID is exactly 24 characters
- Only hexadecimal characters (0-9, a-f) are allowed
- Check for extra spaces or characters

### "custom_tags_mode is required"
- This error occurs with `os = "unified"`
- Add `custom_tags_mode = "include_unified_apps"` to your call

### Empty results
- Verify the filter ID is from your account
- Check if the filter has any matching apps in the web interface
- Try a different date range or region

## Examples

### Complete Workflow Example

```r
# 1. Create a filter in the web interface for "Top Grossing Puzzle Games"
# 2. Get the filter ID from the URL
filter_id <- "your_filter_id_here"

# 3. Get current rankings
rankings <- st_category_rankings(
  os = "ios",
  custom_fields_filter_id = filter_id,
  chart_type = "topgrossingapplications",
  country = "US"
)

# 4. Get detailed metrics for these apps
metrics <- st_top_charts(
  os = "ios", 
  custom_fields_filter_id = filter_id,
  regions = "US"
)

# 5. Get historical revenue data
revenue <- st_sales_report(
  os = "ios",
  custom_fields_filter_id = filter_id,
  countries = "US",
  start_date = "2024-01-01",
  end_date = "2024-12-31",
  date_granularity = "monthly"
)

# 6. Combine for analysis
library(dplyr)
analysis <- rankings %>%
  left_join(metrics, by = c("app_id" = "unified_app_id")) %>%
  left_join(revenue, by = "app_id")
```

### Comparing Filtered vs Unfiltered Results

```r
# Get all games
all_games <- st_category_rankings(
  os = "ios",
  category = 6014,  # Games
  chart_type = "topgrossingapplications",
  limit = 200
)

# Get only your filtered games
filtered_games <- st_category_rankings(
  os = "ios",
  custom_fields_filter_id = filter_id,
  chart_type = "topgrossingapplications",
  limit = 200
)

# Compare
cat("Total games:", nrow(all_games), "\n")
cat("Filtered games:", nrow(filtered_games), "\n")
cat("Filter selected", 
    round(nrow(filtered_games) / nrow(all_games) * 100, 1), 
    "% of games\n")
```

## Best Practices

1. **Cache filter results** - Filter queries can be slow, so cache results when possible
2. **Use specific date ranges** - Narrow date ranges return faster
3. **Limit results appropriately** - Use the `limit` parameter to get only what you need
4. **Test filters first** - Verify a filter returns data in the web interface before using in R
5. **Document filter logic** - Keep notes on what each filter ID represents

## Future Enhancements

While `st_batch_metrics()`, `st_game_summary()`, and `st_top_publishers()` don't currently support custom filters due to API limitations, you can work around this by:

1. Using custom filters with supported functions to get app IDs
2. Passing those app IDs to the unsupported functions

```r
# Workaround example
# 1. Get app IDs from custom filter
filtered_apps <- st_top_charts(
  os = "ios",
  custom_fields_filter_id = filter_id,
  regions = "US",
  limit = 50
)

# 2. Use those IDs with batch metrics
app_ids <- unique(filtered_apps$entities.app_ids.ios)
metrics <- st_batch_metrics(
  os = "ios",
  app_list = app_ids,
  metrics = c("dau", "mau", "revenue"),
  date_range = "ytd",
  countries = "US"
)
```

# Sensor Tower Web Interface to API Parameter Mapping

## URL Analysis

From the provided URL:
```
https://app.sensortower.com/market-analysis/top-apps?os=unified&edit=1&granularity=weekly&start_date=2025-06-29&end_date=2025-07-28&duration=P30D&measure=DAU&comparison_attribute=absolute&country=US&country=AU...&category=0&device=iphone&device=ipad&device=android&metric=downloads&page=1&page_size=25&custom_fields_filter_mode=include_unified_apps&period=day
```

## Parameter Breakdown

### 1. Core Parameters (Direct Mapping)
| Web Parameter | API Parameter | Notes |
|--------------|---------------|-------|
| `os=unified` | `os` | Direct mapping |
| `measure=DAU` | `measure` | Direct mapping (DAU, WAU, MAU, revenue, units) |
| `comparison_attribute=absolute` | `comparison_attribute` | Direct mapping |
| `start_date=2025-06-29` | `date` or `start_date` | Depends on endpoint |
| `end_date=2025-07-28` | `end_date` | Optional in API |
| `category=0` | `category` | 0 means "All Categories" |
| `page_size=25` | `limit` | API uses `limit` instead |
| `page=1` | `offset` | API uses offset (calculate as (page-1)*limit) |

### 2. Country Parameter (Special Handling)
- **Web**: Multiple `country=XX` parameters (appears to be ~100 countries)
- **API**: `regions` parameter accepts comma-separated string or array
- **Note**: The URL shows virtually ALL countries, which might be equivalent to "WW" (worldwide)

### 3. Granularity/Time Parameters (Complex Mapping)
| Web Parameter | Purpose | API Equivalent |
|--------------|---------|----------------|
| `granularity=weekly` | Data aggregation level | `time_range` in API |
| `duration=P30D` | ISO 8601 duration (30 days) | Calculated from start/end dates |
| `period=day` | Unknown - conflicts with granularity? | Possibly ignored |

### 4. Device Parameters (Web-Specific)
- **Web**: `device=iphone&device=ipad&device=android`
- **API**: Uses `device_type` with values: "iphone", "ipad", "total", or none for Android
- **Note**: Having all devices selected in web likely maps to `os=unified` without device restrictions

### 5. Custom Fields Filter Mode
- **Web**: `custom_fields_filter_mode=include_unified_apps`
- **API**: `custom_tags_mode` with same values
- **Purpose**: Controls how unified apps are filtered when using custom filters

### 6. Parameters Not in API
- `edit=1` - UI state parameter (editing mode)
- `metric=downloads` - Seems redundant with `measure=DAU`
- `duration=P30D` - API calculates from date range

## Implementation Recommendations

### 1. Add Country Helper Function
```r
# Convert extensive country list to regions parameter
st_parse_countries <- function(countries) {
  # If more than 50 countries, suggest using "WW"
  if (length(countries) > 50) {
    message("Large country list detected. Consider using 'WW' for worldwide.")
    return("WW")
  }
  # Otherwise return comma-separated
  paste(countries, collapse = ",")
}
```

### 2. Handle "All Categories"
```r
# In st_top_charts and other functions
if (!is.null(category) && category == "0") {
  category <- NULL  # Let API return all categories
  message("Category '0' (All Categories) converted to NULL")
}
```

### 3. Support custom_fields_filter_mode Parameter
The `custom_fields_filter_mode` is already supported as `custom_tags_mode` in the API, but we should document the mapping:

```r
# Web parameter name -> API parameter name
web_to_api_params <- list(
  custom_fields_filter_mode = "custom_tags_mode",
  page_size = "limit",
  country = "regions"
)
```

### 4. URL Parser Function
```r
#' Parse Sensor Tower Web URL into API Parameters
#' 
#' @param url Character string. Sensor Tower web interface URL
#' @return List of API-compatible parameters
#' @export
st_parse_web_url <- function(url) {
  parsed <- httr::parse_url(url)
  query <- parsed$query
  
  # Initialize API parameters
  api_params <- list()
  
  # Direct mappings
  if (!is.null(query$os)) api_params$os <- query$os
  if (!is.null(query$measure)) api_params$measure <- query$measure
  if (!is.null(query$comparison_attribute)) {
    api_params$comparison_attribute <- query$comparison_attribute
  }
  if (!is.null(query$start_date)) api_params$date <- query$start_date
  if (!is.null(query$end_date)) api_params$end_date <- query$end_date
  if (!is.null(query$granularity)) {
    api_params$time_range <- switch(
      query$granularity,
      "daily" = "day",
      "weekly" = "week",
      "monthly" = "month",
      "quarterly" = "quarter",
      query$granularity
    )
  }
  
  # Category (handle "0" as all categories)
  if (!is.null(query$category)) {
    if (query$category != "0") {
      api_params$category <- query$category
    }
  }
  
  # Countries (handle multiple values)
  countries <- query[names(query) == "country"]
  if (length(countries) > 0) {
    api_params$regions <- if (length(countries) > 50) "WW" else paste(countries, collapse = ",")
  }
  
  # Page/limit
  if (!is.null(query$page_size)) api_params$limit <- as.integer(query$page_size)
  if (!is.null(query$page)) {
    page <- as.integer(query$page)
    limit <- as.integer(query$page_size %||% 25)
    api_params$offset <- (page - 1) * limit
  }
  
  # Custom fields filter mode -> custom tags mode
  if (!is.null(query$custom_fields_filter_mode)) {
    api_params$custom_tags_mode <- query$custom_fields_filter_mode
  }
  
  # Extract custom filter ID if present
  if (!is.null(query$custom_fields_filter_id)) {
    api_params$custom_fields_filter_id <- query$custom_fields_filter_id
  }
  
  api_params
}
```

## Example Usage

```r
# Parse the provided URL
url <- "https://app.sensortower.com/market-analysis/top-apps?os=unified&..."
params <- st_parse_web_url(url)

# Use in API call
do.call(st_top_charts, params)
```

## Key Differences Between Web and API

1. **Country Selection**: Web allows individual country selection, API prefers regions or comma-separated list
2. **Pagination**: Web uses page numbers, API uses offset
3. **Time Specification**: Web has redundant time parameters, API is more streamlined
4. **Device Filtering**: Web has explicit device checkboxes, API handles via os/device_type
5. **UI State**: Web URLs contain UI-specific parameters not needed for API

## Recommendations

1. **Implement URL parser**: Add `st_parse_web_url()` to help users transition from web to API
2. **Document parameter mappings**: Add this mapping to package documentation
3. **Handle edge cases**: All categories (0), worldwide (many countries), etc.
4. **Validate combinations**: Some parameter combinations may not be valid in the API

---
title: "Tidy Active User Workflows"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tidy Active User Workflows}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE, purl=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE,
  purl = FALSE
)
```

```{r, purl=FALSE}
library(sensortowerR)
library(dplyr)
library(tidyr)
```

## Overview

`st_active_users()` is a focused, tidy wrapper that returns DAU/WAU/MAU in long format, built on the same underlying machinery as `st_metrics()`.

This makes downstream analysis intuitive with standard tidyverse verbs.

## Basic Pipeline

```{r basic, purl=FALSE}
active_users <- st_active_users(
  os = "unified",
  app_list = c("5ba4585f539ce75b97db6bcb", "67ec0bf3e540b65904256cc4"),
  metrics = c("dau", "mau"),
  date_range = list(start_date = "2025-01-01", end_date = "2025-12-31"),
  countries = c("US", "GB"),
  granularity = "monthly"
)

active_summary <- active_users %>%
  group_by(app_id, original_id, app_name, metric, date, country) %>%
  summarise(total_active_users = sum(value, na.rm = TRUE), .groups = "drop")
```

## Wide Reporting Table

```{r wide, purl=FALSE}
active_wide <- active_users %>%
  group_by(app_id, original_id, app_name, date, country, metric) %>%
  summarise(value = sum(value, na.rm = TRUE), .groups = "drop") %>%
  pivot_wider(names_from = metric, values_from = value)
```

## Join With Sales Data

```{r join-sales, purl=FALSE}
sales <- st_metrics(
  app_id      = c("5ba4585f539ce75b97db6bcb", "67ec0bf3e540b65904256cc4"),
  os          = "unified",
  metrics     = c("revenue", "downloads"),
  countries   = c("US", "GB"),
  date_from   = "2025-01-01",
  date_to     = "2025-12-31",
  granularity = "monthly"
)

sales_wide <- sales %>%
  group_by(app_id, date, country, metric) %>%
  summarise(value = sum(value, na.rm = TRUE), .groups = "drop") %>%
  pivot_wider(names_from = metric, values_from = value)

combined <- active_wide %>%
  left_join(sales_wide, by = c("app_id", "date", "country"))
```

## App-Level KPI Aggregation

```{r kpi, purl=FALSE}
app_kpis <- combined %>%
  mutate(
    rev_per_dau = if_else(!is.na(dau) & dau > 0, revenue / dau, NA_real_),
    rev_per_mau = if_else(!is.na(mau) & mau > 0, revenue / mau, NA_real_)
  ) %>%
  group_by(app_id, original_id, app_name, country) %>%
  summarise(
    avg_dau = mean(dau, na.rm = TRUE),
    avg_mau = mean(mau, na.rm = TRUE),
    total_revenue = sum(revenue, na.rm = TRUE),
    avg_rev_per_dau = mean(rev_per_dau, na.rm = TRUE),
    avg_rev_per_mau = mean(rev_per_mau, na.rm = TRUE),
    .groups = "drop"
  )
```

# Migration Guide for SensorTowerR v0.2.3

## Overview

SensorTowerR v0.2.3 introduces automatic revenue standardization across all functions. This guide helps you migrate existing code to use the new standardized revenue columns.

## What Changed?

Previously, different SensorTowerR functions returned revenue in different units:
- `st_top_charts()`: cents (revenue_absolute)
- `st_sales_report()`: dollars (total_revenue)
- `st_top_publishers()`: cents (revenue_absolute)

Now, all functions provide standardized revenue columns in base currency units (dollars for USD, euros for EUR, etc.).

## Quick Migration Reference

### st_top_charts()

**Before v0.2.3:**
```r
top_games <- st_top_charts(...)
top_games %>%
  mutate(revenue_dollars = revenue_absolute / 100)
```

**After v0.2.3:**
```r
top_games <- st_top_charts(...)
# Use the 'revenue' column directly - it's already in dollars
top_games %>%
  select(app_name, revenue)
```

### st_sales_report()

**No changes needed** - This function already returned values in base currency units.

### st_top_publishers()

**Before v0.2.3:**
```r
publishers <- st_top_publishers(...)
publishers %>%
  mutate(revenue_dollars = revenue_absolute / 100)
```

**After v0.2.3:**
```r
publishers <- st_top_publishers(...)
# Use 'revenue_usd' column directly
publishers %>%
  select(publisher_name, revenue_usd)
```

## Backward Compatibility Pattern

If your code needs to work with both old and new versions:

```r
# For st_top_charts()
process_revenue <- function(data) {
  if ("revenue" %in% names(data)) {
    # v0.2.3+ - use standardized column
    data$revenue_millions <- data$revenue / 1e6
  } else {
    # Pre-v0.2.3 - manual conversion
    data$revenue_millions <- data$revenue_absolute / 100 / 1e6
  }
  return(data)
}

top_games <- st_top_charts(...) %>%
  process_revenue()
```

## Common Migration Scenarios

### Scenario 1: Revenue in Millions

**Old code:**
```r
mutate(revenue_millions = revenue_absolute / 1e8)  # cents to millions
```

**New code:**
```r
mutate(revenue_millions = revenue / 1e6)  # dollars to millions
```

### Scenario 2: Revenue Formatting

**Old code:**
```r
mutate(revenue_fmt = scales::dollar(revenue_absolute / 100))
```

**New code:**
```r
mutate(revenue_fmt = scales::dollar(revenue))
```

### Scenario 3: Gini Coefficient Calculations

**Old code:**
```r
gini_coef <- calculate_gini(data$revenue_absolute / 100)
```

**New code:**
```r
gini_coef <- calculate_gini(data$revenue)
```

## Checking Your Version

```r
# Check if you have v0.2.3+
if (packageVersion("SensorTowerR") >= "0.2.3") {
  message("You have the standardized revenue version!")
} else {
  message("Please update: devtools::install_github('econosopher/SensorTowerR')")
}
```

## Benefits of Upgrading

1. **Consistency**: Same units across all functions
2. **Simplicity**: No manual conversions needed
3. **Clarity**: Clear column names indicate units
4. **Compatibility**: Original columns preserved for backward compatibility

## Need Help?

- Check the revenue units guide: `inst/docs/revenue_units_guide.md`
- See examples: `inst/examples/revenue_standardization_example.R`
- Report issues: https://github.com/econosopher/SensorTowerR/issues

# Custom Filter Integration Strategy for SensorTowerR

## Executive Summary

This document outlines a comprehensive strategy for enhancing custom field filter support across the SensorTowerR package, making it easier for users to leverage Sensor Tower's web-based filter creation capabilities in their R workflows.

## Current State

### What's Implemented
- `st_top_charts()` supports `custom_fields_filter_id` parameter
- Basic validation that either `category` or `custom_fields_filter_id` is provided
- Support for `custom_tags_mode` when `os = "unified"`

### Gaps
- No filter validation functions
- Limited error handling for invalid filter IDs
- No helper functions for filter management
- Other potentially compatible functions don't support custom filters
- No documentation about which endpoints support filters

## Proposed Enhancements

### 1. Core Infrastructure

#### A. Filter Validation Function
```r
#' Validate Custom Field Filter ID
#' 
#' @param filter_id Character string. The filter ID to validate
#' @param test_call Logical. Whether to make a test API call
#' @return List with validation results
#' @export
st_validate_filter <- function(filter_id, test_call = FALSE) {
  # Check format
  valid_format <- grepl("^[a-f0-9]{24}$", filter_id)
  
  result <- list(
    filter_id = filter_id,
    valid_format = valid_format,
    tested = FALSE,
    test_result = NULL
  )
  
  if (test_call && valid_format) {
    # Make minimal API call to test
    test <- tryCatch(
      st_top_charts(
        os = "ios",
        custom_fields_filter_id = filter_id,
        regions = "US",
        limit = 1
      ),
      error = function(e) e
    )
    
    result$tested <- TRUE
    result$test_result <- !inherits(test, "error")
    if (inherits(test, "error")) {
      result$error_message <- test$message
    }
  }
  
  class(result) <- c("st_filter_validation", "list")
  result
}
```

#### B. Filter Management System
```r
# R/st_filter_manager.R

# Environment to store filter mappings
.filter_cache <- new.env(parent = emptyenv())

#' Register a custom filter
#' 
#' @param name Character string. Friendly name for the filter
#' @param filter_id Character string. The Sensor Tower filter ID
#' @param description Character string. Optional description
#' @export
st_register_filter <- function(name, filter_id, description = NULL) {
  if (!st_validate_filter(filter_id)$valid_format) {
    stop("Invalid filter ID format")
  }
  
  .filter_cache[[name]] <- list(
    id = filter_id,
    description = description,
    registered = Sys.time()
  )
  
  invisible(TRUE)
}

#' Get registered filter ID by name
#' 
#' @param name Character string. Filter name
#' @return Character string. Filter ID
#' @export
st_get_filter <- function(name) {
  if (!exists(name, envir = .filter_cache)) {
    stop("Filter '", name, "' not found. Use st_list_filters() to see available filters.")
  }
  .filter_cache[[name]]$id
}

#' List all registered filters
#' 
#' @return Data frame of registered filters
#' @export
st_list_filters <- function() {
  filters <- ls(envir = .filter_cache)
  if (length(filters) == 0) {
    return(data.frame(
      name = character(),
      filter_id = character(),
      description = character(),
      registered = character()
    ))
  }
  
  do.call(rbind, lapply(filters, function(name) {
    f <- .filter_cache[[name]]
    data.frame(
      name = name,
      filter_id = f$id,
      description = f$description %||% "",
      registered = f$registered
    )
  }))
}
```

### 2. Enhanced Error Handling

```r
# utils.R enhancement
handle_filter_error <- function(e, filter_id) {
  if (grepl("422", e$message)) {
    stop(
      "Invalid custom filter ID: ", filter_id, "\n",
      "This could mean:\n",
      "- The filter ID doesn't exist\n",
      "- The filter belongs to a different account\n",
      "- The filter has been deleted\n",
      "To get a valid filter ID, visit app.sensortower.com and create a filter",
      call. = FALSE
    )
  } else if (grepl("403", e$message)) {
    stop(
      "Access denied for filter ID: ", filter_id, "\n",
      "This filter may belong to a different Sensor Tower account",
      call. = FALSE
    )
  } else {
    stop(e)
  }
}
```

### 3. Extend Filter Support to Other Functions

#### A. Identify Compatible Functions
Based on API analysis, these functions could support custom filters:
- `st_sales_report()` - Uses sales_report_estimates endpoint
- `st_batch_metrics()` - Could pass filters through to underlying calls
- Future market analysis functions

#### B. Implementation Pattern
```r
# Add to function parameters
#' @param custom_fields_filter_id Optional. Custom filter ID from Sensor Tower
#' @param custom_tags_mode Optional. Required for unified OS with custom filter

# In function body
if (!is.null(custom_fields_filter_id)) {
  # Validate filter
  if (!grepl("^[a-f0-9]{24}$", custom_fields_filter_id)) {
    stop("Invalid custom_fields_filter_id format")
  }
  
  # Add to query parameters
  query_params$custom_fields_filter_id <- custom_fields_filter_id
  
  # Handle unified OS requirement
  if (os == "unified" && is.null(custom_tags_mode)) {
    stop("custom_tags_mode is required when using custom_fields_filter_id with unified OS")
  }
}
```

### 4. Configuration File Support

```r
#' Load filters from configuration file
#' 
#' @param file Path to YAML or JSON file with filter definitions
#' @export
st_load_filters <- function(file = "sensor_tower_filters.yml") {
  if (!file.exists(file)) {
    stop("Filter configuration file not found: ", file)
  }
  
  # Support both YAML and JSON
  if (grepl("\\.ya?ml$", file)) {
    filters <- yaml::read_yaml(file)
  } else if (grepl("\\.json$", file)) {
    filters <- jsonlite::fromJSON(file)
  } else {
    stop("Unsupported file format. Use .yml, .yaml, or .json")
  }
  
  # Register each filter
  for (name in names(filters)) {
    st_register_filter(
      name = name,
      filter_id = filters[[name]]$id,
      description = filters[[name]]$description
    )
  }
  
  message("Loaded ", length(filters), " filters from ", file)
  invisible(TRUE)
}
```

Example configuration file (sensor_tower_filters.yml):
```yaml
hypercasual_games:
  id: "687df26ac5a19ebcfe817d7f"
  description: "Hypercasual games from top publishers"

strategy_asia:
  id: "789ef34bc6b20fcade928e8a"
  description: "Strategy games popular in Asian markets"

premium_apps:
  id: "456ab78cd9e31gbcef103f2b"
  description: "Premium apps over $4.99"
```

### 5. User Workflow Enhancements

#### A. Interactive Filter Builder
```r
#' Build a Sensor Tower filter URL
#' 
#' @param base_url Base URL for Sensor Tower
#' @param ... Filter parameters
#' @export
st_build_filter_url <- function(
  base_url = "https://app.sensortower.com/top-charts",
  os = "unified",
  category = NULL,
  countries = NULL,
  ...
) {
  # Build URL with parameters
  params <- list(os = os, ...)
  if (!is.null(category)) params$category <- category
  if (!is.null(countries)) params$countries <- paste(countries, collapse = ",")
  
  url <- httr::modify_url(base_url, query = params)
  
  message("Visit this URL to create your filter:")
  message(url)
  message("\nAfter creating the filter, copy the custom_fields_filter_id from the URL")
  
  invisible(url)
}
```

#### B. Filter Testing Utility
```r
#' Test a custom filter
#' 
#' @param filter_id Filter ID to test
#' @param verbose Logical. Print detailed results
#' @export
st_test_filter <- function(filter_id, verbose = TRUE) {
  if (verbose) cat("Testing filter:", filter_id, "\n\n")
  
  # Test on different OS
  results <- list()
  
  for (os in c("ios", "android", "unified")) {
    if (verbose) cat("Testing", os, "... ")
    
    args <- list(
      os = os,
      custom_fields_filter_id = filter_id,
      regions = "US",
      limit = 5
    )
    
    if (os == "unified") {
      args$custom_tags_mode <- "include"
    }
    
    result <- tryCatch(
      do.call(st_top_charts, args),
      error = function(e) e
    )
    
    success <- !inherits(result, "error")
    results[[os]] <- success
    
    if (verbose) {
      cat(ifelse(success, "✅ Success", "❌ Failed"), "\n")
      if (!success) {
        cat("  Error:", result$message, "\n")
      } else if (is.data.frame(result)) {
        cat("  Returned", nrow(result), "apps\n")
      }
    }
  }
  
  invisible(results)
}
```

### 6. Documentation Strategy

#### A. Vignette: "Using Custom Filters"
Create a comprehensive vignette covering:
- What are custom filters
- How to create them in Sensor Tower
- Using filters in SensorTowerR
- Managing multiple filters
- Best practices
- Troubleshooting

#### B. Function Documentation
Update all compatible functions with:
- Parameter documentation for custom_fields_filter_id
- Examples using custom filters
- Links to the vignette

#### C. README Section
Add prominent section about custom filter support with quick examples

### 7. Testing Framework

```r
# tests/testthat/test-custom-filters.R

test_that("Filter validation works correctly", {
  expect_true(st_validate_filter("687df26ac5a19ebcfe817d7f")$valid_format)
  expect_false(st_validate_filter("invalid")$valid_format)
  expect_false(st_validate_filter("687df26ac5a19ebcfe817d7")$valid_format)
})

test_that("Filter registration and retrieval work", {
  st_register_filter("test_filter", "687df26ac5a19ebcfe817d7f", "Test filter")
  expect_equal(st_get_filter("test_filter"), "687df26ac5a19ebcfe817d7f")
  expect_error(st_get_filter("nonexistent"))
})

test_that("Custom filter parameters work in API calls", {
  skip_if_no_auth()
  
  # Test with mock filter ID
  expect_error(
    st_top_charts(
      custom_fields_filter_id = "invalid_id",
      regions = "US"
    ),
    "Invalid custom_fields_filter_id format"
  )
})
```

## Implementation Phases

### Phase 1: Core Infrastructure (Week 1)
- Implement filter validation function
- Add filter management system
- Enhance error handling
- Create basic tests

### Phase 2: Function Integration (Week 2)
- Extend filter support to compatible functions
- Update parameter validation
- Add configuration file support
- Update documentation

### Phase 3: User Experience (Week 3)
- Create interactive utilities
- Write comprehensive vignette
- Add examples to README
- Create tutorial blog post

### Phase 4: Testing and Release (Week 4)
- Comprehensive testing
- User acceptance testing
- Documentation review
- Package release

## Success Metrics

1. **Adoption**: Number of users utilizing custom filters
2. **Error Reduction**: Decrease in filter-related support issues
3. **Efficiency**: Time saved by using pre-configured filters
4. **Coverage**: Percentage of compatible functions supporting filters

## Conclusion

This integration strategy will make SensorTowerR the most comprehensive R interface for Sensor Tower's custom filter functionality, significantly improving user workflow efficiency and reducing the complexity of working with complex data filters.

# Custom Field Filter Investigation Summary

## Overview
This document summarizes the comprehensive investigation into Sensor Tower's custom field filter functionality and its integration into the SensorTowerR package.

## Key Findings

### 1. How Custom Filters Work
- **Creation**: Users create filters in the Sensor Tower web interface (app.sensortower.com)
- **Format**: 24-character hexadecimal string (e.g., "687df26ac5a19ebcfe817d7f")
- **Persistence**: Filter IDs remain valid as long as they exist in the account
- **Account-specific**: Filters are tied to specific Sensor Tower accounts

### 2. API Support
Custom field filters are supported in:
- **Market Analysis endpoints**: `/v1/{os}/sales_report_estimates_comparison_attributes`
- **Top and Trending endpoints**: Including active user metrics
- Currently implemented in `st_top_charts()` function

### 3. Implementation Requirements
- **iOS/Android**: Use `custom_fields_filter_id` parameter directly
- **Unified OS**: Must also provide `custom_tags_mode` parameter
  - Options: "include", "exclude", "include_unified_apps"
- **Category**: Becomes optional when custom filter is provided
- **Other parameters**: Still apply (regions, time_range, etc.)

## What Was Created

### 1. Documentation
- **Custom Fields Filter Guide** (`inst/docs/custom_fields_filter_guide.md`)
  - Comprehensive explanation of how filters work
  - API implementation details
  - Testing strategies
  - Best practices

- **Integration Strategy** (`inst/docs/custom_filter_integration_strategy.md`)
  - Roadmap for enhanced filter support
  - Proposed helper functions
  - Implementation phases
  - Success metrics

### 2. Code Enhancements

#### A. Updated `st_top_charts()` Function
- Made `category` parameter optional when `custom_fields_filter_id` is provided
- Improved error messages for better user guidance

#### B. New Helper Functions (`R/st_filter_helpers.R`)
```r
# Validate filter ID format
st_is_valid_filter_id("687df26ac5a19ebcfe817d7f")

# Test if filter works with API
st_test_filter("687df26ac5a19ebcfe817d7f")

# Generate example IDs for testing
st_generate_example_filter_ids(5)

# Build URL for creating filters
st_build_filter_url(os = "ios", category = 6014)

# Extract filter ID from URL
st_extract_filter_id("https://app.sensortower.com/...?custom_fields_filter_id=...")

# Compare filtered vs unfiltered results
st_compare_filter_results(filter_id, category = 6014)
```

### 3. Examples and Tests

#### A. Basic Usage Example (`examples/custom_filter_example.R`)
- Demonstrates using custom filters with different OS types
- Shows parameter requirements
- Includes visualization example
- Provides troubleshooting guidance

#### B. Comprehensive Test Suite (`inst/test_scripts/test_custom_field_filters.R`)
- Format validation tests
- API functionality tests
- Parameter interaction tests
- Different measure types (revenue, downloads, DAU/WAU/MAU)
- Time range variations
- Error handling tests

#### C. Filter Discovery Helper (`examples/find_test_filter_ids.R`)
- Methods to obtain valid filter IDs
- URL generation for filter creation
- Configuration file templates
- Testing utilities

### 4. Updated Package Documentation
- Enhanced README with custom filter section
- Updated function documentation
- Added to NEWS.md for version 0.7.3

## Testing Results

### What Works
✅ Filter ID format validation  
✅ iOS/Android filter support  
✅ Unified OS with custom_tags_mode  
✅ Different measures (revenue, units, active users)  
✅ Various time ranges  
✅ Parameter validation  

### Known Issues
❌ Example filter ID (687df26ac5a19ebcfe817d7f) may not work with all accounts  
❌ No way to retrieve filter details via API  
❌ Filters are opaque - can't see what they're filtering  

## Best Practices

### 1. Filter Management
```r
# Store filter IDs in configuration
FILTERS <- list(
  hypercasual = "687df26ac5a19ebcfe817d7f",
  strategy_asia = "789ef34bc6b20fcade928e8a"
)

# Use descriptive names
top_games <- st_top_charts(
  custom_fields_filter_id = FILTERS$hypercasual,
  regions = "US"
)
```

### 2. Error Handling
```r
# Always validate before use
if (st_is_valid_filter_id(filter_id)) {
  result <- st_test_filter(filter_id)
  if (result$success) {
    # Use the filter
  }
}
```

### 3. Documentation
- Document what each filter does
- Include the web UI configuration
- Note the creation date
- Track which analyses use which filters

## Future Enhancements

### Phase 1: Immediate (Completed)
- ✅ Basic filter support in st_top_charts()
- ✅ Helper functions for validation
- ✅ Documentation and examples

### Phase 2: Short-term
- [ ] Extend filter support to other compatible functions
- [ ] Add filter caching mechanism
- [ ] Create vignette on custom filters

### Phase 3: Long-term
- [ ] Filter discovery tools
- [ ] Multi-filter support
- [ ] Filter sharing utilities
- [ ] Integration with other Sensor Tower features

## How to Get Started

### For Users
1. Log into app.sensortower.com
2. Create filters in the web interface
3. Copy the filter ID from the URL
4. Use in R:
```r
library(SensorTowerR)

# Test your filter
st_test_filter("your_filter_id_here")

# Use in analysis
data <- st_top_charts(
  custom_fields_filter_id = "your_filter_id_here",
  regions = "US"
)
```

### For Developers
1. Review the helper functions in `R/st_filter_helpers.R`
2. See integration strategy document for enhancement ideas
3. Run test suite to understand functionality
4. Contribute improvements via pull requests

## Conclusion

Custom field filters provide a powerful way to leverage Sensor Tower's web interface filtering capabilities in programmatic workflows. The implementation in SensorTowerR now provides basic support with room for significant enhancements based on user needs and API capabilities.

# Custom Fields Filter Guide for SensorTowerR

## Overview

Custom field filters are a powerful feature in Sensor Tower that allows users to create complex, reusable filters in the web interface and then apply them via the API using a unique filter ID.

## How Custom Field Filters Work

### 1. Filter Creation Process
1. User logs into Sensor Tower web interface (app.sensortower.com)
2. Navigates to the data section (e.g., Top Apps, Market Analysis)
3. Configures filters using the UI:
   - Categories (e.g., specific game genres)
   - Publishers (include/exclude specific publishers)
   - Custom tags/fields
   - Geographic regions
   - Platform specifications
   - Time ranges
   - Revenue/download thresholds
4. The web interface generates a unique filter ID
5. This ID appears in the URL as `custom_fields_filter_id=XXXXXXXXXX`

### 2. Filter ID Structure
- Format: 24-character hexadecimal string (e.g., "687df26ac5a19ebcfe817d7f")
- Persistent: Filter IDs remain valid as long as the filter exists in your account
- Account-specific: Filter IDs are tied to your Sensor Tower account
- Shareable within organization: Team members with the same account access can use the same filter IDs

### 3. API Implementation

The filter ID replaces the need to specify individual filter parameters in API calls:

```r
# Instead of this complex filter specification:
st_top_charts(
  category = 6014,
  regions = c("US", "GB", "JP"),
  # ... many other filter parameters
)

# You can use:
st_top_charts(
  custom_fields_filter_id = "687df26ac5a19ebcfe817d7f",
  regions = "US"  # Still need to specify regions
)
```

## Supported Endpoints

Based on API documentation analysis, `custom_fields_filter_id` is supported in:

### 1. Market Analysis Endpoints
- `/v1/{os}/sales_report_estimates_comparison_attributes` (used by `st_top_charts()`)
- Supports revenue and download metrics
- Can be combined with time range specifications

### 2. Top and Trending Endpoints
- `/v1/{os}/top_and_trending/active_users` (DAU/WAU/MAU metrics)
- Potentially other trending endpoints

## Important Requirements

### 1. Platform-Specific Behavior
- **iOS/Android**: Use `custom_fields_filter_id` directly
- **Unified OS**: Must also provide `custom_tags_mode` parameter
  - Options: "include", "exclude", "include_unified_apps"
  - This determines how the filter applies to unified apps

### 2. Parameter Interactions
- When `custom_fields_filter_id` is provided:
  - `category` becomes optional (filter already contains category info)
  - `regions` is still required (can override filter's region)
  - Time parameters (`date`, `time_range`) still apply
  - `limit` and `offset` work normally

### 3. Error Handling
Common errors and their causes:
- **422 Error**: Invalid filter ID or filter doesn't exist
- **403 Error**: Filter belongs to different account
- **400 Error**: Missing required parameters (e.g., `custom_tags_mode` for unified)

## Testing Strategy

### 1. Validation Tests
```r
# Test 1: Verify filter ID format
is_valid_filter_id <- function(id) {
  grepl("^[a-f0-9]{24}$", id)
}

# Test 2: Compare filtered vs unfiltered results
test_filter_effectiveness <- function(filter_id) {
  # Get data with filter
  filtered <- st_top_charts(
    custom_fields_filter_id = filter_id,
    regions = "US"
  )
  
  # Get unfiltered data
  unfiltered <- st_top_charts(
    category = 6014,  # Games
    regions = "US"
  )
  
  # Filter should return subset
  expect_true(nrow(filtered) <= nrow(unfiltered))
}
```

### 2. Integration Tests
```r
# Test different OS types
test_os_compatibility <- function(filter_id) {
  # iOS test
  ios_result <- tryCatch(
    st_top_charts(
      os = "ios",
      custom_fields_filter_id = filter_id,
      regions = "US"
    ),
    error = function(e) NULL
  )
  
  # Android test
  android_result <- tryCatch(
    st_top_charts(
      os = "android", 
      custom_fields_filter_id = filter_id,
      regions = "US"
    ),
    error = function(e) NULL
  )
  
  # Unified test (requires custom_tags_mode)
  unified_result <- tryCatch(
    st_top_charts(
      os = "unified",
      custom_fields_filter_id = filter_id,
      custom_tags_mode = "include",
      regions = "US"
    ),
    error = function(e) NULL
  )
  
  list(
    ios = !is.null(ios_result),
    android = !is.null(android_result),
    unified = !is.null(unified_result)
  )
}
```

## Best Practices

### 1. Filter Management
- Document filter IDs with their purpose
- Store filter IDs in configuration files
- Create a mapping of filter names to IDs:

```r
# config/filters.R
SENSOR_TOWER_FILTERS <- list(
  hypercasual_games = "687df26ac5a19ebcfe817d7f",
  top_publishers = "789ef34bc6b20fcade928e8a",
  strategy_games_asia = "456ab78cd9e31gbcef103f2b"
)
```

### 2. Error Recovery
```r
safe_filter_query <- function(filter_id, ...) {
  result <- tryCatch(
    st_top_charts(
      custom_fields_filter_id = filter_id,
      ...
    ),
    error = function(e) {
      if (grepl("422", e$message)) {
        warning("Invalid filter ID, falling back to category filter")
        st_top_charts(category = 6014, ...)
      } else {
        stop(e)
      }
    }
  )
  result
}
```

### 3. Performance Optimization
- Cache results when using the same filter repeatedly
- Filters are processed server-side, so complex filters don't impact API performance
- Use filters for commonly repeated queries to reduce parameter complexity

## Future Integration Ideas

### 1. Filter Discovery Function
```r
# Proposed function to validate and describe filters
st_describe_filter <- function(filter_id) {
  # Make a minimal API call to test the filter
  # Return information about what the filter does
}
```

### 2. Filter Builder Helper
```r
# Help users construct filter URLs
st_filter_url <- function(categories = NULL, 
                         publishers = NULL,
                         regions = NULL,
                         ...) {
  base_url <- "https://app.sensortower.com/top-charts"
  # Build URL with parameters
  # User can then create filter in web UI
}
```

### 3. Multi-Filter Support
```r
# Allow combining multiple filters
st_top_charts(
  custom_fields_filter_ids = c("filter1", "filter2"),
  filter_mode = "intersection"  # or "union"
)
```

## Known Limitations

1. **Filter Opacity**: API doesn't return filter details, only applies them
2. **No CRUD Operations**: Can't create/update/delete filters via API
3. **Account Dependency**: Filters are account-specific, not portable
4. **Limited Documentation**: Sensor Tower doesn't provide detailed filter specs

## Recommendations for Package Enhancement

1. **Add filter validation**: Check format before API calls
2. **Improve error messages**: Provide guidance when filter fails
3. **Add examples**: Show common filter use cases
4. **Cache filter results**: Store successful filter IDs
5. **Documentation**: Add filter guide to package vignettes

# Custom Filter Integration Plan for SensorTowerR

## Overview
This document outlines how to integrate custom filter support (`custom_fields_filter_id` and `custom_tags_mode`) across multiple SensorTowerR functions to maximize reusability.

## Functions That Should Support Custom Filters

### 1. Already Implemented
- ✅ `st_top_charts()` - Full support for custom filters

### 2. High Priority - Should Add Support
These functions query similar endpoints and would benefit from custom filter support:

#### `st_category_rankings()`
- **Current**: Fetches official app store rankings by category
- **Enhancement**: Add custom filter to get rankings within a filtered subset
- **Use Case**: "Show me iOS game rankings, but only for games from my custom filter"
- **Implementation**: Add `custom_fields_filter_id` and `custom_tags_mode` parameters

#### `st_sales_report()`
- **Current**: Gets revenue/download data for specific apps
- **Enhancement**: Filter which apps to include in the report
- **Use Case**: "Get sales data only for apps matching my publisher filter"
- **Implementation**: Add filter parameters to the endpoint call

#### `st_game_summary()`
- **Current**: Aggregated game market data by category
- **Enhancement**: Apply custom filters to market analysis
- **Use Case**: "Show market summary only for hypercasual games"
- **Implementation**: Pass filter to aggregation endpoint

### 3. Medium Priority - Could Benefit
#### `st_top_publishers()`
- **Current**: Top publishers by revenue/downloads
- **Enhancement**: Filter to specific publisher groups
- **Use Case**: "Top publishers in my competitive set"

#### `st_batch_metrics()`
- **Current**: Fetches metrics for multiple apps
- **Enhancement**: Pre-filter the app list
- **Use Case**: "Get metrics for all apps matching my filter"

### 4. Functions That Don't Need Custom Filters
These work with specific app IDs and wouldn't benefit from filters:
- ❌ `st_app_info()` - Searches for specific apps
- ❌ `st_app_details()` - Gets details for specific app IDs
- ❌ `st_app_lookup()` - Looks up specific app IDs
- ❌ `st_publisher_apps()` - Gets apps from specific publisher

## Implementation Strategy

### Step 1: Create Shared Parameter Handler
```r
# In utils.R or new file custom_filter_utils.R

#' Add custom filter parameters to query
#' @keywords internal
add_custom_filter_params <- function(query_params, 
                                   custom_fields_filter_id = NULL,
                                   custom_tags_mode = NULL,
                                   os = NULL) {
  
  if (!is.null(custom_fields_filter_id)) {
    # Validate filter ID
    if (!st_is_valid_filter_id(custom_fields_filter_id)) {
      stop("Invalid custom_fields_filter_id format")
    }
    
    query_params$custom_fields_filter_id <- custom_fields_filter_id
    
    # Add custom_tags_mode if unified OS
    if (!is.null(os) && os == "unified") {
      if (is.null(custom_tags_mode)) {
        stop("custom_tags_mode is required when using custom_fields_filter_id with unified OS")
      }
      query_params$custom_tags_mode <- custom_tags_mode
    }
  }
  
  query_params
}
```

### Step 2: Update Function Signatures
Add these parameters to relevant functions:
```r
#' @param custom_fields_filter_id Optional. Character string. ID of a Sensor
#'   Tower custom field filter to apply. Use filter IDs from the web interface.
#' @param custom_tags_mode Optional. Character string. Required if `os` is
#'   'unified' and `custom_fields_filter_id` is provided. Options: "include",
#'   "exclude", "include_unified_apps".
```

### Step 3: Example Implementation for st_category_rankings
```r
st_category_rankings <- function(os,
                               category = NULL,  # Make optional with filter
                               chart_type = NULL,
                               country = "US",
                               date = NULL,
                               limit = 100,
                               offset = 0,
                               custom_fields_filter_id = NULL,
                               custom_tags_mode = NULL,
                               auth_token = NULL) {
  
  # Validate: need either category or custom filter
  if (is.null(category) && is.null(custom_fields_filter_id)) {
    stop("Either 'category' or 'custom_fields_filter_id' must be provided")
  }
  
  # Build query parameters
  query_params <- list(
    auth_token = auth_token,
    category = category,
    chart_type = chart_type,
    country = country,
    limit = limit,
    offset = offset
  )
  
  # Add custom filter params
  query_params <- add_custom_filter_params(
    query_params,
    custom_fields_filter_id,
    custom_tags_mode,
    os
  )
  
  # Continue with API call...
}
```

## Usage Examples

### 1. Category Rankings with Custom Filter
```r
# Get rankings for games in my custom competitive set
rankings <- st_category_rankings(
  os = "ios",
  custom_fields_filter_id = "60746340241bc16eb8a65d76",
  chart_type = "topgrossingapplications",
  country = "US"
)
```

### 2. Sales Report with Custom Filter
```r
# Get revenue data for apps matching my filter
sales <- st_sales_report(
  os = "unified",
  custom_fields_filter_id = "60746340241bc16eb8a65d76",
  custom_tags_mode = "include_unified_apps",
  start_date = "2025-01-01",
  end_date = "2025-01-31"
)
```

### 3. Game Market Summary with Filter
```r
# Market analysis for specific game subset
market <- st_game_summary(
  os = "unified",
  custom_fields_filter_id = "hypercasual_filter_id",
  custom_tags_mode = "include",
  countries = c("US", "GB", "JP")
)
```

### 4. Combining URL Parsing with Custom Filters
```r
# Parse a web URL and use the filter elsewhere
url <- "https://app.sensortower.com/..."
params <- st_parse_web_url(url)

# Use the extracted filter in different functions
if (!is.null(params$custom_fields_filter_id)) {
  # Get rankings
  rankings <- st_category_rankings(
    os = params$os,
    custom_fields_filter_id = params$custom_fields_filter_id,
    custom_tags_mode = params$custom_tags_mode
  )
  
  # Get publisher analysis
  publishers <- st_top_publishers(
    os = params$os,
    custom_fields_filter_id = params$custom_fields_filter_id,
    custom_tags_mode = params$custom_tags_mode
  )
}
```

## Benefits of This Approach

1. **Consistency**: Same parameter names and behavior across all functions
2. **Reusability**: Create one filter, use it everywhere
3. **Flexibility**: Mix and match filters with other parameters
4. **Web Integration**: Seamlessly use filters from web interface
5. **Validation**: Centralized validation logic

## Testing Strategy

### 1. Unit Tests
```r
test_that("Custom filter parameters work across functions", {
  filter_id <- "60746340241bc16eb8a65d76"
  
  # Test with each function
  expect_error(
    st_category_rankings(os = "unified", custom_fields_filter_id = filter_id),
    "custom_tags_mode is required"
  )
  
  expect_s3_class(
    st_category_rankings(
      os = "ios",
      custom_fields_filter_id = filter_id
    ),
    "data.frame"
  )
})
```

### 2. Integration Tests
- Test filter ID validation
- Test parameter combinations
- Test error messages
- Test with real filter IDs

## Documentation Updates

### README Section
```markdown
## Using Custom Filters Across Functions

Custom filters created in the Sensor Tower web interface can be used across multiple functions:

| Function | Custom Filter Support | Use Case |
|----------|---------------------|----------|
| `st_top_charts()` | ✅ Full support | Filter top apps by custom criteria |
| `st_category_rankings()` | ✅ Full support | Rankings within filtered apps |
| `st_sales_report()` | ✅ Full support | Revenue for filtered apps |
| `st_game_summary()` | ✅ Full support | Market analysis with filters |
| `st_top_publishers()` | ✅ Full support | Publishers of filtered apps |
| `st_batch_metrics()` | ✅ Full support | Metrics for filtered app sets |
```

## Implementation Timeline

1. **Phase 1** (Immediate): Update `st_category_rankings()`
2. **Phase 2** (Next): Update `st_sales_report()` and `st_game_summary()`
3. **Phase 3** (Future): Update `st_top_publishers()` and `st_batch_metrics()`
4. **Phase 4** (Ongoing): Add examples and vignettes

## Conclusion

By systematically adding custom filter support across relevant functions, we can make the SensorTowerR package much more powerful for users who leverage Sensor Tower's web interface filters. This creates a seamless workflow between web and R environments.

---
title: "Working with Custom Fields Filters"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with Custom Fields Filters}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE, purl = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE,
  purl = FALSE
)
```

```{r, purl = FALSE}
library(sensortowerR)
library(dplyr)
```

## Introduction

Custom Fields Filters are a powerful feature in Sensor Tower that let you create complex, reusable queries to segment apps across hundreds of attributes. In `sensortowerR` 1.0.0, every filter is built with a single verb — `st_filter()` — which returns an S3 object accepted anywhere a raw filter ID string was accepted before.

## Understanding Custom Fields

Sensor Tower tracks over 400 custom fields, including:

- **Game attributes**: Genre, sub-genre, art style, camera POV
- **Monetization**: Free/paid, IAP, ads, subscriptions
- **Technical**: SDKs used, platform support, device compatibility
- **Performance**: Retention rates, ratings, download counts
- **Metadata**: Release dates, publisher info, update frequency

## Basic workflow

### 1. Discover available fields

```{r, purl = FALSE}
# List every custom field you can filter on
fields <- st_discover_fields()

# See valid values for a specific field
st_custom_fields_values("Primary Genre")
```

### 2. Build a filter

```{r, purl = FALSE}
# Single-criterion filter
puzzle_filter <- st_filter(genre = "Puzzle")

# Multi-criterion filter (all AND-combined)
word_puzzle_free <- st_filter(
  genre        = c("Puzzle", "Word"),
  monetization = "Free"
)

# Date range
new_releases <- st_filter(
  date_from = Sys.Date() - 90,
  date_to   = Sys.Date()
)

# SDK filter
unity_games <- st_filter(sdk = "Unity")

# Publisher filter
ea_games <- st_filter(publisher = "Electronic Arts")
```

### 3. Inspect the filter

```{r, purl = FALSE}
print(puzzle_filter)
as.character(puzzle_filter)  # Server-side filter ID
```

### 4. Use the filter

Every function that accepts a filter takes either a raw 24-hex ID string or an `st_filter` object:

```{r, purl = FALSE}
# Discover apps matching the filter
puzzle_apps <- st_apps(filter = puzzle_filter, limit = 100)

# Top-chart rankings restricted to the filter
top_puzzle <- st_rankings(
  entity = "app",
  os     = "unified",
  filter = puzzle_filter,
  limit  = 25
)

# Raw list of filtered apps (power-user)
raw_hits <- st_get_filtered_apps(filter = puzzle_filter)
```

### 5. Combine filters

Two `st_filter` objects compose with `c()`:

```{r, purl = FALSE}
rpgs       <- st_filter(genre = "RPG")
free_apps  <- st_filter(monetization = "Free")
free_rpgs  <- c(rpgs, free_apps)

apps <- st_apps(filter = free_rpgs)
```

### 6. Reuse an existing filter ID

If you already have a filter ID from the web UI or a previous session:

```{r, purl = FALSE}
existing <- st_filter(filter_id = "5ba4585f539ce75b97db6bcb")
st_apps(filter = existing)
```

## Advanced: arbitrary custom fields

For fields not covered by the named arguments, use `custom_fields = list(...)`:

```{r, purl = FALSE}
complex_filter <- st_filter(
  custom_fields = list(
    "Art Style"       = "3D",
    "Player Count"    = "Multiplayer",
    "Primary Genre"   = "Shooter"
  )
)

top_shooters <- st_rankings(
  entity = "app",
  os     = "ios",
  filter = complex_filter,
  limit  = 50
)
```

## Inspecting and validating filters

```{r, purl = FALSE}
# Does a string look like a valid filter ID?
st_is_valid_filter_id("5ba4585f539ce75b97db6bcb")  # TRUE

# Fetch the canonical filter definition from the API
collection <- st_get_filter_collection("5ba4585f539ce75b97db6bcb")

# Run a smoke-test that the filter yields results
st_test_filter("5ba4585f539ce75b97db6bcb")

# Richer diagnostic output
st_analyze_filter("5ba4585f539ce75b97db6bcb", verbose = TRUE)
```

## Migration from 0.9.x

If you were using `st_filter_by_date()`, `st_filter_by_genre()`, `st_custom_fields_filter()`, `st_combine_filters()`, etc. — all of those are now `.Defunct()` stubs that point at the correct `st_filter()` call:

```{r, purl = FALSE}
# Old (0.9.x)
# puzzle <- st_filter_by_genre(genres = "Puzzle")
# decade <- st_filter_by_date(start_date = "2020-01-01", end_date = "2029-12-31")
# combined <- st_combine_filters(list(puzzle, decade))

# New (1.0.0)
puzzle   <- st_filter(genre = "Puzzle")
decade   <- st_filter(date_from = "2020-01-01", date_to = "2029-12-31")
combined <- c(puzzle, decade)
```

See `vignette("migrating-to-1.0", package = "sensortowerR")` for the complete translation table.

# DAU Sorting Investigation - August 2025

## Problem Statement
When using custom filters from Sensor Tower web interface URLs with DAU sorting, the API was returning apps in incorrect order. NYT Games (5.6M DAU) was appearing at position #26 instead of #1.

## Web Interface Behavior
- URL shows: `/v1/{os}/sales_report_estimates_comparison_attributes` endpoint
- Parameters include: `measure=DAU` in the web URL
- Custom filter ID: `uai=5a39e9681454d22f5a5e75ca` (Word Puzzle filter)
- Results are correctly sorted by DAU in web interface

## Original Implementation (Pre v0.8.1)
The package routed based on measure type:
- `measure = "revenue"` or `"units"` → `sales_report_estimates_comparison_attributes` endpoint
- `measure = "DAU"`, `"WAU"`, `"MAU"` → `top_and_trending/active_users` endpoint

### Issue with active_users endpoint
- Returns `entities.users_absolute` (point-in-time DAU, e.g., 4.3M for NYT)
- Different from `dau_30d_us` (30-day average DAU, e.g., 5.6M for NYT)
- Sorting by `entities.users_absolute` gives wrong order
- The endpoint doesn't seem to respect the measure parameter for sorting

## Current Solution (v0.8.1)
Force all measures to use the `sales_report_estimates_comparison_attributes` endpoint:
```r
# In st_top_charts.R
is_active_users <- FALSE  # Don't use active_users endpoint
is_sales <- TRUE  # Always use sales_report_estimates_comparison_attributes
```

### Workaround Required
- The sales endpoint only accepts `measure = "revenue"` or `"units"` (not "DAU")
- Must use `measure = "revenue"` even when wanting DAU data
- The custom filter handles returning DAU metrics despite revenue measure
- Results are then manually sorted by `dau_30d_us`

## Potential Issues with Current Solution

### Why This Might Be Wrong
1. **API Design Intent**: The existence of separate endpoints suggests they serve different purposes
2. **Parameter Mismatch**: Using `measure = "revenue"` to get DAU data feels like a hack
3. **Future Breaking**: If Sensor Tower fixes their API, this workaround might break

### Alternative Hypothesis
The `active_users` endpoint might be correct but:
1. We're missing a sort parameter (tried `sort_by` but didn't work)
2. The endpoint expects different parameters we haven't discovered
3. The web interface might be doing client-side sorting
4. There might be a third endpoint we haven't found

## Evidence Supporting Current Solution
1. Web interface shows `/sales_report_estimates_comparison_attributes` in network tab
2. Using this endpoint returns correct data (just needs sorting)
3. Results match web interface when sorted by `dau_30d_us`

## Evidence Against Current Solution
1. API returns 422 error when using `measure = "DAU"` with sales endpoint
2. Separate endpoints exist for a reason
3. We're working around the API rather than using it correctly

## How to Test if We Need to Revert

### If Sensor Tower Updates Their API
1. Try using `measure = "DAU"` with sales endpoint - if it starts working, they fixed it
2. Check if active_users endpoint starts returning correctly sorted data
3. Look for new parameters in their API documentation

### Test Commands
```r
# Test if sales endpoint accepts DAU directly (currently fails with 422)
test1 <- tryCatch(
  st_top_charts(
    os = "unified",
    measure = "DAU",  # Currently causes 422 error
    custom_fields_filter_id = "5a39e9681454d22f5a5e75ca",
    category = 7019,
    regions = "US"
  ),
  error = function(e) e
)

# Test if active_users endpoint sorts correctly (currently wrong order)
# Would need to temporarily revert the package changes to test
```

## Recommendation
Keep current solution but monitor for:
1. API documentation updates
2. Changes in web interface behavior  
3. User reports of issues

If Sensor Tower fixes their API to accept `measure = "DAU"` on the sales endpoint, we should update to use that directly rather than the current workaround.

## Related Files
- `/Users/phillip/Documents/vibe_coding_projects/SensorTowerR/R/st_top_charts.R` - Main function
- `/Users/phillip/Documents/vibe_coding_projects/SensorTowerR/R/utils.R` - Parameter preparation
- `/Users/phillip/Documents/vibe_coding_projects/blog_post_source/2025_08_20_puzzle/puzzle_kpi_table.R` - Working example

---
title: "Migrating to sensortowerR 1.0.0"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Migrating to sensortowerR 1.0.0}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE, purl = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", eval = FALSE, purl = FALSE)
```

# Why 1.0.0?

Through versions 0.1 – 0.9 the public surface grew organically to 78 exported functions. Many did the same thing with slightly different argument names (five ways to fetch revenue, three ways to get top charts, two app-lookup paths), and 11 internal helpers had leaked into the namespace. Version 1.0.0 consolidates the API around **four core verbs** — `st_metrics()`, `st_rankings()`, `st_app()` / `st_apps()`, `st_filter()` — and standardizes parameter names across every retained function.

The old names still exist as `.Defunct()` stubs that error with the exact replacement call, so you won't get a cryptic "object not found" — you'll get a one-line migration hint right in your REPL.

## The one rule that affects every script

**Revenue is now returned in dollars by default.** Previously every revenue column was raw integer cents. If you have pipelines that did `revenue / 100` by hand, remove that division. If you have pipelines that rely on integer cents, pass `revenue_unit = "cents"` to restore the old values:

```{r revenue-unit, purl = FALSE}
# Old behavior (integer cents)
sales_cents <- st_metrics(app_id = "...", revenue_unit = "cents")

# New default (numeric dollars)
sales_dollars <- st_metrics(app_id = "...")
```

## Function-by-function translation

### Metrics (revenue, downloads)

| v0.9.x | v1.0.0 |
|---|---|
| `st_sales_report(os = "ios", ios_app_id = id, ...)` | `st_metrics(app_id = id, os = "ios", ...)` |
| `st_unified_sales_report(unified_app_id = id, ...)` | `st_metrics(app_id = id, os = "unified", ...)` (default) |
| `st_batch_metrics(app_list = ids, ...)` | `st_metrics(app_id = ids, ...)` |
| `st_smart_metrics(app_ids = ids, ..., use_cache = TRUE)` | `st_metrics(app_id = ids, ..., cache = TRUE)` (default) |

Argument renames:
- `start_date` / `end_date` → `date_from` / `date_to`
- `date_granularity` → `granularity`
- `measure` → `metrics` (vector)
- `regions` → `countries`
- `app_list` / `ios_app_id` / `android_app_id` / `unified_app_id` → `app_id`

### Rankings

| v0.9.x | v1.0.0 |
|---|---|
| `st_top_charts(measure = "revenue", category = 6014, ...)` | `st_rankings(entity = "app", category = 6014, ...)` |
| `st_top_publishers(...)` | `st_rankings(entity = "publisher", ...)` |
| `st_category_rankings(category = 6014, ...)` | `st_rankings(entity = "category", category = 6014, ...)` |

### App lookup and discovery

| v0.9.x | v1.0.0 |
|---|---|
| `st_app_info(term = "Royal Match")` | `st_apps(query = "Royal Match")` |
| `st_app_lookup(app_id = id)` | `st_app(app_id = id)` |
| `st_app_details(app_ids = ids, os = "ios")` | `st_app(app_id = ids, os = "ios")` |

### Filters

All nine filter-builder functions collapse into a single `st_filter()`:

| v0.9.x | v1.0.0 |
|---|---|
| `st_filter_by_date(start = ..., end = ...)` | `st_filter(date_from = ..., date_to = ...)` |
| `st_filter_by_genre(genre_id = 123)` | `st_filter(genre = 123)` |
| `st_filter_by_monetization(model = "free")` | `st_filter(monetization = "free")` |
| `st_filter_by_publisher(publisher_id = 456)` | `st_filter(publisher = 456)` |
| `st_filter_by_sdk(sdk_name = "unity")` | `st_filter(sdk = "unity")` |
| `st_custom_fields_filter(custom_fields = list(...))` | `st_filter(custom_fields = list(...))` |
| `st_custom_fields_filter_by_id(filter_id = "5f...")` | `st_filter(filter_id = "5f...")` |
| `st_combine_filters(list(a, b))` | `c(a, b)` (S3 method) |
| `st_create_simple_filter(field, value)` | `st_filter(custom_fields = setNames(list(value), field))` |

The returned `st_filter` object is accepted everywhere a raw filter ID string was accepted before — `st_apps(filter = ...)`, `st_rankings(filter = ...)`, `st_get_filtered_apps(filter = ...)`.

### Parameter renames in retained functions

Functions that were already clean in 0.9.x (`st_active_users()`, `st_retention()`, `st_publisher_apps()`, `st_publisher_portfolio()`, `st_session_metrics()`, `st_yoy_metrics()`, `st_demographics()`, `st_game_summary()`, `st_app_enriched()`) keep their identities but have their parameters standardized:

| Old | New |
|---|---|
| `ios_app_id`, `android_app_id`, `unified_app_id`, `unified_id` | `app_id` |
| `app_store`, `platform` | `os` |
| `region`, `regions` | `country` (scalar) / `countries` (vector) |
| `start_date`, `end_date` | `date_from`, `date_to` |
| `date_granularity` | `granularity` |
| `measure`, `metric` | `metrics` |
| `token`, `api_key` | `auth_token` |
| `n`, `max_results` | `limit` |

### Removed exports (now internal)

These were implementation details that were accidentally exported. If you were using them externally, they're still accessible via `sensortowerR:::name()` — but the public API replacements are:

| Removed | Replacement |
|---|---|
| `clean_numeric_column()`, `format_vector()` | Use typed formatters: `format_currency()`, `format_downloads()`, etc. |
| `find_column()`, `select_columns_safe()`, `select_robust()`, `require_column()` | Relied on v0.9 column conventions — v1.0 output schemas are stable, so these aren't needed |
| `validate_columns()`, `validate_top_charts_data()`, `get_column_spec()`, `map_region_columns()`, `try_column_operation()` | Implementation details; no replacement |

## Checking your scripts

The fastest way to find v0.9 calls that need updating:

```{r grep, purl = FALSE}
# From the terminal
system("grep -rn 'st_sales_report\\|st_unified_sales_report\\|st_batch_metrics\\|st_smart_metrics' your_project/")
```

Or just run your script — the `.Defunct()` stubs will error with the exact replacement call for each removed function.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_ratings_review_facets.R
\name{st_ratings_facets}
\alias{st_ratings_facets}
\title{Fetch Rating Metrics from Sensor Tower's Facets API}
\usage{
st_ratings_facets(
  app_ids,
  bundle = c("ratings_incremental", "ratings_cumulative"),
  breakdown = c("app_id", "date"),
  start_date,
  end_date,
  date_granularity = NULL,
  regions = NULL,
  android_localized_estimates = TRUE,
  auth_token = NULL,
  verbose = FALSE
)
}
\arguments{
\item{app_ids}{Character vector of iOS app IDs or Android package names.
Sensor Tower documents a maximum of 1,000 IDs per request.}

\item{bundle}{Character string. Rating bundle to request:
`"ratings_incremental"` or `"ratings_cumulative"`.}

\item{breakdown}{Breakdown fields or a comma-separated breakdown string.
Supported combinations are `"app_id"`, `"app_id,date"`, `"region"`,
`"region,date"`, and `"app_version"`.}

\item{start_date}{Start date in `YYYY-MM-DD` format or as `Date`.}

\item{end_date}{End date in `YYYY-MM-DD` format or as `Date`.}

\item{date_granularity}{Optional date granularity. Required when
`breakdown` includes `date`.}

\item{regions}{Optional character vector of region codes.}

\item{android_localized_estimates}{Logical. Whether to apply Android country
weighting. Defaults to `TRUE` to match the current documented default.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.}

\item{verbose}{Logical. If `TRUE`, prints the request URL with the auth token
redacted.}
}
\value{
A [tibble][tibble::tibble] containing rating metrics such as
  `rating_average_incremental`, `rating_count_incremental`, or their
  cumulative counterparts.
}
\description{
Retrieves rating metrics from Sensor Tower's new facets-based ratings
endpoint. This wrapper targets `"/v1/facets/metrics?facets=ratings"` and
returns a tidy tibble with the live response columns preserved.
}
\details{
Validated against the live Sensor Tower API on March 17, 2026.
}
\examples{
\dontrun{
ratings <- st_ratings_facets(
  app_ids = "553834731",
  bundle = "ratings_incremental",
  breakdown = c("app_id", "date"),
  start_date = "2024-01-01",
  end_date = "2024-01-07",
  date_granularity = "day"
)
}

}
\seealso{
[st_facets_metrics()] for raw facets access
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_filter_helpers.R
\name{st_build_filter_url}
\alias{st_build_filter_url}
\title{Build Sensor Tower Filter URL}
\usage{
st_build_filter_url(
  os = "unified",
  category = NULL,
  countries = NULL,
  base_url = "https://app.sensortower.com/top-charts"
)
}
\arguments{
\item{os}{Character string. Operating system filter. One of "ios", "android", 
or "unified". Defaults to "unified".}

\item{category}{Optional. Category ID to pre-select.}

\item{countries}{Optional. Character vector of country codes to pre-select.}

\item{base_url}{Character string. Base URL for Sensor Tower. Defaults to 
"https://app.sensortower.com/top-charts".}
}
\value{
Character string. The constructed URL.
}
\description{
Constructs a URL for the Sensor Tower web interface where you can create
custom filters. Visit this URL, configure your filters, and then copy the
custom_fields_filter_id from the resulting URL.
}
\examples{
\dontrun{
# Build URL for iOS games in US
url <- st_build_filter_url(os = "ios", category = 6014, countries = "US")

# Open in browser
browseURL(url)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_filter.R
\name{st_filter}
\alias{st_filter}
\alias{print.st_filter}
\alias{format.st_filter}
\alias{as.character.st_filter}
\alias{c.st_filter}
\title{Create a Sensor Tower Filter Object}
\usage{
st_filter(
  date_from = NULL,
  date_to = NULL,
  genre = NULL,
  monetization = NULL,
  publisher = NULL,
  sdk = NULL,
  custom_fields = NULL,
  filter_id = NULL,
  combine = c("and", "or"),
  auth_token = NULL
)

\method{print}{st_filter}(x, ...)

\method{format}{st_filter}(x, ...)

\method{as.character}{st_filter}(x, ...)

\method{c}{st_filter}(..., recursive = FALSE, combine = c("and", "or"))
}
\arguments{
\item{date_from, date_to}{Optional date bounds for a release-date criterion.}

\item{genre}{Optional character vector of game genres.}

\item{monetization}{Optional monetization criteria. Accept either a character
vector such as `c("free", "iap")` or a named list like
`list(free = TRUE, iap = TRUE)`.}

\item{publisher}{Optional character vector of publisher names.}

\item{sdk}{Optional character vector of SDK names.}

\item{custom_fields}{Optional raw custom-fields specification. Can be a
list-of-lists already matching the legacy request format or a named list of
field names and values.}

\item{filter_id}{Optional existing 24-character Sensor Tower filter ID to
wrap.}

\item{combine}{Logical operator metadata for multi-criterion filters: `"and"`
or `"or"`.}

\item{auth_token}{Optional Sensor Tower API token.}

\item{x}{An `st_filter` object (S3 method argument).}

\item{...}{Additional arguments. For `c.st_filter()`, further `st_filter`
objects to combine; ignored by `print.st_filter()`, `format.st_filter()`,
and `as.character.st_filter()`.}

\item{recursive}{Unused; present for S3 method-signature compatibility with
[base::c()].}
}
\value{
An object of class `st_filter`, implemented as a list with elements:
  - `criteria`: normalized criteria supplied to `st_filter()`
  - `combine`: `"and"` or `"or"`
  - `filter_id`: the wrapped or created server-side filter ID
}
\description{
Thin v1.0.0 facade for building or wrapping Sensor Tower custom-field filter
IDs. When criteria are supplied, the function translates them into the legacy
custom-fields request structure and creates a server-side filter ID using the
existing implementation.
}
\examples{
\dontrun{
st_filter(genre = c("Puzzle", "Strategy"))

st_filter(
  date_from = "2024-01-01",
  date_to = "2024-12-31",
  monetization = c("free", "iap")
)

existing <- st_filter(filter_id = "687df26ac5a19ebcfe817d7f")
as.character(existing)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_retention}
\alias{format_retention}
\title{Format retention rates}
\usage{
format_retention(val, digits = 1)
}
\arguments{
\item{val}{Numeric value as decimal (0-1 scale)}

\item{digits}{Number of decimal places (default: 1)}
}
\value{
Formatted retention percentage string
}
\description{
Format retention rates
}
\examples{
format_retention(0.234)    # "23.4\%"
format_retention(0.85)     # "85.0\%"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_filter_helpers.R
\name{st_is_valid_filter_id}
\alias{st_is_valid_filter_id}
\title{Validate Custom Field Filter ID Format}
\usage{
st_is_valid_filter_id(filter_id)
}
\arguments{
\item{filter_id}{Character string. The filter ID to validate}
}
\value{
Logical. TRUE if valid format, FALSE otherwise
}
\description{
Checks if a filter ID matches the expected 24-character hexadecimal format
used by Sensor Tower.
}
\examples{
\dontrun{
# Valid filter ID
st_is_valid_filter_id("687df26ac5a19ebcfe817d7f")  # TRUE

# Invalid filter IDs
st_is_valid_filter_id("invalid")  # FALSE
st_is_valid_filter_id("687df26ac5a19ebcfe817d7")  # FALSE (too short)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_downloads}
\alias{format_downloads}
\title{Format download counts with appropriate suffixes}
\usage{
format_downloads(val)
}
\arguments{
\item{val}{Numeric value to format as downloads}
}
\value{
Formatted download count string
}
\description{
Format download counts with appropriate suffixes
}
\examples{
format_downloads(1234567)      # "1.2M"
format_downloads(1234567890)   # "1.2B"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields_workflow.R
\name{st_custom_fields_workflow}
\alias{st_custom_fields_workflow}
\title{Custom Fields Filter Workflow Helper Functions}
\description{
Helper functions to streamline working with custom fields filters in Sensor Tower.
These functions combine the custom fields endpoints with data retrieval functions
to provide a complete workflow.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields_workflow.R
\name{st_discover_fields}
\alias{st_discover_fields}
\title{Discover Available Custom Fields}
\usage{
st_discover_fields(search_term = NULL, show_values = FALSE, auth_token = NULL)
}
\arguments{
\item{search_term}{Optional. Character string to search for in field names}

\item{show_values}{Logical. Whether to show possible values for each field}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.}
}
\value{
A tibble with custom fields information
}
\description{
Searches and displays available custom fields that can be used for filtering.
This is helpful for discovering what fields are available before creating filters.
}
\examples{
\dontrun{
# Find all game-related fields
game_fields <- st_discover_fields("game")

# Find all date fields
date_fields <- st_discover_fields("date")

# Show all fields with their values
all_fields <- st_discover_fields(show_values = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_url_parser.R
\name{st_extract_url_params}
\alias{st_extract_url_params}
\title{Extract All Parameters from Sensor Tower URL}
\usage{
st_extract_url_params(url)
}
\arguments{
\item{url}{Character string. A Sensor Tower web interface URL}
}
\value{
Data frame with parameter names and values
}
\description{
Extracts and displays all parameters from a Sensor Tower web URL
in a readable format. Useful for understanding complex URLs.
}
\examples{
\dontrun{
url <- "https://app.sensortower.com/market-analysis/top-apps?os=unified&measure=DAU"
params_df <- st_extract_url_params(url)
View(params_df)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_percent}
\alias{format_percent}
\title{Format percentages}
\usage{
format_percent(val, digits = 1)
}
\arguments{
\item{val}{Numeric value to format as percentage (0-100 scale)}

\item{digits}{Number of decimal places (default: 1)}
}
\value{
Formatted percentage string
}
\description{
Format percentages
}
\examples{
format_percent(23.456)    # "23.5\%"
format_percent(0.234, digits = 2)    # "0.23\%"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_ratings_review_facets.R
\name{st_reviews_by_rating_facets}
\alias{st_reviews_by_rating_facets}
\title{Fetch Review Metrics by Rating from Sensor Tower's Facets API}
\usage{
st_reviews_by_rating_facets(
  app_id,
  breakdown = c("date", "review_rating"),
  start_date,
  end_date,
  date_granularity = NULL,
  regions = NULL,
  languages = NULL,
  review_keywords = NULL,
  review_sentiments = NULL,
  review_tags = NULL,
  search_terms = NULL,
  rating_filters = NULL,
  auth_token = NULL,
  verbose = FALSE
)
}
\arguments{
\item{app_id}{Single iOS app ID or Android package name.}

\item{breakdown}{Breakdown fields or a comma-separated breakdown string.
Supported combinations are `"review_rating"`, `"date,review_rating"`,
`"region,review_rating"`, `"language,review_rating"`, and
`"app_version,review_rating"`.}

\item{start_date}{Start date in `YYYY-MM-DD` format or as `Date`.}

\item{end_date}{End date in `YYYY-MM-DD` format or as `Date`.}

\item{date_granularity}{Optional date granularity. Required when
`breakdown` includes `date`.}

\item{regions}{Optional character vector of iOS region codes.}

\item{languages}{Optional character vector of Android language codes.}

\item{review_keywords}{Optional character vector of review-keyword filters.}

\item{review_sentiments}{Optional character vector of sentiment filters.}

\item{review_tags}{Optional character vector of review-tag filters.}

\item{search_terms}{Optional character vector of content search terms.}

\item{rating_filters}{Optional character vector or integer vector of star
filters (`1` through `5`).}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.}

\item{verbose}{Logical. If `TRUE`, prints the request URL with the auth token
redacted.}
}
\value{
A [tibble][tibble::tibble] containing fields such as
  `review_rating`, `review_rating_count`, `review_rating_percentage`, and
  `review_rating_average`.
}
\description{
Retrieves review metrics broken down by star rating from Sensor Tower's new
facets-based review endpoint. This wrapper targets
`"/v1/facets/metrics?facets=reviews_by_rating"`.
}
\details{
Validated against the live Sensor Tower API on March 17, 2026.
}
\examples{
\dontrun{
reviews <- st_reviews_by_rating_facets(
  app_id = "553834731",
  breakdown = c("date", "review_rating"),
  start_date = "2024-01-01",
  end_date = "2024-01-07",
  date_granularity = "day",
  regions = "US"
)
}

}
\seealso{
[st_facets_metrics()] for raw facets access
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_retention_facets.R
\name{st_retention_facets}
\alias{st_retention_facets}
\title{Fetch Retention Metrics from Sensor Tower's Facets API}
\usage{
st_retention_facets(
  app_ids = NULL,
  unified_app_ids = NULL,
  bundle = c("retention_daily", "retention_weekly", "retention_monthly"),
  breakdown = c("date", "app_id"),
  start_date,
  end_date,
  regions = NULL,
  auth_token = NULL,
  verbose = FALSE
)
}
\arguments{
\item{app_ids}{Character vector of platform-specific app IDs. May contain iOS
numeric IDs or Android bundle IDs. Supply exactly one of `app_ids` or
`unified_app_ids`.}

\item{unified_app_ids}{Character vector of Sensor Tower unified app IDs.
Supply exactly one of `app_ids` or `unified_app_ids`.}

\item{bundle}{Character string. Retention bundle to request:
`"retention_daily"`, `"retention_weekly"`, or `"retention_monthly"`.}

\item{breakdown}{Character vector of breakdown fields. Supported combinations
are the ones documented by Sensor Tower:
`"date"`, `"app_id"`, `"unified_app_id"`, `"date,app_id"`,
`"date,unified_app_id"`, `"unified_app_id,app_id"`, and
`"date,unified_app_id,app_id"`.}

\item{start_date}{Start date in `YYYY-MM-DD` format or as `Date`.}

\item{end_date}{End date in `YYYY-MM-DD` format or as `Date`.}

\item{regions}{Optional character vector of region codes. When omitted,
Sensor Tower returns worldwide estimates.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.}

\item{verbose}{Logical. If `TRUE`, prints the request URL with the auth token
redacted.}
}
\value{
A [tibble][tibble::tibble] with one row per requested breakdown
  combination. The response preserves Sensor Tower's live metric column names
  such as `est_retention_d1`, `est_retention_d14`, `est_retention_w52`, or
  `est_retention_m12`.
}
\description{
Retrieves retention curves from Sensor Tower's new facets-based retention
endpoint. This wrapper targets `"/v1/facets/metrics?facets=retention"` and
returns a tidy tibble with the live response columns preserved.
}
\details{
Validated against the live Sensor Tower API on March 17, 2026. The production
response currently returns daily retention columns including
`est_retention_d14` and `est_retention_d365`.
}
\examples{
\dontrun{
retention <- st_retention_facets(
  app_ids = "553834731",
  bundle = "retention_daily",
  breakdown = c("date", "app_id"),
  start_date = "2025-01-01",
  end_date = "2025-01-31"
)
}

}
\seealso{
[st_retention()] for the legacy retention endpoint,
  [st_facets_metrics()] for raw facets access
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_metrics.R
\name{st_metrics}
\alias{st_metrics}
\title{Fetch Sensor Tower Metrics}
\usage{
st_metrics(
  app_id,
  metrics = c("revenue", "downloads"),
  os = "unified",
  countries = "WW",
  date_from = Sys.Date() - 90,
  date_to = Sys.Date(),
  granularity = "daily",
  revenue_unit = c("dollars", "cents"),
  shape = c("long", "wide"),
  cache = TRUE,
  auth_token = NULL
)
}
\arguments{
\item{app_id}{Character scalar or vector of app identifiers. Each entry can be
a Sensor Tower unified app ID, an iOS numeric app ID, or an Android package
name.}

\item{metrics}{Character vector of metrics to return. Supported values are
`"revenue"` and `"downloads"`.}

\item{os}{Character scalar. One of `"ios"`, `"android"`, or `"unified"`.}

\item{countries}{Character vector of 2-letter country codes. Use `"WW"` for
worldwide aggregates.}

\item{date_from, date_to}{Date bounds for the query. Accept `Date` objects or
ISO date strings.}

\item{granularity}{Character scalar. One of `"daily"`, `"weekly"`,
`"monthly"`, or `"quarterly"`.}

\item{revenue_unit}{Character scalar. `"dollars"` (default) returns revenue in
base currency units. `"cents"` returns revenue in cents for compatibility
with the legacy API surface.}

\item{shape}{Character scalar. `"long"` returns one row per metric observation.
`"wide"` returns one row per app/date/country with separate metric columns.}

\item{cache}{Logical. If `TRUE`, use a process-local cache keyed on the
normalized arguments.}

\item{auth_token}{Optional Sensor Tower API token. If `NULL`, falls back to
`SENSORTOWER_AUTH_TOKEN`.}
}
\value{
If `shape = "long"`, a tibble with columns:
  - `app_id`: identifier supplied to `st_metrics()`
  - `os`: normalized operating system
  - `country`: 2-letter country code
  - `date`: observation date
  - `metric`: one of `"revenue"` or `"downloads"`
  - `value`: metric value; revenue is in dollars by default and cents when
    `revenue_unit = "cents"`

If `shape = "wide"`, a tibble with columns:
  - `app_id`
  - `os`
  - `country`
  - `date`
  - one numeric column per requested metric
}
\description{
Thin v1.0.0 facade for sales/download metrics. The function validates a
standardized argument set, dispatches to the legacy implementation that best
matches the request, and then normalizes the result into a stable schema.
}
\examples{
\dontrun{
st_metrics(
  app_id = "553834731",
  os = "ios",
  countries = "US",
  date_from = Sys.Date() - 30,
  date_to = Sys.Date() - 1
)

st_metrics(
  app_id = c("553834731", "com.supercell.clashofclans"),
  os = "unified",
  shape = "wide",
  countries = c("US", "GB"),
  revenue_unit = "cents"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields_workflow.R
\name{st_analyze_filter}
\alias{st_analyze_filter}
\title{Analyze Custom Filter Performance}
\usage{
st_analyze_filter(
  filter_id,
  measure = "DAU",
  regions = "US",
  top_n = 10,
  auth_token = NULL
)
}
\arguments{
\item{filter_id}{Character. The custom fields filter ID to analyze}

\item{measure}{Character. Metric to analyze: "DAU", "revenue", or "units"}

\item{regions}{Character vector. Region codes (default "US")}

\item{top_n}{Integer. Number of top apps to show (default 10)}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.}
}
\value{
A list containing summary statistics and top apps
}
\description{
Provides a summary analysis of apps matching a custom filter, including
top performers, growth metrics, and category breakdown.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_categories.R
\name{st_categories}
\alias{st_categories}
\title{List Available Sensor Tower Categories}
\usage{
st_categories(platform = NULL)
}
\arguments{
\item{platform}{Optional character string. Filter results for a specific
platform ("ios" or "android"). If NULL (default), returns categories for
both platforms.}
}
\value{
A tibble with columns `platform` (character, "ios" or "android"),
  `category_id` (character, e.g., "6014"), and `category_name`
  (character, e.g., "Games").
}
\description{
Returns a tibble of app categories recognized by the Sensor Tower API,
mapping category IDs to category names for different platforms (iOS/Android).
Useful for finding valid inputs for the `category` parameter in other
functions.
}
\examples{
# Get all categories
all_cats <- st_categories()
head(all_cats)

# Get only iOS categories
ios_cats <- st_categories(platform = "ios")
head(ios_cats)

# Find game categories on iOS
ios_games <- subset(st_categories("ios"), grepl("Game", category_name))
head(ios_games)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_sales_report.R
\name{lookup_category_names}
\alias{lookup_category_names}
\title{Helper function to look up category names}
\usage{
lookup_category_names(category_ids, platform = "ios")
}
\arguments{
\item{category_ids}{Character vector of category IDs}

\item{platform}{Character string. "ios" or "android"}
}
\value{
Character vector of category names
}
\description{
Helper function to look up category names
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields_utils.R
\name{st_get_filter_collection}
\alias{st_get_filter_collection}
\title{Get Pre-Built Filter Collections}
\usage{
st_get_filter_collection(
  collection = c("top_genres", "monetization_models", "platform_exclusive",
    "market_segments"),
  auth_token = NULL
)
}
\arguments{
\item{collection}{Character. Name of the collection:
- "top_genres": Major game genres
- "monetization_models": Different monetization approaches
- "platform_exclusive": Platform-specific apps
- "market_segments": Market segment filters}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.}
}
\value{
A named list of filter IDs
}
\description{
Returns commonly used filter IDs for quick access to pre-defined app segments.
}
\examples{
\dontrun{
# Get filter IDs for top game genres
genre_filters <- st_get_filter_collection("top_genres")

# Use a filter from the collection
puzzle_apps <- st_get_filtered_apps(
  filter_id = genre_filters$puzzle,
  measure = "DAU",
  regions = "US"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_publisher_portfolio.R
\name{st_publisher_portfolio}
\alias{st_publisher_portfolio}
\title{Publisher Portfolio Analysis}
\usage{
st_publisher_portfolio(
  publisher = NULL,
  publisher_id = NULL,
  start_date = "2023-01-01",
  end_date = NULL,
  countries = "WW",
  metrics = c("revenue", "downloads", "mau"),
  include_rankings = TRUE,
  include_portfolio_total = TRUE,
  granularity = "yearly",
  min_revenue = 1e+05,
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
  verbose = TRUE,
  use_cache = FALSE,
  cache_dir = NULL
)
}
\arguments{
\item{publisher}{Character. Publisher name to search for (e.g., "Lilith Games",
"Supercell", "King"). The function will search for the publisher and use
the first match.}

\item{publisher_id}{Character. Optional. If provided, skips the publisher
search and uses this unified_publisher_id directly.}

\item{start_date}{Date or character. Start date for metrics (default: "2023-01-01").}

\item{end_date}{Date or character. End date for metrics (default: last day of
previous month).}

\item{countries}{Character. Countries for metrics (default: "WW" for worldwide).}

\item{metrics}{Character vector. Which metrics to fetch. Options: "revenue",
"downloads", "mau". Default: all three.}

\item{include_rankings}{Logical. Whether to fetch subgenre rankings from top
charts. Default: TRUE.}

\item{include_portfolio_total}{Logical. Whether to add a portfolio total row.
Default: TRUE.}

\item{granularity}{Character. How to aggregate the data: "yearly" (default),
"quarterly", or "monthly".}

\item{min_revenue}{Numeric. Minimum revenue threshold to include an app.
Default: 100000 (apps with at least $100K in any year).}

\item{auth_token}{Character. Sensor Tower API token. Defaults to
SENSORTOWER_AUTH_TOKEN environment variable}

\item{verbose}{Logical. Print progress messages. Default: TRUE.}

\item{use_cache}{Logical. Use cached data if available. Default: FALSE.
When TRUE, requires cache_dir to be specified.}

\item{cache_dir}{Character. Directory for cached data. Default: NULL (no caching).
Must be explicitly set to enable caching. Use tempdir() for temporary caching.}
}
\value{
A tibble with portfolio data including:
  - app_name: Game name
  - subgenre: Game sub-genre
  - subgenre_rank: Rank within sub-genre
  - revenue_\{year\}: Revenue by year
  - downloads_\{year\}: Downloads by year
  - mau_\{year\}: Average MAU by year (if requested)
  - revenue_yoy, downloads_yoy, mau_yoy: Year-over-year growth percentages
}
\description{
Fetches comprehensive portfolio data for a publisher including revenue,
downloads, MAU, and rankings. Returns a tidy data frame ready for
visualization or GT table creation.
}
\examples{
\dontrun{
# Simple usage - just provide publisher name
lilith_portfolio <- st_publisher_portfolio("Lilith Games")

# Piped workflow
library(dplyr)

"Supercell" \%>\%
  st_publisher_portfolio(
    start_date = "2023-01-01",
    metrics = c("revenue", "downloads")
  ) \%>\%
  filter(revenue_2024 > 1000000) \%>\%
  arrange(desc(revenue_2024))

# Custom date range and countries
portfolio <- st_publisher_portfolio(
  publisher = "King",
  start_date = "2022-01-01",
  end_date = "2024-12-31",
  countries = c("US", "GB", "DE"),
  metrics = c("revenue", "downloads", "mau"),
  include_rankings = TRUE
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_smart_metrics.R
\name{st_clear_id_cache}
\alias{st_clear_id_cache}
\title{Clear App ID Cache}
\usage{
st_clear_id_cache(disk = TRUE)
}
\arguments{
\item{disk}{Logical. Also remove the on-disk cache file (default TRUE).}
}
\value{
No return value, called for side effects (clearing the cache).
}
\description{
Clears the in-memory and on-disk cache of app ID mappings
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_app_tag.R
\name{st_app_tag}
\alias{st_app_tag}
\title{Fetch Apps by Custom Fields and Tags}
\usage{
st_app_tag(
  app_id_type,
  custom_fields_filter_id,
  name = NULL,
  value = NULL,
  global = TRUE,
  last_known_id = NULL,
  auth_token = NULL,
  base_url = "https://api.sensortower.com"
)
}
\arguments{
\item{app_id_type}{Character string. Operating System. Must be one of
"itunes" (iOS) or "unified". Required.}

\item{custom_fields_filter_id}{Character string. ID of a Sensor Tower
custom field filter. Required. Use the filter ID from relevant endpoint.}

\item{name}{Optional. Character string. Name of Custom or Global Field.
Defaults to "Stock Ticker".}

\item{value}{Optional. Character string. Tag value for custom or global 
field provided. Leave blank to fetch all possible apps.}

\item{global}{Optional. Logical. Filter by global or organization custom 
fields. Defaults to TRUE (false means organization custom fields).}

\item{last_known_id}{Optional. Character string. Supply last_known_id from 
previous request to get next page. Leave blank to get first page.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.}

\item{base_url}{Optional. Character string. The base URL for the API.}
}
\value{
A [tibble][tibble::tibble] with app data including IDs and metadata.
}
\description{
Retrieves apps filtered by custom fields and tags from Sensor Tower.
This function uses the /v1/app_tag/apps endpoint.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\docType{data}
\name{NUMERIC_METRIC_PATTERNS}
\alias{NUMERIC_METRIC_PATTERNS}
\title{Numeric Metric Patterns}
\format{
An object of class \code{character} of length 29.
}
\usage{
NUMERIC_METRIC_PATTERNS
}
\description{
Patterns to identify columns that should be converted to numeric.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\docType{data}
\name{METRIC_MAPPING}
\alias{METRIC_MAPPING}
\title{Metric Mapping for Custom Tags}
\format{
An object of class \code{character} of length 60.
}
\usage{
METRIC_MAPPING
}
\description{
Maps raw API field names (from entities.custom_tags or aggregate_tags)
to cleaner, user-friendly column names.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fetch_unified_data}
\alias{fetch_unified_data}
\title{Fetch and Unified Data from Platforms}
\usage{
fetch_unified_data(
  ios_app_id = NULL,
  android_app_id = NULL,
  start_date,
  end_date,
  countries,
  date_granularity,
  auth_token,
  verbose = FALSE,
  combine_to_unified = TRUE
)
}
\arguments{
\item{ios_app_id}{Character. iOS App ID.}

\item{android_app_id}{Character. Android App ID.}

\item{start_date}{Date. Start date.}

\item{end_date}{Date. End date.}

\item{countries}{Character vector. Country codes.}

\item{date_granularity}{Character. Granularity.}

\item{auth_token}{Character. API token.}

\item{verbose}{Logical. Verbose output.}

\item{combine_to_unified}{Logical. Whether to sum metrics into a unified view.}
}
\value{
A tibble with columns date, country, revenue, downloads, and optionally platform/app_id.
}
\description{
Fetches data from iOS and/or Android and optionally combines them.
Handles missing data from one platform gracefully.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_game_summary.R
\name{process_game_summary_response}
\alias{process_game_summary_response}
\title{Process Game Summary API Response}
\usage{
process_game_summary_response(resp, os)
}
\arguments{
\item{resp}{List. Raw API response from game summary endpoint.}

\item{os}{Character string. Operating system to determine field mappings.}
}
\value{
A processed tibble with descriptive column names.
}
\description{
Internal function to process and enrich game summary API responses.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_session_metrics.R
\name{st_session_metrics}
\alias{st_session_metrics}
\title{Fetch Session Metrics Time Series Data}
\usage{
st_session_metrics(
  unified_app_id = NULL,
  ios_app_id = NULL,
  android_app_id = NULL,
  start_date,
  end_date,
  metrics = c("session_count", "session_duration", "time_spent"),
  regions = "US",
  time_period = "week",
  date_granularity = "monthly",
  os = NULL,
  breakdown = "unified_app_id",
  auth_token = NULL,
  verbose = TRUE
)
}
\arguments{
\item{unified_app_id}{Character string or vector. Sensor Tower unified app ID(s)
(24-character hex). Maximum 100 apps per request.}

\item{ios_app_id}{Character string or vector. iOS app ID(s) for non-unified queries.}

\item{android_app_id}{Character string or vector. Android package name(s) for
non-unified queries.}

\item{start_date}{Date or character string. Start date in "YYYY-MM-DD" format.
Data is available from 2021-01-01 onward.}

\item{end_date}{Date or character string. End date in "YYYY-MM-DD" format.}

\item{metrics}{Character vector. Metrics to retrieve. Options include:
- "time_spent" (average seconds per user per day)
- "total_time_spent" (total seconds across all users)
- "session_duration" (average session length in seconds)
- "session_count" (average sessions per user per day)
- "total_session_count" (total sessions across all users)
Default is c("session_count", "session_duration", "time_spent").}

\item{regions}{Character vector. Region/country codes (e.g., "US", "GB").
Default is "US". Use NULL for all regions.}

\item{time_period}{Character string. Session metrics time period.
Options: "day", "week". Default is "week".
Returns averaged session metrics for each period within a month.}

\item{date_granularity}{Character string. Aggregate data by granularity.
Options: "daily", "weekly", "monthly". Default is "monthly".
Note: "daily" granularity may not be supported by the API for all apps;
use "weekly" or "monthly" if you receive empty results with "daily".}

\item{os}{Character string. Filter by platform for unified apps.
Options: "ios", "android", or NULL for both. Default is NULL.}

\item{breakdown}{Character string. Fields for data aggregation.
Options: "unified_app_id", "app_id", "region". Default is "unified_app_id".}

\item{auth_token}{Character string. Your Sensor Tower API token.
Defaults to environment variable SENSORTOWER_AUTH_TOKEN.}

\item{verbose}{Logical. If TRUE, prints progress messages.}
}
\value{
A [tibble][tibble::tibble] with session metrics including:
  - **unified_app_id** or **app_id**: The app identifier
  - **date**: Date of the data point
  - **time_spent**: Average seconds spent per user per day
  - **total_time_spent**: Total seconds across all users
  - **session_duration**: Average session length in seconds
  - **session_count**: Average sessions per user per day
  - **total_session_count**: Total session count across all users
}
\description{
Retrieves session metrics time series data (session count, session duration,
time spent) for apps from the Sensor Tower Usage Intelligence API.
}
\section{Data Availability}{

- Data is available from 2021-01-01 onward
- Session metrics require Usage Intelligence subscription
- Maximum 100 apps per request
}

\examples{
\dontrun{
# Get session metrics for a unified app
sessions <- st_session_metrics(
  unified_app_id = "5fbc3849d0b8414136857afc",
  start_date = "2024-01-01",
  end_date = "2024-12-01"
)

# Get specific metrics with weekly granularity
sessions <- st_session_metrics(
  unified_app_id = "5fbc3849d0b8414136857afc",
  start_date = "2024-01-01",
  end_date = "2024-03-01",
  metrics = c("session_count", "session_duration"),
  date_granularity = "weekly"
)

# Get session data for Android app directly
sessions <- st_session_metrics(
  android_app_id = "com.example.app",
  start_date = "2024-01-01",
  end_date = "2024-06-01"
)
}

}
\seealso{
[st_retention()] for retention metrics,
  [st_demographics()] for user demographics,
  [st_batch_metrics()] for MAU/DAU/WAU metrics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_filter_utils.R
\name{add_custom_filter_params}
\alias{add_custom_filter_params}
\title{Add Custom Filter Parameters to Query}
\usage{
add_custom_filter_params(
  query_params,
  custom_fields_filter_id = NULL,
  custom_tags_mode = NULL,
  os = NULL
)
}
\arguments{
\item{query_params}{List. Existing query parameters}

\item{custom_fields_filter_id}{Character. Custom filter ID from Sensor Tower}

\item{custom_tags_mode}{Character. Tag mode for unified OS}

\item{os}{Character. Operating system (ios, android, unified)}
}
\value{
Modified query_params list with custom filter parameters added
}
\description{
Internal function to add custom field filter parameters to API query parameters.
Handles validation and OS-specific requirements.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_game_summary.R
\name{map_game_summary_fields}
\alias{map_game_summary_fields}
\title{Map Game Summary Field Names}
\usage{
map_game_summary_fields(data, os)
}
\arguments{
\item{data}{Tibble. Data with abbreviated field names.}

\item{os}{Character string. Operating system to determine mappings.}
}
\value{
Tibble with descriptive field names.
}
\description{
Internal function to map abbreviated API field names to descriptive names.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_yoy_metrics.R
\name{calculate_yoy_growth}
\alias{calculate_yoy_growth}
\title{Calculate year-over-year growth rates}
\usage{
calculate_yoy_growth(yoy_data, baseline_year = NULL)
}
\arguments{
\item{yoy_data}{Output from st_yoy_metrics}

\item{baseline_year}{The year to use as baseline (default: earliest year)}
}
\value{
A tibble with growth rates relative to baseline year
}
\description{
Helper function to calculate YoY growth rates from the output of st_yoy_metrics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_app.R
\name{st_apps}
\alias{st_apps}
\title{Search or Filter Apps}
\usage{
st_apps(
  query = NULL,
  filter = NULL,
  os = "ios",
  country = "US",
  limit = 100,
  auth_token = NULL
)
}
\arguments{
\item{query}{Optional search string for app-name search.}

\item{filter}{Optional `st_filter` object or 24-character filter ID string.}

\item{os}{Operating system context for query searches.}

\item{country}{Two-letter country code used by the filtered-app workflow.}

\item{limit}{Positive integer row limit.}

\item{auth_token}{Optional Sensor Tower API token.}
}
\value{
A tibble returned by the dispatched legacy implementation.
}
\description{
Thin v1.0.0 facade for app discovery by query string or server-side filter.
}
\examples{
\dontrun{
st_apps(query = "Royal Match", os = "unified", limit = 25)

filt <- st_filter(genre = "Puzzle")
st_apps(filter = filt, country = "US", limit = 50)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example_data.R
\name{example_sensortower_data}
\alias{example_sensortower_data}
\title{Example Data for sensortowerR Package}
\usage{
example_sensortower_data(type = "app_info", n = 5)
}
\arguments{
\item{type}{Character string specifying the type of example data to create.
Options: "app_info", "rankings", "details", "metrics", "publisher_apps",
"top_charts", "game_summary"}

\item{n}{Number of example records to generate (default: 5)}
}
\value{
A tibble with example data matching the structure of actual API responses
}
\description{
Creates example data structures that mimic the API responses for testing
and demonstration purposes without requiring actual API calls.
}
\examples{
# Get example app search results
example_apps <- example_sensortower_data("app_info")
head(example_apps)

# Get example ranking data
example_rankings <- example_sensortower_data("rankings", n = 10)
print(example_rankings)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_app.R
\name{st_app}
\alias{st_app}
\title{Fetch App Details}
\usage{
st_app(app_id, os = "unified", fields = NULL, auth_token = NULL)
}
\arguments{
\item{app_id}{Character scalar or vector of app identifiers.}

\item{os}{Operating system context. One of `"ios"`, `"android"`, or
`"unified"`.}

\item{fields}{Optional character vector of enriched fields to keep. When
supplied, the request is routed through `st_app_enriched()`.}

\item{auth_token}{Optional Sensor Tower API token.}
}
\value{
A tibble returned by the dispatched legacy implementation.
}
\description{
Thin v1.0.0 facade for app metadata lookups. By default it dispatches to the
legacy detailed app metadata endpoint. When `fields` is supplied, it dispatches
to `st_app_enriched()` and optionally subsets to the requested fields.
}
\examples{
\dontrun{
st_app("553834731", os = "ios")

st_app(
  c("553834731", "com.supercell.clashofclans"),
  fields = c("revenue_90d_ww", "downloads_30d_ww")
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_filter_utils.R
\name{custom_filter_utils}
\alias{custom_filter_utils}
\title{Custom Filter Utilities}
\description{
Internal functions for handling custom field filters across sensortowerR functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revenue_standardization.R
\name{standardize_revenue_units}
\alias{standardize_revenue_units}
\title{Standardize Revenue Units Across sensortowerR}
\usage{
standardize_revenue_units(data, source = "unknown", target_unit = "base")
}
\arguments{
\item{data}{Data frame containing revenue columns}

\item{source}{Character string indicating the data source}

\item{target_unit}{Character string: "base" (dollars/euros/etc) or "cents"}
}
\value{
Data frame with standardized revenue values
}
\description{
Ensures revenue values have consistent units across all sensortowerR functions.
The Sensor Tower API returns revenue in cents, but for consistency with other
functions like st_sales_report, this converts to base currency units.
}
\details{
The Sensor Tower API inconsistently returns revenue values:
- Top Charts API: revenue in cents
- Sales Report API: revenue in base currency units
- Publishers API: revenue in cents (but converts to base units)

This function ensures consistency by converting all revenue to base units
by default, while preserving the original values for reference.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_app_details.R
\name{process_app_details_response}
\alias{process_app_details_response}
\title{Process App Details API Response}
\usage{
process_app_details_response(resp, os)
}
\arguments{
\item{resp}{Response object from httr2}

\item{os}{Operating system}
}
\value{
A processed tibble with app details
}
\description{
Internal function to process and enrich app details API responses.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_url_parser.R
\name{st_parse_web_url}
\alias{st_parse_web_url}
\title{Parse Sensor Tower Web URL to API Parameters}
\usage{
st_parse_web_url(url, verbose = TRUE)
}
\arguments{
\item{url}{Character string. A Sensor Tower web interface URL}

\item{verbose}{Logical. Whether to print parameter mapping details. 
Defaults to TRUE.}
}
\value{
List of API-compatible parameters suitable for use with st_top_charts()
  and other sensortowerR functions
}
\description{
Converts a Sensor Tower web interface URL into API-compatible parameters
that can be used with sensortowerR functions. This is helpful when you want
to replicate a web query in R.
}
\examples{
\dontrun{
# Parse a web URL
url <- "https://app.sensortower.com/market-analysis/top-apps?os=unified&measure=DAU"
params <- st_parse_web_url(url)

# Use the parameters in an API call
data <- do.call(st_top_charts, params)

# Or modify parameters before using
params$limit <- 50
data <- do.call(st_top_charts, params)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_arpu}
\alias{format_arpu}
\title{Format ARPU (Average Revenue Per User)}
\usage{
format_arpu(val, digits = 2)
}
\arguments{
\item{val}{Numeric ARPU value}

\item{digits}{Number of decimal places (default: 2)}
}
\value{
Formatted ARPU string
}
\description{
Format ARPU (Average Revenue Per User)
}
\examples{
format_arpu(5.234)    # "$5.23"
format_arpu(0.99)     # "$0.99"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{formatting_helpers}
\alias{formatting_helpers}
\title{Formatting Helper Functions}
\description{
Functions for formatting numeric values in a human-readable way,
particularly useful for revenue, download counts, and percentages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_filter_helpers.R
\name{filter_helpers}
\alias{filter_helpers}
\title{Custom Filter Helper Functions}
\description{
Functions to validate, test, and manage Sensor Tower custom field filters
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_filter_utils.R
\name{extract_custom_filter_params}
\alias{extract_custom_filter_params}
\title{Extract Custom Filter from URL Parameters}
\usage{
extract_custom_filter_params(params)
}
\arguments{
\item{params}{List. Parameters from st_parse_web_url()}
}
\value{
List with custom_fields_filter_id and custom_tags_mode (or NULLs)
}
\description{
Helper to extract custom filter parameters from parsed URL parameters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_app_enriched.R
\name{st_app_enriched}
\alias{st_app_enriched}
\title{Fetch Enriched Metrics for Specific Apps}
\usage{
st_app_enriched(
  unified_app_ids,
  os = "unified",
  regions = "WW",
  auth_token = NULL
)
}
\arguments{
\item{unified_app_ids}{Character vector. One or more unified app IDs
(24-character hex strings). Required. Use `st_app_info()` to find these.}

\item{os}{Character string. Operating system context for the request.
Must be "unified" (default), "ios", or "android".}

\item{regions}{Character vector. Region codes for data filtering.
Defaults to "WW" (worldwide).}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable SENSORTOWER_AUTH_TOKEN.}
}
\value{
A [tibble][tibble::tibble] with enriched metrics including:
  - **Identification**: `unified_app_id`, `unified_app_name`
  - **Active Users**: `dau_30d_us`, `dau_30d_ww`, `wau_4w_us`, `wau_4w_ww`,
    `mau_month_us`, `mau_month_ww`
  - **Retention**: `retention_1d_us/ww`, `retention_7d_us/ww`,
    `retention_14d_us/ww`, `retention_30d_us/ww`, `retention_60d_us/ww`
  - **Demographics**: `genders_us`, `genders_ww`, `age_us`, `age_ww`,
    `male_share_us`, `female_share_us`
  - **Revenue/Downloads**: `revenue_30d_ww`, `revenue_90d_ww`, `revenue_alltime_us/ww`,
    `downloads_30d_ww`, `downloads_alltime_us/ww`
  - **Monetization**: `rpd_alltime_us/ww`, `arpu_90d_us/ww`
  - **Launch**: `release_date_us/ww`, `earliest_release_date`
}
\description{
Retrieves comprehensive metrics including retention, MAU, DAU, demographics,
and other aggregate tags for specific apps by their unified app IDs.
}
\details{
This function is designed for the common workflow of:
1. Search for apps by name using `st_app_info()`
2. Get their unified IDs
3. Fetch enriched metrics for those specific apps using this function
}
\section{Recommended Workflow}{

```
# Step 1: Search for apps by name
apps <- st_app_info("Royal Match")

# Step 2: Get unified IDs
app_ids <- apps$unified_app_id

# Step 3: Fetch enriched metrics
metrics <- st_app_enriched(app_ids)
```
}

\section{Data Availability Notes}{

- **IMPORTANT: Geographic Limitations** - All enriched metrics are only available
  for **US market (`_us` suffix)** and **Worldwide aggregates (`_ww` suffix)**.
  Per-country data (e.g., GB, DE, FR, JP) is NOT available through this endpoint.
  For per-country data, use [st_sales_report()] for revenue/downloads or
  [st_batch_metrics()] for MAU/DAU time-series.
- Retention data (D1, D7, D14, D30, D60) is aggregated for the "last quarter" -
  not time-series data. D90 retention is NOT available through the API.
- Demographics (age/gender) are primarily available for US market only.
- Not all metrics are available for all apps - smaller apps may have NA values.
- This returns **snapshot data**, not time-series. For historical trends, use
  [st_batch_metrics()] or [st_sales_report()].
}

\examples{
\dontrun{
# Get enriched data for specific apps
royal_match <- st_app_info("Royal Match")
enriched <- st_app_enriched(royal_match$unified_app_id)

# Get data for multiple apps at once
game_ids <- c("5f16a8019f7b275235017614", "660af7c66237390ce7c829fc")
multi_enriched <- st_app_enriched(game_ids)

# View retention metrics
multi_enriched \%>\%
  select(unified_app_name, starts_with("retention"))
}

}
\seealso{
[st_app_info()] for searching apps by name,
  [st_app_lookup()] for resolving app IDs,
  [st_sales_report()] for time-series revenue/download data,
  [st_batch_metrics()] for time-series DAU/WAU/MAU data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_large_number}
\alias{format_large_number}
\title{Format large numbers with K/M/B suffixes}
\usage{
format_large_number(val, digits = 1, prefix = "")
}
\arguments{
\item{val}{Numeric value to format}

\item{digits}{Number of digits after decimal for millions/billions (default: 1)}

\item{prefix}{Optional prefix (e.g., "$" for currency)}
}
\value{
Formatted string with appropriate suffix
}
\description{
Format large numbers with K/M/B suffixes
}
\examples{
format_large_number(1234567)      # "1.2M"
format_large_number(1234567890)   # "1.2B"
format_large_number(1234567, prefix = "$") # "$1.2M"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_publisher_apps.R
\name{st_get_app_names}
\alias{st_get_app_names}
\title{Get App Names from Publisher Apps Result}
\usage{
st_get_app_names(apps_df, include_canonical = TRUE)
}
\arguments{
\item{apps_df}{A tibble returned by `st_publisher_apps()`.}

\item{include_canonical}{Logical. If TRUE, includes mappings for canonical IDs
that were resolved during aggregation. Defaults to TRUE.}
}
\value{
A tibble with columns `unified_app_id` and `app_name` suitable for
  joining with sales data or other API results.
}
\description{
Helper function to create a name lookup table from the result of
`st_publisher_apps()`. This handles canonical ID mapping automatically,
so you can join sales data (which uses canonical IDs) back to app names.
}
\examples{
\dontrun{
# Get apps with canonical ID resolution
apps <- st_publisher_apps("647eb849d9d91f31a54f1792", aggregate_related = TRUE)

# Get name lookup table
name_lookup <- st_get_app_names(apps)

# Use with sales data
sales <- st_unified_sales_report(apps$unified_app_id, ...)
sales_with_names <- sales \%>\%
  left_join(name_lookup, by = "unified_app_id")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_filter_utils.R
\name{validate_custom_filter_params}
\alias{validate_custom_filter_params}
\title{Validate Custom Filter Parameters}
\usage{
validate_custom_filter_params(
  custom_fields_filter_id = NULL,
  custom_tags_mode = NULL,
  os = NULL,
  require_category = TRUE,
  category = NULL
)
}
\arguments{
\item{custom_fields_filter_id}{Character. Custom filter ID}

\item{custom_tags_mode}{Character. Tag mode}

\item{os}{Character. Operating system}

\item{require_category}{Logical. Whether category is required when no filter}

\item{category}{Category parameter value}
}
\value{
NULL if valid, stops with error if invalid
}
\description{
Checks if custom filter parameters are valid before making API call
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_filter_helpers.R
\name{st_extract_filter_id}
\alias{st_extract_filter_id}
\title{Extract Filter ID from Sensor Tower URL}
\usage{
st_extract_filter_id(url)
}
\arguments{
\item{url}{Character string. A Sensor Tower URL containing custom_fields_filter_id}
}
\value{
Character string. The extracted filter ID, or NULL if not found.
}
\description{
Extracts the custom_fields_filter_id parameter from a Sensor Tower URL.
This is helpful when copying URLs from the web interface.
}
\examples{
\dontrun{
url <- "https://app.sensortower.com/top-charts?custom_fields_filter_id=687df26ac5a19ebcfe817d7f"
filter_id <- st_extract_filter_id(url)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_filter_helpers.R
\name{st_test_filter}
\alias{st_test_filter}
\title{Test a Custom Filter ID}
\usage{
st_test_filter(filter_id, os = "ios", verbose = TRUE, auth_token = NULL)
}
\arguments{
\item{filter_id}{Character string. The filter ID to test}

\item{os}{Character string. Operating system to test with. One of "ios", 
"android", or "unified". Defaults to "ios".}

\item{verbose}{Logical. Whether to print detailed test results. Defaults to TRUE.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable SENSORTOWER_AUTH_TOKEN.}
}
\value{
List with test results including success status and any error messages
}
\description{
Tests whether a custom filter ID works with the Sensor Tower API by making
a minimal test request. This helps verify that the filter exists and is
accessible with your authentication.
}
\examples{
\dontrun{
# Test a filter ID
result <- st_test_filter("687df26ac5a19ebcfe817d7f")

# Test silently
result <- st_test_filter("687df26ac5a19ebcfe817d7f", verbose = FALSE)

# Test with different OS
result <- st_test_filter("687df26ac5a19ebcfe817d7f", os = "unified")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_batch_app_lookup.R
\name{st_batch_app_lookup}
\alias{st_batch_app_lookup}
\title{Batch Look up app information}
\usage{
st_batch_app_lookup(
  app_ids,
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
  verbose = FALSE
)
}
\arguments{
\item{app_ids}{Character vector of unified app IDs.}

\item{auth_token}{Character string. Your Sensor Tower API authentication token.}

\item{verbose}{Logical. Whether to show progress messages. Default is FALSE.}
}
\value{
A \code{tibble} with columns:
  \itemize{
    \item \code{unified_app_id}: The Sensor Tower unified app ID
    \item \code{app_name}: The app's display name
    \item \code{ios_app_id}: iOS app ID if found
    \item \code{android_app_id}: Android app ID if found
    \item \code{icon_url}: URL of the app icon
  }
  Returns empty tibble if no apps can be found or details cannot be fetched.
}
\description{
This function takes a vector of unified app IDs and returns a single
data frame containing the corresponding unified_app_id, app_name,
platform-specific IDs (ios_app_id, android_app_id), and icon_url.
It handles chunking requests to the st_app_details endpoint (limit 100).
}
\examples{
\dontrun{
# Look up multiple games
apps <- st_batch_app_lookup(c("5ba4585f539ce75b97db6bcb", "5d10d6bfecb4db5f93902f23"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fetch_data_core}
\alias{fetch_data_core}
\title{Core Data Fetching Function}
\usage{
fetch_data_core(
  endpoint,
  params,
  auth_token,
  verbose = FALSE,
  enrich_response = TRUE,
  processor = process_response
)
}
\arguments{
\item{endpoint}{Character. API endpoint path (e.g. "ios/sales_report_estimates").}

\item{params}{List. Query parameters.}

\item{auth_token}{Character. API token.}

\item{verbose}{Logical. Whether to print debug messages.}

\item{enrich_response}{Logical. Whether to enrich the response with metadata.}

\item{processor}{Function. Function to process the response. Defaults to process_response.}
}
\value{
A tibble with the results.
}
\description{
A reusable function to handle the common pattern of fetching data from the API.
Handles URL building, request execution, error handling, and response processing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_publisher_apps.R
\name{st_publisher_apps}
\alias{st_publisher_apps}
\title{Get All Apps from a Publisher}
\usage{
st_publisher_apps(
  unified_id = NULL,
  publisher_id = NULL,
  aggregate_related = FALSE,
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
  verbose = TRUE
)
}
\arguments{
\item{unified_id}{Character. Unified ID to resolve apps for. May be either:
- Unified Publisher ID (24-char hex)
- Unified App ID (24-char hex) belonging to a publisher
The API returns the unified publisher and all associated apps in both cases.}

\item{publisher_id}{Deprecated alias for `unified_id`.}

\item{aggregate_related}{Logical. If TRUE, ensures each app's unified_app_id
is the canonical ID that aggregates ALL regional SKUs. This solves the problem
where games like "Watcher of Realms" are published under multiple regional
publishers (Moonton, Vizta Games, Skystone Games, etc.) and may return
different unified_app_ids. When TRUE, the function looks up each app by name
to find the true unified_app_id that combines all regional versions.
Defaults to FALSE for backwards compatibility.}

\item{auth_token}{Character. Your Sensor Tower API authentication token.
Defaults to the value stored in the `SENSORTOWER_AUTH_TOKEN` environment
variable.}

\item{verbose}{Logical. If TRUE, prints progress messages during aggregation.
Defaults to TRUE.}
}
\value{
A [tibble][tibble::tibble] containing details of the apps associated
  with the publisher. The exact columns depend on the API response but often
  include app IDs, names, platform, etc. Returns an empty tibble if the
  publisher ID is invalid, has no apps, or an error occurs.
}
\description{
Retrieves a list of apps associated with a specified unified publisher ID
from the Sensor Tower API. Targets the `/v1/unified/publishers/apps`
endpoint.
}
\section{Solving Regional Publisher Issues}{

Many publishers have regional subsidiaries or partners that publish the same
game under different app IDs in different regions. For example, Moonton's
"Watcher of Realms" is published by Moonton in some regions, Vizta Games in
others, and Skystone Games in others.

When `aggregate_related = TRUE`, this function ensures you get the unified_app_id
that represents the FULL game across all regional publishers, which is required
for accurate revenue/download aggregation via `st_unified_sales_report()`.
}

\section{API Endpoint Used}{

  - `GET /v1/unified/publishers/apps`
}

\examples{
\dontrun{
# Ensure SENSORTOWER_AUTH_TOKEN is set in your environment
# Sys.setenv(SENSORTOWER_AUTH_TOKEN = "your_secure_auth_token_here")

# Basic usage - get publisher's apps
apps_list <- st_publisher_apps(unified_id = "647eb849d9d91f31a54f1792")

# With regional SKU aggregation - ensures canonical unified_app_ids
apps_list <- st_publisher_apps(
  unified_id = "647eb849d9d91f31a54f1792",
  aggregate_related = TRUE
)

# Then use with st_unified_sales_report() for accurate data
sales <- st_unified_sales_report(
  unified_app_id = apps_list$unified_app_id,
  countries = "WW",
  start_date = "2024-01-01",
  end_date = "2024-12-31",
  date_granularity = "monthly"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_get_unified_mapping.R
\name{st_get_unified_mapping}
\alias{st_get_unified_mapping}
\title{Get Unified ID Mapping for Apps}
\usage{
st_get_unified_mapping(
  app_ids,
  app_names = NULL,
  os = "unified",
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN")
)
}
\arguments{
\item{app_ids}{Character vector of app IDs (can be iOS, Android, or unified hex IDs)}

\item{app_names}{Character vector of app names (optional, helps with resolution)}

\item{os}{Character string. Operating system: "ios", "android", or "unified"}

\item{auth_token}{Character string. Sensor Tower API authentication token.
Defaults to environment variable SENSORTOWER_AUTH_TOKEN.}
}
\value{
A data frame with columns:
  - `input_id`: The original ID provided
  - `unified_app_id`: The unified app ID (hex format)
  - `unified_app_name`: The unified app name
  - `ios_app_id`: iOS app ID (if available)
  - `android_app_id`: Android app ID (if available)
  - `publisher_id`: Publisher ID
  - `publisher_name`: Publisher name
}
\description{
Retrieves the mapping between platform-specific app IDs and unified app IDs.
This function handles cases where platform IDs from st_top_charts may not be
directly searchable, using app names as a fallback resolution method.
}
\details{
This function uses an ID-first approach (no name-based resolution):
1. For hex IDs (24-char), uses st_app_lookup to get platform IDs
2. For platform IDs, first tries to look them up via st_app_lookup
3. If direct lookup fails, searches the unified index using the platform ID
   as the term and matches exact IDs within nested ios_apps/android_apps
4. Returns the best available mapping for each app using IDs only

Note: Platform IDs from st_top_charts may be regional or legacy IDs that
aren't directly searchable. In these cases, name-based search provides
the most reliable resolution to unified IDs.
}
\examples{
\dontrun{
# Get mapping with app names for better resolution
mapping <- st_get_unified_mapping(
  app_ids = c("943599237", "com.bandainamcogames.dbzdokkan"),
  app_names = c("Dragon Ball Z Dokkan Battle", "Dragon Ball Z Dokkan Battle"),
  os = "unified"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_smart_metrics.R
\name{st_cache_info}
\alias{st_cache_info}
\title{Show App ID Cache Statistics}
\usage{
st_cache_info()
}
\value{
No return value, called for side effects (displaying cache statistics).
}
\description{
Display information about the current app ID cache
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_category_rankings.R
\name{process_ranking_response}
\alias{process_ranking_response}
\title{Process Category Ranking API Response}
\usage{
process_ranking_response(resp, os, category, country, chart_type, date, limit)
}
\arguments{
\item{resp}{Response object from httr2}

\item{os}{Operating system}

\item{category}{Category ID used in request}

\item{country}{Country code used in request}

\item{chart_type}{Chart type used in request}

\item{date}{Date used in request (may be NULL)}

\item{limit}{Number of results requested}
}
\value{
A processed tibble with ranking data
}
\description{
Internal function to process and enrich category ranking API responses.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_game_summary.R
\name{st_game_summary}
\alias{st_game_summary}
\title{Fetch Game Market Summary Data}
\usage{
st_game_summary(
  categories = 7001,
  countries,
  os,
  date_granularity,
  start_date,
  end_date,
  auth_token = NULL,
  base_url = "https://api.sensortower.com",
  enrich_response = TRUE
)
}
\arguments{
\item{categories}{Character string or numeric vector. Game category IDs to 
analyze. Defaults to 7001 (a popular game category). Use `st_categories()` 
to find valid category IDs.}

\item{countries}{Character vector or comma-separated string. Country codes 
(e.g., `"US"`, `c("US", "GB")`, `"WW"` for worldwide) to analyze. Required.}

\item{os}{Character string. Operating System. Must be one of "ios", "android", 
or "unified". Required. Note: The underlying API only supports "ios" and "android";
when `os = "unified"` this function automatically fetches both platforms and
combines them into a single table with total columns.}

\item{date_granularity}{Character string. Time granularity for aggregation. 
Must be one of "daily", "weekly", "monthly", or "quarterly". Required.}

\item{start_date}{Character string or Date object. Start date for the query 
in "YYYY-MM-DD" format. Required.}

\item{end_date}{Character string or Date object. End date for the query 
in "YYYY-MM-DD" format, inclusive. Required.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.}

\item{base_url}{Optional. Character string. The base URL for the API.}

\item{enrich_response}{Optional. Logical. If `TRUE` (default), enriches
the response with readable column names and processes the data.}
}
\value{
A [tibble][tibble::tibble] with aggregate game market summary data including:
  - **Category information**: Game category details
  - **Geographic data**: Country-level breakdowns
  - **Downloads**: iOS (iPhone + iPad combined) and Android download estimates
  - **Revenue**: iOS (iPhone + iPad combined) and Android revenue estimates,
    expressed in dollars when `enrich_response = TRUE`
  - **Totals (unified only)**: `Total Downloads`, `Total Revenue`
  - **Time series**: Data broken down by specified granularity
  
  **Automatic Data Combination**: For iOS and unified platforms, iPhone and iPad
  data are automatically combined into single "iOS Downloads" and "iOS Revenue" 
  columns for simplified analysis.
}
\description{
Retrieves aggregate download and revenue estimates by game categories,
countries, and date ranges from Sensor Tower's `games_breakdown` endpoint.
Use this for market/category denominator series; do not approximate market
totals by batching top charts, rankings, or large app rosters.
}
\section{API Endpoint Used}{

  - **Game Summary**: `GET /v1/\{os\}/games_breakdown`
    (API only supports `os = "ios"` or `os = "android"`; unified is synthesized)
}

\section{Revenue Units}{

  The raw `games_breakdown` endpoint returns revenue in cents. With
  `enrich_response = TRUE`, this function converts revenue fields to dollars
  before combining devices or platforms. With `enrich_response = FALSE`, raw
  endpoint fields are returned unchanged.
}

\section{Field Mappings and Processing}{

  The API returns abbreviated field names which are automatically mapped to 
  descriptive names and processed:
  - **iOS**: `iu` + `au` = iOS Downloads (combined), `ir` + `ar` = iOS Revenue (combined)
  - **Android**: `u` = Android Downloads, `r` = Android Revenue
  - **Common**: `ca` = Category, `cc` = Country Code, `d` = Date
  
  iPhone and iPad data are automatically combined for simplified analysis.
  For `os = "unified"`, iOS and Android aggregate rows are summed by date and
  country across the requested category basket.
}

\examples{
\dontrun{
# Specific categories and countries
rpg_summary <- st_game_summary(
  categories = c(7001, 7002),
  countries = c("US", "GB", "DE"),
  os = "ios",
  date_granularity = "weekly",
  start_date = "2024-01-01",
  end_date = "2024-03-31"
)

# Monthly summary for iOS games in the US
ios_monthly <- st_game_summary(
  os = "ios",
  countries = "US", 
  date_granularity = "monthly",
  start_date = "2024-01-01",
  end_date = "2024-06-30"
)
}

}
\seealso{
[st_categories()], [st_rankings()], [st_metrics()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_filter_utils.R
\name{create_custom_filter_example}
\alias{create_custom_filter_example}
\title{Create Custom Filter Example}
\usage{
create_custom_filter_example(function_name, os = "ios")
}
\arguments{
\item{function_name}{Character. Name of the function}

\item{os}{Character. Example OS to use}
}
\value{
Character string with example code
}
\description{
Generates example code for using custom filters with a function
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_demographics.R
\name{st_demographics}
\alias{st_demographics}
\title{Fetch Demographics Data for Apps}
\usage{
st_demographics(
  unified_app_id = NULL,
  ios_app_id = NULL,
  android_app_id = NULL,
  os = NULL,
  country = "US",
  date_granularity = "all_time",
  start_date = NULL,
  end_date = NULL,
  auth_token = NULL,
  verbose = TRUE
)
}
\arguments{
\item{unified_app_id}{Character string. Sensor Tower unified app ID (24-character hex).
Will be resolved to platform-specific IDs automatically.}

\item{ios_app_id}{Character string. iOS app ID (numeric, e.g., "1234567890").}

\item{android_app_id}{Character string. Android package name (e.g., "com.example.app").}

\item{os}{Character string. Operating system: "ios" or "android". Required if using
platform-specific IDs. When using unified_app_id, defaults to trying both platforms.}

\item{country}{Character string. Country code (e.g., "US", "GB"). Default is "US".
Only single country supported per request.}

\item{date_granularity}{Character string. Either "all_time" (default) or "quarterly".
All-time data goes back to Q4 2015. Quarterly data begins Q1 2021.}

\item{start_date}{Date or character string. Start date for quarterly data in "YYYY-MM-DD" format.
Ignored for all_time granularity.}

\item{end_date}{Date or character string. End date for quarterly data in "YYYY-MM-DD" format.
Ignored for all_time granularity.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable SENSORTOWER_AUTH_TOKEN.}

\item{verbose}{Logical. If TRUE, prints progress messages.}
}
\value{
A [tibble][tibble::tibble] with demographics metrics including:
  - **app_id**: The platform-specific app ID
  - **os**: Platform (ios or android)
  - **country**: Country code
  - **female_percent**: Percentage of female users (0-100)
  - **male_percent**: Percentage of male users (0-100)
  - **average_age**: Average user age
  - **age_13_17**, **age_18_24**, **age_25_34**, **age_35_44**, **age_45_54**,
    **age_55_64**, **age_65_plus**: Age group percentages
  - **confidence**: Data confidence level
}
\description{
Retrieves user demographics (age and gender breakdown) for specific apps from
the Sensor Tower Usage Intelligence API. This function queries the demographics
endpoint directly using platform-specific app IDs.
}
\section{Data Availability}{

- Quarterly data begins Q1 2021
- All-time data goes back to Q4 2015
- Demographics are primarily available for US market
- Data availability depends on app's user base size
}

\section{Recommended Workflow}{

```
# Step 1: Search for app by name
app <- st_app_info("Royal Match")

# Step 2: Get demographics using unified ID
demographics <- st_demographics(unified_app_id = app$unified_app_id[1])
```
}

\examples{
\dontrun{
# Get demographics for an app using unified ID
demo <- st_demographics(unified_app_id = "5f16a8019f7b275235017614")

# Get demographics for iOS app directly
demo <- st_demographics(
  ios_app_id = "553834731",
  os = "ios",
  country = "US"
)
}

}
\seealso{
[st_app_info()] for searching apps by name,
  [st_app_lookup()] for resolving app IDs,
  [st_retention()] for retention metrics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_active_users.R
\name{st_active_users}
\alias{st_active_users}
\title{Fetch Active User Metrics with a Tidy Long Output}
\usage{
st_active_users(
  os,
  app_list,
  metrics = c("dau", "wau", "mau"),
  date_range = list(start_date = Sys.Date() - 90, end_date = Sys.Date() - 1),
  countries,
  granularity = "monthly",
  parallel = FALSE,
  verbose = TRUE,
  auth_token = NULL,
  max_cores = 2
)
}
\arguments{
\item{os}{Character. Required. Operating system: `"ios"`, `"android"`, or
`"unified"`.}

\item{app_list}{List or data frame containing app information. Same formats
accepted by [st_batch_metrics()].}

\item{metrics}{Character vector. Active-user metrics to return. Must be one
or more of `"dau"`, `"wau"`, `"mau"`. Default is all three.}

\item{date_range}{List with `start_date` and `end_date`.}

\item{countries}{Character vector of country codes.}

\item{granularity}{Character granularity (`"daily"`, `"weekly"`,
`"monthly"`, `"quarterly"`). Default `"monthly"`.}

\item{parallel}{Logical. Whether to use parallel processing.}

\item{verbose}{Logical. Whether to print progress messages.}

\item{auth_token}{Character string. Sensor Tower API token. Defaults to
`SENSORTOWER_AUTH_TOKEN` environment variable.}

\item{max_cores}{Integer. Maximum cores for parallel processing.}
}
\value{
A tibble with columns including `original_id`, `app_name`, `app_id`,
  `app_id_type`, `date`, `country`, `metric`, and `value`.
}
\description{
Lightweight wrapper around [st_batch_metrics()] focused only on active-user
metrics (`dau`, `wau`, `mau`). Returns a tidy long-format tibble that is
designed for straightforward tidyverse chaining.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_yoy_metrics.R
\name{st_yoy_metrics}
\alias{st_yoy_metrics}
\title{Year-over-Year Metrics Comparison}
\usage{
st_yoy_metrics(
  os,
  unified_app_id = NULL,
  ios_app_id = NULL,
  android_app_id = NULL,
  publisher_id = NULL,
  years = NULL,
  period_start,
  period_end,
  metrics = c("revenue", "downloads"),
  countries,
  cache_dir = NULL,
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN"),
  verbose = TRUE,
  granularity,
  use_single_fetch = TRUE
)
}
\arguments{
\item{os}{Character. Required. Operating system: "ios", "android", or "unified".
This determines which platform's data is returned.}

\item{unified_app_id}{Character vector. Sensor Tower unified app ID(s). 
Must be 24-character hex format (e.g., "5ba4585f539ce75b97db6bcb").}

\item{ios_app_id}{Character vector. iOS app ID(s) (e.g., "1234567890").}

\item{android_app_id}{Character vector. Android package name(s) (e.g., "com.example.app").}

\item{publisher_id}{Character vector. Publisher ID(s) (alternative to app IDs).}

\item{years}{Integer vector. Years to compare (e.g., c(2022, 2023, 2024)).
If NULL, uses current year and previous year.}

\item{period_start}{Character string or Date. Start of the comparison period.
Can be "MM-DD" format (e.g., "01-01" for Jan 1) or a full date.
If a full date is provided, only the month and day are used.}

\item{period_end}{Character string or Date. End of the comparison period.
Can be "MM-DD" format (e.g., "03-31" for Mar 31) or a full date.
If a full date is provided, only the month and day are used.}

\item{metrics}{Character vector. Metrics to fetch. Supports "revenue", "downloads", 
"dau", "wau", and "mau". Default is both revenue and downloads.}

\item{countries}{Character vector. Country codes (e.g., "US", "GB", "JP"). Required.}

\item{cache_dir}{Character. Directory for caching API responses (optional).}

\item{auth_token}{Character string. Sensor Tower API token.}

\item{verbose}{Logical. Print progress messages.}

\item{granularity}{Character. Date granularity for the data (e.g., "daily", "monthly").}

\item{use_single_fetch}{Logical. If TRUE, uses a single API call to fetch all data.
Defaults to TRUE for efficiency.}
}
\value{
A tibble in tidy/long format with columns:
  - `app_id`: The app ID used for fetching data
  - `app_id_type`: Type of app ID ("ios", "android", or "unified")
  - `entity_id`: App or publisher ID
  - `entity_name`: App or publisher name
  - `entity_type`: "app" or "publisher"
  - `year`: Year of the data
  - `date_start`: Start date of the period (YYYY-MM-DD)
  - `date_end`: End date of the period (YYYY-MM-DD)
  - `country`: Country code
  - `metric`: The metric name (e.g., "revenue", "downloads", "dau")
  - `value`: Metric value (units depend on metric type)
  - `yoy_change`: Year-over-year change (percentage)
  - `yoy_change_absolute`: Year-over-year change (absolute value)
}
\description{
Fetches metrics for the same date range across multiple years for 
year-over-year comparisons. Allows flexible date ranges and supports
all available metrics including revenue, downloads, and active users.
}
\details{
This function is designed for year-over-year comparisons:

- **Flexible date ranges**: Compare any period (e.g., Q1, specific months, custom ranges)
- **Multiple years**: Compare across 2+ years in a single call
- **Smart date handling**: Automatically handles leap years and invalid dates
- **YoY calculations**: Includes both percentage and absolute change
- **Caching**: Reuses cached data to minimize API calls

The function will apply the same calendar period (month/day range) to each
specified year, making it easy to compare seasonal trends, campaign periods,
or any custom date range across years.
}
\examples{
\dontrun{
# Compare Q1 performance across years
q1_comparison <- st_yoy_metrics(
  os = "ios",
  ios_app_id = "553834731",  # Candy Crush iOS
  years = c(2022, 2023, 2024),
  period_start = "01-01",
  period_end = "03-31",
  countries = "US",
  metrics = c("revenue", "downloads")
)

# Compare holiday season (Nov-Dec) across years
holiday_comparison <- st_yoy_metrics(
  os = "unified",
  unified_app_id = "5ba4585f539ce75b97db6bcb",
  years = c(2021, 2022, 2023),
  period_start = "11-01",
  period_end = "12-31",
  countries = c("US", "GB", "JP"),
  metrics = c("revenue", "downloads", "dau")
)

# Compare specific campaign period using full dates
campaign_comparison <- st_yoy_metrics(
  os = "android", 
  android_app_id = "com.king.candycrushsaga",
  years = NULL,  # Uses current and previous year
  period_start = as.Date("2024-02-14"),  # Valentine's campaign
  period_end = as.Date("2024-02-28"),
  countries = c("US", "GB", "JP"),
  metrics = c("revenue", "downloads", "dau", "wau")
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields.R
\name{st_custom_fields_values}
\alias{st_custom_fields_values}
\title{Get Custom Fields Values}
\usage{
st_custom_fields_values(
  term = NULL,
  auth_token = NULL,
  base_url = "https://api.sensortower.com"
)
}
\arguments{
\item{term}{Optional. Character string. Search term to filter field names.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.}

\item{base_url}{Optional. Character string. The base URL for the API.}
}
\value{
A tibble containing custom fields and their possible values
}
\description{
Retrieves a list of all accessible custom fields and their possible values.
This is useful for discovering what custom fields are available to filter by.
}
\examples{
\dontrun{
# Get all custom fields
fields <- st_custom_fields_values()

# Search for specific fields
date_fields <- st_custom_fields_values(term = "date")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_users}
\alias{format_users}
\title{Format user counts (DAU/MAU/WAU)}
\usage{
format_users(val)
}
\arguments{
\item{val}{Numeric value to format as user count}
}
\value{
Formatted user count string
}
\description{
Format user counts (DAU/MAU/WAU)
}
\examples{
format_users(1234567)      # "1.2M"
format_users(1234)         # "1.2K"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_rankings.R
\name{st_rankings}
\alias{st_rankings}
\title{Fetch Sensor Tower Rankings}
\usage{
st_rankings(
  entity = c("app", "publisher", "category"),
  os = "ios",
  category = NULL,
  country = "US",
  chart_type = "topfreeapplications",
  date = Sys.Date() - 1,
  limit = 100,
  filter = NULL,
  auth_token = NULL
)
}
\arguments{
\item{entity}{Ranking entity to fetch: `"app"`, `"publisher"`, or
`"category"`.}

\item{os}{Operating system. One of `"ios"`, `"android"`, or `"unified"`.}

\item{category}{Optional category identifier forwarded to the legacy ranking
implementation.}

\item{country}{Two-letter country code.}

\item{chart_type}{Chart type for category rankings. Ignored for app and
publisher rankings.}

\item{date}{Ranking date. Accepts a `Date` object or ISO date string.}

\item{limit}{Positive integer row limit.}

\item{filter}{Optional `st_filter` object or 24-character filter ID string.}

\item{auth_token}{Optional Sensor Tower API token.}
}
\value{
A tibble with standardized columns `rank`, `id`, `name`, `os`,
  `category`, `country`, and `date`, plus any entity-specific columns returned
  by the dispatched implementation.
}
\description{
Thin v1.0.0 facade that standardizes the app, publisher, and category ranking
entrypoints while dispatching to the existing implementations.
}
\examples{
\dontrun{
st_rankings(entity = "app", os = "ios", category = 6014, country = "US")

st_rankings(
  entity = "category",
  os = "android",
  category = "game",
  chart_type = "topgrossing",
  country = "GB"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/custom_filter_utils.R
\name{get_custom_filter_docs}
\alias{get_custom_filter_docs}
\title{Get Custom Filter Documentation}
\usage{
get_custom_filter_docs()
}
\value{
Character vector with roxygen2 parameter documentation
}
\description{
Returns standardized documentation for custom filter parameters
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_url_parser.R
\name{st_build_web_url}
\alias{st_build_web_url}
\title{Create Sensor Tower Web URL from Parameters}
\usage{
st_build_web_url(
  os = "unified",
  measure = "revenue",
  category = NULL,
  regions = "US",
  start_date = NULL,
  end_date = NULL,
  custom_fields_filter_id = NULL,
  custom_tags_mode = NULL,
  ...
)
}
\arguments{
\item{os}{Operating system}

\item{measure}{Measure type}

\item{category}{Category ID}

\item{regions}{Region codes (converted to country parameters)}

\item{start_date}{Start date}

\item{end_date}{End date}

\item{custom_fields_filter_id}{Custom filter ID}

\item{custom_tags_mode}{Custom tags mode}

\item{...}{Additional parameters}
}
\value{
Character string URL
}
\description{
Builds a Sensor Tower web interface URL from API parameters.
This is the reverse of st_parse_web_url().
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields_workflow.R
\name{st_get_filtered_apps}
\alias{st_get_filtered_apps}
\title{Get Top Apps with Custom Filter}
\usage{
st_get_filtered_apps(
  field_name = NULL,
  field_values = NULL,
  filter_id = NULL,
  measure = "DAU",
  regions = "US",
  date = NULL,
  end_date = NULL,
  limit = 100,
  enrich_response = TRUE,
  auth_token = NULL,
  ...
)
}
\arguments{
\item{field_name}{Character. Name of the custom field to filter by (or NULL to use filter_id)}

\item{field_values}{Character vector. Values to filter for (or NULL to use filter_id)}

\item{filter_id}{Character. Existing filter ID to use (alternative to field_name/values)}

\item{measure}{Character. Metric to measure: "DAU", "WAU", "MAU", "revenue", or "units"}

\item{regions}{Character vector. Region codes (e.g., "US", "WW")}

\item{date}{Character or Date. Start date for the query}

\item{end_date}{Optional. Character or Date. End date for the query}

\item{limit}{Integer. Maximum number of apps to return (default 100)}

\item{enrich_response}{Logical. Whether to enrich with additional metrics}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.}

\item{...}{Additional parameters passed to st_top_charts}
}
\value{
A tibble with top apps data
}
\description{
Retrieves top apps using a custom fields filter. This combines filter creation
with data retrieval in a single workflow.
}
\examples{
\dontrun{
# Get top Word games by DAU
word_games <- st_get_filtered_apps(
  field_name = "Game Sub-genre",
  field_values = "Word",
  measure = "DAU",
  regions = "US",
  limit = 20
)

# Use existing filter ID
apps <- st_get_filtered_apps(
  filter_id = "603697f4241bc16eb8570d37",
  measure = "revenue",
  regions = "US"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_market_share}
\alias{format_market_share}
\title{Format market share as percentage}
\usage{
format_market_share(val, digits = 1)
}
\arguments{
\item{val}{Numeric value as decimal (0-1 scale)}

\item{digits}{Number of decimal places (default: 1)}
}
\value{
Formatted market share percentage string
}
\description{
Format market share as percentage
}
\examples{
format_market_share(0.234)    # "23.4\%"
format_market_share(0.05)     # "5.0\%"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields.R
\name{st_custom_fields}
\alias{st_custom_fields}
\title{Custom Fields Filter Functions}
\description{
Functions to work with Sensor Tower custom fields filters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_helpers.R
\name{format_currency}
\alias{format_currency}
\title{Format currency values with appropriate suffixes}
\usage{
format_currency(val, digits = 2)
}
\arguments{
\item{val}{Numeric value to format as currency}

\item{digits}{Number of digits after decimal for millions/billions (default: 2)}
}
\value{
Formatted currency string
}
\description{
Format currency values with appropriate suffixes
}
\examples{
format_currency(1234567)      # "$1.23M"
format_currency(1234567890)   # "$1.23B"
format_currency(123)          # "$123"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{st_clear_app_cache}
\alias{st_clear_app_cache}
\title{Clear App Name Cache}
\usage{
st_clear_app_cache()
}
\value{
No return value, called for side effects (clearing the cache).
}
\description{
Clears the internal cache of app name lookups. Useful for testing or when
you want to refresh app name data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_market_metrics.R
\name{st_market_metrics}
\alias{st_market_metrics}
\title{Fetch Aggregate Sensor Tower Market Metrics}
\usage{
st_market_metrics(
  category,
  countries,
  os,
  date_from,
  date_to,
  granularity = "monthly",
  metrics = c("revenue", "downloads"),
  revenue_unit = c("dollars", "cents"),
  shape = c("long", "wide"),
  auth_token = NULL
)
}
\arguments{
\item{category}{Character or numeric vector of Sensor Tower game category IDs.
iOS and Android use different category identifiers; pass the appropriate
platform category for platform-specific calls, or the full platform basket
for `os = "unified"`.}

\item{countries}{Character vector of 2-letter country codes. Use `"WW"` for
worldwide.}

\item{os}{Character scalar. One of `"ios"`, `"android"`, or `"unified"`.}

\item{date_from, date_to}{Date bounds for the query. Accept `Date` objects or
ISO date strings.}

\item{granularity}{Character scalar. One of `"daily"`, `"weekly"`,
`"monthly"`, or `"quarterly"`.}

\item{metrics}{Character vector. One or both of `"revenue"` and `"downloads"`.}

\item{revenue_unit}{Character scalar. `"dollars"` returns revenue in base
currency units. `"cents"` returns revenue in cents.}

\item{shape}{Character scalar. `"long"` returns one row per metric
observation. `"wide"` returns metric columns.}

\item{auth_token}{Optional Sensor Tower API token. If `NULL`, falls back to
`SENSORTOWER_AUTH_TOKEN`.}
}
\value{
If `shape = "long"`, a tibble with columns `date`, `country`, `category_id`,
`os`, `metric`, and `value`. If `shape = "wide"`, a tibble with columns
`date`, `country`, `category_id`, `os`, and requested metric columns.
}
\description{
Tidy facade for market/category revenue and download denominators. This
function routes to Sensor Tower's aggregate `games_breakdown` endpoint through
[st_game_summary()] and never approximates market totals by enumerating apps,
rankings, top charts, or title cohorts.
}
\section{Denominator Safety}{

Use `st_market_metrics()` for game category market denominators. Use
[st_rankings()], [st_apps()], and custom-field top-chart workflows for
discovery, leaderboards, and title cohorts only. True Game IQ subgenre market
denominators require a Sensor Tower market-size export or another confirmed
aggregate market endpoint.
}

\examples{
\dontrun{
casino_market <- st_market_metrics(
  category = c("7006", "game_casino"),
  countries = "WW",
  os = "unified",
  date_from = "2025-01-01",
  date_to = "2025-12-31",
  granularity = "monthly"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_facets_metrics.R
\name{st_facets_metrics}
\alias{st_facets_metrics}
\title{Low-Level Access to Sensor Tower Facets Metrics}
\usage{
st_facets_metrics(
  query = character(),
  params = list(),
  auth_token = NULL,
  host = c("api", "app"),
  verbose = FALSE
)
}
\arguments{
\item{query}{Character vector of raw query fragments appended verbatim after
encoded `params`. This is useful when Sensor Tower documents nested or
repeated parameters that are awkward to express as a regular named R list.
Example: `c("filters[app_ids][]=553834731")`.}

\item{params}{Named list of regular query parameters. Vector values are
serialized as comma-separated strings to match Sensor Tower's
`style=form, explode=false` usage in the OpenAPI spec.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable `SENSORTOWER_AUTH_TOKEN`.}

\item{host}{Character string. Which host to target: `"api"` (default) or
`"app"`. The `app` host uses `https://app.sensortower.com/api`.}

\item{verbose}{Logical. If `TRUE`, prints the request URL with the auth token
redacted.}
}
\value{
A parsed JSON response. Rectangular top-level responses are returned
  as a tibble; nested responses are returned as a named list.
}
\description{
Performs a GET request against Sensor Tower's `"/v1/facets/metrics"` route.
This helper stays intentionally low-level so package code can work with new
facets-based endpoints while higher-level wrappers are added incrementally.
}
\details{
As of March 17, 2026, the facets route is live and the retention contract is
validated against production. The machine-readable Sensor Tower docs remain
gated behind a signed-in web session, so this helper still provides a stable
escape hatch for raw or partially documented facets requests.
}
\examples{
\dontrun{
# Retention request using regular query parameters
response <- st_facets_metrics(
  params = list(
    facets = "retention",
    bundle = "retention_daily",
    breakdown = c("date", "app_id"),
    start_date = "2025-01-01",
    end_date = "2025-01-31",
    app_ids = "553834731"
  )
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_api_diagnostics.R
\name{st_api_diagnostics}
\alias{st_api_diagnostics}
\title{Diagnose API Issues}
\usage{
st_api_diagnostics(
  app_id,
  verbose = TRUE,
  auth_token = Sys.getenv("SENSORTOWER_AUTH_TOKEN")
)
}
\arguments{
\item{app_id}{Character string. The app ID to diagnose (can be unified, iOS, or Android)}

\item{verbose}{Logical. Show detailed diagnostic output. Default is TRUE.}

\item{auth_token}{Character string. Your Sensor Tower API authentication token.}
}
\value{
A list with diagnostic results including:
  - `id_type`: Detected ID type
  - `platform_ids`: Resolved platform-specific IDs
  - `endpoint_results`: Results from testing various endpoints
  - `recommendations`: Suggested approach for this app
}
\description{
This function helps diagnose common API issues by testing various ID formats
and endpoints to determine the best approach for fetching data.
}
\examples{
\dontrun{
# Diagnose Star Trek Fleet Command
diagnosis <- st_api_diagnostics("5ba4585f539ce75b97db6bcb")

# Check iOS app
diagnosis <- st_api_diagnostics("1427744264")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_gt_dashboard.R
\name{st_gt_dashboard}
\alias{st_gt_dashboard}
\title{Create FiveThirtyEight-styled GT Dashboard from Top Charts Data}
\usage{
st_gt_dashboard(
  data,
  title = "Top Mobile Games",
  subtitle = NULL,
  ranking_metric = "revenue_180d_ww",
  show_demographics = TRUE,
  show_engagement = TRUE,
  show_retention = TRUE,
  retention_region = "us",
  show_rpd = TRUE,
  bar_charts = TRUE,
  bar_chart_columns = NULL,
  heatmap_retention = TRUE,
  compact_mode = TRUE,
  width = 1800,
  height = 700,
  save_path = NULL,
  icon_cache_dir = "inst/images/app_icons",
  raw = FALSE,
  color_scheme = list(revenue = "#FF6600", downloads = "#008FD5", engagement = "#9C27B0",
    rpd = "#4CAF50", retention_low = "#FFCDD2", retention_mid = "#C8E6C9", retention_high
    = "#4CAF50")
)
}
\arguments{
\item{data}{Data frame from st_top_charts() or similar Sensor Tower function}

\item{title}{Character string for the table title (default: "Top Mobile Games")}

\item{subtitle}{Character string for subtitle. If NULL, auto-generates based on data}

\item{ranking_metric}{Character string specifying which metric to use for ranking.
Options: "revenue_180d_ww", "revenue_30d_ww", "downloads_180d_ww", "downloads_30d_ww", etc.
(default: "revenue_180d_ww")}

\item{show_demographics}{Logical, whether to show demographic columns (age, gender)
(default: TRUE)}

\item{show_engagement}{Logical, whether to show engagement metrics (DAU, WAU, MAU)
(default: TRUE)}

\item{show_retention}{Logical, whether to show retention metrics
(default: TRUE)}

\item{retention_region}{Character string for retention region ("us", "ww", etc.)
(default: "us")}

\item{show_rpd}{Logical, whether to show Revenue Per Download
(default: TRUE)}

\item{bar_charts}{Logical, whether to show bar chart visualizations
(default: TRUE)}

\item{bar_chart_columns}{Character vector of column patterns to add bar charts to.
If NULL, applies to all numeric columns except RPD and retention.}

\item{heatmap_retention}{Logical, whether to apply heatmap to retention columns
(default: TRUE)}

\item{compact_mode}{Logical, whether to use compact row heights
(default: TRUE)}

\item{width}{Numeric, table width in pixels (default: 1800)}

\item{height}{Numeric, table height in pixels (default: 700)}

\item{save_path}{Character string, path to save the table image. If NULL, returns GT object}

\item{icon_cache_dir}{Character string, directory to cache app icons
(default: "inst/images/app_icons")}

\item{raw}{Logical, whether to return a minimally styled table without 
custom formatting, bar charts, or heatmaps (default: FALSE)}

\item{color_scheme}{List with color codes for different metric types:
- revenue: Revenue metrics color (default: "#FF6600")
- downloads: Downloads metrics color (default: "#008FD5")
- engagement: Engagement metrics color (default: "#9C27B0")
- rpd: RPD metrics color (default: "#4CAF50")
- retention_low: Low retention color (default: "#FFCDD2")
- retention_mid: Mid retention color (default: "#C8E6C9")
- retention_high: High retention color (default: "#4CAF50")}
}
\value{
GT object (if save_path is NULL) or saves image and returns path
}
\description{
Creates a professional, FiveThirtyEight-themed GT table dashboard from 
Sensor Tower top charts data with customizable styling and metric options.
}
\examples{
\dontrun{
# Basic usage - one line after st_top_charts()
top_rpgs <- st_top_charts(category = 7014, measure = "revenue")
st_gt_dashboard(top_rpgs)

# Raw mode for minimal styling
st_gt_dashboard(top_rpgs, raw = TRUE)

# Customize the dashboard
st_gt_dashboard(
  top_rpgs,
  title = "Top Role-Playing Games Q4 2024",
  ranking_metric = "revenue_30d_ww",
  show_retention = FALSE,
  save_path = "dashboard.png"
)

# Change color scheme
st_gt_dashboard(
  top_rpgs,
  color_scheme = list(
    revenue = "#E74C3C",
    downloads = "#3498DB",
    engagement = "#9B59B6"
  )
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_id_cache.R
\docType{data}
\name{id_cache}
\alias{id_cache}
\alias{.sensortowerR_env}
\title{Intelligent App ID Cache and Resolution System}
\format{
An object of class \code{environment} of length 0.
}
\usage{
.sensortowerR_env
}
\description{
Internal functions to manage a persistent cache of app ID mappings,
reducing API calls and improving performance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_retention.R
\name{st_retention}
\alias{st_retention}
\title{Fetch Retention Data for Apps}
\usage{
st_retention(
  unified_app_id = NULL,
  ios_app_id = NULL,
  android_app_id = NULL,
  os = NULL,
  country = "US",
  date_granularity = "all_time",
  start_date = NULL,
  end_date = NULL,
  auth_token = NULL,
  verbose = TRUE
)
}
\arguments{
\item{unified_app_id}{Character string. Sensor Tower unified app ID (24-character hex).
Will be resolved to platform-specific IDs automatically.}

\item{ios_app_id}{Character string. iOS app ID (numeric, e.g., "1234567890").}

\item{android_app_id}{Character string. Android package name (e.g., "com.example.app").}

\item{os}{Character string. Operating system: "ios" or "android". Required if using
platform-specific IDs. When using unified_app_id, defaults to "ios" but will
try both platforms.}

\item{country}{Character string. Country code (e.g., "US", "GB"). Default is "US".
Only single country supported per request.}

\item{date_granularity}{Character string. Either "all_time" (default) or "quarterly".
All-time data goes back to Q4 2015. Quarterly data begins Q1 2021.}

\item{start_date}{Date or character string. Start date for quarterly data in "YYYY-MM-DD" format.
Ignored for all_time granularity.}

\item{end_date}{Date or character string. End date for quarterly data in "YYYY-MM-DD" format.
Ignored for all_time granularity.}

\item{auth_token}{Optional. Character string. Your Sensor Tower API token.
Defaults to environment variable SENSORTOWER_AUTH_TOKEN.}

\item{verbose}{Logical. If TRUE, prints progress messages.}
}
\value{
A [tibble][tibble::tibble] with retention metrics including:
  - **app_id**: The platform-specific app ID
  - **os**: Platform (ios or android)
  - **country**: Country code
  - **retention_d1** through **retention_d90**: Retention percentages (0-1 scale)
  - **confidence**: Data confidence level (red=low, yellow=medium, green=high)
  - **baseline_downloads**: Total downloads in baseline period
  - **baseline_start_date**, **baseline_end_date**: Dates for baseline period
}
\description{
Retrieves retention metrics (D1-D90) for specific apps from the Sensor Tower
Usage Intelligence API. This function queries the retention endpoint directly
using platform-specific app IDs.
}
\section{Data Availability}{

- Quarterly data begins Q1 2021
- All-time data goes back to Q4 2015
- Data is only available for apps with sufficient user base
- Confidence levels: red (<=3), yellow (4-6), green (>=7)
}

\section{Recommended Workflow}{

```
# Step 1: Search for app by name
app <- st_app_info("Royal Match")

# Step 2: Get retention data using unified ID
retention <- st_retention(unified_app_id = app$unified_app_id[1])
```
}

\examples{
\dontrun{
# Get retention for an app using unified ID
retention <- st_retention(unified_app_id = "5f16a8019f7b275235017614")

# Get retention for iOS app directly
retention <- st_retention(
  ios_app_id = "553834731",
  os = "ios",
  country = "US"
)

# Get quarterly retention data
retention <- st_retention(
  unified_app_id = "5f16a8019f7b275235017614",
  date_granularity = "quarterly",
  start_date = "2024-01-01",
  end_date = "2024-09-30"
)
}

}
\seealso{
[st_app_info()] for searching apps by name,
  [st_app_lookup()] for resolving app IDs,
  [st_demographics()] for user demographics data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/st_custom_fields_utils.R
\name{st_custom_fields_utils}
\alias{st_custom_fields_utils}
\title{Custom Fields Utility Functions}
\description{
Utility functions for common custom fields filtering scenarios in Sensor Tower.
These functions provide pre-built filters for frequently used queries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defunct.R
\name{sensortowerR-defunct}
\alias{sensortowerR-defunct}
\alias{st_sales_report}
\alias{st_unified_sales_report}
\alias{st_batch_metrics}
\alias{st_smart_metrics}
\alias{st_top_charts}
\alias{st_top_publishers}
\alias{st_category_rankings}
\alias{st_app_info}
\alias{st_app_lookup}
\alias{st_app_details}
\alias{st_filter_by_date}
\alias{st_filter_by_genre}
\alias{st_filter_by_monetization}
\alias{st_filter_by_publisher}
\alias{st_filter_by_sdk}
\alias{st_custom_fields_filter}
\alias{st_custom_fields_filter_by_id}
\alias{st_combine_filters}
\alias{st_create_simple_filter}
\alias{st_compare_filter_results}
\alias{st_generate_example_filter_ids}
\title{Defunct Functions In sensortowerR}
\usage{
st_sales_report(...)

st_unified_sales_report(...)

st_batch_metrics(...)

st_smart_metrics(...)

st_top_charts(...)

st_top_publishers(...)

st_category_rankings(...)

st_app_info(...)

st_app_lookup(...)

st_app_details(...)

st_filter_by_date(...)

st_filter_by_genre(...)

st_filter_by_monetization(...)

st_filter_by_publisher(...)

st_filter_by_sdk(...)

st_custom_fields_filter(...)

st_custom_fields_filter_by_id(...)

st_combine_filters(...)

st_create_simple_filter(...)

st_compare_filter_results(...)

st_generate_example_filter_ids(...)
}
\description{
Legacy entrypoints retained only as defunct stubs for the v1.0.0 transition.
}
\keyword{internal}
